Package: ccar3
Title: Canonical Correlation Analysis via Reduced Rank Regression
Version: 0.1.2
Date: 2026-06-12
Authors@R: c(
    person("Claire", "Donnat", , "cdonnat@uchicago.edu", 
           role = c("aut", "cre"),
           comment = c(ORCID = "https://orcid.org/0000-0001-7079-8060")),
    person("Elena", "Tuzhilina", , "elena.tuzhilina@utoronto.ca", 
           role = "aut",
           comment = c(ORCID = "https://orcid.org/0000-0002-1898-6010")),
    person("Zixuan", "Wu", , "zixuanwu@uchicago.edu", 
           role = "aut",
           comment = c(ORCID = "https://orcid.org/0009-0006-4745-0000")))
Author: Claire Donnat [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-7079-8060>),
  Elena Tuzhilina [aut] (ORCID: <https://orcid.org/0000-0002-1898-6010>),
  Zixuan Wu [aut] (ORCID: <https://orcid.org/0009-0006-4745-0000>)
Maintainer: Claire Donnat <cdonnat@uchicago.edu>
Description: Canonical correlation analysis (CCA) via reduced-rank regression with support for regularization and cross-validation. Several methods for estimating CCA in high-dimensional settings are implemented. The first set of methods, cca_rrr() (and variants: cca_group_rrr() and cca_graph_rrr()), assumes that one dataset is high-dimensional and the other is low-dimensional, while the second, ecca() (for Efficient CCA) assumes that both datasets are high-dimensional. For both methods, standard l1 regularization as well as group-lasso regularization are available. cca_graph_rrr further supports total variation regularization when there is a known graph structure among the variables of the high-dimensional dataset. In this case, the loadings of the canonical directions of the high-dimensional dataset are assumed  to be smooth on the graph. For more details see Donnat and Tuzhilina (2024)  <doi:10.48550/arXiv.2405.19539> and Wu, Tuzhilina and Donnat (2025) <doi:10.48550/arXiv.2507.11160>.
Depends: R (>= 3.5.0)
Imports: methods, magrittr, tidyr, dplyr, foreach, pracma, corpcor,
        matrixStats, RSpectra
Suggests: codetools, SMUT, igraph, testthat (>= 3.0.0), pkgload,
        rrpack, Matrix, glmnet, CCA, CVXR, PMA, doParallel, crayon
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-06-12 19:18:09 UTC; clairedonnat
Repository: CRAN
Date/Publication: 2026-06-13 06:10:02 UTC
Built: R 4.5.2; ; 2026-06-13 10:13:29 UTC; unix

Package: ccar3
Title: Canonical Correlation Analysis via Reduced Rank Regression
Version: 0.1.2
Date: 2026-06-12
Authors@R: c(
    person("Claire", "Donnat", , "cdonnat@uchicago.edu", 
           role = c("aut", "cre"),
           comment = c(ORCID = "https://orcid.org/0000-0001-7079-8060")),
    person("Elena", "Tuzhilina", , "elena.tuzhilina@utoronto.ca", 
           role = "aut",
           comment = c(ORCID = "https://orcid.org/0000-0002-1898-6010")),
    person("Zixuan", "Wu", , "zixuanwu@uchicago.edu", 
           role = "aut",
           comment = c(ORCID = "https://orcid.org/0009-0006-4745-0000")))
Author: Claire Donnat [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-7079-8060>),
  Elena Tuzhilina [aut] (ORCID: <https://orcid.org/0000-0002-1898-6010>),
  Zixuan Wu [aut] (ORCID: <https://orcid.org/0009-0006-4745-0000>)
Maintainer: Claire Donnat <cdonnat@uchicago.edu>
Description: Canonical correlation analysis (CCA) via reduced-rank regression with support for regularization and cross-validation. Several methods for estimating CCA in high-dimensional settings are implemented. The first set of methods, cca_rrr() (and variants: cca_group_rrr() and cca_graph_rrr()), assumes that one dataset is high-dimensional and the other is low-dimensional, while the second, ecca() (for Efficient CCA) assumes that both datasets are high-dimensional. For both methods, standard l1 regularization as well as group-lasso regularization are available. cca_graph_rrr further supports total variation regularization when there is a known graph structure among the variables of the high-dimensional dataset. In this case, the loadings of the canonical directions of the high-dimensional dataset are assumed  to be smooth on the graph. For more details see Donnat and Tuzhilina (2024)  <doi:10.48550/arXiv.2405.19539> and Wu, Tuzhilina and Donnat (2025) <doi:10.48550/arXiv.2507.11160>.
Depends: R (>= 3.5.0)
Imports: methods, magrittr, tidyr, dplyr, foreach, pracma, corpcor,
        matrixStats, RSpectra
Suggests: codetools, SMUT, igraph, testthat (>= 3.0.0), pkgload,
        rrpack, Matrix, glmnet, CCA, CVXR, PMA, doParallel, crayon
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-06-12 19:18:09 UTC; clairedonnat
Repository: CRAN
Date/Publication: 2026-06-13 06:10:02 UTC

# Generated by roxygen2: do not edit by hand

export(FPR)
export(SCCA_Parkhomenko)
export(SparseCCA)
export(TNR)
export(TPR)
export(Witten.CV)
export(cca_graph_rrr)
export(cca_graph_rrr_cv)
export(cca_group_rrr)
export(cca_group_rrr_cv)
export(cca_rrr)
export(cca_rrr_cv)
export(ecca)
export(ecca.cv)
export(get_edge_incidence)
export(principal_angles)
export(regular_cca)
export(sinTheta)
export(sparse_CCA_benchmarks)
export(subdistance)
importFrom(foreach,"%do%")
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(magrittr,"%>%")
importFrom(stats,cor)
importFrom(stats,cov)
importFrom(stats,median)
importFrom(stats,rnorm)
importFrom(stats,sd)
importFrom(stats,var)
importFrom(tidyr,replace_na)

#' @title Additional Benchmarks for Sparse CCA Methods
#' @param X_train Matrix of predictors (n x p)
#' @param Y_train  Matrix of responses (n x q)
#' @param S Optional covariance matrix (default is NULL, which computes it from X_train and Y_train)
#' @param rank Target rank for the CCA (default is 2)
#' @param method.type Type of method to use for Sparse CCA (default is "FIT_SAR_CV"). Choices include "FIT_SAR_BIC", "FIT_SAR_CV", "Witten_Perm", "Witten.CV",  and "SCCA_Parkhomenko".
#' @param kfolds Number of cross-validation folds (default is 5)
#' @param lambdax Vector of sparsity parameters for X (default is a sequence from 0 to 1 with step 0.1)
#' @param lambday Vector of sparsity parameters for Y (default is a sequence from 0 to 1 with step 0.1)
#' @param standardize Standardize (center and scale) the data matrices X and Y (default is TRUE) before analysis
#' 
#' 
#'
#' @return A matrix (p+q)x r containing the canonical directions for X and Y.
#' @export
sparse_CCA_benchmarks <- function(X_train, Y_train, S=NULL, 
                              rank=2, kfolds=5, method.type="FIT_SAR_CV",
                              lambdax = 10^seq(from=-3,to=2,length=10),
                              lambday = c(0, 1e-7, 1e-6, 1e-5),
                              standardize = TRUE){

  X_train = as.matrix(data.frame(X_train) %>% dplyr::mutate_all(~tidyr::replace_na(., mean(., na.rm = TRUE))))
  Y_train = as.matrix(data.frame(Y_train) %>% dplyr::mutate_all(~tidyr::replace_na(., mean(., na.rm = TRUE))))
  p1 <- dim(X_train)[2]
  p2 <- dim(Y_train)[2]
  p <- p1 + p2;
  n <- nrow(X_train)
  pp <- c(p1,p2);
  if(is.null(S)){
    S = cov(cbind(X_train, Y_train))
  }
  
  if (method.type=="FIT_SAR_BIC"){

    if (!requireNamespace("glmnet", quietly = TRUE) ||
        !requireNamespace("CCA", quietly = TRUE)) {
      stop("Packages 'glmnet' and 'CCA' must be installed to use the SAR approach.",
          call. = FALSE)
    }


    method<-SparseCCA(X=X_train,Y=Y_train,rank=rank,
                           lambdaAseq=lambdax,
                           lambdaBseq=lambday,
                           max.iter=100, conv=10^-2,
                           selection.criterion=1, n.cv=kfolds,
                           standardize=standardize)
    a_estimate = rbind(method$uhat, method$vhat)
    
  }
  if(method.type=="FIT_SAR_CV"){

    if (!requireNamespace("glmnet", quietly = TRUE) ||
        !requireNamespace("CCA", quietly = TRUE)) {
      stop("Packages 'glmnet' and 'CCA' must be installed to use the SAR approach.",
          call. = FALSE)
    }
    method<-SparseCCA(X=X_train,Y=Y_train,rank=rank,
                          lambdaAseq=lambdax,
                          lambdaBseq=lambday,
                          max.iter=100,conv=10^-2, selection.criterion=2, 
                          n.cv=kfolds,
                          standardize=standardize)
    a_estimate = rbind(method$uhat, method$vhat)
    
  }
  if (method.type=="Witten_Perm"){

    if (!requireNamespace("PMA", quietly = TRUE)) {
      stop("Package  'PMA' must be installed to use the Witten approach.",
          call. = FALSE)
    }
    Witten_Perm <- PMA::CCA.permute(x=X_train,z=Y_train,
                               typex="standard",typez="standard", 
                               penaltyxs =lambdax[which(lambdax < 1)],
                               penaltyzs = lambday[which(lambday < 1)],
                               standardize = standardize,
                               nperms=50)
    
    method<-PMA::CCA(x=X_train, z=Y_train, typex="standard",typez="standard",K=rank,
                         penaltyx=Witten_Perm$bestpenaltyx,penaltyz=Witten_Perm$bestpenaltyz,trace=FALSE,
                         standardize = standardize)
    a_estimate = rbind(method$u, method$v)
  }
  if(method.type=="Witten.CV"){

    if (!requireNamespace("PMA", quietly = TRUE)) {
      stop("Package  'PMA' must be installed to use the Witten approach.",
          call. = FALSE)
    }
    Witten_CV <- Witten.CV(
      X = X_train, Y = Y_train, n.cv = 5, rank = rank,
      lambdax = lambdax[which(lambdax < 1)],
      lambday = lambday[which(lambdax < 1)],
      standardize = standardize
    )
    
    method <-PMA::CCA(x=X_train,z=Y_train,typex="standard",typez="standard",
                 K=rank,penaltyx=Witten_CV$lambdax.opt,
                 penaltyz=Witten_CV$lambday.opt,trace=FALSE,
                 standardize = standardize)
    a_estimate = rbind(method$u, method$v)
    
  }
  if(method.type=="SCCA_Parkhomenko"){
    method<- SCCA_Parkhomenko(x.data=X_train, y.data=Y_train, Krank=rank,
                              lambda.v.seq = lambdax[which(lambdax < 2)],
                              lambda.u.seq = lambday[which(lambday < 2)],
                              standardize = standardize)
    a_estimate = rbind(method$uhat, method$vhat)
    
  }
  a_estimate <- gca_to_cca(a_estimate, S, pp)
  return(a_estimate)
}



gca_to_cca <-
  function(a_estimate, S, pp){
    p1 = pp[1];
    p2 = pp[2];
    p = p1 + p2;
    nnz_indices = which(apply(a_estimate, 1, function(x) sqrt(sum(x^2))) >0)
    nnz_indices_x = nnz_indices[which(nnz_indices<(p1+1))]
    nnz_indices_y = nnz_indices[which(nnz_indices>(p1))]
    ### Make sure things are normalized
    if (length(which(nnz_indices<(p1+1)))>0){
      sigmaxhat = S[nnz_indices_x,nnz_indices_x];
      a_estimate[nnz_indices_x,] = a_estimate[nnz_indices_x,] %*% pracma::sqrtm(t(a_estimate[nnz_indices_x,]) %*% sigmaxhat %*% a_estimate[nnz_indices_x,])$Binv;
    }
    if (length(nnz_indices_y)>0){
      sigmayhat = S[nnz_indices_y,nnz_indices_y];
      a_estimate[nnz_indices_y,] = a_estimate[nnz_indices_y,] %*% pracma::sqrtm(t(a_estimate[nnz_indices_y,]) %*% sigmayhat %*% a_estimate[nnz_indices_y,])$Binv;
    }
    
    u_estimate = a_estimate[1:p1,]
    v_estimate = a_estimate[(p1+1):p,] 
    l = list("U" = u_estimate, "V" = v_estimate)
    return(l)
  }

library(ccar3)
test_that("SAR works", {
  set.seed(123)
  r = 3
  skip_if_not_installed("glmnet")
  skip_if_not_installed("CCA")
  gen = generate_example_sparse_U(n=1000, p1=100, p2=10,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y
  
  result <- sparse_CCA_benchmarks(X,
                              Y, S=NULL, 
                              rank=r, kfolds=5, method.type = "FIT_SAR_CV",
                              lambdax = 10^seq(from=-3,to=,length=10),
                              lambday = 10^seq(from=-8,to=-1,length=8))
  
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], dim(X)[2])
  expect_equal(dim(result$V)[1], dim(Y)[2])

  
})


test_that("SCCA_Parkhomenko works", {
  set.seed(123)
  r = 3
  gen = generate_example_sparse_U(n=1000, p1=100, p2=10,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y
  
  result <- sparse_CCA_benchmarks(X,
                                  Y, S=NULL, 
                                  rank=r, kfolds=5, method.type = "SCCA_Parkhomenko",
                                  lambdax = 10^seq(from=-3,to=,length=10),
                                  lambday = 10^seq(from=-8,to=-1,length=8))
  
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], dim(X)[2])
  expect_equal(dim(result$V)[1], dim(Y)[2])
  
  
})



test_that("Witten.CV works", {
  skip_if_not_installed("PMA")
  set.seed(123)
  r = 3
  gen = generate_example_sparse_U(n=1000, p1=100, p2=10,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y
  
  result <- sparse_CCA_benchmarks(X,
                                  Y, S=NULL, 
                                  rank=r, kfolds=5, method.type = "Witten.CV",
                                  lambdax = 10^seq(from=-3,to=,length=10),
                                  lambday = 10^seq(from=-8,to=-1,length=8))
  
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], dim(X)[2])
  expect_equal(dim(result$V)[1], dim(Y)[2])
  
  
})


test_that("Witten_Perm works", {
  skip_if_not_installed("PMA")
  set.seed(123)
  r = 3
  gen = generate_example_sparse_U(n=1000, p1=100, p2=10,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y
  
  result <- sparse_CCA_benchmarks(X,
                                  Y, S=NULL, 
                                  rank=r, kfolds=5, method.type = "Witten_Perm",
                                  lambdax = 10^seq(from=-3,to=0,length=10),
                                  lambday = 10^seq(from=-8,to=0,length=10))
  
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], dim(X)[2])
  expect_equal(dim(result$V)[1], dim(Y)[2])
  
  
})

#' Function to perform Sparse CCA based on Wilms and Croux (2018)
#'  REFERENCE Wilms, I., & Croux, C. (2018). Sparse canonical correlation analysis using alternating regressions. Journal of Computational and Graphical Statistics, 27(1), 1-10.
#' @param X Matrix of predictors (n x p)
#' @param Y Matrix of responses (n x q)
#' @param lambdaAseq Vector of sparsity parameters for X (default is a sequence from 0 to 1 with step 0.1)
#' @param lambdaBseq Vector of sparsity parameters for Y (default is a sequence from 0 to 1 with step 0.1)
#' @param rank Number of canonical components to extract
#' @param selection.criterion Criterion for selecting the optimal tuning parameter (1 for minimizing difference between test and training sample correlation, 2 for maximizing test sample correlation)
#' @param n.cv Number of cross-validation folds (default is 5)
#' @param A.initial Initial value for the canonical vector A (default is NULL, which uses a canonical ridge solution)
#' @param B.initial Initial value for the canonical vector B (default is NULL, which uses a canonical ridge solution)
#' @param max.iter Maximum number of iterations for convergence (default is 20)
#' @param conv Convergence threshold (default is 1e-2)
#' @param standardize Standardize (center and scale) the data matrices X and Y (default is TRUE) before analysis
#' @return A list with elements:
#' \describe{
#'   \item{U}{Canonical direction matrix for X (p x r)}
#'   \item{V}{Canonical direction matrix for Y (q x r)}
#'   \item{loss}{Mean squared error of prediction}
#'   \item{cor}{Canonical covariances}
#' }
#' @export
SparseCCA <- function(X, Y, lambdaAseq=seq(from=1, to=0.01, by=-0.01),
                    lambdaBseq=seq(from=1, to=0.01, by=-0.01),
                    rank, selection.criterion=1, n.cv=5, A.initial=NULL,
                    B.initial=NULL, max.iter=20, conv=10^-2, standardize=TRUE){
  ### Function to perform Sparse Canonical Correlation Analysis using alternating regressions

  ### INPUT
  # X                   : (nxp) data matrix
  # Y                   : (nxq) data matrix
  # lambdaAseq          : grid of sparsity parameters for lambdaA
  # lambdaBseq          : grid of sparsity parameters for lambdaB
  # rank                : number of canonical vector pairs to extract
  # selection.criterion : 1 for BIC,  2 for Cross-validation to select sparsity parameters
  # n.cv                : n.cv-fold cross-validation
  # A.initial           : starting value for the canonical vector A
  # B.initial           : starting value for the canonical vector B
  # max.iter            : maximum number of iterations
  # conv                : tolerance value for convergence

  ### OUTPUT
  # ALPHA               : (pxr) estimated canonical vectors correpsonding to the first data set
  # BETA                : (qxr) estimated canonical vectors correpsonding to the second data set
  # cancors             : r estimated canonical correlations
  # U_ALL               : (nxr) estimated canonical variates corresponding to the first data set
  # V_ALL               : (nxr) estimated canonical variates corresponding to the second data set
  # lamdbaA             : value of the sparsity parameter lambdaA
  # lamdbaB             : value of the sparsity parameter lambdaB
  # it                  : number of iterations


  ### STORE RESULTS
  ALPHA_ALL <- matrix(NA, ncol=rank, nrow=ncol(X))
  BETA_ALL <- matrix(NA, ncol=rank, nrow=ncol(Y))
  U_ALL <- matrix(NA, ncol=rank, nrow=nrow(X))
  V_ALL <- matrix(NA, ncol=rank, nrow=nrow(Y))
  cancors <- matrix(NA, ncol=rank, nrow=1)
  lambdaA_ALL <- matrix(NA, ncol=rank, nrow=max.iter)
  lambdaB_ALL <- matrix(NA, ncol=rank, nrow=max.iter)
  iterations <- matrix(NA, ncol=rank, nrow=1)
  obj.it <- matrix(NA, ncol=rank, nrow=max.iter+1)



  ### START CODE
  if (standardize){
    X = scale(X, center = TRUE, scale=TRUE)
    Y = scale(Y, center = TRUE, scale=TRUE)
  }

  # Starting Values: Canonical Ridge Solution
  if(is.null(A.initial)){
    cancor_regpar <- estim.regul_crossvalidation(X, Y, n.cv=n.cv,
                                                 lambda1grid = lambdaAseq,
                                                 lambda2grid = lambdaBseq)
    cancor_regul <- CCA::rcc(X, Y, cancor_regpar$lambda1.optim, cancor_regpar$lambda2.optim)
    A.initial_ridge <- matrix(cancor_regul$xcoef[, 1:rank], ncol=rank, nrow=ncol(X))
    B.initial_ridge <- matrix(cancor_regul$ycoef[, 1:rank], ncol=rank, nrow=ncol(Y))
    A.initial <- apply(A.initial_ridge, 2, NORMALIZATION_UNIT)
    B.initial <- apply(B.initial_ridge, 2, NORMALIZATION_UNIT)
  }

  # Sequentially extract the r canonical vector pairs
  for (i.r in 1:rank){

    if (i.r==1){ # for r=1: start from original data sets
      X_data <- X
      Y_data <- Y
    }

    # STARTING VALUES: canonical ridge
    A.STARTING <- matrix(A.initial[, i.r], ncol=1)
    B.STARTING <- matrix(B.initial[, i.r], ncol=1)

    # CONVERGENCE CRITERION
    obj.initial <- mean((X_data%*%matrix(A.STARTING, ncol=1)-Y_data%*%matrix(B.STARTING, ncol=1))^2)
    obj.it[1, i.r] <- obj.initial

    # INITIALIZE CONVERGENCE PARAMETERS
    it <- 1
    diff.obj <- conv*10

    # FROM i until convergence: canonical vectors
    while( (it<max.iter) & (diff.obj>conv) ){

      # Estimating A conditional on B
      FIT.A <- alternating.regression(Xreg=X_data, Yreg=Y_data%*%B.STARTING,
                                    lambdaseq=lambdaAseq,
                                    selection.criterion=selection.criterion,
                                    n.cv=n.cv)
      AHAT_FINAL <- FIT.A$COEF_FINAL
      lambdaA_ALL[it, i.r] <- FIT.A$LAMBDA_FINAL

      # Estimating B conditional on A
      FIT.B <- alternating.regression(Xreg=Y_data,
                                    Yreg=X_data%*%AHAT_FINAL,
                                    lambdaseq=lambdaBseq,
                                    selection.criterion=selection.criterion,
                                    n.cv=n.cv)
      BHAT_FINAL <- FIT.B$COEF_FINAL
      lambdaB_ALL[it, i.r] <- FIT.B$LAMBDA_FINAL

      # Check convergence
      obj.new <- mean((X_data%*%matrix(AHAT_FINAL, ncol=1)-Y_data%*%matrix(BHAT_FINAL, ncol=1))^2)
      obj.it[it+1, i.r] <- obj.new
      diff.obj <- abs(obj.new-obj.initial)/max(obj.initial, 1e-4)

      # Updated starting values
      B.STARTING <- BHAT_FINAL
      A.STARTING <- AHAT_FINAL
      obj.initial <- obj.new
      it <- it+1
    } # end while-loop

    # Number of ITERATIONS
    iterations[1, i.r] <- it

    # CANONICAL VARIATES after convergence
    Uhat <- X_data%*%AHAT_FINAL
    Vhat <- Y_data%*%BHAT_FINAL


    # Express canonical vectors in terms of ORIGINAL DATA MATRICES
    if (i.r==1){# FIRST DIMENSION

      # Final estimates of canonical vectors,  variates and canonical correlation
      ALPHA_ALL[, i.r] <- AHAT_FINAL
      BETA_ALL[, i.r] <- BHAT_FINAL
      U_ALL[, i.r] <- Uhat
      V_ALL[, i.r] <- Vhat
      cancors[1, i.r] <- cor(Uhat, Vhat)

      # Deflated data matrices
      X_data <-  round(X_data  - Uhat%*%solve(t(Uhat)%*%Uhat)%*%t(Uhat)%*%X_data, 10)
      Y_data <-  round(Y_data - Vhat%*%solve(t(Vhat)%*%Vhat)%*%t(Vhat)%*%Y_data, 10)

      # Sparsity parameters
      lambdaA_FINAL <- FIT.A$LAMBDA_FINAL
      lambdaB_FINAL <- FIT.B$LAMBDA_FINAL


    } else {# HIGHER ORDER DIMENSIONS

      # A expressed in terms of original data set X
      FIT.Aorig <- alternating.regression(Yreg=Uhat, Xreg=X, lambdaseq=lambdaAseq,
                                        selection.criterion=selection.criterion,
                                        n.cv=n.cv)
      ALPHAhat <- FIT.Aorig$COEF_FINAL
      lambdaA_FINAL <- FIT.Aorig$LAMBDA_FINAL

      # B expressed in terms of original data set Y
      FIT.Borig <- alternating.regression(Yreg=Vhat, Xreg=Y,
                                        lambdaseq=lambdaBseq,
                                        selection.criterion=selection.criterion,
                                        n.cv=n.cv)
      BETAhat <- FIT.Borig$COEF_FINAL
      lambdaB_FINAL <- FIT.Borig$LAMBDA_FINAL


      # Final estimates of canonical vectors,  variates and canonical correlation
      ALPHA_ALL[, i.r] <- ALPHAhat
      BETA_ALL[, i.r] <- BETAhat
      Uhat <- X%*%ALPHAhat
      Vhat <- Y%*%BETAhat
      U_ALL[, i.r] <- Uhat
      V_ALL[, i.r] <- Vhat
      cancors[1, i.r] <- cor(Uhat, Vhat)

      # Deflated data matrices: regress original data sets on all previously found canonical variates
      X_data <-  X  - U_ALL[, 1:i.r]%*%solve(t(U_ALL[, 1:i.r])%*%U_ALL[, 1:i.r])%*%t(U_ALL[, 1:i.r])%*%X
      Y_data <-  Y -  V_ALL[, 1:i.r]%*%solve(t(V_ALL[, 1:i.r])%*%V_ALL[, 1:i.r])%*%t(V_ALL[, 1:i.r])%*%Y
    }
  } # END FOR-LOOP

  ##OUTPUT
  out <- list(uhat=ALPHA_ALL, vhat=BETA_ALL, cancors=cancors, U_ALL=U_ALL, V_ALL=V_ALL, lambdaA=lambdaA_FINAL, lambdaB=lambdaB_FINAL, it=iterations)

}



alternating.regression <- function(Xreg, Yreg,
                                 lambdaseq=seq(from=1, to=0.01, by=-0.01),
                                 selection.criterion=1, n.cv=5){
  ### Function to perform sparse alternating regression

  ### INPUT
  #Xreg               : design matrix
  #Yreg               : response
  #lambdaseq          : sequence of sparsity parameters
  #selection.criterion: 1 for BIC,  2 for Cross-validation to select sparsity parameter
  #n.cv               : n.cv-fold cross-validation

  ### OUTPUT
  #COEF_FINAL         : estimated coefficients
  #LAMBDA_FINAL       : optimal sparsity parameter


  ##Standardize
  Xreg = scale(Xreg, center = TRUE, scale=TRUE)
  Xreg_st <- matrix(Xreg, ncol=ncol(Xreg))
  for (i.variable in 1:ncol(Xreg)){
    if (is.na(apply(Xreg_st, 2, sum)[i.variable])==T) {
      Xreg_st[, i.variable] <- 0}
  }

  ##LASSO FIT
  LASSOFIT <- glmnet::glmnet(y=Yreg, x=Xreg_st, family="gaussian", lambda=lambdaseq, intercept=T)
  if (is.integer(which(LASSOFIT$df!=0)) && length(which(LASSOFIT$df!=0)) == 0L) {
    # Smaller lambda sequence necessary
    LASSOFIT <- glmnet::glmnet(y=Yreg, x=Xreg_st, family="gaussian", intercept=T)
    COEFhat <- matrix(LASSOFIT$beta[, which(LASSOFIT$df!=0)[1:length(lambdaseq)]], nrow=ncol(Xreg_st)) # estimated coefficients
    LAMBDA <- LASSOFIT$lambda[which(LASSOFIT$df!=0)[1:length(lambdaseq)]] # lambda values
  } else {
    COEFhat <- matrix(LASSOFIT$beta[, which(LASSOFIT$df!=0)], nrow=ncol(Xreg_st)) # estimated coefficients
    LAMBDA <- LASSOFIT$lambda[which(LASSOFIT$df!=0)] # lambda values
  }

  ##Selection of sparsity parameter
  if (selection.criterion==1){ #BIC
    BICvalues <- apply(COEFhat, 2, BIC, Y.data=Yreg, X.data=Xreg_st) # BIC
    COEF_FINAL <- matrix(COEFhat[, which.min(BICvalues)], ncol=1) # Final coefficient estimates
    COEF_FINAL[which(apply(Xreg, 2, sd)!=0), ] <- COEF_FINAL[which(apply(Xreg, 2, sd)!=0), ]/apply(Xreg, 2, sd)[which(apply(Xreg, 2, sd)!=0)]
    COEF_FINAL <- apply(COEF_FINAL, 2, NORMALIZATION_UNIT)
    LAMBDA_FINAL <- LAMBDA[which.min(BICvalues)]

  } else {
    if (selection.criterion==2){ #Cross-validation
      n = nrow(Xreg_st)
      n.cv.sample <- trunc(n/n.cv)
      whole.sample <- seq(1, n)
      X.data <- Xreg_st
      Y.data <- Yreg
      cvscore <- matrix(NA, ncol=length(LAMBDA), nrow=n.cv)
      for (i in 1:n.cv){
        testing.sample <- whole.sample[((i-1)*n.cv.sample+1):(i*n.cv.sample)]
        training.sample <- whole.sample[-(((i-1)*n.cv.sample+1):(i*n.cv.sample))]
        Xcv = X.data[training.sample,  ]
        Ycv = Y.data[training.sample,  ]
        LASSOFIT_cv <- glmnet::glmnet(y=Ycv, x=Xcv, family="gaussian", lambda=LAMBDA)
        COEF_cv <- LASSOFIT_cv$beta
        cvscore[i, ] <- apply(COEF_cv, 2, testsample.correlation, Xdata=X.data[testing.sample,  ], yscore = Y.data[testing.sample,  ] )
      }

      CVscore.mean <- apply(cvscore, 2, mean) # cv score
      COEF_FINAL <- matrix(COEFhat[, which.max(CVscore.mean)], ncol=1) # Final coefficient estimates
      COEF_FINAL[which(apply(Xreg, 2, sd)!=0), ] <- COEF_FINAL[which(apply(Xreg, 2, sd)!=0), ]/apply(Xreg, 2, sd)[which(apply(Xreg, 2, sd)!=0)]
      COEF_FINAL <- apply(COEF_FINAL, 2, NORMALIZATION_UNIT)
      LAMBDA_FINAL <- LAMBDA[which.max(CVscore.mean)]
    } else {
      stop("selection.criterion needs to be equal to 1 (BIC),  2 or 3 (Cross-validation)")
    }
  }

  ##OUTPUT
  out <- list(COEF_FINAL=COEF_FINAL, LAMBDA_FINAL=LAMBDA_FINAL)
}

NORMALIZATION_UNIT <- function(U){
  ### Normalize a vector U to have norm one
  if (is.null(dim(U))) U <- matrix(U, ncol = 1)


  length.U <- as.numeric(sqrt(t(U)%*%U))
  if(length.U==0){length.U <- 1}
  Uunit <- U/length.U
}

BIC <- function(U, Y.data, X.data){
  ### Calculate value of Bayesian information Criterion

  NEGLOGLIK <- sum(diag((1/nrow(Y.data))*(t(Y.data-X.data%*%U)%*%(Y.data-X.data%*%U))))
  BIC <- 2*NEGLOGLIK+ length(which(U!=0))*log(nrow(Y.data))
  return(BIC)
}

testsample.correlation <- function(U, Xdata, yscore){
  ### Calculate correlation in test sample
  if (is.null(dim(U))) U <- matrix(U, ncol = 1)

  xscore = Xdata%*%U
  if (all(U==0)) {
    return(0)
  } else {
    return(abs(cor(xscore, yscore)))
  }
}


############################################################
# CANONICAL RIDGE (used as starting value in SAR algorithm #
############################################################

estim.regul_crossvalidation <- function (X,  Y,  lambda1grid = NULL,
                                         lambda2grid = NULL, n.cv=5){

  
  
  if (is.null(lambda1grid)) {
    lambda1grid <- seq(0.001, 1, length = 5)
  } else {
    lambda1grid <- as.vector(lambda1grid)
  }
  
  if (is.null(lambda2grid)) {
    lambda2grid <- seq(0.001, 1, length = 5)
  } else {
    lambda2grid <- as.vector(lambda2grid)
  }
  
  lambda.grid <- expand.grid(lambda1=lambda1grid, lambda2=lambda2grid)
  
  cvscores <- mapply(function(l1, l2) {
    RCC_crossvalidation(X=X, Y=Y, lambda1=l1, lambda2=l2, n.cv=n.cv)$cv
  }, lambda.grid$lambda1, lambda.grid$lambda2)
  
  best_idx <- which.max(cvscores)
  lambda1.optim <- lambda.grid$lambda1[best_idx]
  lambda2.optim <- lambda.grid$lambda2[best_idx]
  cv.optim <- cvscores[best_idx]
  
  out <- list(lambda1.optim=lambda1.optim, lambda2.optim=lambda2.optim, cv.optim=cv.optim)
  return(out)
}


RCC_crossvalidation <- function (X,  Y,  lambda1,  lambda2, n.cv) {
# AUXILIARY FUNCTION CANONICAL RIDGE: n.cv-fold cross-validation
  ### INPUT
  # X                   : (nxp) data matrix
  # Y                   : (nxq) data matrix
  # lambda1             : tuning parameter for lambda1
  # lambda2             : tuning parameter for lambda2
  # n.cv                : n.cv-fold cross-validation

  ### OUTPUT
  # cv                  : value test sample canonical correlation

  n = nrow(X)
  ncvsample <- trunc(n/n.cv)
  ncvindex <- matrix(rep(1:n.cv), nrow=1)

  CVfit=apply(ncvindex, MARGIN=2, FUN=cvfunction, n=n,
              Xmatrix=X, Ymatrix=Y, lambda1=lambda1,
              lambda2=lambda2, ncvsample=ncvsample)
  cv <- sum(CVfit)/n.cv
  out=list(cv=cv)
}

cvfunction <- function(U, n, Xmatrix, Ymatrix, lambda1, lambda2, ncvsample){
 # AUXILIARY FUNCTION CANONICAL RIDGE: Canonical ridge fit
  cv <- 0
  whole.sample <- seq(1, n)
  testing.sample <- whole.sample[((U-1)*ncvsample+1):(U*ncvsample)]
  training.sample <- whole.sample[-(((U-1)*ncvsample+1):(U*ncvsample))]
  Xcv = Xmatrix[training.sample,  ]
  Ycv = Ymatrix[training.sample,  ]
  res = CCA::rcc(Xcv,  Ycv,  lambda1,  lambda2)

  xscore = Xmatrix[testing.sample,  ] %*% res$xcoef[,  1]
  yscore = Ymatrix[testing.sample,  ] %*% res$ycoef[,  1]
  cv <-  cv + abs(cor(xscore, yscore, use="pairwise"))
  return(cv)
}

#########################################
# AUXILIARLY FUNCTIONS SIMULATION STUDY #
#########################################

#' @title Metrics for subspaces
#' @description Calculate principal angles between subspace spanned by the columns of a and the subspace spanned by the columns of b
#' @param a A matrix whose columns span a subspace.
#' @param b A matrix whose columns span a subspace.
#' @return  a vector of principal angles (in radians)
#' @examples
#' a <- matrix(rnorm(9), 3, 3)
#' b <- matrix(rnorm(9), 3, 3)
#' principal_angles(a, b)
#' @export
principal_angles <- function(a, b){
  ### Calculate principal angles between subspace spanned by the columns of a and the subspace spanned by the columns of b

  angles=matrix(0, ncol=ncol(a), nrow=1)
  qa= qr.Q(qr(a))
  qb= qr.Q(qr(b))
  C=svd(t(qa)%*%qb)$d
  rkA=qr(a)$rank;rkB=qr(b)$rank
  if (rkA<=rkB){
    B = qb - qa%*%(t(qa)%*%qb);
  } else {B = qa - qb%*%(t(qb)%*%qa)}
  S=svd(B)$d
  S=sort(S)

  for (i in 1:min(rkA, rkB)){
    if (C[i]^2 < 0.5) {angles[1, i]=acos(C[i])}
    else if (S[i]^2 <=0.5) {angles[1, i]=asin(S[i])}
  }
  angles=t(angles)

  ##OUTPPUT
  out=list(angles=angles)
  return(out)
}




#' @title True Positive Rate (TPR)
#' 
#' @description This is a function that compares the structure of two matrices A and B.  It outputs the number of entries that A and B have in common that are different from zero. A and B need to have the same number of rows and columns
#' @param A A matrix (the estimator).
#' @param B A matrix (assumed to be the ground truth).
#' @param tol tolerance for declaring the entries non zero.
#' @return True Positive Rate (nb of values that are non zero in both A and B / (nb of values that are non zero in A))
#' @examples
#' A <- matrix(c(1, 0, 0, 1, 1, 0), nrow = 2)
#' B <- matrix(c(1, 0, 1, 1, 0, 0), nrow = 2)
#' TPR(A, B)  
#' @export
TPR  <-  function(A, B, tol=1e-4){

  A[which(abs(A) <tol)] =0
  B[which(abs(B) <tol)] =0
  out  <-  sum((A!=0)*(B!=0))/max(1,sum(A!=0))
  return(out)
}



#' @title False Positive Rate (TPR)
#' @description This is a function that compares the structure of two matrices A and B. It outputs the number of entries where A  is not zero but Bis. A and B need to have the same number of rows and columns
#' @param A A matrix.
#' @param B A matrix (assumed to be the ground truth).
#' @param tol tolerance for declaring the entries non zero.
#' @return False Positive Rate (nb of values that are non zero in A and  zero in B / (nb of values that are non zero in A))
#' @examples
#' A <- matrix(c(1, 0, 0, 1, 1, 0), nrow = 2)
#' B <- matrix(c(1, 0, 1, 1, 0, 0), nrow = 2)
#' FPR(A, B)  
#' @export
FPR  <-  function(A, B, tol=1e-4){
  A[which(abs(A) <tol)] =0
  B[which(abs(B) <tol)] =0
  out  <-  sum((A!=0)*(B==0))/max(1,sum(A!=0))
  return(out)
}



#' True Negative Rate (TNR)
#' 
#' @description This is a function that compares the structure of two matrices A and B. It outputs the number of entries where A and B are both 0. A and B need to have the same number of rows and columns
#' @param A A matrix.
#' @param B A matrix (assumed to be the ground truth)..
#' @param tol tolerance for declaring the entries non zero.
#' @return True Negative Rate (nb of values that are zero in A and  zero in B / (nb of values that are  zero in A))
#' @export
TNR  <-  function(A, B, tol=1e-4){
  # This is a function that compares the structure of two matrices A and B
  # It outputs the number of entries that A and B have in common that are zero #
  # A and B need to have the same number of rows and columns
  A[which(abs(A) <tol)] =0
  B[which(abs(B) <tol)] =0
  out  <-  sum((A==0)*(B==0))/max(1,sum(A==0))
  return(out)
}



#' @title Subdistance between subspaces
#' @description Calculate subdistance between subspace spanned by the columns of a and the subspace spanned by the columns of b
#' @param A A matrix whose columns span a subspace.
#' @param B A matrix whose columns span a subspace.
#' @return subdistance between the two subspaces spanned by the matrices A and B, defined as min(O orthogonal) ||AO-B||_F
#' @export
subdistance <- function(A, B){
  svdresult = svd(t(A) %*% B);
  U = svdresult$u
  V = svdresult$v
  O = U %*% t(V);
  l = norm(A %*% O-B, type = c('F'));
  return(l)
}



#' @title SinTheta distance between subspaces
#' @description Calculate the distance spanned by the columns of A and the subspace spanned by the columns of B, defined as ||UU^T - VV^T||_F / sqrt(2)
#' @param U A matrix whose columns span a subspace.
#' @param V A matrix whose columns span a subspace.
#' @return sinTheta distance between the two subspaces spanned by the matrices A and B, defined as ||UU^T - VV^T||_F / sqrt(2)
#' @export
sinTheta<- function(U, V){
  l = 1/sqrt(2) * norm(U %*% t(U)-V %*% t(V), type = c('F'));
  return(l)
}

library(ccar3)

test_that("cca_rrr returns correct output structure", {
  X <- matrix(rnorm(100*10), 100, 10)
  Y <- matrix(rnorm(100*5), 100, 5)
  result <- cca_rrr(X, Y, r = 2)
  
  expect_type(result, "list")
  expect_true(all(c("U", "V", "loss", "cor") %in% names(result)))
  expect_equal(dim(result$U)[2], 2)
  expect_equal(dim(result$V)[2], 2)
  expect_equal(dim(result$U)[1], 10)
  expect_equal(dim(result$V)[1], 5)
})


test_that("cca_rrr returns the correct answer", {

  ##### Generate toy example data
  set.seed(123)
  r = 3
  gen = generate_example_sparse_U(n=10000, p1=100, p2=10,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y

  result <- cca_rrr(X, Y, r = r, Sx=NULL,  Sy=NULL, lambda=0, highdim=TRUE,
                    LW_Sy = TRUE, solver = "ADMM")
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.15)  
  expect_true(subdistance(result$V, gen$v) < 0.15)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
})





test_that("cca_rrr computes the same solutions across solvers", {
  set.seed(123)
  skip_if_not_installed("CVXR")


  r = 3
  gen = generate_example_sparse_U(n=10000, p1=100, p2=10,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y  

  result2 <- cca_rrr(X, Y, r = r, Sx=NULL,  Sy=NULL, lambda=0.1, rho=10,
                     highdim=TRUE,
                     LW_Sy = TRUE, solver="ADMM")
  print("Done ADMM")
  result3 <- cca_rrr(X, Y, r = r, Sx=NULL,  Sy=NULL, lambda=0.1,
                    highdim=TRUE,
                    LW_Sy = TRUE, solver="CVX")
  print("Done CVXR")
  
  expect_true(subdistance(result2$U, gen$u) <0.2 )
  expect_true(subdistance(result3$U, gen$u) <0.2 )
})


test_that("cca_rrr exposes matrix-free ADMM tuning arguments", {
  set.seed(42)
  X <- matrix(rnorm(80 * 12), 80, 12)
  Y <- matrix(rnorm(80 * 4), 80, 4)

  result <- cca_rrr(
    X, Y, r = 2,
    highdim = TRUE,
    solver = "ADMM",
    matrix_free_threshold = 1L,
    cg_tol = 1e-5,
    cg_maxiter = 50
  )

  expect_type(result, "list")
  expect_equal(dim(result$U), c(12, 2))
  expect_equal(dim(result$V), c(4, 2))
})

test_that("cca_rrr can skip internal preprocessing", {
  set.seed(21)
  X <- scale(matrix(rnorm(90 * 10), 90, 10), center = TRUE, scale = FALSE)
  Y <- scale(matrix(rnorm(90 * 4), 90, 4), center = TRUE, scale = FALSE)

  result <- cca_rrr(
    X, Y, r = 2,
    highdim = TRUE,
    solver = "ADMM",
    standardize = FALSE,
    preprocess = FALSE
  )

  expect_type(result, "list")
  expect_equal(dim(result$U), c(10, 2))
  expect_equal(dim(result$V), c(4, 2))
})

test_that("cca_rrr supports preprocess modes and logical backward compatibility", {
  set.seed(33)
  X <- matrix(rnorm(120 * 8), 120, 8)
  Y <- matrix(rnorm(120 * 4), 120, 4)

  fit_center_mode <- cca_rrr(
    X, Y, r = 2,
    highdim = TRUE,
    solver = "ADMM",
    preprocess = "center",
    LW_Sy = FALSE,
    niter = 200
  )
  fit_center_logical <- cca_rrr(
    X, Y, r = 2,
    highdim = TRUE,
    solver = "ADMM",
    standardize = FALSE,
    preprocess = TRUE,
    LW_Sy = FALSE,
    niter = 200
  )

  X_centered <- scale(X, center = TRUE, scale = FALSE)
  Y_centered <- scale(Y, center = TRUE, scale = FALSE)
  fit_none_char <- cca_rrr(
    X_centered, Y_centered, r = 2,
    highdim = TRUE,
    solver = "ADMM",
    preprocess = "none",
    LW_Sy = FALSE,
    niter = 200
  )
  fit_none_logical <- cca_rrr(
    X_centered, Y_centered, r = 2,
    highdim = TRUE,
    solver = "ADMM",
    standardize = FALSE,
    preprocess = FALSE,
    LW_Sy = FALSE,
    niter = 200
  )

  expect_equal(fit_center_mode$B_opt, fit_center_logical$B_opt, tolerance = 1e-8)
  expect_equal(fit_none_char$B_opt, fit_none_logical$B_opt, tolerance = 1e-8)
})

library(ccar3)

test_that("cca_rrr returns correct output structure", {
  X <- matrix(rnorm(100*10), 100, 10)
  Y <- matrix(rnorm(100*5), 100, 5)
  result <- cca_rrr(X, Y, r = 2)
  
  expect_type(result, "list")
  expect_true(all(c("U", "V", "loss", "cor") %in% names(result)))
  expect_equal(dim(result$U)[2], 2)
  expect_equal(dim(result$V)[2], 2)
  expect_equal(dim(result$U)[1], 10)
  expect_equal(dim(result$V)[1], 5)
})


test_that("cca_rrr returns the correct answer", {

  ##### Generate toy example data
  set.seed(123)
  r = 3
  gen = generate_example_sparse_U(n=10000, p1=100, p2=10,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y

  result <- cca_rrr(X, Y, r = r, Sx=NULL,  Sy=NULL, lambda=0, highdim=TRUE,
                    LW_Sy = TRUE, solver = "ADMM")
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.15)  
  expect_true(subdistance(result$V, gen$v) < 0.15)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
})





test_that("cca_rrr computes the same solutions across solvers", {
  set.seed(123)
  skip_if_no_cvxr_conic_solver()


  r = 3
  gen = generate_example_sparse_U(n=10000, p1=100, p2=10,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y  

  result2 <- cca_rrr(X, Y, r = r, Sx=NULL,  Sy=NULL, lambda=0.1, rho=10,
                     highdim=TRUE,
                     LW_Sy = TRUE, solver="ADMM")
  print("Done ADMM")
  result3 <- cca_rrr(X, Y, r = r, Sx=NULL,  Sy=NULL, lambda=0.1,
                    highdim=TRUE,
                    LW_Sy = TRUE, solver="CVX")
  print("Done CVXR")
  
  expect_true(subdistance(result2$U, gen$u) <0.2 )
  expect_true(subdistance(result3$U, gen$u) <0.2 )
})


test_that("cca_rrr exposes matrix-free ADMM tuning arguments", {
  set.seed(42)
  X <- matrix(rnorm(80 * 12), 80, 12)
  Y <- matrix(rnorm(80 * 4), 80, 4)

  result <- cca_rrr(
    X, Y, r = 2,
    highdim = TRUE,
    solver = "ADMM",
    matrix_free_threshold = 1L,
    cg_tol = 1e-5,
    cg_maxiter = 50
  )

  expect_type(result, "list")
  expect_equal(dim(result$U), c(12, 2))
  expect_equal(dim(result$V), c(4, 2))
})

test_that("cca_rrr can skip internal preprocessing", {
  set.seed(21)
  X <- scale(matrix(rnorm(90 * 10), 90, 10), center = TRUE, scale = FALSE)
  Y <- scale(matrix(rnorm(90 * 4), 90, 4), center = TRUE, scale = FALSE)

  result <- cca_rrr(
    X, Y, r = 2,
    highdim = TRUE,
    solver = "ADMM",
    standardize = FALSE,
    preprocess = FALSE
  )

  expect_type(result, "list")
  expect_equal(dim(result$U), c(10, 2))
  expect_equal(dim(result$V), c(4, 2))
})

test_that("cca_rrr supports preprocess modes and logical backward compatibility", {
  set.seed(33)
  X <- matrix(rnorm(120 * 8), 120, 8)
  Y <- matrix(rnorm(120 * 4), 120, 4)

  fit_center_mode <- cca_rrr(
    X, Y, r = 2,
    highdim = TRUE,
    solver = "ADMM",
    preprocess = "center",
    LW_Sy = FALSE,
    niter = 200
  )
  fit_center_logical <- cca_rrr(
    X, Y, r = 2,
    highdim = TRUE,
    solver = "ADMM",
    standardize = FALSE,
    preprocess = TRUE,
    LW_Sy = FALSE,
    niter = 200
  )

  X_centered <- scale(X, center = TRUE, scale = FALSE)
  Y_centered <- scale(Y, center = TRUE, scale = FALSE)
  fit_none_char <- cca_rrr(
    X_centered, Y_centered, r = 2,
    highdim = TRUE,
    solver = "ADMM",
    preprocess = "none",
    LW_Sy = FALSE,
    niter = 200
  )
  fit_none_logical <- cca_rrr(
    X_centered, Y_centered, r = 2,
    highdim = TRUE,
    solver = "ADMM",
    standardize = FALSE,
    preprocess = FALSE,
    LW_Sy = FALSE,
    niter = 200
  )

  expect_equal(fit_center_mode$B_opt, fit_center_logical$B_opt, tolerance = 1e-8)
  expect_equal(fit_none_char$B_opt, fit_none_logical$B_opt, tolerance = 1e-8)
})

library(ccar3)

test_that("cca_graph returns correct output structure", {
  skip_if_not_installed("igraph")

  set.seed(123)
  r = 2
  gen = generate_example_graph(n=500, p1=10, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  Gamma = gen$Gamma
  result <- cca_graph_rrr(X, Y, Gamma = Gamma, r = r)
  
  expect_type(result, "list")
  expect_true(all(c("U", "V", "loss", "cor") %in% names(result)))
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], dim(X)[2])
  expect_equal(dim(result$V)[1], dim(Y)[2])
})


test_that("cca_graph returns the correct answer", {
  skip_if_not_installed("igraph")


  ##### Generate toy example data
  set.seed(123)
  r = 2
  gen = generate_example_graph(n=500, p1=10, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  Gamma = gen$Gamma
  result <- cca_graph_rrr(X, Y, Gamma = Gamma, lambda = 0.025, r = r)
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.5)  
  expect_true(subdistance(result$V, gen$v) < 0.5)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  
})

test_that("cca_graph_rrr accepts sparse Gamma and preprocessing modes", {
  skip_if_not_installed("igraph")
  skip_if_not_installed("Matrix")

  set.seed(123)
  r <- 2
  gen <- generate_example_graph(
    n = 300, p1 = 10, p2 = 5,
    r_pca = 3,
    nnzeros = 5,
    theta = diag(c(0.9, 0.8)),
    lambda_pca = 1,
    r = r, overlapping_amount = 1,
    normalize_diagonal = TRUE,
    n_new = 1000
  )

  Gamma_sparse <- Matrix::Matrix(gen$Gamma, sparse = TRUE)
  res_none <- cca_graph_rrr(gen$X, gen$Y, Gamma = Gamma_sparse, lambda = 0.02, r = r, preprocess = "none")
  res_center <- cca_graph_rrr(gen$X, gen$Y, Gamma = Gamma_sparse, lambda = 0.02, r = r, preprocess = "center")
  res_scale <- cca_graph_rrr(gen$X, gen$Y, Gamma = Gamma_sparse, lambda = 0.02, r = r, preprocess = "scale")

  expect_equal(dim(res_none$U), c(ncol(gen$X), r))
  expect_equal(dim(res_center$U), c(ncol(gen$X), r))
  expect_equal(dim(res_scale$U), c(ncol(gen$X), r))
})

library(ccar3)

test_that("cca_group returns correct output structure", {
  set.seed(123)
  r = 2
  gen = generate_example_group(n=10000, p1=100, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  groups = gen$groups

  result <- cca_group_rrr(X, Y, lambda=0, groups=groups, r = r)
  expect_type(result, "list")
  expect_true(all(c("U", "V", "loss", "cor") %in% names(result)))
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], 100)
  expect_equal(dim(result$V)[1], 5)
})


test_that("cca_group returns the correct answer", {

  ##### Generate toy example data
  set.seed(123)
  r = 3
  gen = generate_example_group(n=10000, p1=100, p2=5, 
                                  r_pca = 3,
                                  nnzeros = 5,
                                  theta = diag(c(0.9,  0.85,  0.8)),
                                 lambda_pca = 1,
                                  r = r, overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                 n_new = 5000)
  X = gen$X
  Y = gen$Y
  result <- cca_group_rrr(X, Y, r = r, group=gen$groups, Sx=NULL,  Sy=NULL, lambda=0.01,
                    LW_Sy = TRUE)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.15)  
  expect_true(subdistance(result$V, gen$v) < 0.15)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  

})





test_that("cca_group computes the same solutions across solvers", {
  skip_if_no_cvxr_conic_solver()
  set.seed(123)
  r = 2
  gen = generate_example_group(n=500, p1=100, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.85,  0.8)[1:r]),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  X = gen$X
  Y = gen$Y                          
  result1 <- cca_group_rrr(X, Y, r = r, groups=gen$groups, Sx=NULL,  Sy=NULL, 
                          lambda=0.1, LW_Sy = TRUE, solver="ADMM")
  result2 <- cca_group_rrr(X, Y, r = r, 
                          groups=gen$groups, 
                          Sx=NULL,  Sy=NULL, lambda=0.1,
                          LW_Sy = TRUE, solver="CVXR")
  
  expect_true(subdistance(result1$U, gen$u) <0.35)
  expect_true(subdistance(result2$U, gen$u) <0.35)
  
  expect_true(subdistance(result1$V, gen$v) <0.35)
  expect_true(subdistance(result2$V, gen$v) <0.35)
})

test_that("cca_group_rrr supports preprocess modes and p > n", {
  set.seed(42)
  n <- 40
  p <- 80
  q <- 5
  r <- 2
  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n * q), n, q)
  groups <- split(seq_len(p), ceiling(seq_len(p) / 5))

  fit_none <- cca_group_rrr(X, Y, groups = groups, r = r, lambda = 0.01,
                            preprocess = "none", LW_Sy = FALSE, niter = 300)
  fit_center <- cca_group_rrr(X, Y, groups = groups, r = r, lambda = 0.01,
                              preprocess = "center", LW_Sy = FALSE, niter = 300)
  fit_scale <- cca_group_rrr(X, Y, groups = groups, r = r, lambda = 0.01,
                             preprocess = "scale", LW_Sy = FALSE, niter = 300)

  expect_equal(dim(fit_none$U), c(p, r))
  expect_equal(dim(fit_center$U), c(p, r))
  expect_equal(dim(fit_scale$U), c(p, r))
})

test_that("cca_group_rrr supports matrix-free ADMM solves", {
  set.seed(7)
  n <- 50
  p <- 70
  q <- 4
  r <- 2
  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n * q), n, q)
  groups <- split(seq_len(p), ceiling(seq_len(p) / 5))

  fit <- cca_group_rrr(
    X, Y, groups = groups, r = r, lambda = 0.01,
    preprocess = "center", LW_Sy = FALSE, niter = 200,
    matrix_free_threshold = 1L,
    cg_tol = 1e-5,
    cg_maxiter = 50
  )

  expect_equal(dim(fit$U), c(p, r))
  expect_equal(dim(fit$V), c(q, r))
  expect_true(all(is.finite(fit$cor)))
})

library(ccar3)

test_that("cca_group returns correct output structure", {
  set.seed(123)
  r = 2
  gen = generate_example_group(n=10000, p1=100, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  groups = gen$groups

  result <- cca_group_rrr(X, Y, lambda=0, groups=groups, r = r)
  expect_type(result, "list")
  expect_true(all(c("U", "V", "loss", "cor") %in% names(result)))
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], 100)
  expect_equal(dim(result$V)[1], 5)
})


test_that("cca_group returns the correct answer", {

  ##### Generate toy example data
  set.seed(123)
  r = 3
  gen = generate_example_group(n=10000, p1=100, p2=5, 
                                  r_pca = 3,
                                  nnzeros = 5,
                                  theta = diag(c(0.9,  0.85,  0.8)),
                                 lambda_pca = 1,
                                  r = r, overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                 n_new = 5000)
  X = gen$X
  Y = gen$Y
  result <- cca_group_rrr(X, Y, r = r, group=gen$groups, Sx=NULL,  Sy=NULL, lambda=0.01,
                    LW_Sy = TRUE)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.15)  
  expect_true(subdistance(result$V, gen$v) < 0.15)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  

})





test_that("cca_group computes the same solutions across solvers", {
  skip_if_not_installed("CVXR")
  set.seed(123)
  r = 2
  gen = generate_example_group(n=500, p1=100, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.85,  0.8)[1:r]),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  X = gen$X
  Y = gen$Y                          
  result1 <- cca_group_rrr(X, Y, r = r, groups=gen$groups, Sx=NULL,  Sy=NULL, 
                          lambda=0.1, LW_Sy = TRUE, solver="ADMM")
  result2 <- cca_group_rrr(X, Y, r = r, 
                          groups=gen$groups, 
                          Sx=NULL,  Sy=NULL, lambda=0.1,
                          LW_Sy = TRUE, solver="CVXR")
  
  expect_true(subdistance(result1$U, gen$u) <0.35)
  expect_true(subdistance(result2$U, gen$u) <0.35)
  
  expect_true(subdistance(result1$V, gen$v) <0.35)
  expect_true(subdistance(result2$V, gen$v) <0.35)
})

test_that("cca_group_rrr supports preprocess modes and p > n", {
  set.seed(42)
  n <- 40
  p <- 80
  q <- 5
  r <- 2
  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n * q), n, q)
  groups <- split(seq_len(p), ceiling(seq_len(p) / 5))

  fit_none <- cca_group_rrr(X, Y, groups = groups, r = r, lambda = 0.01,
                            preprocess = "none", LW_Sy = FALSE, niter = 300)
  fit_center <- cca_group_rrr(X, Y, groups = groups, r = r, lambda = 0.01,
                              preprocess = "center", LW_Sy = FALSE, niter = 300)
  fit_scale <- cca_group_rrr(X, Y, groups = groups, r = r, lambda = 0.01,
                             preprocess = "scale", LW_Sy = FALSE, niter = 300)

  expect_equal(dim(fit_none$U), c(p, r))
  expect_equal(dim(fit_center$U), c(p, r))
  expect_equal(dim(fit_scale$U), c(p, r))
})

test_that("cca_group_rrr supports matrix-free ADMM solves", {
  set.seed(7)
  n <- 50
  p <- 70
  q <- 4
  r <- 2
  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n * q), n, q)
  groups <- split(seq_len(p), ceiling(seq_len(p) / 5))

  fit <- cca_group_rrr(
    X, Y, groups = groups, r = r, lambda = 0.01,
    preprocess = "center", LW_Sy = FALSE, niter = 200,
    matrix_free_threshold = 1L,
    cg_tol = 1e-5,
    cg_maxiter = 50
  )

  expect_equal(dim(fit$U), c(p, r))
  expect_equal(dim(fit$V), c(q, r))
  expect_true(all(is.finite(fit$cor)))
})

gca_to_cca <-
  function(a_estimate, S, pp){
    p1 = pp[1];
    p2 = pp[2];
    p = p1 + p2;
    nnz_indices = which(apply(a_estimate, 1, function(x) sqrt(sum(x^2))) >0)
    nnz_indices_x = nnz_indices[which(nnz_indices<(p1+1))]
    nnz_indices_y = nnz_indices[which(nnz_indices>(p1))]
    ### Make sure things are normalized
    if (length(which(nnz_indices<(p1+1)))>0){
       sigmaxhat = S[nnz_indices_x,nnz_indices_x];
       a_estimate[nnz_indices_x,] = a_estimate[nnz_indices_x,] %*% pracma::sqrtm(t(a_estimate[nnz_indices_x,]) %*% sigmaxhat %*% a_estimate[nnz_indices_x,])$Binv;
     }
     if (length(nnz_indices_y)>0){
       sigmayhat = S[nnz_indices_y,nnz_indices_y];
       a_estimate[nnz_indices_y,] = a_estimate[nnz_indices_y,] %*% pracma::sqrtm(t(a_estimate[nnz_indices_y,]) %*% sigmayhat %*% a_estimate[nnz_indices_y,])$Binv;
     }
    
    u_estimate = a_estimate[1:p1,]
    v_estimate = a_estimate[(p1+1):p,] 
    l = list("U" = u_estimate, "V" = v_estimate)
    return(l)
  }


# Set up a parallel backend with graceful fallbacks.
#
# Attempts to create a parallel cluster, first trying the efficient FORK
# method (on Unix-like systems), then falling back to PSOCK, and finally
# returning NULL if all attempts fail.

cleanup_parallel_backend <- function(cl = NULL, verbose = FALSE) {
  if (!is.null(cl)) {
    try(parallel::stopCluster(cl), silent = !verbose)
  }

  if (requireNamespace("foreach", quietly = TRUE)) {
    foreach::registerDoSEQ()
  }

  if (requireNamespace("doParallel", quietly = TRUE) &&
      exists("stopImplicitCluster", envir = asNamespace("doParallel"), inherits = FALSE)) {
    try(doParallel::stopImplicitCluster(), silent = !verbose)
  }

  invisible(NULL)
}

.safe_mean_na <- function(x) {
  out <- mean(x, na.rm = TRUE)
  if (is.nan(out)) NA_real_ else out
}

.safe_sd_na <- function(x) {
  out <- stats::sd(x, na.rm = TRUE)
  if (is.nan(out)) NA_real_ else out
}

.create_cv_folds <- function(n, k) {
  n <- as.integer(n)
  k <- as.integer(k)

  if (length(n) != 1L || is.na(n) || n < 2L) {
    stop("`n` must be a single integer >= 2.", call. = FALSE)
  }
  if (length(k) != 1L || is.na(k) || k < 2L) {
    stop("`k` must be a single integer >= 2.", call. = FALSE)
  }

  k <- min(k, n)
  fold_ids <- sample(rep(seq_len(k), length.out = n))
  unname(split(seq_len(n), fold_ids))
}

.find_local_global_names <- function(fun) {
  if (!requireNamespace("codetools", quietly = TRUE)) {
    return(character())
  }

  env <- environment(fun)
  if (is.null(env)) {
    return(character())
  }

  globals <- tryCatch(codetools::findGlobals(fun, merge = FALSE), error = function(e) NULL)
  if (is.null(globals)) {
    return(character())
  }

  names <- unique(c(globals$variables, globals$functions))
  names <- names[nzchar(names) & !grepl("^[[:punct:]]+$", names)]

  binding_env <- function(name, start) {
    cur <- start
    while (!identical(cur, emptyenv())) {
      if (exists(name, envir = cur, inherits = FALSE)) {
        return(cur)
      }
      cur <- parent.env(cur)
    }
    NULL
  }

  keep <- vapply(names, function(name) {
    where <- binding_env(name, env)
    if (is.null(where) || identical(where, baseenv())) {
      return(FALSE)
    }

    if ("ccar3" %in% loadedNamespaces()) {
      ccar3_ns <- asNamespace("ccar3")
      if (exists(name, envir = ccar3_ns, inherits = FALSE)) {
        return(FALSE)
      }
    }

    env_name <- environmentName(where)
    identical(where, globalenv()) || identical(env_name, "")
  }, logical(1))

  names[keep]
}

initialize_parallel_workers <- function(cl,
                                        pkg = "ccar3",
                                        pkg_path = NULL,
                                        verbose = FALSE) {
  if (is.null(cl)) {
    return(invisible(NULL))
  }

  parallel::clusterEvalQ(cl, {
    Sys.setenv(
      KMP_SHM_DISABLE = 1,
      KMP_SHM_DIR = "/tmp",
      TMPDIR = "/tmp",
      OMP_NUM_THREADS = 1,
      OPENBLAS_NUM_THREADS = 1,
      MKL_NUM_THREADS = 1,
      VECLIB_MAXIMUM_THREADS = 1
    )
    NULL
  })

  if (is.null(pkg_path)) {
    option_path <- getOption("ccar3_pkg_path", default = NULL)
    if (!is.null(option_path) &&
        nzchar(option_path) &&
        file.exists(file.path(option_path, "DESCRIPTION")) &&
        file.exists(file.path(option_path, "NAMESPACE"))) {
      pkg_path <- option_path
    }
  }

  if (is.null(pkg_path)) {
    env_path <- Sys.getenv("CCAR3_PKG_PATH", unset = "")
    if (nzchar(env_path) &&
        file.exists(file.path(env_path, "DESCRIPTION")) &&
        file.exists(file.path(env_path, "NAMESPACE"))) {
      pkg_path <- env_path
    }
  }

  if (is.null(pkg_path)) {
    ns_path <- tryCatch({
      if (requireNamespace("pkgload", quietly = TRUE) &&
          pkg %in% loadedNamespaces() &&
          pkgload::is_dev_package(pkg)) {
        getNamespaceInfo(asNamespace(pkg), "path")
      } else {
        ""
      }
    }, error = function(e) "")

    if (nzchar(ns_path) &&
        file.exists(file.path(ns_path, "DESCRIPTION")) &&
        file.exists(file.path(ns_path, "NAMESPACE"))) {
      pkg_path <- ns_path
    }
  }

  if (is.null(pkg_path)) {
    wd <- tryCatch(normalizePath(getwd(), mustWork = FALSE), error = function(e) "")
    if (nzchar(wd) &&
        file.exists(file.path(wd, "DESCRIPTION")) &&
        file.exists(file.path(wd, "NAMESPACE"))) {
      pkg_path <- wd
    }
  }

  if (!is.null(pkg_path) && nzchar(pkg_path)) {
    if (requireNamespace("pkgload", quietly = TRUE)) {
      if (verbose) {
        cat("Initializing workers from source package at", pkg_path, "\n")
      }
      parallel::clusterCall(cl, function(path) {
        pkgload::load_all(
          path,
          quiet = TRUE,
          export_all = FALSE,
          helpers = FALSE,
          attach_testthat = FALSE
        )
        NULL
      }, pkg_path)
    } else {
      if (verbose) {
        cat("Initializing workers by sourcing R files from", pkg_path, "\n")
      }
      parallel::clusterCall(cl, function(path) {
        r_dir <- file.path(path, "R")
        r_files <- c(
          file.path(r_dir, "helpers.r"),
          file.path(r_dir, "utils.R"),
          sort(list.files(r_dir, pattern = "\\.[Rr]$", full.names = TRUE))
        )
        r_files <- unique(r_files[file.exists(r_files)])
        for (f in r_files) {
          source(f, local = .GlobalEnv)
        }
        NULL
      }, pkg_path)
    }
  } else {
    if (verbose) {
      cat("Initializing workers with installed package", pkg, "\n")
    }
    parallel::clusterCall(cl, function(pkg_name) {
      suppressPackageStartupMessages(library(pkg_name, character.only = TRUE))
      NULL
    }, pkg)
  }

  invisible(NULL)
}

setup_parallel_backend <- function(num_cores = NULL, verbose = FALSE) {

  # 1. Determine the number of cores to use
  if (is.null(num_cores)) {
    # For SLURM, read the allocated CPUs. Fallback for local use.
    n_cores_str <- Sys.getenv("SLURM_CPUS_PER_TASK")
    num_cores <- if (n_cores_str == "") (parallel::detectCores() - 1) else as.integer(n_cores_str)
  }
  # Ensure we have at least 1 core
  num_cores <- max(1, num_cores)
  if (verbose){
     cat(paste("\nAttempting to set up parallel backend with", num_cores, "cores.\n"))
  }
  cl <- NULL
  sysname <- Sys.info()[["sysname"]]
  prefer_fork <- .Platform$OS.type == "unix" && !identical(sysname, "Darwin")

  if (prefer_fork) {
    if (verbose) {
      cat("Unix-like system detected. Trying FORK backend first...\n")
    }
    cl <- tryCatch(
      parallel::makeCluster(num_cores, type = "FORK"),
      error = function(e_fork) {
        if (requireNamespace("crayon", quietly = TRUE)) {
          cat(crayon::yellow("Initial FORK backend setup failed: ", e_fork$message, "\n"))
        } else {
          cat("Initial FORK backend setup failed:", e_fork$message, "\n")
        }
        NULL
      }
    )
  } else if (verbose) {
    if (identical(sysname, "Darwin")) {
      cat("macOS detected. Using PSOCK backend to avoid orphaned FORK workers on interrupts...\n")
    } else {
      cat("Using PSOCK backend...\n")
    }
  }

  if (is.null(cl)) {
    if (verbose && prefer_fork) {
      cat("Attempting fallback PSOCK backend...\n")
    }
    cl <- tryCatch(
      parallel::makeCluster(num_cores, type = "PSOCK"),
      error = function(e_psock) {
        if (requireNamespace("crayon", quietly = TRUE)) {
          cat(crayon::red("PSOCK setup also failed: ", e_psock$message, "\n"))
        } else {
          cat("PSOCK setup also failed:", e_psock$message, "\n")
        }
        NULL
      }
    )
  }

  return(cl)
}

run_lambda_jobs <- function(lambdas,
                            run_one,
                            parallelize = TRUE,
                            nb_cores = NULL,
                            packages = character(),
                            exports = NULL,
                            verbose = FALSE) {
  if (!parallelize) {
    return(lapply(lambdas, run_one))
  }

  if (!requireNamespace("doParallel", quietly = TRUE)) {
    stop("Package 'doParallel' must be installed to use the parallelization option.",
         call. = FALSE)
  }

  if (!requireNamespace("crayon", quietly = TRUE)) {
    stop("Package 'crayon' must be installed to use the parallelization option.",
         call. = FALSE)
  }

  cl <- setup_parallel_backend(nb_cores, verbose = verbose)
  if (is.null(cl)) {
    warning("Parallel setup failed; running CV serially.", immediate. = TRUE)
    return(lapply(lambdas, run_one))
  }

  initialize_parallel_workers(cl, verbose = verbose)
  local_env <- environment(run_one)
  local_names <- if (is.null(local_env)) character() else ls(envir = local_env, all.names = TRUE)
  local_names <- setdiff(local_names, "run_one")
  if (length(local_names) > 0) {
    parallel::clusterExport(cl, varlist = local_names, envir = local_env)
  }

  global_names <- unique(c(exports, .find_local_global_names(run_one)))
  global_names <- setdiff(global_names, c("run_one", local_names))
  global_names <- global_names[vapply(global_names, exists, logical(1), envir = globalenv(), inherits = FALSE)]
  if (length(global_names) > 0) {
    parallel::clusterExport(cl, varlist = global_names, envir = globalenv())
  }

  parallel::clusterExport(cl, varlist = "run_one", envir = parent.frame())
  doParallel::registerDoParallel(cl)
  on.exit(cleanup_parallel_backend(cl), add = TRUE)

  foreach::foreach(
    lambda = lambdas,
    .packages = unique(c(packages, "foreach"))
  ) %dopar% run_one(lambda)
}

gca_to_cca <-
  function(a_estimate, S, pp){
    p1 = pp[1];
    p2 = pp[2];
    p = p1 + p2;
    nnz_indices = which(apply(a_estimate, 1, function(x) sqrt(sum(x^2))) >0)
    nnz_indices_x = nnz_indices[which(nnz_indices<(p1+1))]
    nnz_indices_y = nnz_indices[which(nnz_indices>(p1))]
    ### Make sure things are normalized
    if (length(which(nnz_indices<(p1+1)))>0){
       sigmaxhat = S[nnz_indices_x,nnz_indices_x];
       a_estimate[nnz_indices_x,] = a_estimate[nnz_indices_x,] %*% pracma::sqrtm(t(a_estimate[nnz_indices_x,]) %*% sigmaxhat %*% a_estimate[nnz_indices_x,])$Binv;
     }
     if (length(nnz_indices_y)>0){
       sigmayhat = S[nnz_indices_y,nnz_indices_y];
       a_estimate[nnz_indices_y,] = a_estimate[nnz_indices_y,] %*% pracma::sqrtm(t(a_estimate[nnz_indices_y,]) %*% sigmayhat %*% a_estimate[nnz_indices_y,])$Binv;
     }
    
    u_estimate = a_estimate[1:p1,]
    v_estimate = a_estimate[(p1+1):p,] 
    l = list("U" = u_estimate, "V" = v_estimate)
    return(l)
  }


# Set up a parallel backend with graceful fallbacks.
#
# Attempts to create a parallel cluster, first trying the efficient FORK
# method (on Unix-like systems), then falling back to PSOCK, and finally
# returning NULL if all attempts fail.

cleanup_parallel_backend <- function(cl = NULL, verbose = FALSE) {
  if (!is.null(cl)) {
    try(parallel::stopCluster(cl), silent = !verbose)
  }

  if (requireNamespace("foreach", quietly = TRUE)) {
    foreach::registerDoSEQ()
  }

  if (requireNamespace("doParallel", quietly = TRUE) &&
      exists("stopImplicitCluster", envir = asNamespace("doParallel"), inherits = FALSE)) {
    try(doParallel::stopImplicitCluster(), silent = !verbose)
  }

  invisible(NULL)
}

.safe_mean_na <- function(x) {
  out <- mean(x, na.rm = TRUE)
  if (is.nan(out)) NA_real_ else out
}

.safe_sd_na <- function(x) {
  out <- stats::sd(x, na.rm = TRUE)
  if (is.nan(out)) NA_real_ else out
}

.truthy_env <- function(name) {
  value <- tolower(Sys.getenv(name, unset = "false"))
  value %in% c("true", "1", "yes")
}

.cran_core_limit_active <- function() {
  .truthy_env("_R_CHECK_LIMIT_CORES_")
}

.limit_parallel_cores <- function(num_cores) {
  if (length(num_cores) == 0L) {
    num_cores <- NA_integer_
  } else {
    num_cores <- suppressWarnings(as.integer(num_cores[[1L]]))
  }
  if (is.na(num_cores) || !is.finite(num_cores)) {
    num_cores <- 1L
  }

  num_cores <- max(1L, num_cores)
  if (.cran_core_limit_active()) {
    num_cores <- min(num_cores, 2L)
  }

  num_cores
}

.cvxr_preferred_solver <- function(preferred = c("CLARABEL", "ECOS", "SCS")) {
  if (!nzchar(system.file(package = "CVXR"))) {
    return(NULL)
  }

  installed_solvers <- tryCatch(
    getExportedValue("CVXR", "installed_solvers")(),
    error = function(e) character()
  )
  installed_solvers <- toupper(as.character(installed_solvers))
  preferred <- toupper(preferred)

  solver <- preferred[preferred %in% installed_solvers]
  if (length(solver) == 0L) {
    return(NULL)
  }

  solver[[1L]]
}

.cvxr_mosek_error <- function(e) {
  grepl("mosek|Rmosek|mosek_attachbuilder", conditionMessage(e), ignore.case = TRUE)
}

.stop_unusable_mosek <- function() {
  stop(
    "CVXR selected MOSEK, but Rmosek is not fully configured. ",
    "Install a non-MOSEK CVXR solver such as CLARABEL, ECOS, or SCS, ",
    "or use solver = 'ADMM'.",
    call. = FALSE
  )
}

.create_cv_folds <- function(n, k) {
  n <- as.integer(n)
  k <- as.integer(k)

  if (length(n) != 1L || is.na(n) || n < 2L) {
    stop("`n` must be a single integer >= 2.", call. = FALSE)
  }
  if (length(k) != 1L || is.na(k) || k < 2L) {
    stop("`k` must be a single integer >= 2.", call. = FALSE)
  }

  k <- min(k, n)
  fold_ids <- sample(rep(seq_len(k), length.out = n))
  unname(split(seq_len(n), fold_ids))
}

.find_local_global_names <- function(fun) {
  if (!requireNamespace("codetools", quietly = TRUE)) {
    return(character())
  }

  env <- environment(fun)
  if (is.null(env)) {
    return(character())
  }

  globals <- tryCatch(codetools::findGlobals(fun, merge = FALSE), error = function(e) NULL)
  if (is.null(globals)) {
    return(character())
  }

  names <- unique(c(globals$variables, globals$functions))
  names <- names[nzchar(names) & !grepl("^[[:punct:]]+$", names)]

  binding_env <- function(name, start) {
    cur <- start
    while (!identical(cur, emptyenv())) {
      if (exists(name, envir = cur, inherits = FALSE)) {
        return(cur)
      }
      cur <- parent.env(cur)
    }
    NULL
  }

  keep <- vapply(names, function(name) {
    where <- binding_env(name, env)
    if (is.null(where) || identical(where, baseenv())) {
      return(FALSE)
    }

    if ("ccar3" %in% loadedNamespaces()) {
      ccar3_ns <- asNamespace("ccar3")
      if (exists(name, envir = ccar3_ns, inherits = FALSE)) {
        return(FALSE)
      }
    }

    env_name <- environmentName(where)
    identical(where, globalenv()) || identical(env_name, "")
  }, logical(1))

  names[keep]
}

initialize_parallel_workers <- function(cl,
                                        pkg = "ccar3",
                                        pkg_path = NULL,
                                        verbose = FALSE) {
  if (is.null(cl)) {
    return(invisible(NULL))
  }

  parallel::clusterEvalQ(cl, {
    Sys.setenv(
      KMP_SHM_DISABLE = 1,
      KMP_SHM_DIR = "/tmp",
      TMPDIR = "/tmp",
      OMP_NUM_THREADS = 1,
      OPENBLAS_NUM_THREADS = 1,
      MKL_NUM_THREADS = 1,
      VECLIB_MAXIMUM_THREADS = 1
    )
    NULL
  })

  if (is.null(pkg_path)) {
    option_path <- getOption("ccar3_pkg_path", default = NULL)
    if (!is.null(option_path) &&
        nzchar(option_path) &&
        file.exists(file.path(option_path, "DESCRIPTION")) &&
        file.exists(file.path(option_path, "NAMESPACE"))) {
      pkg_path <- option_path
    }
  }

  if (is.null(pkg_path)) {
    env_path <- Sys.getenv("CCAR3_PKG_PATH", unset = "")
    if (nzchar(env_path) &&
        file.exists(file.path(env_path, "DESCRIPTION")) &&
        file.exists(file.path(env_path, "NAMESPACE"))) {
      pkg_path <- env_path
    }
  }

  if (is.null(pkg_path)) {
    ns_path <- tryCatch({
      if (requireNamespace("pkgload", quietly = TRUE) &&
          pkg %in% loadedNamespaces() &&
          pkgload::is_dev_package(pkg)) {
        getNamespaceInfo(asNamespace(pkg), "path")
      } else {
        ""
      }
    }, error = function(e) "")

    if (nzchar(ns_path) &&
        file.exists(file.path(ns_path, "DESCRIPTION")) &&
        file.exists(file.path(ns_path, "NAMESPACE"))) {
      pkg_path <- ns_path
    }
  }

  if (is.null(pkg_path)) {
    wd <- tryCatch(normalizePath(getwd(), mustWork = FALSE), error = function(e) "")
    if (nzchar(wd) &&
        file.exists(file.path(wd, "DESCRIPTION")) &&
        file.exists(file.path(wd, "NAMESPACE"))) {
      pkg_path <- wd
    }
  }

  if (!is.null(pkg_path) && nzchar(pkg_path)) {
    if (requireNamespace("pkgload", quietly = TRUE)) {
      if (verbose) {
        cat("Initializing workers from source package at", pkg_path, "\n")
      }
      parallel::clusterCall(cl, function(path) {
        pkgload::load_all(
          path,
          quiet = TRUE,
          export_all = FALSE,
          helpers = FALSE,
          attach_testthat = FALSE
        )
        NULL
      }, pkg_path)
    } else {
      if (verbose) {
        cat("Initializing workers by sourcing R files from", pkg_path, "\n")
      }
      parallel::clusterCall(cl, function(path) {
        r_dir <- file.path(path, "R")
        r_files <- c(
          file.path(r_dir, "helpers.r"),
          file.path(r_dir, "utils.R"),
          sort(list.files(r_dir, pattern = "\\.[Rr]$", full.names = TRUE))
        )
        r_files <- unique(r_files[file.exists(r_files)])
        for (f in r_files) {
          source(f, local = .GlobalEnv)
        }
        NULL
      }, pkg_path)
    }
  } else {
    if (verbose) {
      cat("Initializing workers with installed package", pkg, "\n")
    }
    parallel::clusterCall(cl, function(pkg_name) {
      suppressPackageStartupMessages(library(pkg_name, character.only = TRUE))
      NULL
    }, pkg)
  }

  invisible(NULL)
}

setup_parallel_backend <- function(num_cores = NULL, verbose = FALSE) {

  # 1. Determine the number of cores to use
  if (is.null(num_cores)) {
    # For SLURM, read the allocated CPUs. Fallback for local use.
    n_cores_str <- Sys.getenv("SLURM_CPUS_PER_TASK")
    num_cores <- if (n_cores_str == "") (parallel::detectCores() - 1) else as.integer(n_cores_str)
  }
  # Ensure we have at least 1 core
  num_cores <- .limit_parallel_cores(num_cores)
  if (verbose){
     cat(paste("\nAttempting to set up parallel backend with", num_cores, "cores.\n"))
  }
  cl <- NULL
  sysname <- Sys.info()[["sysname"]]
  prefer_fork <- .Platform$OS.type == "unix" && !identical(sysname, "Darwin")

  if (prefer_fork) {
    if (verbose) {
      cat("Unix-like system detected. Trying FORK backend first...\n")
    }
    cl <- tryCatch(
      parallel::makeCluster(num_cores, type = "FORK"),
      error = function(e_fork) {
        if (requireNamespace("crayon", quietly = TRUE)) {
          cat(crayon::yellow("Initial FORK backend setup failed: ", e_fork$message, "\n"))
        } else {
          cat("Initial FORK backend setup failed:", e_fork$message, "\n")
        }
        NULL
      }
    )
  } else if (verbose) {
    if (identical(sysname, "Darwin")) {
      cat("macOS detected. Using PSOCK backend to avoid orphaned FORK workers on interrupts...\n")
    } else {
      cat("Using PSOCK backend...\n")
    }
  }

  if (is.null(cl)) {
    if (verbose && prefer_fork) {
      cat("Attempting fallback PSOCK backend...\n")
    }
    cl <- tryCatch(
      parallel::makeCluster(num_cores, type = "PSOCK"),
      error = function(e_psock) {
        if (requireNamespace("crayon", quietly = TRUE)) {
          cat(crayon::red("PSOCK setup also failed: ", e_psock$message, "\n"))
        } else {
          cat("PSOCK setup also failed:", e_psock$message, "\n")
        }
        NULL
      }
    )
  }

  return(cl)
}

run_lambda_jobs <- function(lambdas,
                            run_one,
                            parallelize = TRUE,
                            nb_cores = NULL,
                            packages = character(),
                            exports = NULL,
                            verbose = FALSE) {
  if (!parallelize) {
    return(lapply(lambdas, run_one))
  }

  if (!requireNamespace("doParallel", quietly = TRUE)) {
    stop("Package 'doParallel' must be installed to use the parallelization option.",
         call. = FALSE)
  }

  if (!requireNamespace("crayon", quietly = TRUE)) {
    stop("Package 'crayon' must be installed to use the parallelization option.",
         call. = FALSE)
  }

  cl <- setup_parallel_backend(nb_cores, verbose = verbose)
  if (is.null(cl)) {
    warning("Parallel setup failed; running CV serially.", immediate. = TRUE)
    return(lapply(lambdas, run_one))
  }

  initialize_parallel_workers(cl, verbose = verbose)
  local_env <- environment(run_one)
  local_names <- if (is.null(local_env)) character() else ls(envir = local_env, all.names = TRUE)
  local_names <- setdiff(local_names, "run_one")
  if (length(local_names) > 0) {
    parallel::clusterExport(cl, varlist = local_names, envir = local_env)
  }

  global_names <- unique(c(exports, .find_local_global_names(run_one)))
  global_names <- setdiff(global_names, c("run_one", local_names))
  global_names <- global_names[vapply(global_names, exists, logical(1), envir = globalenv(), inherits = FALSE)]
  if (length(global_names) > 0) {
    parallel::clusterExport(cl, varlist = global_names, envir = globalenv())
  }

  parallel::clusterExport(cl, varlist = "run_one", envir = parent.frame())
  doParallel::registerDoParallel(cl)
  on.exit(cleanup_parallel_backend(cl), add = TRUE)

  foreach::foreach(
    lambda = lambdas,
    .packages = unique(c(packages, "foreach"))
  ) %dopar% run_one(lambda)
}

# environment to store whether we've already warned
.ccar3_internal_env <- new.env(parent = emptyenv())
.ccar3_internal_env$smut_warned <- FALSE


matmul <- function(A, B) {
  if (requireNamespace("SMUT", quietly = TRUE)) {
    # Use the fast C++ multiplication from SMUT
    SMUT::eigenMapMatMult(A, B)
  } else {
      # If not installed, warn once per session
      if (!.ccar3_internal_env$smut_warned) {
        packageStartupMessage("Using base R %*% for matrix multiplication. 
          Install 'SMUT' for faster performance.")
        .ccar3_internal_env$smut_warned <- TRUE
      }
    # Fallback to base R multiplication
    A %*% B
  }
}

compute_sqrt_inv <- function(S, threshold = 1e-4) {
  svd_S <- svd(S)
  svd_S$u %*% diag(sapply(svd_S$d, function(x) ifelse(x > threshold, 1 / sqrt(x), 0))) %*% t(svd_S$v)
}

compute_sqrt <- function(S, threshold = 1e-4) {
  svd_S <- svd(S)
  svd_S$u %*% diag(sapply(svd_S$d, function(x) ifelse(x > threshold, sqrt(x), 0))) %*% t(svd_S$u)
}

soft_thresh = function(A, lambda){
  #A * pmax(1 - lambda / abs(A), 0)
  sign(A) * pmax(abs(A) - lambda, 0)
}


fnorm = function(A){
  sqrt(sum(A^2))
}



soft_thresh_group = function(A, lambda){
  norm_A <- sqrt(sum(A^2))
  if (norm_A == 0) return(A)
  A * pmax(0, 1 - lambda / norm_A)
}



soft_thresh2 <- function(A, lambda){
  norm_A <- sqrt(sum(A^2))
  if(norm_A == 0){
    return(A)
  }
  result = A * pmax(1 - lambda/(norm_A), 0)
  return(result)
}


rmat = function(n, p){
  matrix(rnorm(n * p), n, p)
}

#' Function to perform Sparse CCA based on Waaijenborg et al. (2008)
#'  REFERENCE Parkhomenko et al. (2009), "Sparse Canonical Correlation Anlaysis with Application to Genomic Data Integration" in  Statistical Applications in Genetics and Molecular Biology, Volume 8, Issue 1, Article 1
#'
#' @param x.data Matrix of predictors (n x p)
#' @param y.data Matrix of responses (n x q)
#' @param n.cv Number of cross-validation folds (default is 5)
#' @param lambda.u.seq Vector of sparsity parameters for X (default is a sequence from 0 to 1 with step 0.1)
#' @param lambda.v.seq Vector of sparsity parameters for Y (default is a sequence from 0 to 1 with step 0.1)
#' @param Krank Number of canonical components to extract
#' @param standardize Standardize (center and scale) the data matrices X and Y (default is TRUE) before analysis
#' @return A list with elements:
#' \describe{
#'   \item{U}{Canonical direction matrix for X (p x r)}
#'   \item{V}{Canonical direction matrix for Y (q x r)}
#'   \item{cor}{Canonical correlations}
#' }
#' @export

SCCA_Parkhomenko<-function(x.data,y.data,n.cv=5,lambda.v.seq=seq(0, 0.2, by=0.02), lambda.u.seq=seq(0, 0.2, by=0.02),Krank=1, standardize=TRUE){
  ### Function to perform Sparse CCA based on Parkhomenko et al. (2009)
  # REFERENCE Parkhomenko et al. (2009), "Sparse Canonical Correlation Anlaysis with Application to Genomic Data Integration" in  Statistical Applications in Genetics and Molecular Biology, Volume 8, Issue 1, Article 1
  
  ### INPUT
  # x.data:         1st datamatrix n times p
  # y.data:         2nd datamatrix n times q
  # n.cv:           n.cv-fold cross-validation
  # lambda.v.seq:   Possible values of sparseness parameters for data Y. Lower bouns should be 0, upper bound can be increased to 2.
  # lambda.u.seq:   Possible values of sparseness parameters for data X. Lower bouns should be 0, upper bound can be increased to 2.
  # Krank:          rank approximation singular value decompostion
  
  
  ### OUTPUT
  # a:              canonical vectors accosiated to matrix X
  # b:              Canonical vectors associated to matrix Y
  # cancor:         Canonical Correlations
  #lambda.uopt:     value of sparsity parameter lambdau
  #lambda.vopt      value of sparsity parameter lamdbav

  # START CODE
  
  # SCCA to obtain first pair of canonical vectors

  if (standardize){
    x.data <- scale(x.data, center = TRUE, scale=TRUE)
    y.data <-  scale(y.data, center = TRUE, scale=TRUE)
  }
  
  SCCA_FIT_K1 <- SCCA_Parkhomenko_K1(x.data=x.data,y.data=y.data,n.cv=n.cv,
                                     lambda.v.seq=lambda.v.seq, lambda.u.seq=lambda.u.seq)
  
  
  if (Krank==1){ #1 canonical vector pair
    ##OUTPUT
    out<-list(a=SCCA_FIT_K1$a,b=SCCA_FIT_K1$b,cancor=SCCA_FIT_K1$cancor,lambda.uopt=SCCA_FIT_K1$best.lambda.u,lambda.vopt=SCCA_FIT_K1$best.lambda.v)

  }   else{ #more than 1 canonical vector pair
    
    ### STORE RESULTS
    a_vectors<-matrix(data=0,ncol=Krank,nrow=ncol(x.data))
    b_vectors<-matrix(data=0,ncol=Krank,nrow=ncol(y.data))
    
    
    # First pair of canonical vectors
    a_vectors[,1]<- SCCA_FIT_K1$a
    b_vectors[,1]<- SCCA_FIT_K1$b 
    
    # Obtain higher-order pairs of canonical vectors
    for (irank in 2:Krank){ 
      if (irank==2){
        SCCA_FIT<-SCCA_Parkhomenko_Khigher(x.data=x.data,y.data=y.data,k=SCCA_FIT_K1$K_residual,best.lambda.u=max(SCCA_FIT_K1$best.lambda.u),best.lambda.v=max(SCCA_FIT_K1$best.lambda.v))
        a_vectors[,irank]<- SCCA_FIT$a 
        b_vectors[,irank]<- SCCA_FIT$b
        D<-diag(c(SCCA_FIT_K1$cancor,SCCA_FIT$cancor))
        
        # Work on residual space to extract higher order canoncical vectors
        K_hat<-a_vectors[,1:irank]%*%D%*%t(b_vectors[,1:irank])
        K_residual<-SCCA_FIT_K1$k-K_hat
      }   else {
        SCCA_FIT<-SCCA_Parkhomenko_Khigher(x.data=x.data,y.data=y.data,k=K_residual,best.lambda.u=max(SCCA_FIT_K1$best.lambda.u),best.lambda.v=max(SCCA_FIT_K1$best.lambda.v))
        a_vectors[,irank]<- SCCA_FIT$a 
        b_vectors[,irank]<- SCCA_FIT$b
        D<-diag(c(diag(D),SCCA_FIT$cancor))
        K_hat<-a_vectors[,1:irank]%*%D%*%t(b_vectors[,1:irank])
        K_residual<-SCCA_FIT_K1$k-K_hat
      } 
    }
    
    ##OUTPUT
    out <- list(uhat=a_vectors,vhat=b_vectors,cancor=c(diag(D)),lambda.uopt=SCCA_FIT_K1$best.lambda.u,lambda.vopt=SCCA_FIT_K1$best.lambda.v)
  } 
}


SCCA_Parkhomenko_K1<-function(x.data,y.data,n.cv=5,lambda.v.seq=seq(0, 0.2, by=0.02), 
                              lambda.u.seq=seq(0, 0.2, by=0.02)){ # Sparse CCA for first dimension
  # CODE taken from http://www.uhnres.utoronto.ca/labs/tritchler/
  
  ### INPUT
  # x.data:         1st datamatrix n times p
  # y.data:         2nd datamatrix n times q
  # n.cv:           number of cross-validation steps to select sparseness parameters
  # lambda.v.seq:   Possible values of sparseness parameters for data Y. Lower bouns should be 0, upper bound can be increased to 2.
  # lambda.u.seq:   Possible values of sparseness parameters for data X. Lower bouns should be 0, upper bound can be increased to 2.
  
  ### OUTPUT
  # v:              right singular vector associated to matrix Y
  # u:              left singular vector associated to matrix X
  # cancor:         first canonical correlation
  # a:              canonical vectors accosiated to matrix X
  # b:              Canonical vectors associated to matrix Y
  # K_hat:          rank 1 approximation of matrix K
  # K_residual:     K-K_hat
  # k:              matrix K (sample covariance matrix)
  # best.lambda.u:  optimal sparseness parameter for matrix X
  # best.lambda.v:  optimal sparseness parameter for matrix Y
  
  
  # ANALYSIS
  # Dimensions of input-dataset
  n.sample<-nrow(x.data) # Sample size
  p<-ncol(x.data) # Number of variables in X
  q<-ncol(y.data) # Number of variables in Y
  
  # Settings sparseness parameters
  n.lambdas.u <-  length(lambda.u.seq)
  n.lambdas.v <-  length(lambda.v.seq)
  lambda.v.matrix <- matrix(rep(lambda.v.seq, n.lambdas.u), nrow=n.lambdas.u, byrow=TRUE)
  lambda.u.matrix <- matrix(rep(lambda.u.seq, n.lambdas.v), nrow=n.lambdas.u, byrow=FALSE)
  ones.p <- rep(1, p)/p
  ones.q <- rep(1, q)/q
  
  
  # SCCA Analysis
  n.cv.sample <- trunc(n.sample/n.cv)
  whole.sample <- seq(1, n.sample)
  
  predict.corr.scca <- matrix(0, nrow=n.lambdas.u, ncol=n.lambdas.v)  # This matrix will contain average test sample correlation for each combination of sparseness parameters
  
  #_______Cross-validation to select optimal combination of sparseness parameters____________
  for (i.cv in 1:n.cv)
  {
    testing.sample <- whole.sample[((i.cv-1)*n.cv.sample+1):(i.cv*n.cv.sample)]
    training.sample <- whole.sample[!whole.sample%in%testing.sample]
    
    k <- sample.sigma12.function(x.data[training.sample, ], y.data[training.sample, ])
    
    # Get starting values for singular vectors
    # as column and row means from matrix K
    u.initial <- k %*% ones.q
    u.initial <- u.initial /sqrt(as.numeric(t(u.initial)%*%u.initial))
    v.initial <- t(k) %*% ones.p
    v.initial <- v.initial /sqrt(as.numeric(t(v.initial)%*%v.initial))
    
    # _______________Data for Predicted correlation (testing sample)_________________
    
    x.predict <- x.data[testing.sample, ]
    y.predict <- y.data[testing.sample, ]
    
    # Standardize data
    x.predict <- x.predict - mean(x.predict)
    y.predict <- y.predict - mean(y.predict)
    
    sigma11.predict <- var(x.predict)
    sigma22.predict <- var(y.predict)
    
    x.predict <- x.predict %*% diag( 1/sqrt(diag(sigma11.predict)) )
    y.predict <- y.predict %*% diag( 1/sqrt(diag(sigma22.predict)) )
    
    
    # ____________Loops for sparseness parameter combinations__________
    for(j.lambda.v in 1:n.lambdas.v)
    {
      
      flag.na <- 0
      
      for(j.lambda.u in 1:n.lambdas.u)
      {
        lambda.v <- lambda.v.seq[j.lambda.v]  # sparseness parameter for Y
        lambda.u <- lambda.u.seq[j.lambda.u]  # sparseness parameter for X
        
        if(flag.na==0)
        {
          uv <- scca.function(k, u.initial, v.initial, lambda.u, lambda.v)
          
          vj <- uv$v.new
          uj <- uv$u.new
          
          # Calculate predicted correlation for SCCA
          predict.corr.scca[j.lambda.u, j.lambda.v] <- predict.corr.scca[j.lambda.u, j.lambda.v] + abs(cor(x.predict%*%uj, y.predict%*%vj))
          if(is.na(predict.corr.scca[j.lambda.u, j.lambda.v])) flag.na <- 1
        }  # close if
        
        if(flag.na==1)
        {
          predict.corr.scca[j.lambda.u:n.lambdas.u, j.lambda.v] <- predict.corr.scca[j.lambda.u:n.lambdas.u, j.lambda.v] + NA
          break
        }
        
      }  # close loop on lambda.u
    }  # close loop on lambda.v
    
  }	# close cross-validation loop
  
  # ______________Identify optimal sparseness parameter combination___________  	
  
  predict.corr.scca[is.na(predict.corr.scca)] <- 0
  predict.corr.scca <- predict.corr.scca/n.cv
  
  best.predict.corr.scca <- max(abs(predict.corr.scca), na.rm=TRUE)
  best.lambda.v <- lambda.v.matrix[predict.corr.scca==best.predict.corr.scca]
  best.lambda.u <- lambda.u.matrix[predict.corr.scca==best.predict.corr.scca]  

  best.lambda.u <- best.lambda.u[1]
  best.lambda.v <- best.lambda.v[1]
  
  
  k <- sample.sigma12.function(x.data, y.data)
  
  # Get starting values for singular vectors
  # as column and row means from matrix K
  u.initial <- k %*% ones.q
  u.initial <- u.initial /sqrt(as.numeric(t(u.initial)%*%u.initial))
  v.initial <- t(k) %*% ones.p
  v.initial <- v.initial /sqrt(as.numeric(t(v.initial)%*%v.initial))
  
  uv <- scca.function(k, u.initial, v.initial, best.lambda.u, best.lambda.v)
  
  vj <- uv$v.new  # sparse singular vector (canonical vector for Y)
  uj <- uv$u.new	# sparse singular vector (canonical vector for X)
  
  corr.scca <- abs(cor(x.data%*%uj, y.data%*%vj))	# canonical correlation for X and Y data
  
  # Canoncial vectors
  a<-diag(1/apply(x.data,2,sd),p)%*%uj
  b<-diag(1/apply(y.data,2,sd),q)%*%vj
  
  # rank 1 approximation of K matrix
  K_hat<-a%*%corr.scca%*%t(b)
  K_residual<-k-K_hat
  
  out<-list(v=vj,u=uj,cancor=corr.scca,a=a,b=b,K_hat=K_hat,K_residual=K_residual,k=k,best.lambda.u=best.lambda.u,best.lambda.v=best.lambda.v)
}

SCCA_Parkhomenko_Khigher<-function(x.data,y.data,k,best.lambda.u,best.lambda.v){ # Sparse CCA for higher-order dimensions
  ### INPUT
  # x.data:         1st datamatrix n times p
  # y.data:         2nd datamatrix n times q
  # k:              K-K_hat
  # best.lambda.u:  optimal sparseness parameter for matrix X
  # best.lambda.v:  optimal sparseness parameter for matrix Y
  
  ### OUTPUT
  # v:              right singular vector associated to matrix Y
  # u:              left singular vector associated to matrix X
  # cancor:         first canonical correlation
  # a:              canonical vectors accosiated to matrix X
  # b:              Canonical vectors associated to matrix Y
  # k:              matrix K (sample covariance matrix)
  
  # ANALYSIS
  # Dimensions of input-dataset
  p<-ncol(x.data) # Number of variables in X
  q<-ncol(y.data) # Number of variables in Y
  ones.p <- rep(1, p)/p
  ones.q <- rep(1, q)/q
  
  # Get starting values for singular vectors
  # as column and row means from matrix K
  u.initial <- k %*% ones.q
  u.initial <- u.initial /sqrt(as.numeric(t(u.initial)%*%u.initial))
  v.initial <- t(k) %*% ones.p
  v.initial <- v.initial /sqrt(as.numeric(t(v.initial)%*%v.initial))
  
  uv <- scca.function(k, u.initial, v.initial, best.lambda.u, best.lambda.v)
  
  vj <- uv$v.new  # sparse singular vector (canonical vector for Y)
  uj <- uv$u.new  # sparse singular vector (canonical vector for X)
  
  corr.scca <- abs(cor(x.data%*%uj, y.data%*%vj))	# canonical correlation for X and Y data
  
  
  a<-diag(1/apply(x.data,2,sd),p)%*%uj
  b<-diag(1/apply(y.data,2,sd),q)%*%vj
  
  
  out<-list(v=vj,u=uj,cancor=corr.scca,a=a,b=b,k=k)
}


####################################################################################
####################### Code From Parkhomenko (2009) ###############################
####################################################################################
# Algorithm for Sparse Canonical Correlation Analysis
# Parkhomenko, Elena; Tritchler, David; and Beyene, Joseph (2008) "Sparse Canonical Correlation Analysis with Application to Genomic Data Integration," Statistical Applications in Genetics and Molecular Biology: Vol. 7 : Iss. 1, Article 1.  
# Available at: http://www.bepress.com/sagmb/vol7/iss1/1 


# This function computes sparse version of the first singular vectors of matrix K

# The number of selected variables, i.e. variables with non-zero entries 
# in computed singular vectors, is controlled by the sparseness parameters. 
# Increasing the sparseness parameter will decrease the number of selected 
# variables.


# Parameters:
# k is p by q covariance matrix for standardized data sets X and Y
# u.initial p by 1 vector of starting values for left singular vector
# v.initial q by 1 vector of starting values for right singular vector
# lambda.u is the sparseness parameter for left singular vector
# lambda.v is the sparseness parameter for right singular vector

scca.function <- function(k, u.initial, v.initial, lambda.u, lambda.v)
{
  i <- 0		# number of iterations used by SCCA
  eps <- 0.001	# convergence criterion 
  max.iter <- 50	# maximum nuber of iterations
  diff.u <- eps*10	
  diff.v <- eps*10
  
  while ((i < max.iter) & ((diff.u > eps) || (diff.v > eps)) )
  {
    i <- i+1
    
    # Update left singular vector
    
    vx <-  k %*% v.initial
    length.vx <- as.numeric(sqrt(t(vx)%*%vx))
    if(length.vx==0) length.vx <- 1
    vx <- vx / length.vx
    u.new <- abs(vx) - 0.5*lambda.u
    u.new <- (u.new + abs(u.new))/2
    u.new <- u.new*sign(vx)
    length.u.new <- as.numeric(sqrt(t(u.new)%*%u.new))
    if(length.u.new==0) length.u.new <- 1
    u.new <- u.new / length.u.new
    
    
    # Update right singular vector
    
    ux <- t(k) %*% u.new
    length.ux <- as.numeric(sqrt(t(ux)%*%ux))
    if(length.ux==0) length.ux <- 1
    ux <- ux / length.ux
    v.new <- abs(ux) - 0.5*lambda.v
    v.new <- (v.new + abs(v.new))/2
    v.new <- v.new * sign(ux)
    length.v.new <- as.numeric(sqrt(t(v.new)%*%v.new))
    if(length.v.new==0) length.v.new <- 1
    v.new <- v.new / length.v.new
    
    
    # Convergence measures
    
    diff.v <- max(abs(v.initial - v.new))
    diff.u <- max(abs(u.initial - u.new))
    
    v.initial <- v.new
    u.initial <- u.new
  }
  
  # Report the results:
  # u.new is computed left singular vector
  # v.new is computed right singular vector
  # i is the number of iterations used by SCCA
  
  list(u.new=u.new, v.new=v.new, i=i)
}


############  SOURCE FOR COVARIANCE MATRIX FUNCTION   #############################
# see source("sample_cov_function.R") above
###################################################################################
# Calculating sample covariance function
sample.sigma12.function <- function(x, y)
{
  x <- x - mean(x)
  y <- y - mean(y)
  
  # Sample variance-covariance matrices 
  sigma11 <- var(x)
  sigma22 <- var(y)
  
  x <- x %*% diag( 1/sqrt(diag(sigma11)) )
  y <- y %*% diag( 1/sqrt(diag(sigma22)) )
  
  return(cov(x,y))
  
  #   xstdize<-stdize(x)
  #   ystdize<-stdize(y)
  #   return(cov(xstdize,ystdize))
  
}

library(pracma)


generate_example_sparse_U <- function(n, p1, p2,
                                      r_pca = 3,
                                      nnzeros = 5,
                                      theta = diag(c(0.9,  0.8)),
                                      lambda_pca = 1,
                                      r = 2, overlapping_amount = 0,
                                      normalize_diagonal = TRUE,
                                      n_new = 50000) {
  ###
  # This generates a dataset (X,Y) from a multivariate normal where Sigma_XX  and 
  # Sigma_YY have a correlation structure, and Sigma_XY = U Lambda V^T is rank r on a set of nnzeros rows, 0 elsewhere.
  # The number of rows (resp. Columns) on which Sigma_XX (resp. Sigma_YY)  and Sigma_XY
  # overlap is controlled by overlapping_amount
  # n: number of samples
  # p1: nb of features for X 
  # p2: nb of features for Y
  # nnzeros: number of non zero rows of U and V
  # r: rank of Sigma_XY
  # theta : canonical correlation (must be a list of length r)
  # r_pca: rank of the PCA (used in the creation of Sigma_XX and SigmaYY)
  # lambda_pca: also used to create Sigma_XX as Sigma_{XX} = U_X Lambda_pca U_X^T, and setting diag(Sigma_XX) to 1
  # Returns:
  # S : empirical covariance matrix X^TY
  # Sigma: underlying (population) covariance Sigma_{XY}
  # u: ground truth for u
  # v: ground truth for v
  ###
  p_tot <- p1 + p2
  pp <- c(p1, p2)
  #print("We have:")
  #print(c(p_tot, nnzeros <= min(p1, p2), nnzeros))
  #print('--------------------------------------')
  #print('Generating data ...')
  Sigma <- diag(p1 + p2)
  s <- (1):(nnzeros)
  #print(paste0('Number of non zeros is: ', nnzeros))
  start <- ceiling((1 - overlapping_amount) * nnzeros)
  s_pca  <- (start + 1) : (start + nnzeros)
  s_pca2  <- 1:p2
  Lambda_pca <- rep(lambda_pca, r_pca)
  # generate vcovariance matrix for X and Y
  
  if (r_pca > 0){
    u1 = matrix(0, p1, r_pca)
    u1[s_pca, ] <- matrix(runif(n = nnzeros * r_pca, max = 3, min=0), nrow = nnzeros, ncol = r_pca)
    # Normalize u1
    u1[s_pca,] <- 0.5 * diag(1/ sqrt(apply(u1[s_pca, ]^2, 1, sum))) %*% u1[s_pca,]
    u1[s_pca, 1:r_pca] <- u1[s_pca,1:r_pca] #%*% (pracma::sqrtm(t(u1[s_pca,1:r_pca]) %*% u1[s_pca, 1:r_pca])$Binv)
    T1 <- u1 %*% diag(Lambda_pca) %*% t(u1)
    if (normalize_diagonal){
      diag(T1) <- 1
      Sigma[1:p1, 1:p1] <- T1
    }else{
      Sigma[1:p1, 1:p1] <- Sigma[1:p1, 1:p1] + T1 
    } 
    ### Same for Sigma_y
    u2 <- matrix(runif( p2 * r_pca, max = 1, min=0), nrow=p2)
    u2 <- 0.5 * diag(1/ sqrt(apply(u2^2, 1, sum))) %*% u2
    #u2[s_pca2, 1:r_pca] <- u2[s_pca2, 1:r_pca] #%*% (pracma::sqrtm(t(u2[s_pca2, 1:r_pca]) %*% u2[s_pca2, 1:r_pca])$Binv)
    T2 <- u2 %*% t(u2)
    if (normalize_diagonal){
      diag(T2) <- 1
      Sigma[(p1 + 1):(p1 + p2), (p1 + 1):(p1 + p2)] <- T2
    }else{
      Sigma[(p1 + 1):(p1 + p2), (p1 + 1):(p1 + p2)] <- T2 + Sigma[(p1 + 1):(p1 + p2), (p1 + 1) : (p1 + p2)]
    }
  }
  Sigmax <- Sigma[1:p1,1:p1];
  Sigmay <- Sigma[(p1+1):p_tot,(p1+1):p_tot];
  
  ### Generate cross covariance
  Tss <- Sigmax[s,s]
  prod <- matrix(0, pp[1], r)
  prod[s, 1:r] <- as.matrix(runif( nnzeros * r,max = 3, min=1), nrow=nnzeros)  * as.matrix(sample(c(-1,1), nnzeros*r, replace=TRUE), nrow=nnzeros)
  u <- prod %*% (pracma::sqrtm(t(prod[s, 1:r]) %*% Tss %*% prod[s, 1:r])$Binv)
  
  
  Tss_v <- Sigma[(p1 + 1):(p1 + p2), (p1 + 1):(p1 + p2)] 
  prod_v <- matrix(0, pp[2], r)
  prod_v[, 1:r] <- as.matrix(runif( p2 * r , max = 3, min=1), nrow=p2)  * as.matrix(sample(c(-1,1), p2 * r, replace=TRUE), nrow=p2)
  v <- prod_v %*% (pracma::sqrtm(t(prod_v[, 1:r]) %*% Tss_v %*% prod_v[, 1:r])$Binv)

  
  #Sigma[(p1 + 1) :(p1 + p2), 1:p1] <- Sigmax %*% u %*% theta %*% t(v) %*% Sigmay
  Sigma[1:p1, (p1 + 1):(p1 + p2)] <- Sigmax %*% u %*% theta %*% t(v) %*% Sigmay
  Sigma[(p1 + 1) :(p1 + p2), 1:p1]  <-  Sigmay %*% v%*% theta %*% t(u) %*% Sigmax
  
  #Generate Multivariate Normal Data According to Sigma
  sqrt_Sigma <- pracma::sqrtm(Sigma)$B
  Data <- matrix(rnorm(n = (n + n_new) * p_tot), ncol=p_tot) %*% sqrt_Sigma
  X <- Data[1:n, 1:p1]
  Y <- Data[1:n, (p1 + 1):(p1 + p2)]
  Xnew <- Data[(n+1):(n_new + n), 1:p1]
  Ynew <- Data[(n+1):(n_new + n), (p1 + 1):(p1 + p2)]
  #print('Data generated.')
  #print('--------------------------------------')
  Mask <- matrix(0, p_tot, p_tot)
  idx1 <- 1:pp[1]
  idx2 <- (pp[1] + 1):(pp[1] + pp[2])
  Mask[idx1, idx1] <- matrix(1, pp[1], pp[1])
  Mask[idx2, idx2] <- matrix(1, pp[2], pp[2])
  Sigma0 <- Sigma * Mask
  S <- cov(Data)
  sigma0hat <- S * Mask
  # Generate ground truth canonical vectors
  #Sigma_X_inv <- pracma::sqrtm(Sigma[1:p1, 1:p1])$Binv
  #Sigma_Y_inv <-  pracma::sqrtm(Sigma[(p1+1):(p_tot), (p1+1):(p_tot)])$Binv
  #GT = svd(Sigma_X_inv %*% Sigma[1:p1, (p1 + 1):p_tot] %*% Sigma_Y_inv, nu = r, nv = r)
  return(list(Sigma=Sigma, Sigma0=Sigma0,
              S = S, sigma0hat =  sigma0hat, Mask= Mask,
              X=X, Y = Y, Data=Data, 
              u=u,#pracma::sqrtm(Sigma[1:p1, 1:p1])$Binv %*% GT$u, 
              v=v, #pracma::sqrtm(Sigma[(p1+1):(p1+p2), (p1+1):(p1+p2)])$Binv %*% GT$v, 
              Xnew = Xnew, Ynew=Ynew,
              Sigmax=Sigmax, Sigmay=Sigmay
  ))
  
}



generate_example_group <- function(n, p1, p2, 
                                  r_pca = 3,
                                  nnzeros = 5,
                                  theta = diag(c(0.9,  0.8)),
                                  lambda_pca = 1,
                                  r = 2, overlapping_amount = 0,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) {
  ###
  # This generates a dataset (X,Y) from a multivariate normal where Sigma_XX  and 
  # Sigma_YY have a correlation structure, and Sigma_XY = U Lambda V^T is rank r on a set of nnzeros rows, 0 elsewhere.
  # The number of rows (resp. Columns) on which Sigma_XX (resp. Sigma_YY)  and Sigma_XY
  # overlap is controlled by overlapping_amount
  # n: number of samples
  # p1: nb of features for X 
  # p2: nb of features for Y
  # nnzeros: number of non zero rows of U and V
  # r: rank of Sigma_XY
  # theta : canonical correlation (must be a list of length r)
  # r_pca: rank of the PCA (used in the creation of Sigma_XX and SigmaYY)
  # lambda_pca: also used to create Sigma_XX as Sigma_{XX} = U_X Lambda_pca U_X^T, and setting diag(Sigma_XX) to 1
  # Returns:
  # S : empirical covariance matrix X^TY
  # Sigma: underlying (population) covariance Sigma_{XY}
  # u: ground truth for u 
  # v: ground truth for v
  ###
  p_tot <- p1 + p2
  pp <- c(p1, p2)
  #print("We have:")
  #print(c(p_tot, nnzeros <= min(p1, p2), nnzeros))
  #print('--------------------------------------')
  #print('Generating data ...')
  Sigma <- diag(p1 + p2)

  #print(paste0('Number of non zeros is: ', nnzeros))
  start = 1
  s_pca  <- 1:min(p2, 20)
  s_pca2  <- 1:min(p2, 20)
  #if (start + nnzeros > q){
  #  s_pca2 <- 1:nnzeros
  #}
  Lambda_pca <- rep(lambda_pca, r_pca)
  # generate covariance matrix for X and Y
  groups <- ccar3:::.create_cv_folds(p1, ceiling(p1 / 5)) ### creates groups of equal size
  s <- c()
  for (i in 1:nnzeros){
    s <- c(s, groups[[i]])
  }
  s_q = 1:p2
  if (r_pca >0){
    u1 = matrix(0, p1, r_pca)
    u1[s_pca, ] <- matrix(runif(n = length(s_pca) * r_pca, max = 3, min=1), 
                          nrow = length(s_pca), ncol = r_pca) * matrix(sample(c(-1,1), length(s_pca) * r_pca, replace=TRUE), nrow=length(s_pca), ncol=r_pca)
    # Normalize u1
    #print(dim(u1))
    #print(c(s_pca, r_pca))
    u1[s_pca, 1:r_pca] <- u1[s_pca,1:r_pca] %*% (pracma::sqrtm(t(u1[s_pca,1:r_pca]) %*% u1[s_pca, 1:r_pca])$Binv)
    T1 <- u1 %*% diag(Lambda_pca) %*% t(u1)
    if (normalize_diagonal){
      diag(T1) <- 1
      Sigma[1:p1, 1:p1] <- T1
    }else{
      Sigma[1:p1, 1:p1] <- Sigma[1:p1, 1:p1] + T1 
    }
  
     ### Same for Sigma_y
    u2 <- matrix(0, pp[2], r_pca)
    u2[s_pca2, 1:r_pca] <- matrix(runif( nnzeros * r_pca,max = 3, min=1), nrow=nnzeros)  * matrix(sample(c(-1,1), nnzeros*r_pca, replace=TRUE), nrow=nnzeros, ncol=r_pca)
    u2[s_pca2, 1:r_pca] <- u2[s_pca2,1:r_pca] %*% (pracma::sqrtm(t(u2[s_pca2,1:r_pca]) %*% u2[s_pca2,1:r_pca])$Binv)
    T2 <- u2 %*% diag(Lambda_pca) %*% t(u2)
    if (normalize_diagonal){
      diag(T2) <- 1
      Sigma[(p1 + 1):(p1 + p2), (p1 + 1):(p1 + p2)] <- T2
    }else{
      Sigma[(p1 + 1):(p1 + p2), (p1 + 1):(p1 + p2)] <- T2 + Sigma[(p1 + 1):(p1 + p2), (p1 + 1) : (p1 + p2)]
    }
  }
  Sigmax = Sigma[1:p1,1:p1];
  Sigmay = Sigma[(p1+1):p_tot,(p1+1):p_tot];

  ### Generate cross covariance
  Tss <- Sigma[s,s]
  u <- matrix(0, pp[1], r)
  for (i in 1:nnzeros){
    u[groups[[i]], 1:r] <- as.matrix(runif(r * length(groups[[i]]),
                                           max = 3, min=1), 
                                     nrow=length(groups[[i]]))
  }
  
  u <- u %*% (pracma::sqrtm(t(u[s, 1:r]) %*% Tss %*% u[s, 1:r])$Binv)
  Tss_v <- Sigma[(p1 + 1):(p1 + p2), (p1 + 1):(p1 + p2)][s_q, s_q]
  v <- matrix(0, pp[2], r)
  v[s_q, 1:r] <- as.matrix(runif( length(s_q) * r,max = 3, min=1), nrow=nnzeros)  * as.matrix(sample(c(-1,1), length(s_q) *r, replace=TRUE), nrow=nnzeros)
  v <- v %*% (pracma::sqrtm(t(v[s_q, 1:r]) %*% Tss_v %*% v[s_q, 1:r])$Binv)
  Sigma[(p1 + 1) :(p1 + p2), 1:p1] <- Sigmay %*%  v  %*% theta %*% t(u) %*% Sigmax
  Sigma[1:p1, (p1 + 1):(p1 + p2)] <- Sigmax %*%  u  %*% theta %*% t(v) %*% Sigmay


  
  #Generate Multivariate Normal Data According to Sigma
  sqrt_Sigma <- pracma::sqrtm(Sigma)$B
  Data <- matrix(rnorm(n = (n + n_new) * p_tot), ncol=p_tot) %*% sqrt_Sigma
  X <- Data[1:n, 1:p1]
  Y <- Data[1:n, (p1 + 1):(p1 + p2)]
  Xnew <- Data[(n+1):(n_new + n), 1:p1]
  Ynew <- Data[(n+1):(n_new + n), (p1 + 1):(p1 + p2)]
  #print('Data generated.')
  #print('--------------------------------------')
  Mask <- matrix(0, p_tot, p_tot)
  idx1 <- 1:pp[1]
  idx2 <- (pp[1] + 1):(pp[1] + pp[2])
  Mask[idx1, idx1] <- matrix(1, pp[1], pp[1])
  Mask[idx2, idx2] <- matrix(1, pp[2], pp[2])
  Sigma0 <- Sigma * Mask
  S <- cov(Data)
  sigma0hat <- S * Mask
  # Generate ground truth canonical vectors
  Sigma_X_inv <- solve(Sigma[1:p1, 1:p1])
  Sigma_Y_inv <-  solve(Sigma[(p1 + 1):(p_tot), (p1 + 1):(p_tot)])
  #GT = svd(Sigma_X_inv %*% Sigma[1:p1, (p1 + 1):p_tot] %*% Sigma_Y_inv, nu = r, nv = r)
  return(list(Sigma=Sigma, Sigma0=Sigma0,
              S = S, sigma0hat =  sigma0hat, Mask= Mask,
              X=X, Y = Y, Data=Data, u=u, v=v, 
              groups = groups,
              Xnew=Xnew,
              Ynew=Ynew,
              Sigmax=Sigmax, Sigmay=Sigmay
  ))
}



generate_example_graph <- function(n, p1 = 10,
                                   type_graph="2d-grid",
                                            p2, order = 2,
                                            r_pca = 3,
                                            nnzeros = 5,
                                            do_plot = FALSE,
                                            theta = diag(c(0.9,  0.8)),
                                            lambda_pca = 1,
                                            nnzeros_pca = 10,
                                            r = 2, overlapping_amount = 0,
                                            normalize_diagonal = TRUE,
                                   gen.using.gamma = FALSE,
                                            n_new = 50000) {
  ###
  # This generates a dataset (X,Y) from a multivariate normal where Sigma_XX  and 
  # Sigma_YY have a correlation structure, and Sigma_XY = U Lambda V^T is rank r on a set of nnzeros rows, 0 elsewhere.
  # The number of rows (resp. Columns) on which Sigma_XX (resp. Sigma_YY)  and Sigma_XY
  # overlap is controlled by overlapping_amount
  # n: number of samples
  # p1: nb of features for X 
  # p2: nb of features for Y
  # nnzeros: number of non zero rows of U and V
  # r: rank of Sigma_XY
  # theta : canonical correlation (must be a list of length r)
  # r_pca: rank of the PCA (used in the creation of Sigma_XX and SigmaYY)
  # lambda_pca: also used to create Sigma_XX as Sigma_{XX} = U_X Lambda_pca U_X^T, and setting diag(Sigma_XX) to 1
  # Returns:
  # S : empirical covariance matrix X^TY
  # Sigma: underlying (population) covariance Sigma_{XY}
  # u: ground truth for u
  # v: ground truth for v
  ###

  if (!requireNamespace("igraph", quietly = TRUE)) {
  stop("Package 'igraph' must be installed to use generate_example_graph().", call. = FALSE)
  }

  if (type_graph == "2d-grid"){
    g <- igraph::make_lattice(c(p1, p1))
  }else{
    g <- igraph::sample_pa(p1, power=1.2)
  }
  prop_g <- get_edge_incidence(g, weight = 1)
  p1<-length(igraph::V(g))
  p_tot <- p1 + p2
  
  pp <- c(p1, p2)
  #print("We have:")
  #print(c(p_tot, nnzeros <= min(p1, p2), nnzeros))
  #print('--------------------------------------')
  #print('Generating data ...')
  Sigma <- diag(p1 + p2)
  s <- sample(1:p1, nnzeros)
  
  #print(paste0('Number of non zeros is: ', nnzeros))
  start <- 1
  nnzeros_pca = 30
  s_pca  <- (start + 1) : (start + nnzeros_pca)
  s_pca2  <- 1:p2
  Lambda_pca <- rep(lambda_pca, r_pca)
  # generate vcovariance matrix for X and Y
  
  if (r_pca > 0){
    u1 = matrix(0, p1, r_pca)
    u1[s_pca, ] <- matrix(runif(n = nnzeros_pca * r_pca, max = 3, min=0), 
                          nrow = nnzeros_pca, ncol = r_pca)
    # Normalize u1
    u1[s_pca,] <- 0.5 * diag(1/ sqrt(apply(u1[s_pca, ]^2, 1, sum))) %*% u1[s_pca,]
    u1[s_pca, 1:r_pca] <- u1[s_pca,1:r_pca] #%*% (pracma::sqrtm(t(u1[s_pca,1:r_pca]) %*% u1[s_pca, 1:r_pca])$Binv)
    T1 <- u1 %*% diag(Lambda_pca) %*% t(u1)
    if (normalize_diagonal){
      diag(T1) <- 1
      Sigma[1:p1, 1:p1] <- T1
    }else{
      Sigma[1:p1, 1:p1] <- Sigma[1:p1, 1:p1] + T1 
    } 
    ### Same for Sigma_y
    u2 <- matrix(runif( p2 * r_pca, max = 1, min=0), nrow=p2)
    u2 <- 0.5 * diag(1/ sqrt(apply(u2^2, 1, sum))) %*% u2
    #u2[s_pca2, 1:r_pca] <- u2[s_pca2, 1:r_pca] #%*% (pracma::sqrtm(t(u2[s_pca2, 1:r_pca]) %*% u2[s_pca2, 1:r_pca])$Binv)
    T2 <- u2 %*% t(u2)
    if (normalize_diagonal){
      diag(T2) <- 1
      Sigma[(p1 + 1):(p1 + p2), (p1 + 1):(p1 + p2)] <- T2
    }else{
      Sigma[(p1 + 1):(p1 + p2), (p1 + 1):(p1 + p2)] <- T2 + Sigma[(p1 + 1):(p1 + p2), (p1 + 1) : (p1 + p2)]
    }
  }
  Sigmax <- Sigma[1:p1,1:p1];
  Sigmay <- Sigma[(p1+1):p_tot,(p1+1):p_tot];
  
  ### Generate cross covariance

  Tss <- Sigmax
  #svd_gamma = svd(prop_g)
  #Gamma_dagger = svd_gamma$v[,1:(length(svd_gamma$d) -1)] %*% diag(1/svd_gamma$d[1:(length(svd_gamma$d) -1)]) %*% t(svd_gamma$u[,1:(length(svd_gamma$d) -1)] )
  
  if (gen.using.gamma){
    prod <- matrix(0, nrow(prop_g), r)
    Gamma_dagger = pinv(prop_g)
    s <- sample(1:nrow(prop_g), nnzeros)
    prod[s, 1:r] <- as.matrix(runif( length(s) * r,max = 3, min=1), nrow=length(s))  * as.matrix(sample(c(-1,1), length(s) * r, replace=TRUE), nrow=length(s))
    prod =  Gamma_dagger %*% prod
  }else{
    prod <- matrix(0, p1, r)
    s <- sample(1:p1, nnzeros)
    for (ss in s){
      ### sample neighborhood
      ind <- as.numeric(igraph::neighborhood(g, order=order, nodes = ss)[[1]])
      for(rr in 1:r){
        prod[ind, rr] <- runif(1, max = 3, min=1)  *  sample(c(-1,1),1)
      }
      
    }
  }
  
  ####
  u <- prod %*% (pracma::sqrtm(t(prod[, 1:r]) %*% Tss %*% prod[, 1:r])$Binv)
  
  

  Tss_v <- Sigmay
  prod_v <- matrix(0, pp[2], r)
  prod_v[, 1:r] <- as.matrix(runif( p2 * r , max = 3, min=1), nrow=p2)  * as.matrix(sample(c(-1,1), p2 * r, replace=TRUE), nrow=p2)
  v <- prod_v %*% (pracma::sqrtm(t(prod_v[, 1:r]) %*% Tss_v %*% prod_v[, 1:r])$Binv)
  
  #Sigma[(p1 + 1) :(p1 + p2), 1:p1] <- Sigmax %*% u %*% theta %*% t(v) %*% Sigmay
  Sigma[1:p1, (p1 + 1):(p1 + p2)] <- Sigmax %*% u %*% theta %*% t(v) %*% Sigmay
  Sigma[(p1 + 1) :(p1 + p2), 1:p1]  <- t(Sigma[1:p1, (p1 + 1):(p1 + p2)] )
  
  #Generate Multivariate Normal Data According to Sigma
  sqrt_Sigma <- pracma::sqrtm(Sigma)$B
  Data <- matrix(rnorm(n = (n + n_new) * p_tot), ncol=p_tot) %*% sqrt_Sigma
  X <- Data[1:n, 1:p1]
  Y <- Data[1:n, (p1 + 1):(p1 + p2)]
  Xnew <- Data[(n+1):(n_new + n), 1:p1]
  Ynew <- Data[(n+1):(n_new + n), (p1 + 1):(p1 + p2)]
  #print('Data generated.')
  #print('--------------------------------------')
  Mask <- matrix(0, p_tot, p_tot)
  idx1 <- 1:pp[1]
  idx2 <- (pp[1] + 1):(pp[1] + pp[2])
  Mask[idx1, idx1] <- matrix(1, pp[1], pp[1])
  Mask[idx2, idx2] <- matrix(1, pp[2], pp[2])
  Sigma0 <- Sigma * Mask
  S <- cov(Data)
  sigma0hat <- S * Mask
  # Generate ground truth canonical vectors
  #Sigma_X_inv <- pracma::sqrtm(Sigma[1:p1, 1:p1])$Binv
  #Sigma_Y_inv <-  pracma::sqrtm(Sigma[(p1+1):(p_tot), (p1+1):(p_tot)])$Binv
  #GT = svd(Sigma_X_inv %*% Sigma[1:p1, (p1 + 1):p_tot] %*% Sigma_Y_inv, nu = r, nv = r)
  return(list(Sigma=Sigma, Sigma0=Sigma0,
              S = S, sigma0hat =  sigma0hat, Mask= Mask,
              X=X, Y = Y, Data=Data, 
              u=u,#pracma::sqrtm(Sigma[1:p1, 1:p1])$Binv %*% GT$u, 
              v=v, #pracma::sqrtm(Sigma[(p1+1):(p1+p2), (p1+1):(p1+p2)])$Binv %*% GT$v, 
              Xnew = Xnew, Ynew=Ynew,
              Sigmax=Sigmax, Sigmay=Sigmay,
              Gamma=prop_g
  ))
  
}



generate_example_none_trivial_pca <- function(n, p1, p2,
                                              r_pca = 3,
                                              nnzeros = 5,
                                              theta = diag(c(0.9,  0.8)),
                                              lambda_pca = 1,
                                              r = 2, overlapping_amount = 0,
                                              normalize_diagonal = TRUE) {
  ###
  # This generates a dataset (X,Y) from a multivariate normal where Sigma_XX  and 
  # Sigma_YY have a correlation structure, and Sigma_XY = U Lambda V^T is rank r on a set of nnzeros rows, 0 elsewhere.
  # The number of rows (resp. Columns) on which Sigma_XX (resp. Sigma_YY)  and Sigma_XY
  # overlap is controlled by overlapping_amount
  # n: number of samples
  # p1: nb of features for X 
  # p2: nb of features for Y
  # nnzeros: number of non zero rows of U and V
  # r: rank of Sigma_XY
  # theta : canonical correlation (must be a list of length r)
  # r_pca: rank of the PCA (used in the creation of Sigma_XX and SigmaYY)
  # lambda_pca: also used to create Sigma_XX as Sigma_{XX} = U_X Lambda_pca U_X^T, and setting diag(Sigma_XX) to 1
  # Returns:
  # S : empirical covariance matrix X^TY
  # Sigma: underlying (population) covariance Sigma_{XY}
  # u: ground truth for u 
  # v: ground truth for v
  ###
  p_tot <- p1 + p2
  pp <- c(p1, p2)
  print("We have:")
  print(c(p_tot, nnzeros <= min(p1, p2), nnzeros))
  print('--------------------------------------')
  print('Generating data ...')
  Sigma <- diag(p1 + p2)
  s <- (1):(nnzeros)
  print(paste0('Number of non zeros is: ', nnzeros))
  start <- ceiling((1 - overlapping_amount) * nnzeros)
  s_pca  <- (start + 1) : (start + nnzeros)
  s_pca2  <- s_pca
  Lambda_pca <- rep(lambda_pca, r_pca)
  # generate vcovariance matrix for X and Y
  u1 = matrix(0, p1, r_pca)
  u1[s_pca, ] <- matrix(runif(n = nnzeros * r_pca, max = 3, min=1), nrow = nnzeros, ncol = r_pca) * matrix(sample(c(-1,1), nnzeros * r_pca, replace=TRUE), nrow=nnzeros, ncol=r_pca)
  # Normalize u1
  u1[s_pca, 1:r_pca] <- u1[s_pca,1:r_pca] %*% (sqrtm(t(u1[s_pca,1:r_pca]) %*% u1[s_pca, 1:r_pca])$Binv)
  T1 <- u1 %*% diag(Lambda_pca) %*% t(u1)
  if (normalize_diagonal){
    diag(T1) <- 1
    Sigma[1:p1, 1:p1] <- T1
  }else{
    Sigma[1:p1, 1:p1] <- Sigma[1:p1, 1:p1] + T1 
  }

   ### Same for Sigma_y
  u2 <- matrix(0, pp[2], r_pca)
  u2[s_pca2, 1:r_pca] <- matrix(runif( nnzeros * r_pca,max = 3, min=1), nrow=nnzeros)  * matrix(sample(c(-1,1), nnzeros*r_pca, replace=TRUE), nrow=nnzeros, ncol=r_pca)
  u2[s_pca2, 1:r_pca] <- u2[s_pca2,1:r_pca] %*% (sqrtm(t(u2[s_pca2,1:r_pca]) %*% u2[s_pca2,1:r_pca])$Binv)
  T2 <- u2 %*% diag(Lambda_pca) %*% t(u2)
  if (normalize_diagonal){
    diag(T2) <- 1
    Sigma[(p1 + 1):(p1 + p2), (p1 + 1):(p1 + p2)] <- T2
  }else{
    Sigma[(p1 + 1):(p1 + p2), (p1 + 1):(p1 + p2)] <- T2 + Sigma[(p1 + 1):(p1 + p2), (p1 + 1) : (p1 + p2)]
  }
  Sigmax = Sigma[1:p1,1:p1];
  Sigmay = Sigma[(p1+1):p_tot,(p1+1):p_tot];

  ### Generate cross covariance
  Tss <- Sigma[s,s]
  u <- matrix(0, pp[1], r)
  u[s, 1:r] <- as.matrix(runif( nnzeros * r,max = 3, min=1), nrow=nnzeros)  * as.matrix(sample(c(-1,1), nnzeros*r, replace=TRUE), nrow=nnzeros)
  u <- u %*% (sqrtm(t(u[s, 1:r]) %*% Tss %*% u[s, 1:r])$Binv)
  Tss_v <- Sigma[(p1 + 1):(p1 + p2), (p1 + 1):(p1 + p2)][s, s]
  v <- matrix(0, pp[2], r)
  v[s, 1:r] <- as.matrix(runif( nnzeros * r,max = 3, min=1), nrow=nnzeros)  * as.matrix(sample(c(-1,1), nnzeros*r, replace=TRUE), nrow=nnzeros)
  v <- v %*% (sqrtm(t(v[s, 1:r]) %*% Tss_v %*% v[s, 1:r])$Binv)
  Sigma[(p1 + 1) :(p1 + p2), 1:p1] <- Sigmay %*%  v  %*% theta %*% t(u) %*% Sigmax
  Sigma[1:p1, (p1 + 1):(p1 + p2)] <- Sigmax %*%  u  %*% theta %*% t(v) %*% Sigmay


  
  #Generate Multivariate Normal Data According to Sigma
  sqrt_Sigma <- sqrtm(Sigma)$B
  Data <- matrix(rnorm(n = 2 * n * p_tot), ncol=p_tot) %*% sqrt_Sigma
  X <- Data[1:n, 1:p1]
  Y <- Data[1:n, (p1 + 1):(p1 + p2)]
  Xnew <- Data[(n+1):(2*n), 1:p1]
  Ynew <- Data[(n+1):(2*n), (p1 + 1):(p1 + p2)]
  print('Data generated.')
  print('--------------------------------------')
  Mask <- matrix(0, p_tot, p_tot)
  idx1 <- 1:pp[1]
  idx2 <- (pp[1] + 1):(pp[1] + pp[2])
  Mask[idx1, idx1] <- matrix(1, pp[1], pp[1])
  Mask[idx2, idx2] <- matrix(1, pp[2], pp[2])
  Sigma0 <- Sigma * Mask
  S <- cov(Data[1:n, ])
  sigma0hat <- S * Mask
  # Generate ground truth canonical vectors
  Sigma_X_inv <- solve(Sigma[1:p1, 1:p1])
  Sigma_Y_inv <-  solve(Sigma[(p1 + 1):(p_tot), (p1 + 1):(p_tot)])
  GT = svd(Sigma_X_inv %*% Sigma[1:p1, (p1 + 1):p_tot] %*% Sigma_Y_inv, nu = r, nv = r)
  return(list(Sigma=Sigma, Sigma0=Sigma0,
              S = S, sigma0hat =  sigma0hat, Mask= Mask,
              X=X, Y = Y, Data=Data[1:n, ], u=GT$u, v=GT$v,     Xnew = Xnew, Ynew=Ynew,
              Sigmax=Sigmax, Sigmay=Sigmay
  ))
}

cran_limits_cores <- function() {
  value <- tolower(Sys.getenv("_R_CHECK_LIMIT_CORES_", unset = "false"))
  value %in% c("true", "1", "yes")
}

skip_if_cran_limits_cores <- function() {
  testthat::skip_if(
    cran_limits_cores(),
    "CRAN limits simultaneous parallel worker processes."
  )
}

skip_if_no_cvxr_conic_solver <- function() {
  testthat::skip_if_not_installed("CVXR")

  installed_solvers <- tryCatch(
    getExportedValue("CVXR", "installed_solvers")(),
    error = function(e) character()
  )
  installed_solvers <- toupper(as.character(installed_solvers))
  preferred <- c("CLARABEL", "ECOS", "SCS")

  testthat::skip_if(
    !any(preferred %in% installed_solvers),
    "CVXR has no non-MOSEK conic solver available."
  )
}

.resolve_group_preprocess_mode <- function(standardize = NULL, preprocess = NULL) {
  if (is.logical(preprocess) && length(preprocess) == 1L) {
    if (isTRUE(preprocess)) {
      preprocess <- NULL
    } else {
      return("none")
    }
  }

  if (!is.null(preprocess)) {
    return(match.arg(preprocess, c("scale", "center", "none")))
  }

  if (is.character(standardize)) {
    return(match.arg(standardize, c("scale", "center", "none")))
  }

  if (is.logical(standardize) && length(standardize) == 1L) {
    return(if (isTRUE(standardize)) "scale" else "center")
  }

  if (is.numeric(standardize) && length(standardize) == 1L && is.finite(standardize)) {
    return(if (isTRUE(as.logical(standardize))) "scale" else "center")
  }

  if (is.null(standardize)) {
    return("center")
  }

  stop("`standardize` must be logical or one of: 'scale', 'center', 'none'.", call. = FALSE)
}

.preprocess_group_matrix <- function(M, mode) {
  M <- as.matrix(M)
  if (mode == "none") {
    return(M)
  }

  M <- scale(M, center = TRUE, scale = identical(mode, "scale"))
  M <- as.matrix(M)
  M[!is.finite(M)] <- 0
  M
}


# Helper: ADMM-based group sparse solver
solve_group_rrr_admm <- function(X, tilde_Y, groups, lambda, Sx = NULL,
                                 rho = 1, niter = 1e4,
                                 thresh = 0.00001, thresh_0 = 1e-6, verbose = FALSE,
                                 invSx_apply = NULL,
                                 matrix_free_threshold = 4000L,
                                 cg_tol = 1e-6,
                                 cg_maxiter = NULL) {
  n <- nrow(X)
  p <- ncol(X)
  q <- ncol(tilde_Y)
  prod_xy <- crossprod(X, tilde_Y) / n

  if (is.null(invSx_apply)) {
    invSx_info <- make_invSx_apply(
      X, Sx, rho, verbose = verbose,
      matrix_free_threshold = matrix_free_threshold,
      cg_tol = cg_tol,
      cg_maxiter = cg_maxiter
    )
    invSx_apply <- invSx_info$apply
    if (verbose) message("ADMM linear solver strategy: ", invSx_info$strategy)
  }
  
  U <- matrix(0, p, q)
  Z <- matrix(0, p, q)
  
  for (i in seq_len(niter)) {
    U_old <- U; Z_old <- Z
    B <- invSx_apply(prod_xy + rho * (Z - U))
    Z <- B + U
    
    for (g in seq_along(groups)) {
      idx <- groups[[g]]
      norm_grp <- sqrt(sum(Z[idx, ]^2))
      if (norm_grp < lambda * sqrt(length(idx)) / rho) {
        Z[idx, ] <- 0
      } else {
        Z[idx, ] <- (1 - (lambda * sqrt(length(idx)) / rho) / norm_grp) * Z[idx, ]
      }
    }
    
    U <- U + B - Z
    
    if (verbose) cat("ADMM iter", i, "Primal: ", norm(Z - B), "Dual:", norm(Z_old - Z), "\n")
    if (max(norm(Z - B), norm(Z_old - Z)) < thresh) break
  }
  
  B_opt <- B
  B_opt[abs(B_opt) < thresh_0] <- 0
  return(B_opt)
}


# Helper: CVXR-based group sparse solver
solve_group_rrr_cvxr <- function(X, tilde_Y, groups, lambda, thresh_0 = 1e-6) {

  if (!nzchar(system.file(package = "CVXR"))) {
    stop("Package 'CVXR' must be installed to use the CVX solver.",
         call. = FALSE)
  }

  p <- ncol(X); q <- ncol(tilde_Y)
  n <- nrow(X)
  cvxr_variable <- getExportedValue("CVXR", "Variable")
  cvxr_norm <- getExportedValue("CVXR", "norm")
  cvxr_minimize <- getExportedValue("CVXR", "Minimize")
  cvxr_sum_squares <- getExportedValue("CVXR", "sum_squares")
  cvxr_problem <- getExportedValue("CVXR", "Problem")
  cvxr_psolve <- getExportedValue("CVXR", "psolve")
  cvxr_value <- getExportedValue("CVXR", "value")

  B <- cvxr_variable(c(p, q))
  
  penalty_exprs <- lapply(groups, function(g) {
    cvxr_norm(B[g, , drop = FALSE], "F")
  })
  penalty <- Reduce(`+`, penalty_exprs)  # or do.call("+", penalty_exprs)
  
  objective <- cvxr_minimize(
    1/n * cvxr_sum_squares(tilde_Y - X %*% B) + lambda * penalty
  )
  
  prob <- cvxr_problem(objective)
  res <- cvxr_psolve(prob)

  B_opt <- cvxr_value(B)
  B_opt[abs(B_opt) < thresh_0] <- 0
  return(B_opt)
}



#' Group-Sparse Canonical Correlation via Reduced-Rank Regression
#'
#' Performs group-sparse reduced-rank regression for CCA using either ADMM or CVXR solvers.
#'
#' @param X Predictor matrix (n x p)
#' @param Y Response matrix (n x q)
#' @param groups List of index vectors defining groups of predictors
#' @param Sx Optional covariance matrix for X; if NULL computed internally
#' @param Sy Optional covariance matrix for Y; if NULL computed internally
#' @param Sxy Optional cross covariance matrix for X and Y; if NULL computed internally
#' @param lambda Regularization parameter
#' @param r Target rank
#' @param standardize Backward-compatible preprocessing flag: TRUE = `"scale"`, FALSE = `"center"`.
#' @param preprocess Preprocessing mode. One of `"scale"` (center + scale), `"center"` (center only), or `"none"`.
#' @param LW_Sy Whether to apply Ledoit-Wolf shrinkage to Sy (default TRUE)
#' @param solver Either "ADMM" or "CVXR". The `"CVXR"` backend requires the
#'   optional package `CVXR`.
#' @param rho ADMM parameter
#' @param niter Maximum number of ADMM iterations
#' @param thresh Convergence threshold for ADMM
#' @param thresh_0 tolerance for declaring entries non-zero
#' @param matrix_free_threshold For ADMM: when both `n` and `p` are at least this value, use a matrix-free conjugate-gradient solve instead of forming a dense linear system.
#' @param cg_tol Relative tolerance for the matrix-free conjugate-gradient solve used in ADMM.
#' @param cg_maxiter Maximum iterations for the matrix-free conjugate-gradient solve. Defaults to `min(p, 1000)`.
#' @param verbose Print diagnostics
#'
#' @return A list with elements:
#' \describe{
#'   \item{U}{Canonical direction matrix for X (p x r)}
#'   \item{V}{Canonical direction matrix for Y (q x r)}
#'   \item{cor}{Canonical covariances}
#'   \item{loss}{The prediction error 1/n * \| XU - YV\|^2}
#'   \item{Lambda}{Canonical correlations}
#'   \item{B_opt}{Estimated reduced-rank coefficient matrix}
#' }
#' @export
cca_group_rrr <- function(X, Y, groups, 
                          Sx = NULL, Sy = NULL, Sxy = NULL, 
                          lambda = 0,  r,
                          standardize = FALSE,
                          preprocess = NULL,
                          LW_Sy = TRUE, solver = "ADMM",
                          rho = 1, niter = 1e4, thresh = 1e-4, thresh_0=1e-6,
                          matrix_free_threshold = 4000L,
                          cg_tol = 1e-6,
                          cg_maxiter = NULL,
                          verbose = FALSE) {
  preprocess_mode <- .resolve_group_preprocess_mode(standardize = standardize, preprocess = preprocess)
  X <- .preprocess_group_matrix(X, preprocess_mode)
  Y <- .preprocess_group_matrix(Y, preprocess_mode)

  n <- nrow(X)
  p <- ncol(X)
  q <- ncol(Y)
  
  if (n < min(p, q)) warning("Both X and Y are high-dimensional; method may be unstable.")

  if (is.null(Sy)) {
    Sy <- crossprod(Y) / n
    if (LW_Sy) Sy <- as.matrix(corpcor::cov.shrink(Y, verbose=verbose))
  }
  
  svd_Sy <- svd(as.matrix(Sy))
  sqrt_inv_Sy <- svd_Sy$u %*% diag(ifelse(svd_Sy$d > 1e-4, 1 / sqrt(svd_Sy$d), 0)) %*% t(svd_Sy$v)
  tilde_Y <- Y %*% sqrt_inv_Sy
  
  B_opt <- switch(solver,
                  "ADMM" = solve_group_rrr_admm(X, tilde_Y, groups=groups, lambda=lambda,
                                                Sx = Sx,
                                                rho=rho, niter=niter, 
                                                thresh=thresh, thresh_0=thresh_0, verbose=verbose,
                                                matrix_free_threshold = matrix_free_threshold,
                                                cg_tol = cg_tol,
                                                cg_maxiter = cg_maxiter),
                  "CVXR" = solve_group_rrr_cvxr(X, tilde_Y, groups, lambda, thresh_0=thresh_0),
                  stop("Unsupported solver: choose either 'ADMM' or 'CVXR'")
  )

  post <- .postprocess_rrr_fit(
    B_opt, X, Y, sqrt_inv_Sy, r,
    ridge_whiten = max(thresh_0, 1e-8),
    thresh_0 = thresh_0,
    verbose = verbose
  )

  list(
    U = post$U,
    V = post$V,
    loss = post$loss,
    cor = post$cor,
    Lambda = post$Lambda,
    B_opt = B_opt
  )
}



# Cross-validated loss for group-penalized CCA
cca_group_rrr_cv_folds <- function(X, Y, groups, Sx = NULL, Sy = NULL, kfolds = 5, 
                                   lambda = 0.01, r = 2, 
                                   preprocess = NULL,
                                   LW_Sy = FALSE, solver = "ADMM", 
                                   rho = 1, niter = 1e4, thresh = 1e-4,
                                   thresh_0 = 1e-6,
                                   matrix_free_threshold = 4000L,
                                   cg_tol = 1e-6,
                                   cg_maxiter = NULL,
                                   verbose = FALSE,
                                   folds = NULL,
                                   return_fold_values = FALSE) {
  preprocess_mode <- .resolve_group_preprocess_mode(standardize = FALSE, preprocess = preprocess)
  X <- .preprocess_group_matrix(X, preprocess_mode)
  Y <- .preprocess_group_matrix(Y, preprocess_mode)
  if (is.null(folds)) {
    folds <- .create_cv_folds(nrow(Y), kfolds)
  }

  rmse <- vapply(seq_along(folds), function(i) {
    n <- nrow(X)
    X_train <- X[-folds[[i]], ]; Y_train <- Y[-folds[[i]], ]
    X_val <- X[folds[[i]], ]; Y_val <- Y[folds[[i]], ]
    n_train <- n - nrow(X_val)
    
    if (!is.null(Sx)) {
      Sx_train <- (n * Sx - crossprod(X_val)) / n_train
    } else {
      Sx_train <- NULL
    }

    if (!is.null(Sy)) {
      Sy_train <- (n * Sy - crossprod(Y_val)) / n_train
    } else {
      Sy_train <- NULL
    }

    tryCatch({
      fit <- cca_group_rrr(X_train, Y_train, groups, Sx = Sx_train, Sy = Sy_train,
                           lambda = lambda, r = r, 
                           standardize = FALSE, preprocess = "none",
                           LW_Sy = LW_Sy, solver = solver,
                           rho = rho, niter = niter, thresh = thresh, thresh_0 = thresh_0,
                           matrix_free_threshold = matrix_free_threshold,
                           cg_tol = cg_tol,
                           cg_maxiter = cg_maxiter,
                           verbose = FALSE)
      mean((X_val %*% fit$U - Y_val %*% fit$V)^2)
    }, error = function(e) {
      message("Error in fold ", i, ": ", conditionMessage(e))
      NA_real_
    })
  }, numeric(1))
  
  if (return_fold_values) return(rmse)
  if (all(is.na(rmse))) return(1e8)
  mean(rmse, na.rm = TRUE)
}


#' Group-Sparse Canonical Correlation via Reduced-Rank Regression with CV
#'
#' Performs group-sparse reduced-rank regression for CCA using either ADMM or CVXR solvers.
#'
#' @param X Predictor matrix (n x p)
#' @param Y Response matrix (n x q)
#' @param groups List of index vectors defining groups of predictors
#' @param lambdas Grid of regularization parameters to try out
#' @param r Target rank
#' @param kfolds Nb of folds for the CV procedure
#' @param parallelize Whether to use parallel processing (default is FALSE)
#' @param standardize Backward-compatible preprocessing flag: TRUE = `"scale"`, FALSE = `"center"`.
#' @param preprocess Preprocessing mode. One of `"scale"` (center + scale), `"center"` (center only), or `"none"`.
#' @param LW_Sy Whether to apply Ledoit-Wolf shrinkage to Sy (default TRUE)
#' @param solver Either "ADMM" or "CVXR". The `"CVXR"` backend requires the
#'   optional package `CVXR`.
#' @param rho ADMM parameter
#' @param niter Maximum number of ADMM iterations
#' @param thresh Convergence threshold for ADMM
#' @param verbose Print diagnostics
#' @param thresh_0 tolerance for declaring entries non-zero
#' @param matrix_free_threshold For ADMM: when both `n` and `p` are at least this value, use a matrix-free conjugate-gradient solve instead of forming a dense linear system.
#' @param cg_tol Relative tolerance for the matrix-free conjugate-gradient solve used in ADMM.
#' @param cg_maxiter Maximum iterations for the matrix-free conjugate-gradient solve. Defaults to `min(p, 1000)`.
#' @param nb_cores Number of cores to use for parallelization (default is all available cores minus 1)
#'
#' @return A list with elements:
#' \describe{
#'   \item{U}{Canonical direction matrix for X (p x r)}
#'   \item{V}{Canonical direction matrix for Y (q x r)}
#'   \item{lambda}{Optimal regularisation parameter lambda chosen by CV}
#'   \item{rmse}{Mean squared error of prediction (as computed in the CV)}
#'   \item{cor}{Canonical covariances}
#'   \item{lambda_x}{Alias of the selected `lambda`}
#'   \item{lambda_x_se}{Foldwise standard error at the selected `lambda`}
#'   \item{lambda_y}{Placeholder for symmetry with two-penalty interfaces}
#'   \item{lambda_y_se}{Placeholder for symmetry with two-penalty interfaces}
#'   \item{resultsx}{Backward-compatible alias of `cv_summary`}
#'   \item{cv_summary}{Data frame with one row per lambda containing mean RMSE and its foldwise standard error}
#'   \item{cv_folds}{Data frame with fold-level RMSE values for each lambda}
#'   \item{Lambda}{Canonical correlations from the final fit}
#'   \item{B}{Estimated reduced-rank coefficient matrix from the final fit}
#'   \item{fit}{Final fit at the selected lambda}
#' }
#' @importFrom foreach foreach %dopar%
#' @export
cca_group_rrr_cv <- function(X, Y, groups, r = 2, 
                             lambdas = 10^seq(-3, 1.5, length.out = 10),
                             kfolds = 5, parallelize = FALSE, standardize = FALSE,
                             preprocess = NULL,
                             LW_Sy = TRUE, solver = "ADMM", rho = 1,
                             thresh_0 = 0,
                             niter = 1e4, thresh = 1e-4,
                             matrix_free_threshold = 4000L,
                             cg_tol = 1e-6,
                             cg_maxiter = NULL,
                             verbose = FALSE,
                             nb_cores = NULL) {
  preprocess_mode <- .resolve_group_preprocess_mode(standardize = standardize, preprocess = preprocess)
  
  if (nrow(X) < min(ncol(X), ncol(Y))) {
    warning("Both X and Y are high-dimensional; method may be unstable.")
  }
  
  # Apply preprocessing once before CV. Each fold reuses transformed matrices.
  X <- .preprocess_group_matrix(X, preprocess_mode)
  Y <- .preprocess_group_matrix(Y, preprocess_mode)
  
  n <- nrow(X)
  Sx <- NULL
  Sy <- if (LW_Sy) NULL else crossprod(Y) / n

  # Rebind the current source definitions into a local environment so PSOCK
  # workers use the updated functions rather than any stale installed copy.
  .rrr_cg_solve <- .rrr_cg_solve
  .rrr_chol_jitter <- .rrr_chol_jitter
  .rrr_invsqrt_psd <- .rrr_invsqrt_psd
  .rrr_whiten_factor <- .rrr_whiten_factor
  .rrr_safe_rank_svd <- .rrr_safe_rank_svd
  make_invSx_apply <- make_invSx_apply
  .postprocess_rrr_fit <- .postprocess_rrr_fit
  .resolve_group_preprocess_mode <- .resolve_group_preprocess_mode
  .preprocess_group_matrix <- .preprocess_group_matrix
  .create_cv_folds <- .create_cv_folds
  solve_group_rrr_admm <- solve_group_rrr_admm
  solve_group_rrr_cvxr <- solve_group_rrr_cvxr
  cca_group_rrr <- cca_group_rrr
  cca_group_rrr_cv_folds <- cca_group_rrr_cv_folds
  safe_mean_na <- .safe_mean_na
  safe_sd_na <- .safe_sd_na
  environment(make_invSx_apply) <- environment()
  environment(.postprocess_rrr_fit) <- environment()
  environment(solve_group_rrr_admm) <- environment()
  environment(solve_group_rrr_cvxr) <- environment()
  environment(cca_group_rrr) <- environment()
  environment(cca_group_rrr_cv_folds) <- environment()
  
  run_cv <- function(lambda) {
    fold_values <- cca_group_rrr_cv_folds(
      X, Y, groups, Sx = Sx, Sy = Sy, kfolds = kfolds,
      lambda = lambda, r = r,
      preprocess = "none",
      LW_Sy = LW_Sy, solver = solver,
      rho = rho, niter = niter, thresh = thresh, thresh_0 = thresh_0,
      matrix_free_threshold = matrix_free_threshold,
      cg_tol = cg_tol,
      cg_maxiter = cg_maxiter,
      folds = folds,
      return_fold_values = TRUE
    )

    list(
      summary = data.frame(
        lambda = lambda,
        rmse = safe_mean_na(fold_values),
        se = safe_sd_na(fold_values) / sqrt(sum(!is.na(fold_values))),
        stringsAsFactors = FALSE
      ),
      fold = data.frame(
        lambda = lambda,
        fold = seq_along(fold_values),
        rmse = fold_values,
        stringsAsFactors = FALSE
      )
    )
  }

  folds <- .create_cv_folds(n, kfolds)
  
    if (parallelize) {
      if (!requireNamespace("doParallel", quietly = TRUE)) {
      stop("Package 'doParallel' must be installed to use the parallelization option.",
          call. = FALSE)
      }

      if (!requireNamespace("crayon", quietly = TRUE)) {
      stop("Package 'crayon' must be installed to use the parallelization option.",
          call. = FALSE)
      }
    # --- GRACEFUL PARALLEL SETUP ---
      cl <- setup_parallel_backend(nb_cores)
      
      if (!is.null(cl)) {
        # If the cluster was created successfully, register it and plan to stop it
        if (verbose){
          cat(crayon::green("Parallel backend successfully registered.\n"))
        }

        initialize_parallel_workers(cl, verbose = verbose)
        doParallel::registerDoParallel(cl)
        on.exit(cleanup_parallel_backend(cl), add = TRUE)
      } else {
        # If setup_parallel_backend returned NULL, print a warning and proceed serially
        warning("All parallel setup attempts failed. Proceeding in serial mode.", immediate. = TRUE)
        parallelize <- FALSE # Ensure %dopar% runs serially
    }
   }

   if (parallelize) {


    parallel_packages <- if (solver == "CVXR") c("Matrix", "CVXR") else c("Matrix")
    results_list <- foreach(
      lambda = lambdas,
      .packages = parallel_packages,
      .export = c(
        "cca_group_rrr_cv_folds", "cca_group_rrr",
        "solve_group_rrr_admm", "solve_group_rrr_cvxr",
        "make_invSx_apply", ".rrr_cg_solve", ".rrr_chol_jitter",
        ".rrr_invsqrt_psd", ".rrr_whiten_factor", ".rrr_safe_rank_svd",
        ".postprocess_rrr_fit", ".resolve_group_preprocess_mode",
        ".preprocess_group_matrix", ".create_cv_folds",
        ".safe_mean_na", ".safe_sd_na"
      )
    ) %dopar% run_cv(lambda)
  } else {
    results_list <- lapply(lambdas, run_cv)
  }
  
  cv_summary <- dplyr::bind_rows(lapply(results_list, `[[`, "summary"))
  cv_folds <- dplyr::bind_rows(lapply(results_list, `[[`, "fold"))

  cv_summary$rmse[is.na(cv_summary$rmse) | cv_summary$rmse == 0] <- 1e8
  cv_summary <- cv_summary %>% dplyr::filter(rmse > 1e-5)
  
  opt_lambda <- cv_summary$lambda[which.min(cv_summary$rmse)]
  if (is.na(opt_lambda)) opt_lambda <- 0.1
  
  final <- cca_group_rrr(X, Y, groups, Sx = Sx, Sy = Sy, lambda = opt_lambda,
                         r = r, preprocess = "none", thresh = thresh, thresh_0 = thresh_0,
                         LW_Sy = LW_Sy, solver = solver, rho = rho, niter = niter,
                         matrix_free_threshold = matrix_free_threshold,
                         cg_tol = cg_tol,
                         cg_maxiter = cg_maxiter,
                         verbose=verbose)
  
  list(
    U = final$U,
    V = final$V,
    lambda = opt_lambda,
    rmse = cv_summary$rmse,
    cor = final$cor,
    lambda_x = opt_lambda,
    lambda_x_se = cv_summary$se[match(opt_lambda, cv_summary$lambda)],
    lambda_y = NA_real_,
    lambda_y_se = NA_real_,
    resultsx = cv_summary,
    cv_summary = cv_summary,
    cv_folds = cv_folds,
    Lambda = final$Lambda,
    B = final$B_opt,
    fit = final
  )
}

library(ccar3)

expected_cv_names <- c(
  "U", "V", "lambda", "rmse", "cor",
  "lambda_x", "lambda_x_se", "lambda_y", "lambda_y_se",
  "resultsx", "cv_summary", "cv_folds", "Lambda", "B", "fit"
)

test_that("cca_group_rrr_cv returns correct output structure", {
  set.seed(123)
  r = 2
  gen = generate_example_group(n=10000, p1=100, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  groups = gen$groups

  result <- cca_group_rrr_cv(X, Y, groups=groups, r = 2, kfolds=3, parallelize = FALSE)
  
  expect_type(result, "list")
  expect_identical(names(result), expected_cv_names)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], 100)
  expect_equal(dim(result$V)[1], 5)
  expect_true(all(c("lambda", "rmse", "se") %in% names(result$cv_summary)))
  expect_true(all(c("lambda", "fold", "rmse") %in% names(result$cv_folds)))
})



test_that("cca_group_rrr_cv can run in parallel", {
  skip_if_not_installed("doParallel")
  set.seed(123)
  r = 2
  gen = generate_example_group(n=500, p1=100, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  groups = gen$groups
  result <- cca_group_rrr_cv(X, Y, groups=groups, r = 2, kfolds=3,  parallelize = TRUE)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.35)  
  expect_true(subdistance(result$V, gen$v) < 0.35)
  expect_true(nrow(result$cv_summary) > 0)
  expect_true(nrow(result$cv_folds) > 0)
})


test_that("cca_group_rrr_cv returns the correct answer", {

  ##### Generate toy example data
  set.seed(123)
  r = 2
  gen = generate_example_group(n=500, p1=100, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  groups = gen$groups
  result <- cca_group_rrr_cv(
    X, Y, groups = groups, r = r,
    solver = "ADMM", lambdas = 10^seq(-5, 1.5, length.out = 30),
    kfolds = 3, parallelize = FALSE,
    LW_Sy = TRUE
  )
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.35)  
  expect_true(subdistance(result$V, gen$v) < 0.35)
  expect_true(nrow(result$cv_summary) > 0)
  expect_true(nrow(result$cv_folds) > 0)
})

test_that("cca_group_rrr_cv supports preprocess modes and p > n", {
  set.seed(42)
  n <- 50
  p <- 100
  q <- 5
  r <- 2
  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n * q), n, q)
  groups <- split(seq_len(p), ceiling(seq_len(p) / 5))

  result <- cca_group_rrr_cv(
    X, Y, groups = groups, r = r,
    kfolds = 3, parallelize = FALSE, lambdas = c(0.001, 0.01),
    preprocess = "center", LW_Sy = FALSE, niter = 300
  )

  expect_type(result, "list")
  expect_identical(names(result), expected_cv_names)
})

test_that("cca_group_rrr_cv forwards matrix-free ADMM controls", {
  set.seed(99)
  n <- 60
  p <- 90
  q <- 4
  r <- 2
  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n * q), n, q)
  groups <- split(seq_len(p), ceiling(seq_len(p) / 5))

  result <- cca_group_rrr_cv(
    X, Y, groups = groups, r = r,
    lambdas = c(0.001, 0.01), kfolds = 3, parallelize = FALSE,
    preprocess = "center", LW_Sy = FALSE, niter = 150,
    matrix_free_threshold = 1L,
    cg_tol = 1e-5,
    cg_maxiter = 40
  )

  expect_type(result, "list")
  expect_equal(dim(result$U), c(p, r))
  expect_equal(dim(result$V), c(q, r))
})

test_that("cca_group_rrr_cv_folds honors preprocessing modes on raw inputs", {
  set.seed(71)
  X <- matrix(rnorm(72 * 12), 72, 12)
  Y <- matrix(rnorm(72 * 4), 72, 4)
  groups <- split(seq_len(ncol(X)), ceiling(seq_len(ncol(X)) / 3))
  folds <- ccar3:::.create_cv_folds(nrow(X), 3)

  fold_values <- ccar3:::cca_group_rrr_cv_folds(
    X, Y, groups = groups,
    Sx = NULL, Sy = NULL,
    kfolds = 3, lambda = 0.01, r = 2,
    preprocess = "center",
    LW_Sy = FALSE,
    niter = 200,
    folds = folds,
    return_fold_values = TRUE
  )

  expect_length(fold_values, length(folds))
  expect_false(all(is.na(fold_values)))
})

library(ccar3)

expected_cv_names <- c(
  "U", "V", "lambda", "rmse", "cor",
  "lambda_x", "lambda_x_se", "lambda_y", "lambda_y_se",
  "resultsx", "cv_summary", "cv_folds", "Lambda", "B", "fit"
)

test_that("cca_group_rrr_cv returns correct output structure", {
  set.seed(123)
  r = 2
  gen = generate_example_group(n=10000, p1=100, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  groups = gen$groups

  result <- cca_group_rrr_cv(X, Y, groups=groups, r = 2, kfolds=3, parallelize = FALSE)
  
  expect_type(result, "list")
  expect_identical(names(result), expected_cv_names)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], 100)
  expect_equal(dim(result$V)[1], 5)
  expect_true(all(c("lambda", "rmse", "se") %in% names(result$cv_summary)))
  expect_true(all(c("lambda", "fold", "rmse") %in% names(result$cv_folds)))
})



test_that("cca_group_rrr_cv can run in parallel", {
  skip_if_not_installed("doParallel")
  skip_if_cran_limits_cores()
  set.seed(123)
  r = 2
  gen = generate_example_group(n=500, p1=100, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  groups = gen$groups
  result <- cca_group_rrr_cv(X, Y, groups=groups, r = 2, kfolds=3,  parallelize = TRUE)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.35)  
  expect_true(subdistance(result$V, gen$v) < 0.35)
  expect_true(nrow(result$cv_summary) > 0)
  expect_true(nrow(result$cv_folds) > 0)
})


test_that("cca_group_rrr_cv returns the correct answer", {

  ##### Generate toy example data
  set.seed(123)
  r = 2
  gen = generate_example_group(n=500, p1=100, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  groups = gen$groups
  result <- cca_group_rrr_cv(
    X, Y, groups = groups, r = r,
    solver = "ADMM", lambdas = 10^seq(-5, 1.5, length.out = 30),
    kfolds = 3, parallelize = FALSE,
    LW_Sy = TRUE
  )
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.35)  
  expect_true(subdistance(result$V, gen$v) < 0.35)
  expect_true(nrow(result$cv_summary) > 0)
  expect_true(nrow(result$cv_folds) > 0)
})

test_that("cca_group_rrr_cv supports preprocess modes and p > n", {
  set.seed(42)
  n <- 50
  p <- 100
  q <- 5
  r <- 2
  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n * q), n, q)
  groups <- split(seq_len(p), ceiling(seq_len(p) / 5))

  result <- cca_group_rrr_cv(
    X, Y, groups = groups, r = r,
    kfolds = 3, parallelize = FALSE, lambdas = c(0.001, 0.01),
    preprocess = "center", LW_Sy = FALSE, niter = 300
  )

  expect_type(result, "list")
  expect_identical(names(result), expected_cv_names)
})

test_that("cca_group_rrr_cv forwards matrix-free ADMM controls", {
  set.seed(99)
  n <- 60
  p <- 90
  q <- 4
  r <- 2
  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n * q), n, q)
  groups <- split(seq_len(p), ceiling(seq_len(p) / 5))

  result <- cca_group_rrr_cv(
    X, Y, groups = groups, r = r,
    lambdas = c(0.001, 0.01), kfolds = 3, parallelize = FALSE,
    preprocess = "center", LW_Sy = FALSE, niter = 150,
    matrix_free_threshold = 1L,
    cg_tol = 1e-5,
    cg_maxiter = 40
  )

  expect_type(result, "list")
  expect_equal(dim(result$U), c(p, r))
  expect_equal(dim(result$V), c(q, r))
})

test_that("cca_group_rrr_cv_folds honors preprocessing modes on raw inputs", {
  set.seed(71)
  X <- matrix(rnorm(72 * 12), 72, 12)
  Y <- matrix(rnorm(72 * 4), 72, 4)
  groups <- split(seq_len(ncol(X)), ceiling(seq_len(ncol(X)) / 3))
  folds <- ccar3:::.create_cv_folds(nrow(X), 3)

  fold_values <- ccar3:::cca_group_rrr_cv_folds(
    X, Y, groups = groups,
    Sx = NULL, Sy = NULL,
    kfolds = 3, lambda = 0.01, r = 2,
    preprocess = "center",
    LW_Sy = FALSE,
    niter = 200,
    folds = folds,
    return_fold_values = TRUE
  )

  expect_length(fold_values, length(folds))
  expect_false(all(is.na(fold_values)))
})

#' Sparse CCA by Witten and Tibshirani (2009)
#'
#' @param X Matrix of predictors (n x p)
#' @param Y Matrix of responses (n x q)
#' @param n.cv Number of cross-validation folds (default is 5)
#' @param lambdax Vector of sparsity parameters for X (default is a sequence from 0 to 1 with step 0.1)
#' @param lambday Vector of sparsity parameters for Y (default is a sequence from 0 to 1 with step 0.1)
#' @param standardize Standardize (center and scale) the data matrices X and Y (default is TRUE) before analysis
#' @param rank Number of canonical components to extract
#'
#' @return the appropriate levels of regularisation
#' @export

Witten.CV<-function(X,Y,n.cv=5,
                    rank,
                    lambdax=matrix(seq(from=0,to=1,by=0.1),nrow=1),
                    lambday=matrix(seq(from=0,to=1,by=0.1),nrow=1),
                    standardize = TRUE){ 


  n = nrow(X)
  n.cv.sample<-trunc(n/n.cv)
  whole.sample<-seq(1,n)
  lambdax=matrix(lambdax,nrow=1)
  lambday=matrix(lambday,nrow=1)

  if (standardize){
    X = scale(X, center = TRUE, scale=TRUE)
    Y = scale(Y, center = TRUE, scale=TRUE)
  }else{
    X = scale(X, center = TRUE, scale=FALSE)
    Y = scale(Y, center = TRUE, scale=FALSE)
  }
  
  cvscore<-array(NA,c(length(lambday),length(lambdax),n.cv)) #lambdax in columns, lambday in rows

  
  for (i in 1:n.cv){
    testing.sample<-whole.sample[((i-1)*n.cv.sample+1):(i*n.cv.sample)]
    training.sample<-whole.sample[-(((i-1)*n.cv.sample+1):(i*n.cv.sample))]
    Xcv = X[training.sample, ]
    Ycv = Y[training.sample, ]
    Xtest=X[testing.sample,]
    Ytest=Y[testing.sample,]

    cvscore[,,i]<-apply(lambdax,2, Witten.cv.lambdax, Xtrain=Xcv,Ytrain=Ycv,Xtest=Xtest,Ytest=Ytest,lambday=lambday)
  }
  cvscore.vec<-c(apply(cvscore,c(1,2),mean))

  
  LAMBDAX<-c(matrix(rep(lambdax,length(lambday)),byrow=T,nrow=length(lambday),ncol=length(lambdax)))
  LAMBDAY<-c(matrix(rep(lambday,length(lambdax)),byrow=F,nrow=length(lambday),ncol=length(lambdax)))
  
  lambdax.opt<-LAMBDAX[which.max(cvscore.vec)]
  lambday.opt<-LAMBDAY[which.max(cvscore.vec)]

  ### OUTPUT
  out <- list(lambdax.opt = lambdax.opt, lambday.opt = lambday.opt)
  return(out)
}


Witten.cv.lambdax<-function(U,Xtrain,Ytrain,Xtest,Ytest,lambday){ #AUXILIARY FUNCTION
  testcorrelations<-apply(lambday,2,Witten.cv.lambday,lambdaxfixed=U,Xtrain=Xtrain,Ytrain=Ytrain,Xtest=Xtest,Ytest=Ytest)
  return(testcorrelations)
}

Witten.cv.lambday<-function(V,Xtrain,Ytrain,Xtest,Ytest,
                            lambdaxfixed, r=1){ #AUXILIARY FUNCTION
  #print(lambdaxfixed)
  Fit.Witten<-PMA::CCA(x=Xtrain,z=Ytrain,typex="standard",typez="standard",K=r,penaltyx=lambdaxfixed,penaltyz=V,trace=F)
  return(sum(diag(abs(cor(Xtest%*%Fit.Witten$u, Ytest%*%Fit.Witten$v)))))
}  

.rrr_chol_jitter <- function(G, jitter = 1e-8, max_tries = 6) {
  p <- nrow(G)
  for (k in 0:max_tries) {
    out <- tryCatch(
      chol(G + (10^k) * jitter * diag(p)),
      error = function(e) NULL
    )
    if (!is.null(out)) return(out)
  }
  stop("chol() failed even with jitter.")
}

.rrr_invsqrt_psd <- function(G, eps = 1e-8) {
  G <- (G + t(G)) / 2
  ed <- eigen(G, symmetric = TRUE)
  vals <- pmax(ed$values, eps)
  ed$vectors %*% diag(1 / sqrt(vals), nrow = length(vals)) %*% t(ed$vectors)
}

.rrr_whiten_factor <- function(G, ridge = 1e-8, verbose = FALSE) {
  if (any(!is.finite(G))) {
    if (verbose) cat("\nNon-finite entries in whitening Gram matrix; sanitizing.\n")
    G[!is.finite(G)] <- 0
    G <- (G + t(G)) / 2 + ridge * diag(nrow(G))
  }

  tryCatch({
    R <- .rrr_chol_jitter(G, jitter = ridge)
    backsolve(R, diag(nrow(G)))
  }, error = function(e) {
    if (verbose) cat("\nchol_jitter failed; using eigen whitening fallback.\n")
    .rrr_invsqrt_psd(G, eps = ridge)
  })
}

.rrr_safe_rank_svd <- function(M, k, prefer_sparse = FALSE) {
  if (k < 1) return(NULL)
  nr <- nrow(M)
  nc <- ncol(M)
  if (nr < 1 || nc < 1) return(NULL)
  k <- min(k, nr, nc)

  out <- NULL

  if (requireNamespace("RSpectra", quietly = TRUE) && k < min(nr, nc)) {
    out <- tryCatch({
      if (prefer_sparse && requireNamespace("Matrix", quietly = TRUE)) {
        RSpectra::svds(Matrix::Matrix(M, sparse = TRUE), k)
      } else {
        RSpectra::svds(M, k)
      }
    }, error = function(e) NULL)

    if (!is.null(out)) {
      ok <- all(is.finite(out$u)) && all(is.finite(out$v)) && all(is.finite(out$d))
      if (!ok) out <- NULL
    }
  }

  if (is.null(out)) {
    out <- tryCatch(svd(as.matrix(M), nu = k, nv = k), error = function(e) NULL)
    if (is.null(out)) return(NULL)
    out$u <- out$u[, seq_len(k), drop = FALSE]
    out$v <- out$v[, seq_len(k), drop = FALSE]
    out$d <- out$d[seq_len(k)]
  }

  out
}

.resolve_preprocess_mode <- function(standardize = NULL, preprocess = NULL) {
  if (is.logical(preprocess) && length(preprocess) == 1L) {
    if (isTRUE(preprocess)) {
      preprocess <- NULL
    } else {
      return("none")
    }
  }

  if (!is.null(preprocess)) {
    return(match.arg(preprocess, c("scale", "center", "none")))
  }

  if (is.character(standardize)) {
    return(match.arg(standardize, c("scale", "center", "none")))
  }

  if (is.logical(standardize) && length(standardize) == 1L) {
    return(if (isTRUE(standardize)) "scale" else "center")
  }

  if (is.numeric(standardize) && length(standardize) == 1L && is.finite(standardize)) {
    return(if (isTRUE(as.logical(standardize))) "scale" else "center")
  }

  if (is.null(standardize)) {
    return("center")
  }

  stop("`standardize` must be logical or one of: 'scale', 'center', 'none'.", call. = FALSE)
}

.preprocess_matrix <- function(M, mode) {
  M <- as.matrix(M)
  if (mode == "none") {
    return(M)
  }

  M <- scale(M, center = TRUE, scale = identical(mode, "scale"))
  M <- as.matrix(M)
  M[!is.finite(M)] <- 0
  M
}

.rrr_match_mode <- function(mode) {
  if (missing(mode) || is.null(mode) || length(mode) == 0L || all(is.na(mode))) {
    return("sqrtm_norm")
  }

  mode <- tolower(as.character(mode[[1]]))
  if (identical(mode, "new")) {
    mode <- "sqrtm_norm"
  } else if (identical(mode, "old")) {
    mode <- "product_norm"
  }

  match.arg(mode, c("sqrtm_norm", "product_norm"))
}

.rrr_match_cv_metric <- function(cv_metric) {
  if (missing(cv_metric) || is.null(cv_metric) || length(cv_metric) == 0L || all(is.na(cv_metric))) {
    metric <- "mse"
  } else {
    metric <- tolower(as.character(cv_metric[[1]]))
  }

  if (metric %in% c("mse", "rmse", "prediction")) {
    return("mse")
  }

  if (metric %in% c("correlation", "cor", "corr", "association")) {
    return("correlation")
  }

  stop(
    "Unsupported `cv_metric`. Use one of: 'mse' or 'correlation'.",
    call. = FALSE
  )
}

.rrr_correlation_score <- function(X_scores, Y_scores) {
  X_scores <- as.matrix(X_scores)
  Y_scores <- as.matrix(Y_scores)

  if (nrow(X_scores) == 0L || nrow(Y_scores) == 0L ||
      ncol(X_scores) == 0L || ncol(Y_scores) == 0L) {
    return(0)
  }

  x_sd <- apply(X_scores, 2, stats::sd)
  y_sd <- apply(Y_scores, 2, stats::sd)

  keep_x <- which(is.finite(x_sd) & x_sd > 0)
  keep_y <- which(is.finite(y_sd) & y_sd > 0)

  if (length(keep_x) == 0L || length(keep_y) == 0L) {
    return(0)
  }

  C <- stats::cor(
    X_scores[, keep_x, drop = FALSE],
    Y_scores[, keep_y, drop = FALSE]
  )
  C[!is.finite(C)] <- 0

  sum(svd(C, nu = 0, nv = 0)$d)
}


.postprocess_rrr_fit <- function(Bhat, X, Y, sqrt_inv_Sy, r,
                                 ridge_whiten = 0, thresh_0 = 1e-6,
                                 verbose = FALSE) {
  n <- nrow(X)
  p <- ncol(X)
  q <- ncol(Y)

  if (all(abs(Bhat) < thresh_0)) {
    XU <- matrix(0, n, r)
    YV <- matrix(0, n, r)
    return(list(
      U = matrix(0, p, r),
      V = matrix(0, q, r),
      Lambda = rep(0, r),
      cor = rep(0, r),
      loss = mean((YV - XU)^2)
    ))
  }

  r_eff <- min(r, nrow(Bhat), ncol(Bhat))
  SVDs <- .rrr_safe_rank_svd(Bhat, r_eff, prefer_sparse = TRUE)
  if (is.null(SVDs)) {
    stop("SVD failed during reduced-rank postprocessing.", call. = FALSE)
  }

  U0 <- SVDs$u
  V0 <- sqrt_inv_Sy %*% SVDs$v

  XU0 <- X %*% U0
  YV0 <- Y %*% V0

  GX <- crossprod(XU0) / n + ridge_whiten * diag(r_eff)
  GY <- crossprod(YV0) / n + ridge_whiten * diag(r_eff)

  U <- U0 %*% .rrr_whiten_factor(GX, ridge = ridge_whiten, verbose = verbose)
  V <- V0 %*% .rrr_whiten_factor(GY, ridge = ridge_whiten, verbose = verbose)
  

  XU <- X %*% U
  YV <- Y %*% V
  cor <- diag(crossprod(XU, YV) / n)

  neg <- cor < 0
  if (any(neg)) {
    V[, neg] <- -V[, neg, drop = FALSE]
    YV[, neg] <- -YV[, neg, drop = FALSE]
    cor[neg] <- -cor[neg]
  }

  ord <- order(cor, decreasing = TRUE)
  cor <- cor[ord]
  U <- U[, ord, drop = FALSE]
  V <- V[, ord, drop = FALSE]
  XU <- XU[, ord, drop = FALSE]
  YV <- YV[, ord, drop = FALSE]

  if (r_eff < r) {
    U <- cbind(U, matrix(0, p, r - r_eff))
    V <- cbind(V, matrix(0, q, r - r_eff))
    cor <- c(cor, rep(0, r - r_eff))
  }

  list(
    U = U,
    V = V,
    Lambda = cor,
    cor = cor,
    loss = mean((YV - XU)^2)
  )
}


.postprocess_rrr_fit_old <- function(Bhat, X, Y, sqrt_inv_Sy, r,
                                 ridge_whiten = 0, thresh_0 = 1e-6,
                                 verbose = FALSE) {
  n <- nrow(X)
  p <- ncol(X)
  q <- ncol(Y)

  if (all(abs(Bhat) < thresh_0)) {
    XU <- matrix(0, n, r)
    YV <- matrix(0, n, r)
    return(list(
      U = matrix(0, p, r),
      V = matrix(0, q, r),
      Lambda = rep(0, r),
      cor = rep(0, r),
      loss = mean((YV - XU)^2)
    ))
  }

  r_eff <- min(r, nrow(Bhat), ncol(Bhat))
  active_rows <- which(rowSums(Bhat^2) > 0)
  Sx_active = crossprod(X[, active_rows, drop = FALSE]) / n
  SVDs <- .rrr_safe_rank_svd(Bhat, r_eff, prefer_sparse = TRUE)
  if (is.null(SVDs)) {
    stop("SVD failed during reduced-rank postprocessing.", call. = FALSE)
  }
  if (length(active_rows) > r - 1) {
      sqrt_Sx <- compute_sqrt(Sx_active)
      sol <- svd(sqrt_Sx %*% Bhat[active_rows, ])
      V <- sqrt_inv_Sy %*% sol$v[, 1:r]
      inv_D <- diag(sapply(sol$d[1:r], function(d) ifelse(d < 1e-4, 0, 1 / d)))
      U <- Bhat %*% sol$v[, 1:r] %*% inv_D
    } else {
      U <- matrix(0, p, r)
      V <- matrix(0, q, r)
    }

  XU <- X %*% U
  YV <- Y %*% V
  cor <- diag(crossprod(XU, YV) / n)
  neg <- cor < 0
  if (any(neg)) {
    V[, neg] <- -V[, neg, drop = FALSE]
    YV[, neg] <- -YV[, neg, drop = FALSE]
    cor[neg] <- -cor[neg]
  }

  ord <- order(cor, decreasing = TRUE)
  cor <- cor[ord]
  U <- U[, ord, drop = FALSE]
  V <- V[, ord, drop = FALSE]
  XU <- XU[, ord, drop = FALSE]
  YV <- YV[, ord, drop = FALSE]

  if (r_eff < r) {
    U <- cbind(U, matrix(0, p, r - r_eff))
    V <- cbind(V, matrix(0, q, r - r_eff))
    cor <- c(cor, rep(0, r - r_eff))
  }

  list(
    U = U,
    V = V,
    Lambda = cor,
    cor = cor,
    loss = mean((YV - XU)^2)
  )
}


.rrr_cg_solve <- function(A_mul, b, M_inv = NULL, tol = 1e-6,
                          maxiter = NULL, x0 = NULL) {
  p <- length(b)
  if (is.null(maxiter)) {
    maxiter <- min(p, 1000L)
  }

  x <- if (is.null(x0)) numeric(p) else as.numeric(x0)
  r <- as.numeric(b) - as.numeric(A_mul(x))
  bnorm <- sqrt(sum(b^2))
  if (!is.finite(bnorm) || bnorm < .Machine$double.eps) {
    return(x)
  }

  z <- if (is.null(M_inv)) r else M_inv(r)
  rz_old <- drop(crossprod(r, z))
  if (!is.finite(rz_old) || rz_old <= 0) {
    return(x)
  }

  p_dir <- z
  tol_abs <- tol * max(1, bnorm)

  for (iter in seq_len(maxiter)) {
    Ap <- as.numeric(A_mul(p_dir))
    denom <- drop(crossprod(p_dir, Ap))
    if (!is.finite(denom) || abs(denom) < .Machine$double.eps) {
      break
    }

    alpha <- rz_old / denom
    x <- x + alpha * p_dir
    r <- r - alpha * Ap
    if (sqrt(sum(r^2)) <= tol_abs) {
      break
    }

    z <- if (is.null(M_inv)) r else M_inv(r)
    rz_new <- drop(crossprod(r, z))
    if (!is.finite(rz_new) || rz_new <= 0) {
      break
    }

    beta <- rz_new / rz_old
    p_dir <- z + beta * p_dir
    rz_old <- rz_new
  }

  x
}

# Helper: efficient (Sx + rho I)^{-1} * W without forcing an explicit Sx build.
# For very large n and p, use a matrix-free CG solve that only applies X and t(X).
make_invSx_apply <- function(X, Sx = NULL, rho, ridge = 1e-8, verbose = FALSE,
                             matrix_free_threshold = 4000L,
                             cg_tol = 1e-6,
                             cg_maxiter = NULL) {
  n <- nrow(X)
  p <- ncol(X)
  if (n < 1 || p < 1) stop("X must have positive dimensions.")

  solve_from_factor <- function(G) {
    R <- .rrr_chol_jitter(G, jitter = ridge)
    function(B) backsolve(R, forwardsolve(t(R), as.matrix(B)))
  }

  if (!is.null(Sx)) {
    if (verbose) message("Using direct Cholesky solve with supplied Sx.")
    solve_gram <- solve_from_factor(Sx + rho * diag(p))
    return(list(
      apply = solve_gram,
      strategy = "direct",
      uses_woodbury = FALSE
    ))
  }

  if (min(n, p) >= matrix_free_threshold) {
    if (verbose) message("Using matrix-free CG solve for ADMM update.")

    diag_precond <- rho + colSums(X^2) / n
    diag_precond[!is.finite(diag_precond) | diag_precond < ridge] <- ridge

    A_mul <- function(v) {
      as.numeric(rho * v + crossprod(X, X %*% v) / n)
    }
    M_inv <- function(v) {
      as.numeric(v / diag_precond)
    }

    last_sol <- NULL
    return(list(
      apply = function(W) {
        W <- as.matrix(W)
        out <- matrix(0, p, ncol(W))
        if (!is.null(last_sol) && identical(dim(last_sol), dim(out))) {
          out[,] <- last_sol
        }

        for (j in seq_len(ncol(W))) {
          out[, j] <- .rrr_cg_solve(
            A_mul, W[, j],
            M_inv = M_inv,
            tol = cg_tol,
            maxiter = cg_maxiter,
            x0 = out[, j]
          )
        }

        last_sol <<- out
        out
      },
      strategy = "matrix_free_cg",
      uses_woodbury = FALSE
    ))
  }

  if (p <= n) {
    if (verbose) message("Using direct Cholesky solve from X without precomputing Sx upstream.")
    solve_gram <- solve_from_factor(crossprod(X) / n + rho * diag(p))
    return(list(
      apply = solve_gram,
      strategy = "direct",
      uses_woodbury = FALSE
    ))
  }

  if (verbose) message("Using Woodbury solve for ADMM update.")
  M <- diag(n) + tcrossprod(X) / (n * rho)
  R <- tryCatch(chol(M + ridge * diag(n)), error = function(e) NULL)

  solve_M <- if (!is.null(R)) {
    function(B) backsolve(R, forwardsolve(t(R), B))
  } else {
    if (verbose) message("Cholesky failed in Woodbury solve; using solve().")
    function(B) solve(M, B)
  }

  list(
    apply = function(W) {
      XW <- (X %*% W) / n
      tmp <- solve_M(XW)
      (1 / rho) * W - (1 / rho^2) * crossprod(X, tmp)
    },
    strategy = "woodbury",
    uses_woodbury = TRUE
  )
}



# Helper: ADMM-based group sparse solver
solve_rrr_admm <- function(X, tilde_Y, Sx = NULL, lambda, rho=1, niter=10, thresh, verbose = FALSE, thresh_0 = 0,
                           invSx_apply = NULL,
                           matrix_free_threshold = 4000L,
                           cg_tol = 1e-6,
                           cg_maxiter = NULL) {
  p <- ncol(X); 
  q <- ncol(tilde_Y)
  n <- nrow(X)

  if (is.null(invSx_apply)) {
    invSx_info <- make_invSx_apply(
      X, Sx, rho, verbose = verbose,
      matrix_free_threshold = matrix_free_threshold,
      cg_tol = cg_tol,
      cg_maxiter = cg_maxiter
    )
    invSx_apply <- invSx_info$apply
    if (verbose) message("ADMM linear solver strategy: ", invSx_info$strategy)
  }

  U <- Z <- matrix(0, p, q)
  
  prod_xy <- crossprod(X, tilde_Y) / n ### size is p x q, so okay since q <<< n
  for (i in seq_len(niter)) {
    U_old <- U; Z_old <- Z
    B <- invSx_apply(prod_xy + rho * (Z - U))
    Z <- B + U
    norm_col <- sqrt(rowSums(Z^2))
    shrinkage <- pmax(0, 1 - (lambda / rho) / norm_col)
    shrinkage[is.nan(shrinkage)] <- 0
    Z <- sweep(Z, 1, shrinkage, '*')
    U <- U + B - Z
    if (verbose) cat("ADMM iter", i, "Primal: ", norm(Z - B), "Dual:", norm(Z_old - Z), "\n")
    if (max(c(norm(Z - B) / sqrt(p), norm(Z_old - Z) / sqrt(p))) < thresh) break
  }
  B_opt <- B
  
  B_opt[abs(B_opt) < thresh_0] <- 0
  return(B_opt)
}


# Helper: CVXR-based group sparse solver
solve_rrr_cvxr <- function(X, tilde_Y, lambda, thresh_0=1e-6,
                           rho = 1, niter = 1e4, thresh = 1e-4,
                           verbose = FALSE,
                           matrix_free_threshold = 4000L,
                           cg_tol = 1e-6,
                           cg_maxiter = NULL) {
  if (!nzchar(system.file(package = "CVXR"))) {
    stop("Package 'CVXR' must be installed to use the CVX solver.",
         call. = FALSE)
  }

  p <- ncol(X); q <- ncol(tilde_Y)
  n <- nrow(X)

  fallback_admm <- function(reason = NULL) {
    if (verbose && !is.null(reason)) {
      message("Falling back to ADMM from CVXR: ", reason)
    }
    solve_rrr_admm(
      X, tilde_Y, Sx = NULL, lambda = lambda, rho = rho,
      niter = niter, thresh = thresh, thresh_0 = thresh_0,
      verbose = verbose,
      matrix_free_threshold = matrix_free_threshold,
      cg_tol = cg_tol,
      cg_maxiter = cg_maxiter
    )
  }

  # CVXR becomes impractical quickly because canonicalization densifies the
  # quadratic objective. For larger problems, the ADMM path is both faster and
  # dramatically more memory efficient.
  if (n * p * q > 5e6) {
    return(fallback_admm("problem size is too large for CVXR canonicalization"))
  }

  cvxr_variable <- getExportedValue("CVXR", "Variable")
  cvxr_norm <- getExportedValue("CVXR", "norm")
  cvxr_minimize <- getExportedValue("CVXR", "Minimize")
  cvxr_sum_squares <- getExportedValue("CVXR", "sum_squares")
  cvxr_problem <- getExportedValue("CVXR", "Problem")
  cvxr_psolve <- getExportedValue("CVXR", "psolve")

  B <- cvxr_variable(c(p, q))
  
  row_penalty <- Reduce(
    `+`,
    lapply(seq_len(p), function(j) cvxr_norm(B[j, , drop = FALSE], 2))
  )
  objective <- cvxr_minimize(
    1 / n * cvxr_sum_squares(tilde_Y - X %*% B) + lambda * row_penalty
  )
  B_opt <- tryCatch({
    problem <- cvxr_problem(objective)
    result <- cvxr_psolve(problem)
    result$getValue(B)
  }, error = function(e) {
    msg <- conditionMessage(e)
    if (grepl("memory|vector memory|sparse->dense|cannot allocate", msg, ignore.case = TRUE)) {
      return(fallback_admm(msg))
    }
    stop(e)
  })
  
  B_opt[abs(B_opt) < thresh_0] <- 0
  return(B_opt)
}


#' Canonical Correlation Analysis via Reduced Rank Regression (RRR)
#'
#' Estimates canonical directions using various RRR solvers and penalties.
#'
#' @param X Matrix of predictors.
#' @param Y Matrix of responses.
#' @param Sx Optional X covariance matrix.
#' @param Sy Optional Y covariance matrix.
#' @param lambda Regularization parameter.
#' @param r Rank of the solution.
#' @param mode Mode for postprocessing the RRR solution. One of `"sqrtm_norm"`
#'   (default) or `"product_norm"`. Legacy aliases `"new"` and `"old"` are
#'   also accepted. The former whitens the canonical variates to have identity
#'   covariance, while the latter does not whiten and instead returns the raw
#'   SVD factors of the RRR solution. The `"product_norm"` mode may be more
#'   interpretable in some cases but can yield canonical variates with very
#'   different scales and is not guaranteed to be numerically stable when the
#'   RRR solution is very low-rank or nearly low-rank.
#' @param highdim Boolean for high-dimensional regime.
#' @param solver Solver type: "rrr", "CVX", or "ADMM". The `"CVX"` backend
#'   requires the optional package `CVXR`.
#' @param LW_Sy Whether to use Ledoit-Wolf shrinkage for Sy.
#' @param standardize Backward-compatible preprocessing flag: TRUE = `"scale"`, FALSE = `"center"`.
#' @param preprocess Preprocessing mode. One of `"scale"` (center + scale), `"center"` (center only), or `"none"`.
#'   Logical values are still accepted for backward compatibility: `TRUE` uses `standardize`, `FALSE` skips preprocessing.
#' @param rho ADMM parameter.
#' @param niter Maximum number of iterations for ADMM.
#' @param thresh Convergence threshold.
#' @param thresh_0 For the ADMM solver: Set entries whose absolute value is below this to 0 (default 1e-6).
#' @param matrix_free_threshold For ADMM: when both `n` and `p` are at least this value, use a matrix-free conjugate-gradient solve instead of forming a dense linear system.
#' @param cg_tol Relative tolerance for the matrix-free conjugate-gradient solve used in ADMM.
#' @param cg_maxiter Maximum iterations for the matrix-free conjugate-gradient solve. Defaults to `min(p, 1000)`.
#' @param verbose Logical for verbose output.
#'
#' @return A list with elements:
#' \itemize{
#'   \item U: Canonical direction matrix for X (p x r)
#'   \item V: Canonical direction matrix for Y (q x r)
#'   \item cor: Canonical covariances
#'   \item loss: The prediction error 1/n * || XU - YV ||^2
#' }

#' @export
cca_rrr <- function(X, Y, Sx=NULL, Sy=NULL,
                    lambda = 0, 
                    r, highdim=TRUE, 
                    solver="ADMM",
                    LW_Sy = TRUE,
                    mode = "sqrtm_norm",
                    standardize = FALSE,
                    preprocess = NULL,
                    rho=1,
                    niter=1e4,
                    thresh = 1e-4, thresh_0 = 0,
                    matrix_free_threshold = 4000L,
                    cg_tol = 1e-6,
                    cg_maxiter = NULL,
                    verbose=FALSE) {
  
  preprocess_mode <- .resolve_preprocess_mode(standardize = standardize, preprocess = preprocess)
  mode <- .rrr_match_mode(mode)
  X <- .preprocess_matrix(X, preprocess_mode)
  Y <- .preprocess_matrix(Y, preprocess_mode)

  n <- nrow(X)
  p <- ncol(X)
  
  if (is.null(Sx) && !highdim) Sx <- crossprod(X) / n
  if (is.null(Sy)) {
    Sy <- crossprod(Y) / n
    if (LW_Sy) Sy <- as.matrix(corpcor::cov.shrink(Y, verbose=verbose))
  }
  
  sqrt_inv_Sy <- compute_sqrt_inv(Sy)
  tilde_Y <- Y %*% sqrt_inv_Sy
  
  if (!highdim) {
    if (verbose) print("Not using highdim")
    B_fit <- solve(Sx, crossprod(X, tilde_Y) / n)
  } else {
    if (solver == "CVX") {
      if(verbose){ print("Using CVX solver")}
      B_fit <- solve_rrr_cvxr(
        X, tilde_Y, lambda, thresh_0 = thresh_0,
        rho = rho, niter = niter, thresh = thresh, verbose = verbose,
        matrix_free_threshold = matrix_free_threshold,
        cg_tol = cg_tol,
        cg_maxiter = cg_maxiter
      )
    } else if (solver == "ADMM") {
      if(verbose){print("Using ADMM solver")}
      B_fit <- solve_rrr_admm(X, tilde_Y, Sx = Sx,
                              lambda=lambda, rho=rho, niter=niter, 
                              thresh=thresh, thresh_0=thresh_0,
                              verbose = verbose,
                              matrix_free_threshold = matrix_free_threshold,
                              cg_tol = cg_tol,
                              cg_maxiter = cg_maxiter)
    } else {
      if(verbose){print("Using gglasso solver")}
        if (!requireNamespace("rrpack", quietly = TRUE)) {
          stop("Package 'rrpack' must be installed to use the rrpack solver.",
              call. = FALSE)
        }
      fit <- rrpack::cv.srrr(tilde_Y, X, nrank = r, method = "glasso", nfold = 2,
                              modstr = list("lamA" = rep(lambda, 10), "nlam" = 10))
      B_fit <- fit$coef
    }
  }
  
  B_fit[abs(B_fit) < thresh_0] <- 0
  if(mode =="product_norm"){
    post <- .postprocess_rrr_fit_old(
    B_fit, X, Y, sqrt_inv_Sy, r,
    ridge_whiten = max(thresh_0, 1e-18),
    thresh_0 = thresh_0,
    verbose = verbose
  )

  }else{
    post <- .postprocess_rrr_fit(
    B_fit, X, Y, sqrt_inv_Sy, r,
    ridge_whiten = max(thresh_0, 1e-18),
    thresh_0 = thresh_0,
    verbose = verbose
  )

  }


  list(U = post$U, V = post$V, Lambda = post$Lambda, loss = post$loss, cor = post$cor, B_opt = B_fit)
}




#' Cross-validated Canonical Correlation Analysis via RRR
#'
#' Performs cross-validation to select optimal lambda, fits CCA_rrr.
#' Canonical Correlation Analysis via Reduced Rank Regression (RRR)
#' @param X Matrix of predictors.
#' @param Y Matrix of responses.
#' @param r Rank of the solution.
#' @param kfolds Number of folds for cross-validation.
#' @param lambdas Sequence of lambda values for cross-validation.
#' @param parallelize Logical; should cross-validation be parallelized?
#' @param mode Mode for postprocessing the RRR solution. One of `"sqrtm_norm"`
#'   (default) or `"product_norm"`. Legacy aliases `"new"` and `"old"` are
#'   also accepted. The former whitens the canonical variates to have identity
#'   covariance, while the latter does not whiten and instead returns the raw
#'   SVD factors of the RRR solution. The `"product_norm"` mode may be more
#'   interpretable in some cases but can yield canonical variates with very
#'   different scales and is not guaranteed to be numerically stable when the
#'   RRR solution is very low-rank or nearly low-rank.
#' @param solver Solver type: "rrr", "CVX", or "ADMM". The `"CVX"` backend
#'   requires the optional package `CVXR`.
#' @param LW_Sy Whether to use Ledoit-Wolf shrinkage for Sy.
#' @param standardize Backward-compatible preprocessing flag: TRUE = `"scale"`, FALSE = `"center"`.
#' @param preprocess Preprocessing mode. One of `"scale"` (center + scale), `"center"` (center only), or `"none"`.
#'   Logical values are still accepted for backward compatibility: `TRUE` uses `standardize`, `FALSE` skips preprocessing.
#' Data are preprocessed once up front, and fold fits reuse those transformed matrices without re-centering or re-scaling inside each fold.
#' @param rho ADMM parameter.
#' @param niter Maximum number of iterations for ADMM.
#' @param thresh Convergence threshold.
#' @param cv_metric Cross-validation metric. Use `"mse"` to minimize held-out prediction
#'   error or `"correlation"` to maximize held-out association between `X %*% U`
#'   and `Y %*% V`.
#' @param verbose Logical for verbose output.
#' @param thresh_0 tolerance for declaring entries non-zero
#' @param matrix_free_threshold For ADMM: when both `n` and `p` are at least this value, use a matrix-free conjugate-gradient solve instead of forming a dense linear system.
#' @param cg_tol Relative tolerance for the matrix-free conjugate-gradient solve used in ADMM.
#' @param cg_maxiter Maximum iterations for the matrix-free conjugate-gradient solve. Defaults to `min(p, 1000)`.
#' @param nb_cores Number of cores to use for parallelization. Defaults to min(kfolds, available cores minus 1).
#'
#' @return A list with elements:
#' \describe{
#'   \item{U}{Canonical direction matrix for X (p x r)}
#'   \item{V}{Canonical direction matrix for Y (q x r)}
#'   \item{lambda}{Optimal regularisation parameter lambda chosen by CV}
#'   \item{rmse}{Backward-compatible optimization objective. For `cv_metric = "mse"`
#'   this is the held-out mean squared error; for `cv_metric = "correlation"` it is
#'   `-cv_score` so that smaller still means better.}
#'   \item{cv_score}{Raw held-out cross-validation score averaged across folds.}
#'   \item{cv_metric}{The metric used to score lambdas during cross-validation.}
#'   \item{cor}{Canonical correlations at the selected lambda}
#'   \item{lambda_x}{Alias of the selected `lambda`}
#'   \item{lambda_x_se}{Foldwise standard error at the selected `lambda`}
#'   \item{lambda_y}{Placeholder for symmetry with two-penalty interfaces}
#'   \item{lambda_y_se}{Placeholder for symmetry with two-penalty interfaces}
#'   \item{resultsx}{Backward-compatible alias of `cv_summary`}
#'   \item{cv_summary}{Data frame with one row per lambda containing the mean CV score
#'   and its foldwise standard error.}
#'   \item{cv_folds}{Data frame with fold-level CV scores for each lambda.}
#'   \item{Lambda}{Canonical correlations from the final fit}
#'   \item{B}{Estimated reduced-rank coefficient matrix from the final fit}
#'   \item{fit}{Final fit at the selected lambda}
#' }
#' @importFrom foreach foreach %dopar%
#' @importFrom foreach foreach %do%
#' @export
cca_rrr_cv <- function(X, Y, 
                       r=2, 
                       lambdas=10^seq(-3, 1.5, length.out = 100),
                       kfolds=10,
                       solver="ADMM",
                       mode = "sqrtm_norm",
                       parallelize = FALSE,
                       LW_Sy = TRUE,
                       standardize = FALSE,
                       preprocess = NULL,
                       cv_metric = "mse",
                       rho=1,
                       thresh_0=0,
                       niter=1e4,
                       matrix_free_threshold = 4000L,
                       cg_tol = 1e-6,
                       cg_maxiter = NULL,
                       thresh = 1e-4, verbose=FALSE,
                       nb_cores = NULL) {
  preprocess_mode <- .resolve_preprocess_mode(standardize = standardize, preprocess = preprocess)
  mode <- .rrr_match_mode(mode)
  cv_metric <- .rrr_match_cv_metric(cv_metric)
  X <- .preprocess_matrix(X, preprocess_mode)
  Y <- .preprocess_matrix(Y, preprocess_mode)

  n <- nrow(X)
  Sx <- NULL
  Sy <- if (LW_Sy) NULL else crossprod(Y) / n
  folds <- .create_cv_folds(n, kfolds)
  safe_mean_na <- .safe_mean_na
  safe_sd_na <- .safe_sd_na
  cv_function <- function(lambda) {
    fold_values <- cca_rrr_cv_folds(
      X, Y, Sx = Sx, Sy = Sy, kfolds = kfolds,
      LW_Sy = LW_Sy,
      preprocess = "none",
      cv_metric = cv_metric,
      mode = mode,
      lambda = lambda, r = r, solver = solver,
      rho = rho, niter = niter, thresh = thresh,
      matrix_free_threshold = matrix_free_threshold,
      cg_tol = cg_tol,
      cg_maxiter = cg_maxiter,
      thresh_0 = thresh_0, folds = folds,
      return_fold_values = TRUE
    )

    fold_objective <- if (cv_metric == "mse") fold_values else -fold_values

    list(
      summary = data.frame(
        lambda = lambda,
        rmse = safe_mean_na(fold_objective),
        cv_score = safe_mean_na(fold_values),
        cv_metric = cv_metric,
        se = safe_sd_na(fold_values) / sqrt(sum(!is.na(fold_values))),
        stringsAsFactors = FALSE
      ),
      fold = data.frame(
        lambda = lambda,
        fold = seq_along(fold_values),
        rmse = fold_objective,
        cv_score = fold_values,
        cv_metric = cv_metric,
        stringsAsFactors = FALSE
      )
    )
  }

  foreach <- foreach::foreach
  `%dopar%` <- foreach::`%dopar%`
  
  if (parallelize && solver %in% c("CVX", "CVXR", "ADMM")) {

    if (!requireNamespace("doParallel", quietly = TRUE)) {
    stop("Package 'doParallel' must be installed to use the parallelization option.",
         call. = FALSE)
    }

    if (!requireNamespace("crayon", quietly = TRUE)) {
    stop("Package 'crayon' must be installed to use the parallelization option.",
         call. = FALSE)
    }

    # --- GRACEFUL PARALLEL SETUP ---
      available_cores_str <- Sys.getenv("SLURM_CPUS_PER_TASK")
      available_cores <- if (available_cores_str == "") {
        parallel::detectCores()
      } else {
        as.integer(available_cores_str)
      }
      available_cores <- max(1L, available_cores)
      if (is.null(nb_cores)) {
        nb_cores_effective <- min(kfolds, length(lambdas), available_cores - 1L)
      } else {
        nb_cores_effective <- min(as.integer(nb_cores), kfolds, length(lambdas))
      }
      nb_cores_effective <- max(1L, nb_cores_effective)
      cl <- setup_parallel_backend(nb_cores_effective)
      
      if (!is.null(cl)) {
        # If the cluster was created successfully, register it and plan to stop it
        if (verbose){
          cat(crayon::green("Parallel backend successfully registered.\n"))
        }
        initialize_parallel_workers(cl, verbose = verbose)
        doParallel::registerDoParallel(cl)
        on.exit(cleanup_parallel_backend(cl), add = TRUE)
      } else {
        # If setup_parallel_backend returned NULL, print a warning and proceed serially
        warning("All parallel setup attempts failed. Proceeding in serial mode.", immediate. = TRUE)
        parallelize <- FALSE # Ensure %dopar% runs serially
    }
   }

  if (parallelize && solver %in% c("CVX", "CVXR", "ADMM")) {
    rrr_parallel_exports <- c(
      "cca_rrr_cv_folds", "cca_rrr", "solve_rrr_admm", "solve_rrr_cvxr",
      "make_invSx_apply", ".rrr_cg_solve", ".rrr_chol_jitter", ".resolve_preprocess_mode",
      ".preprocess_matrix", ".create_cv_folds",
      ".rrr_invsqrt_psd", ".rrr_whiten_factor", ".rrr_safe_rank_svd",
      ".postprocess_rrr_fit", "compute_sqrt_inv", ".rrr_match_cv_metric",
      ".rrr_correlation_score",
      ".safe_mean_na", ".safe_sd_na"
    )
    if (solver  %in% c("CVX", "CVXR" )){
        if (!nzchar(system.file(package = "CVXR"))) {
         stop("Package 'CVXR' must be installed to use the CVXR/CVX solver.",
              call. = FALSE)
       }

      results_list <- foreach(
        lambda = lambdas,
        .packages = c("CVXR", "Matrix"),
        .export = rrr_parallel_exports
      ) %dopar% {
        cv_function(lambda)
      }

    }else{
      results_list <- foreach(
        lambda = lambdas,
        .export = rrr_parallel_exports
      ) %dopar% {
        cv_function(lambda)
      }
    }
    
  } else {
    results_list <- lapply(lambdas, cv_function)
  }
  
  cv_summary <- dplyr::bind_rows(lapply(results_list, `[[`, "summary"))
  cv_folds <- dplyr::bind_rows(lapply(results_list, `[[`, "fold"))

  if (cv_metric == "mse") {
    cv_summary <- cv_summary %>% 
      dplyr::mutate(rmse = ifelse(is.na(rmse) | rmse == 0, 1e8, rmse)) %>%
      dplyr::filter(rmse > 1e-5)
  } else {
    cv_summary <- cv_summary[is.finite(cv_summary$cv_score), , drop = FALSE]
  }
  
  resultsx <- cv_summary

  if (nrow(cv_summary) == 0L) {
    stop("No valid lambda values were found during cross-validation.", call. = FALSE)
  }

  opt_lambda <- if (cv_metric == "mse") {
    cv_summary$lambda[which.min(cv_summary$rmse)]
  } else {
    cv_summary$lambda[which.max(cv_summary$cv_score)]
  }
  opt_lambda <- ifelse(is.na(opt_lambda), 0.1, opt_lambda)

  final <- cca_rrr(X, Y, Sx=NULL, Sy=NULL, 
                   lambda=opt_lambda, r=r,
                   highdim=TRUE, solver=solver,
                   mode =mode,
                   standardize=FALSE, preprocess="none",
                   LW_Sy=LW_Sy, rho=rho, niter=niter, 
                   matrix_free_threshold = matrix_free_threshold,
                   cg_tol = cg_tol,
                   cg_maxiter = cg_maxiter,
                   thresh=thresh, thresh_0=thresh_0,
                   verbose=verbose)


  return(list(U = final$U, 
       V = final$V,
       lambda = opt_lambda,
       rmse = cv_summary$rmse,
       cv_score = cv_summary$cv_score,
       cv_metric = cv_metric,
       cor = final$cor,
       lambda_x = opt_lambda,
       lambda_x_se = cv_summary$se[match(opt_lambda, cv_summary$lambda)],
       lambda_y = NA_real_,
       lambda_y_se = NA_real_,
       resultsx = resultsx,
       cv_summary = cv_summary,
       cv_folds = cv_folds,
       Lambda = final$Lambda,
       B = final$B_opt,
       fit = final
       ))
}




cca_rrr_cv_folds <- function(X, Y, Sx, Sy, kfolds=5,
                             lambda=0.01,
                             r=2,
                             solver = "ADMM",
                             mode = "sqrtm_norm",
                             standardize = FALSE,
                             preprocess = "none",
                             cv_metric = "mse",
                             rho=1,
                             LW_Sy = TRUE,
                             niter=1e4,
                             thresh_0=0,
                             matrix_free_threshold = 4000L,
                             cg_tol = 1e-6,
                             cg_maxiter = NULL,
                             thresh = 1e-4,
                             folds = NULL,
                             return_fold_values = FALSE) {
  if (is.null(folds)) {
    folds <- .create_cv_folds(nrow(Y), kfolds)
  }

  preprocess_mode <- .resolve_preprocess_mode(standardize = FALSE, preprocess = preprocess)
  mode <- .rrr_match_mode(mode)
  cv_metric <- .rrr_match_cv_metric(cv_metric)
  X <- .preprocess_matrix(X, preprocess_mode)
  Y <- .preprocess_matrix(Y, preprocess_mode)

  rmse <- vapply(seq_along(folds), function(i) {
    n <- nrow(X)
    X_train <- X[-folds[[i]], ]; Y_train <- Y[-folds[[i]], ]
    X_val <- X[folds[[i]], ]; Y_val <- Y[folds[[i]], ]
    n_train <- n - nrow(X_val)
    
    if (is.null(Sx) == FALSE) {
      Sx_train <- (n * Sx - crossprod(X_val)) / n_train
    } else {
      Sx_train <- NULL
    }

    if (is.null(Sy) == FALSE) {
      Sy_train <- (n * Sy - crossprod(Y_val)) / n_train
    } else {
      Sy_train <- NULL
    }
    
    tryCatch({
      final <- cca_rrr(X_train, Y_train, 
                       Sx=Sx_train, Sy=Sy_train, highdim=TRUE,
                       mode =mode,
                       lambda=lambda, r=r, solver=solver,
                       LW_Sy=LW_Sy, standardize=FALSE, preprocess="none",
                       rho=rho, niter=niter, 
                       matrix_free_threshold = matrix_free_threshold,
                       cg_tol = cg_tol,
                       cg_maxiter = cg_maxiter,
                       thresh=thresh, thresh_0=thresh_0,
                       verbose=FALSE)
      if (cv_metric == "mse") {
        mean((X_val %*% final$U - Y_val %*% final$V)^2)
      } else {
        .rrr_correlation_score(X_val %*% final$U, Y_val %*% final$V)
      }
    }, error = function(e) {
      message("Error in fold ", i, ": ", conditionMessage(e))
      NA_real_
    })
  }, numeric(1))
  
  if (return_fold_values) return(rmse)
  if (all(is.na(rmse))) {
    return(if (cv_metric == "mse") 1e8 else NA_real_)
  }
  mean(rmse, na.rm = TRUE)
}

library(ccar3)

expected_cv_names <- c(
  "U", "V", "lambda", "rmse", "cv_score", "cv_metric", "cor",
  "lambda_x", "lambda_x_se", "lambda_y", "lambda_y_se",
  "resultsx", "cv_summary", "cv_folds", "Lambda", "B", "fit"
)

test_that("cca_rrr_cv returns correct output structure", {
  set.seed(123)
  r = 2
  gen = generate_example_sparse_U(n=100, p1=50, p2=10,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y  

  result <- cca_rrr_cv(X, Y, r = r, kfolds=3, parallelize = FALSE)
  
  expect_type(result, "list")
  expect_identical(names(result), expected_cv_names)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], 50)
  expect_equal(dim(result$V)[1], 10)
  expect_true(all(c("lambda", "rmse", "se") %in% names(result$cv_summary)))
  expect_true(all(c("lambda", "fold", "rmse") %in% names(result$cv_folds)))
})

test_that("cca_rrr_cv can run in parallel", {
    skip_if_not_installed("doParallel")
  set.seed(123)
  r = 3
  gen = generate_example_sparse_U(n=100, p1=50, p2=10,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y  

  result <- cca_rrr_cv(X, Y, r = r, kfolds=3, parallelize = TRUE)
  
  expect_type(result, "list")
  expect_identical(names(result), expected_cv_names)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], 50)
  expect_equal(dim(result$V)[1], 10)
})


test_that("cca_rrr_cv returns the correct answer", {

  ##### Generate toy example data
  set.seed(123)
  r = 2
  gen = generate_example_sparse_U(n=5000, p1=30, p2=5,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y
  result <- cca_rrr_cv(X, Y, r = r,
                      solver = "ADMM", lambdas=10^seq(-5, 1.5, length.out = 10),
                      kfolds=3, parallelize = TRUE,
                      LW_Sy = FALSE)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.2)  
  expect_true(subdistance(result$V, gen$v) < 0.2)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_true(nrow(result$cv_summary) > 0)
  expect_true(nrow(result$cv_folds) > 0)
})

test_that("cca_rrr_cv_folds honors preprocessing modes on raw inputs", {
  set.seed(17)
  X <- matrix(rnorm(75 * 12), 75, 12)
  Y <- matrix(rnorm(75 * 4), 75, 4)
  folds <- ccar3:::.create_cv_folds(nrow(X), 3)

  fold_values <- ccar3:::cca_rrr_cv_folds(
    X, Y, Sx = NULL, Sy = NULL,
    kfolds = 3, lambda = 0.01, r = 2,
    solver = "ADMM",
    preprocess = "center",
    LW_Sy = FALSE,
    niter = 200,
    folds = folds,
    return_fold_values = TRUE
  )

  expect_length(fold_values, length(folds))
  expect_false(all(is.na(fold_values)))
})

library(ccar3)

expected_cv_names <- c(
  "U", "V", "lambda", "rmse", "cv_score", "cv_metric", "cor",
  "lambda_x", "lambda_x_se", "lambda_y", "lambda_y_se",
  "resultsx", "cv_summary", "cv_folds", "Lambda", "B", "fit"
)

test_that("cca_rrr_cv returns correct output structure", {
  set.seed(123)
  r = 2
  gen = generate_example_sparse_U(n=100, p1=50, p2=10,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y  

  result <- cca_rrr_cv(X, Y, r = r, kfolds=3, parallelize = FALSE)
  
  expect_type(result, "list")
  expect_identical(names(result), expected_cv_names)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], 50)
  expect_equal(dim(result$V)[1], 10)
  expect_true(all(c("lambda", "rmse", "se") %in% names(result$cv_summary)))
  expect_true(all(c("lambda", "fold", "rmse") %in% names(result$cv_folds)))
})

test_that("cca_rrr_cv can run in parallel", {
  skip_if_not_installed("doParallel")
  skip_if_cran_limits_cores()
  set.seed(123)
  r = 3
  gen = generate_example_sparse_U(n=100, p1=50, p2=10,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y  

  result <- cca_rrr_cv(X, Y, r = r, kfolds=3, parallelize = TRUE)
  
  expect_type(result, "list")
  expect_identical(names(result), expected_cv_names)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], 50)
  expect_equal(dim(result$V)[1], 10)
})


test_that("cca_rrr_cv returns the correct answer", {
  skip_if_cran_limits_cores()

  ##### Generate toy example data
  set.seed(123)
  r = 2
  gen = generate_example_sparse_U(n=5000, p1=30, p2=5,
                                  r_pca = 5,
                                  nnzeros = 10,
                                  theta = diag(seq(0.9, 0.85, length.out = r)),
                                  lambda_pca = 1,
                                  r = r,  overlapping_amount = 1,
                                  normalize_diagonal = TRUE,
                                  n_new = 5000) 
  X = gen$X
  Y = gen$Y
  result <- cca_rrr_cv(X, Y, r = r,
                      solver = "ADMM", lambdas=10^seq(-5, 1.5, length.out = 10),
                      kfolds=3, parallelize = TRUE,
                      LW_Sy = FALSE)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.2)  
  expect_true(subdistance(result$V, gen$v) < 0.2)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_true(nrow(result$cv_summary) > 0)
  expect_true(nrow(result$cv_folds) > 0)
})

test_that("cca_rrr_cv_folds honors preprocessing modes on raw inputs", {
  set.seed(17)
  X <- matrix(rnorm(75 * 12), 75, 12)
  Y <- matrix(rnorm(75 * 4), 75, 4)
  folds <- ccar3:::.create_cv_folds(nrow(X), 3)

  fold_values <- ccar3:::cca_rrr_cv_folds(
    X, Y, Sx = NULL, Sy = NULL,
    kfolds = 3, lambda = 0.01, r = 2,
    solver = "ADMM",
    preprocess = "center",
    LW_Sy = FALSE,
    niter = 200,
    folds = folds,
    return_fold_values = TRUE
  )

  expect_length(fold_values, length(folds))
  expect_false(all(is.na(fold_values)))
})

.resolve_group_preprocess_mode <- function(standardize = NULL, preprocess = NULL) {
  if (is.logical(preprocess) && length(preprocess) == 1L) {
    if (isTRUE(preprocess)) {
      preprocess <- NULL
    } else {
      return("none")
    }
  }

  if (!is.null(preprocess)) {
    return(match.arg(preprocess, c("scale", "center", "none")))
  }

  if (is.character(standardize)) {
    return(match.arg(standardize, c("scale", "center", "none")))
  }

  if (is.logical(standardize) && length(standardize) == 1L) {
    return(if (isTRUE(standardize)) "scale" else "center")
  }

  if (is.numeric(standardize) && length(standardize) == 1L && is.finite(standardize)) {
    return(if (isTRUE(as.logical(standardize))) "scale" else "center")
  }

  if (is.null(standardize)) {
    return("center")
  }

  stop("`standardize` must be logical or one of: 'scale', 'center', 'none'.", call. = FALSE)
}

.preprocess_group_matrix <- function(M, mode) {
  M <- as.matrix(M)
  if (mode == "none") {
    return(M)
  }

  M <- scale(M, center = TRUE, scale = identical(mode, "scale"))
  M <- as.matrix(M)
  M[!is.finite(M)] <- 0
  M
}


# Helper: ADMM-based group sparse solver
solve_group_rrr_admm <- function(X, tilde_Y, groups, lambda, Sx = NULL,
                                 rho = 1, niter = 1e4,
                                 thresh = 0.00001, thresh_0 = 1e-6, verbose = FALSE,
                                 invSx_apply = NULL,
                                 matrix_free_threshold = 4000L,
                                 cg_tol = 1e-6,
                                 cg_maxiter = NULL) {
  n <- nrow(X)
  p <- ncol(X)
  q <- ncol(tilde_Y)
  prod_xy <- crossprod(X, tilde_Y) / n

  if (is.null(invSx_apply)) {
    invSx_info <- make_invSx_apply(
      X, Sx, rho, verbose = verbose,
      matrix_free_threshold = matrix_free_threshold,
      cg_tol = cg_tol,
      cg_maxiter = cg_maxiter
    )
    invSx_apply <- invSx_info$apply
    if (verbose) message("ADMM linear solver strategy: ", invSx_info$strategy)
  }
  
  U <- matrix(0, p, q)
  Z <- matrix(0, p, q)
  
  for (i in seq_len(niter)) {
    U_old <- U; Z_old <- Z
    B <- invSx_apply(prod_xy + rho * (Z - U))
    Z <- B + U
    
    for (g in seq_along(groups)) {
      idx <- groups[[g]]
      norm_grp <- sqrt(sum(Z[idx, ]^2))
      if (norm_grp < lambda * sqrt(length(idx)) / rho) {
        Z[idx, ] <- 0
      } else {
        Z[idx, ] <- (1 - (lambda * sqrt(length(idx)) / rho) / norm_grp) * Z[idx, ]
      }
    }
    
    U <- U + B - Z
    
    if (verbose) cat("ADMM iter", i, "Primal: ", norm(Z - B), "Dual:", norm(Z_old - Z), "\n")
    if (max(norm(Z - B), norm(Z_old - Z)) < thresh) break
  }
  
  B_opt <- B
  B_opt[abs(B_opt) < thresh_0] <- 0
  return(B_opt)
}


# Helper: CVXR-based group sparse solver
solve_group_rrr_cvxr <- function(X, tilde_Y, groups, lambda, thresh_0 = 1e-6) {

  if (!nzchar(system.file(package = "CVXR"))) {
    stop("Package 'CVXR' must be installed to use the CVX solver.",
         call. = FALSE)
  }

  p <- ncol(X); q <- ncol(tilde_Y)
  n <- nrow(X)
  cvxr_variable <- getExportedValue("CVXR", "Variable")
  cvxr_norm <- getExportedValue("CVXR", "norm")
  cvxr_minimize <- getExportedValue("CVXR", "Minimize")
  cvxr_sum_squares <- getExportedValue("CVXR", "sum_squares")
  cvxr_problem <- getExportedValue("CVXR", "Problem")
  cvxr_psolve <- getExportedValue("CVXR", "psolve")
  cvxr_value <- getExportedValue("CVXR", "value")
  cvxr_solver <- .cvxr_preferred_solver()

  B <- cvxr_variable(c(p, q))
  
  penalty_exprs <- lapply(groups, function(g) {
    cvxr_norm(B[g, , drop = FALSE], "F")
  })
  penalty <- Reduce(`+`, penalty_exprs)  # or do.call("+", penalty_exprs)
  
  objective <- cvxr_minimize(
    1/n * cvxr_sum_squares(tilde_Y - X %*% B) + lambda * penalty
  )
  
  prob <- cvxr_problem(objective)
  tryCatch(
    if (is.null(cvxr_solver)) {
      cvxr_psolve(prob)
    } else {
      cvxr_psolve(prob, solver = cvxr_solver)
    },
    error = function(e) {
      if (.cvxr_mosek_error(e)) {
        .stop_unusable_mosek()
      }
      stop(e)
    }
  )

  B_opt <- cvxr_value(B)
  B_opt[abs(B_opt) < thresh_0] <- 0
  return(B_opt)
}



#' Group-Sparse Canonical Correlation via Reduced-Rank Regression
#'
#' Performs group-sparse reduced-rank regression for CCA using either ADMM or CVXR solvers.
#'
#' @param X Predictor matrix (n x p)
#' @param Y Response matrix (n x q)
#' @param groups List of index vectors defining groups of predictors
#' @param Sx Optional covariance matrix for X; if NULL computed internally
#' @param Sy Optional covariance matrix for Y; if NULL computed internally
#' @param Sxy Optional cross covariance matrix for X and Y; if NULL computed internally
#' @param lambda Regularization parameter
#' @param r Target rank
#' @param standardize Backward-compatible preprocessing flag: TRUE = `"scale"`, FALSE = `"center"`.
#' @param preprocess Preprocessing mode. One of `"scale"` (center + scale), `"center"` (center only), or `"none"`.
#' @param LW_Sy Whether to apply Ledoit-Wolf shrinkage to Sy (default TRUE)
#' @param solver Either "ADMM" or "CVXR". The `"CVXR"` backend requires the
#'   optional package `CVXR`.
#' @param rho ADMM parameter
#' @param niter Maximum number of ADMM iterations
#' @param thresh Convergence threshold for ADMM
#' @param thresh_0 tolerance for declaring entries non-zero
#' @param matrix_free_threshold For ADMM: when both `n` and `p` are at least this value, use a matrix-free conjugate-gradient solve instead of forming a dense linear system.
#' @param cg_tol Relative tolerance for the matrix-free conjugate-gradient solve used in ADMM.
#' @param cg_maxiter Maximum iterations for the matrix-free conjugate-gradient solve. Defaults to `min(p, 1000)`.
#' @param verbose Print diagnostics
#'
#' @return A list with elements:
#' \describe{
#'   \item{U}{Canonical direction matrix for X (p x r)}
#'   \item{V}{Canonical direction matrix for Y (q x r)}
#'   \item{cor}{Canonical covariances}
#'   \item{loss}{The prediction error 1/n * \| XU - YV\|^2}
#'   \item{Lambda}{Canonical correlations}
#'   \item{B_opt}{Estimated reduced-rank coefficient matrix}
#' }
#' @export
cca_group_rrr <- function(X, Y, groups, 
                          Sx = NULL, Sy = NULL, Sxy = NULL, 
                          lambda = 0,  r,
                          standardize = FALSE,
                          preprocess = NULL,
                          LW_Sy = TRUE, solver = "ADMM",
                          rho = 1, niter = 1e4, thresh = 1e-4, thresh_0=1e-6,
                          matrix_free_threshold = 4000L,
                          cg_tol = 1e-6,
                          cg_maxiter = NULL,
                          verbose = FALSE) {
  preprocess_mode <- .resolve_group_preprocess_mode(standardize = standardize, preprocess = preprocess)
  X <- .preprocess_group_matrix(X, preprocess_mode)
  Y <- .preprocess_group_matrix(Y, preprocess_mode)

  n <- nrow(X)
  p <- ncol(X)
  q <- ncol(Y)
  
  if (n < min(p, q)) warning("Both X and Y are high-dimensional; method may be unstable.")

  if (is.null(Sy)) {
    Sy <- crossprod(Y) / n
    if (LW_Sy) Sy <- as.matrix(corpcor::cov.shrink(Y, verbose=verbose))
  }
  
  svd_Sy <- svd(as.matrix(Sy))
  sqrt_inv_Sy <- svd_Sy$u %*% diag(ifelse(svd_Sy$d > 1e-4, 1 / sqrt(svd_Sy$d), 0)) %*% t(svd_Sy$v)
  tilde_Y <- Y %*% sqrt_inv_Sy
  
  B_opt <- switch(solver,
                  "ADMM" = solve_group_rrr_admm(X, tilde_Y, groups=groups, lambda=lambda,
                                                Sx = Sx,
                                                rho=rho, niter=niter, 
                                                thresh=thresh, thresh_0=thresh_0, verbose=verbose,
                                                matrix_free_threshold = matrix_free_threshold,
                                                cg_tol = cg_tol,
                                                cg_maxiter = cg_maxiter),
                  "CVXR" = solve_group_rrr_cvxr(X, tilde_Y, groups, lambda, thresh_0=thresh_0),
                  stop("Unsupported solver: choose either 'ADMM' or 'CVXR'")
  )

  post <- .postprocess_rrr_fit(
    B_opt, X, Y, sqrt_inv_Sy, r,
    ridge_whiten = max(thresh_0, 1e-8),
    thresh_0 = thresh_0,
    verbose = verbose
  )

  list(
    U = post$U,
    V = post$V,
    loss = post$loss,
    cor = post$cor,
    Lambda = post$Lambda,
    B_opt = B_opt
  )
}



# Cross-validated loss for group-penalized CCA
cca_group_rrr_cv_folds <- function(X, Y, groups, Sx = NULL, Sy = NULL, kfolds = 5, 
                                   lambda = 0.01, r = 2, 
                                   preprocess = NULL,
                                   LW_Sy = FALSE, solver = "ADMM", 
                                   rho = 1, niter = 1e4, thresh = 1e-4,
                                   thresh_0 = 1e-6,
                                   matrix_free_threshold = 4000L,
                                   cg_tol = 1e-6,
                                   cg_maxiter = NULL,
                                   verbose = FALSE,
                                   folds = NULL,
                                   return_fold_values = FALSE) {
  preprocess_mode <- .resolve_group_preprocess_mode(standardize = FALSE, preprocess = preprocess)
  X <- .preprocess_group_matrix(X, preprocess_mode)
  Y <- .preprocess_group_matrix(Y, preprocess_mode)
  if (is.null(folds)) {
    folds <- .create_cv_folds(nrow(Y), kfolds)
  }

  rmse <- vapply(seq_along(folds), function(i) {
    n <- nrow(X)
    X_train <- X[-folds[[i]], ]; Y_train <- Y[-folds[[i]], ]
    X_val <- X[folds[[i]], ]; Y_val <- Y[folds[[i]], ]
    n_train <- n - nrow(X_val)
    
    if (!is.null(Sx)) {
      Sx_train <- (n * Sx - crossprod(X_val)) / n_train
    } else {
      Sx_train <- NULL
    }

    if (!is.null(Sy)) {
      Sy_train <- (n * Sy - crossprod(Y_val)) / n_train
    } else {
      Sy_train <- NULL
    }

    tryCatch({
      fit <- cca_group_rrr(X_train, Y_train, groups, Sx = Sx_train, Sy = Sy_train,
                           lambda = lambda, r = r, 
                           standardize = FALSE, preprocess = "none",
                           LW_Sy = LW_Sy, solver = solver,
                           rho = rho, niter = niter, thresh = thresh, thresh_0 = thresh_0,
                           matrix_free_threshold = matrix_free_threshold,
                           cg_tol = cg_tol,
                           cg_maxiter = cg_maxiter,
                           verbose = FALSE)
      mean((X_val %*% fit$U - Y_val %*% fit$V)^2)
    }, error = function(e) {
      message("Error in fold ", i, ": ", conditionMessage(e))
      NA_real_
    })
  }, numeric(1))
  
  if (return_fold_values) return(rmse)
  if (all(is.na(rmse))) return(1e8)
  mean(rmse, na.rm = TRUE)
}


#' Group-Sparse Canonical Correlation via Reduced-Rank Regression with CV
#'
#' Performs group-sparse reduced-rank regression for CCA using either ADMM or CVXR solvers.
#'
#' @param X Predictor matrix (n x p)
#' @param Y Response matrix (n x q)
#' @param groups List of index vectors defining groups of predictors
#' @param lambdas Grid of regularization parameters to try out
#' @param r Target rank
#' @param kfolds Nb of folds for the CV procedure
#' @param parallelize Whether to use parallel processing (default is FALSE)
#' @param standardize Backward-compatible preprocessing flag: TRUE = `"scale"`, FALSE = `"center"`.
#' @param preprocess Preprocessing mode. One of `"scale"` (center + scale), `"center"` (center only), or `"none"`.
#' @param LW_Sy Whether to apply Ledoit-Wolf shrinkage to Sy (default TRUE)
#' @param solver Either "ADMM" or "CVXR". The `"CVXR"` backend requires the
#'   optional package `CVXR`.
#' @param rho ADMM parameter
#' @param niter Maximum number of ADMM iterations
#' @param thresh Convergence threshold for ADMM
#' @param verbose Print diagnostics
#' @param thresh_0 tolerance for declaring entries non-zero
#' @param matrix_free_threshold For ADMM: when both `n` and `p` are at least this value, use a matrix-free conjugate-gradient solve instead of forming a dense linear system.
#' @param cg_tol Relative tolerance for the matrix-free conjugate-gradient solve used in ADMM.
#' @param cg_maxiter Maximum iterations for the matrix-free conjugate-gradient solve. Defaults to `min(p, 1000)`.
#' @param nb_cores Number of cores to use for parallelization (default is all available cores minus 1)
#'
#' @return A list with elements:
#' \describe{
#'   \item{U}{Canonical direction matrix for X (p x r)}
#'   \item{V}{Canonical direction matrix for Y (q x r)}
#'   \item{lambda}{Optimal regularisation parameter lambda chosen by CV}
#'   \item{rmse}{Mean squared error of prediction (as computed in the CV)}
#'   \item{cor}{Canonical covariances}
#'   \item{lambda_x}{Alias of the selected `lambda`}
#'   \item{lambda_x_se}{Foldwise standard error at the selected `lambda`}
#'   \item{lambda_y}{Placeholder for symmetry with two-penalty interfaces}
#'   \item{lambda_y_se}{Placeholder for symmetry with two-penalty interfaces}
#'   \item{resultsx}{Backward-compatible alias of `cv_summary`}
#'   \item{cv_summary}{Data frame with one row per lambda containing mean RMSE and its foldwise standard error}
#'   \item{cv_folds}{Data frame with fold-level RMSE values for each lambda}
#'   \item{Lambda}{Canonical correlations from the final fit}
#'   \item{B}{Estimated reduced-rank coefficient matrix from the final fit}
#'   \item{fit}{Final fit at the selected lambda}
#' }
#' @importFrom foreach foreach %dopar%
#' @export
cca_group_rrr_cv <- function(X, Y, groups, r = 2, 
                             lambdas = 10^seq(-3, 1.5, length.out = 10),
                             kfolds = 5, parallelize = FALSE, standardize = FALSE,
                             preprocess = NULL,
                             LW_Sy = TRUE, solver = "ADMM", rho = 1,
                             thresh_0 = 0,
                             niter = 1e4, thresh = 1e-4,
                             matrix_free_threshold = 4000L,
                             cg_tol = 1e-6,
                             cg_maxiter = NULL,
                             verbose = FALSE,
                             nb_cores = NULL) {
  preprocess_mode <- .resolve_group_preprocess_mode(standardize = standardize, preprocess = preprocess)
  
  if (nrow(X) < min(ncol(X), ncol(Y))) {
    warning("Both X and Y are high-dimensional; method may be unstable.")
  }
  
  # Apply preprocessing once before CV. Each fold reuses transformed matrices.
  X <- .preprocess_group_matrix(X, preprocess_mode)
  Y <- .preprocess_group_matrix(Y, preprocess_mode)
  
  n <- nrow(X)
  Sx <- NULL
  Sy <- if (LW_Sy) NULL else crossprod(Y) / n

  # Rebind the current source definitions into a local environment so PSOCK
  # workers use the updated functions rather than any stale installed copy.
  .rrr_cg_solve <- .rrr_cg_solve
  .rrr_chol_jitter <- .rrr_chol_jitter
  .rrr_invsqrt_psd <- .rrr_invsqrt_psd
  .rrr_whiten_factor <- .rrr_whiten_factor
  .rrr_safe_rank_svd <- .rrr_safe_rank_svd
  make_invSx_apply <- make_invSx_apply
  .postprocess_rrr_fit <- .postprocess_rrr_fit
  .resolve_group_preprocess_mode <- .resolve_group_preprocess_mode
  .preprocess_group_matrix <- .preprocess_group_matrix
  .create_cv_folds <- .create_cv_folds
  solve_group_rrr_admm <- solve_group_rrr_admm
  solve_group_rrr_cvxr <- solve_group_rrr_cvxr
  cca_group_rrr <- cca_group_rrr
  cca_group_rrr_cv_folds <- cca_group_rrr_cv_folds
  safe_mean_na <- .safe_mean_na
  safe_sd_na <- .safe_sd_na
  environment(make_invSx_apply) <- environment()
  environment(.postprocess_rrr_fit) <- environment()
  environment(solve_group_rrr_admm) <- environment()
  environment(solve_group_rrr_cvxr) <- environment()
  environment(cca_group_rrr) <- environment()
  environment(cca_group_rrr_cv_folds) <- environment()
  
  run_cv <- function(lambda) {
    fold_values <- cca_group_rrr_cv_folds(
      X, Y, groups, Sx = Sx, Sy = Sy, kfolds = kfolds,
      lambda = lambda, r = r,
      preprocess = "none",
      LW_Sy = LW_Sy, solver = solver,
      rho = rho, niter = niter, thresh = thresh, thresh_0 = thresh_0,
      matrix_free_threshold = matrix_free_threshold,
      cg_tol = cg_tol,
      cg_maxiter = cg_maxiter,
      folds = folds,
      return_fold_values = TRUE
    )

    list(
      summary = data.frame(
        lambda = lambda,
        rmse = safe_mean_na(fold_values),
        se = safe_sd_na(fold_values) / sqrt(sum(!is.na(fold_values))),
        stringsAsFactors = FALSE
      ),
      fold = data.frame(
        lambda = lambda,
        fold = seq_along(fold_values),
        rmse = fold_values,
        stringsAsFactors = FALSE
      )
    )
  }

  folds <- .create_cv_folds(n, kfolds)
  
    if (parallelize) {
      if (!requireNamespace("doParallel", quietly = TRUE)) {
      stop("Package 'doParallel' must be installed to use the parallelization option.",
          call. = FALSE)
      }

      if (!requireNamespace("crayon", quietly = TRUE)) {
      stop("Package 'crayon' must be installed to use the parallelization option.",
          call. = FALSE)
      }
    # --- GRACEFUL PARALLEL SETUP ---
      cl <- setup_parallel_backend(nb_cores)
      
      if (!is.null(cl)) {
        # If the cluster was created successfully, register it and plan to stop it
        if (verbose){
          cat(crayon::green("Parallel backend successfully registered.\n"))
        }

        initialize_parallel_workers(cl, verbose = verbose)
        doParallel::registerDoParallel(cl)
        on.exit(cleanup_parallel_backend(cl), add = TRUE)
      } else {
        # If setup_parallel_backend returned NULL, print a warning and proceed serially
        warning("All parallel setup attempts failed. Proceeding in serial mode.", immediate. = TRUE)
        parallelize <- FALSE # Ensure %dopar% runs serially
    }
   }

   if (parallelize) {


    parallel_packages <- if (solver == "CVXR") c("Matrix", "CVXR") else c("Matrix")
    results_list <- foreach(
      lambda = lambdas,
      .packages = parallel_packages,
      .export = c(
        "cca_group_rrr_cv_folds", "cca_group_rrr",
        "solve_group_rrr_admm", "solve_group_rrr_cvxr",
        "make_invSx_apply", ".rrr_cg_solve", ".rrr_chol_jitter",
        ".rrr_invsqrt_psd", ".rrr_whiten_factor", ".rrr_safe_rank_svd",
        ".postprocess_rrr_fit", ".resolve_group_preprocess_mode",
        ".preprocess_group_matrix", ".create_cv_folds",
        ".safe_mean_na", ".safe_sd_na"
      )
    ) %dopar% run_cv(lambda)
  } else {
    results_list <- lapply(lambdas, run_cv)
  }
  
  cv_summary <- dplyr::bind_rows(lapply(results_list, `[[`, "summary"))
  cv_folds <- dplyr::bind_rows(lapply(results_list, `[[`, "fold"))

  cv_summary$rmse[is.na(cv_summary$rmse) | cv_summary$rmse == 0] <- 1e8
  cv_summary <- cv_summary %>% dplyr::filter(rmse > 1e-5)
  
  opt_lambda <- cv_summary$lambda[which.min(cv_summary$rmse)]
  if (is.na(opt_lambda)) opt_lambda <- 0.1
  
  final <- cca_group_rrr(X, Y, groups, Sx = Sx, Sy = Sy, lambda = opt_lambda,
                         r = r, preprocess = "none", thresh = thresh, thresh_0 = thresh_0,
                         LW_Sy = LW_Sy, solver = solver, rho = rho, niter = niter,
                         matrix_free_threshold = matrix_free_threshold,
                         cg_tol = cg_tol,
                         cg_maxiter = cg_maxiter,
                         verbose=verbose)
  
  list(
    U = final$U,
    V = final$V,
    lambda = opt_lambda,
    rmse = cv_summary$rmse,
    cor = final$cor,
    lambda_x = opt_lambda,
    lambda_x_se = cv_summary$se[match(opt_lambda, cv_summary$lambda)],
    lambda_y = NA_real_,
    lambda_y_se = NA_real_,
    resultsx = cv_summary,
    cv_summary = cv_summary,
    cv_folds = cv_folds,
    Lambda = final$Lambda,
    B = final$B_opt,
    fit = final
  )
}

.rrr_chol_jitter <- function(G, jitter = 1e-8, max_tries = 6) {
  p <- nrow(G)
  for (k in 0:max_tries) {
    out <- tryCatch(
      chol(G + (10^k) * jitter * diag(p)),
      error = function(e) NULL
    )
    if (!is.null(out)) return(out)
  }
  stop("chol() failed even with jitter.")
}

.rrr_invsqrt_psd <- function(G, eps = 1e-8) {
  G <- (G + t(G)) / 2
  ed <- eigen(G, symmetric = TRUE)
  vals <- pmax(ed$values, eps)
  ed$vectors %*% diag(1 / sqrt(vals), nrow = length(vals)) %*% t(ed$vectors)
}

.rrr_whiten_factor <- function(G, ridge = 1e-8, verbose = FALSE) {
  if (any(!is.finite(G))) {
    if (verbose) cat("\nNon-finite entries in whitening Gram matrix; sanitizing.\n")
    G[!is.finite(G)] <- 0
    G <- (G + t(G)) / 2 + ridge * diag(nrow(G))
  }

  tryCatch({
    R <- .rrr_chol_jitter(G, jitter = ridge)
    backsolve(R, diag(nrow(G)))
  }, error = function(e) {
    if (verbose) cat("\nchol_jitter failed; using eigen whitening fallback.\n")
    .rrr_invsqrt_psd(G, eps = ridge)
  })
}

.rrr_safe_rank_svd <- function(M, k, prefer_sparse = FALSE) {
  if (k < 1) return(NULL)
  nr <- nrow(M)
  nc <- ncol(M)
  if (nr < 1 || nc < 1) return(NULL)
  k <- min(k, nr, nc)

  out <- NULL

  if (requireNamespace("RSpectra", quietly = TRUE) && k < min(nr, nc)) {
    out <- tryCatch({
      if (prefer_sparse && requireNamespace("Matrix", quietly = TRUE)) {
        RSpectra::svds(Matrix::Matrix(M, sparse = TRUE), k)
      } else {
        RSpectra::svds(M, k)
      }
    }, error = function(e) NULL)

    if (!is.null(out)) {
      ok <- all(is.finite(out$u)) && all(is.finite(out$v)) && all(is.finite(out$d))
      if (!ok) out <- NULL
    }
  }

  if (is.null(out)) {
    out <- tryCatch(svd(as.matrix(M), nu = k, nv = k), error = function(e) NULL)
    if (is.null(out)) return(NULL)
    out$u <- out$u[, seq_len(k), drop = FALSE]
    out$v <- out$v[, seq_len(k), drop = FALSE]
    out$d <- out$d[seq_len(k)]
  }

  out
}

.resolve_preprocess_mode <- function(standardize = NULL, preprocess = NULL) {
  if (is.logical(preprocess) && length(preprocess) == 1L) {
    if (isTRUE(preprocess)) {
      preprocess <- NULL
    } else {
      return("none")
    }
  }

  if (!is.null(preprocess)) {
    return(match.arg(preprocess, c("scale", "center", "none")))
  }

  if (is.character(standardize)) {
    return(match.arg(standardize, c("scale", "center", "none")))
  }

  if (is.logical(standardize) && length(standardize) == 1L) {
    return(if (isTRUE(standardize)) "scale" else "center")
  }

  if (is.numeric(standardize) && length(standardize) == 1L && is.finite(standardize)) {
    return(if (isTRUE(as.logical(standardize))) "scale" else "center")
  }

  if (is.null(standardize)) {
    return("center")
  }

  stop("`standardize` must be logical or one of: 'scale', 'center', 'none'.", call. = FALSE)
}

.preprocess_matrix <- function(M, mode) {
  M <- as.matrix(M)
  if (mode == "none") {
    return(M)
  }

  M <- scale(M, center = TRUE, scale = identical(mode, "scale"))
  M <- as.matrix(M)
  M[!is.finite(M)] <- 0
  M
}

.rrr_match_mode <- function(mode) {
  if (missing(mode) || is.null(mode) || length(mode) == 0L || all(is.na(mode))) {
    return("sqrtm_norm")
  }

  mode <- tolower(as.character(mode[[1]]))
  if (identical(mode, "new")) {
    mode <- "sqrtm_norm"
  } else if (identical(mode, "old")) {
    mode <- "product_norm"
  }

  match.arg(mode, c("sqrtm_norm", "product_norm"))
}

.rrr_match_cv_metric <- function(cv_metric) {
  if (missing(cv_metric) || is.null(cv_metric) || length(cv_metric) == 0L || all(is.na(cv_metric))) {
    metric <- "mse"
  } else {
    metric <- tolower(as.character(cv_metric[[1]]))
  }

  if (metric %in% c("mse", "rmse", "prediction")) {
    return("mse")
  }

  if (metric %in% c("correlation", "cor", "corr", "association")) {
    return("correlation")
  }

  stop(
    "Unsupported `cv_metric`. Use one of: 'mse' or 'correlation'.",
    call. = FALSE
  )
}

.rrr_correlation_score <- function(X_scores, Y_scores) {
  X_scores <- as.matrix(X_scores)
  Y_scores <- as.matrix(Y_scores)

  if (nrow(X_scores) == 0L || nrow(Y_scores) == 0L ||
      ncol(X_scores) == 0L || ncol(Y_scores) == 0L) {
    return(0)
  }

  x_sd <- apply(X_scores, 2, stats::sd)
  y_sd <- apply(Y_scores, 2, stats::sd)

  keep_x <- which(is.finite(x_sd) & x_sd > 0)
  keep_y <- which(is.finite(y_sd) & y_sd > 0)

  if (length(keep_x) == 0L || length(keep_y) == 0L) {
    return(0)
  }

  C <- stats::cor(
    X_scores[, keep_x, drop = FALSE],
    Y_scores[, keep_y, drop = FALSE]
  )
  C[!is.finite(C)] <- 0

  sum(svd(C, nu = 0, nv = 0)$d)
}


.postprocess_rrr_fit <- function(Bhat, X, Y, sqrt_inv_Sy, r,
                                 ridge_whiten = 0, thresh_0 = 1e-6,
                                 verbose = FALSE) {
  n <- nrow(X)
  p <- ncol(X)
  q <- ncol(Y)

  if (all(abs(Bhat) < thresh_0)) {
    XU <- matrix(0, n, r)
    YV <- matrix(0, n, r)
    return(list(
      U = matrix(0, p, r),
      V = matrix(0, q, r),
      Lambda = rep(0, r),
      cor = rep(0, r),
      loss = mean((YV - XU)^2)
    ))
  }

  r_eff <- min(r, nrow(Bhat), ncol(Bhat))
  SVDs <- .rrr_safe_rank_svd(Bhat, r_eff, prefer_sparse = TRUE)
  if (is.null(SVDs)) {
    stop("SVD failed during reduced-rank postprocessing.", call. = FALSE)
  }

  U0 <- SVDs$u
  V0 <- sqrt_inv_Sy %*% SVDs$v

  XU0 <- X %*% U0
  YV0 <- Y %*% V0

  GX <- crossprod(XU0) / n + ridge_whiten * diag(r_eff)
  GY <- crossprod(YV0) / n + ridge_whiten * diag(r_eff)

  U <- U0 %*% .rrr_whiten_factor(GX, ridge = ridge_whiten, verbose = verbose)
  V <- V0 %*% .rrr_whiten_factor(GY, ridge = ridge_whiten, verbose = verbose)
  

  XU <- X %*% U
  YV <- Y %*% V
  cor <- diag(crossprod(XU, YV) / n)

  neg <- cor < 0
  if (any(neg)) {
    V[, neg] <- -V[, neg, drop = FALSE]
    YV[, neg] <- -YV[, neg, drop = FALSE]
    cor[neg] <- -cor[neg]
  }

  ord <- order(cor, decreasing = TRUE)
  cor <- cor[ord]
  U <- U[, ord, drop = FALSE]
  V <- V[, ord, drop = FALSE]
  XU <- XU[, ord, drop = FALSE]
  YV <- YV[, ord, drop = FALSE]

  if (r_eff < r) {
    U <- cbind(U, matrix(0, p, r - r_eff))
    V <- cbind(V, matrix(0, q, r - r_eff))
    cor <- c(cor, rep(0, r - r_eff))
  }

  list(
    U = U,
    V = V,
    Lambda = cor,
    cor = cor,
    loss = mean((YV - XU)^2)
  )
}


.postprocess_rrr_fit_old <- function(Bhat, X, Y, sqrt_inv_Sy, r,
                                 ridge_whiten = 0, thresh_0 = 1e-6,
                                 verbose = FALSE) {
  n <- nrow(X)
  p <- ncol(X)
  q <- ncol(Y)

  if (all(abs(Bhat) < thresh_0)) {
    XU <- matrix(0, n, r)
    YV <- matrix(0, n, r)
    return(list(
      U = matrix(0, p, r),
      V = matrix(0, q, r),
      Lambda = rep(0, r),
      cor = rep(0, r),
      loss = mean((YV - XU)^2)
    ))
  }

  r_eff <- min(r, nrow(Bhat), ncol(Bhat))
  active_rows <- which(rowSums(Bhat^2) > 0)
  Sx_active = crossprod(X[, active_rows, drop = FALSE]) / n
  SVDs <- .rrr_safe_rank_svd(Bhat, r_eff, prefer_sparse = TRUE)
  if (is.null(SVDs)) {
    stop("SVD failed during reduced-rank postprocessing.", call. = FALSE)
  }
  if (length(active_rows) > r - 1) {
      sqrt_Sx <- compute_sqrt(Sx_active)
      sol <- svd(sqrt_Sx %*% Bhat[active_rows, ])
      V <- sqrt_inv_Sy %*% sol$v[, 1:r]
      inv_D <- diag(sapply(sol$d[1:r], function(d) ifelse(d < 1e-4, 0, 1 / d)))
      U <- Bhat %*% sol$v[, 1:r] %*% inv_D
    } else {
      U <- matrix(0, p, r)
      V <- matrix(0, q, r)
    }

  XU <- X %*% U
  YV <- Y %*% V
  cor <- diag(crossprod(XU, YV) / n)
  neg <- cor < 0
  if (any(neg)) {
    V[, neg] <- -V[, neg, drop = FALSE]
    YV[, neg] <- -YV[, neg, drop = FALSE]
    cor[neg] <- -cor[neg]
  }

  ord <- order(cor, decreasing = TRUE)
  cor <- cor[ord]
  U <- U[, ord, drop = FALSE]
  V <- V[, ord, drop = FALSE]
  XU <- XU[, ord, drop = FALSE]
  YV <- YV[, ord, drop = FALSE]

  if (r_eff < r) {
    U <- cbind(U, matrix(0, p, r - r_eff))
    V <- cbind(V, matrix(0, q, r - r_eff))
    cor <- c(cor, rep(0, r - r_eff))
  }

  list(
    U = U,
    V = V,
    Lambda = cor,
    cor = cor,
    loss = mean((YV - XU)^2)
  )
}


.rrr_cg_solve <- function(A_mul, b, M_inv = NULL, tol = 1e-6,
                          maxiter = NULL, x0 = NULL) {
  p <- length(b)
  if (is.null(maxiter)) {
    maxiter <- min(p, 1000L)
  }

  x <- if (is.null(x0)) numeric(p) else as.numeric(x0)
  r <- as.numeric(b) - as.numeric(A_mul(x))
  bnorm <- sqrt(sum(b^2))
  if (!is.finite(bnorm) || bnorm < .Machine$double.eps) {
    return(x)
  }

  z <- if (is.null(M_inv)) r else M_inv(r)
  rz_old <- drop(crossprod(r, z))
  if (!is.finite(rz_old) || rz_old <= 0) {
    return(x)
  }

  p_dir <- z
  tol_abs <- tol * max(1, bnorm)

  for (iter in seq_len(maxiter)) {
    Ap <- as.numeric(A_mul(p_dir))
    denom <- drop(crossprod(p_dir, Ap))
    if (!is.finite(denom) || abs(denom) < .Machine$double.eps) {
      break
    }

    alpha <- rz_old / denom
    x <- x + alpha * p_dir
    r <- r - alpha * Ap
    if (sqrt(sum(r^2)) <= tol_abs) {
      break
    }

    z <- if (is.null(M_inv)) r else M_inv(r)
    rz_new <- drop(crossprod(r, z))
    if (!is.finite(rz_new) || rz_new <= 0) {
      break
    }

    beta <- rz_new / rz_old
    p_dir <- z + beta * p_dir
    rz_old <- rz_new
  }

  x
}

# Helper: efficient (Sx + rho I)^{-1} * W without forcing an explicit Sx build.
# For very large n and p, use a matrix-free CG solve that only applies X and t(X).
make_invSx_apply <- function(X, Sx = NULL, rho, ridge = 1e-8, verbose = FALSE,
                             matrix_free_threshold = 4000L,
                             cg_tol = 1e-6,
                             cg_maxiter = NULL) {
  n <- nrow(X)
  p <- ncol(X)
  if (n < 1 || p < 1) stop("X must have positive dimensions.")

  solve_from_factor <- function(G) {
    R <- .rrr_chol_jitter(G, jitter = ridge)
    function(B) backsolve(R, forwardsolve(t(R), as.matrix(B)))
  }

  if (!is.null(Sx)) {
    if (verbose) message("Using direct Cholesky solve with supplied Sx.")
    solve_gram <- solve_from_factor(Sx + rho * diag(p))
    return(list(
      apply = solve_gram,
      strategy = "direct",
      uses_woodbury = FALSE
    ))
  }

  if (min(n, p) >= matrix_free_threshold) {
    if (verbose) message("Using matrix-free CG solve for ADMM update.")

    diag_precond <- rho + colSums(X^2) / n
    diag_precond[!is.finite(diag_precond) | diag_precond < ridge] <- ridge

    A_mul <- function(v) {
      as.numeric(rho * v + crossprod(X, X %*% v) / n)
    }
    M_inv <- function(v) {
      as.numeric(v / diag_precond)
    }

    last_sol <- NULL
    return(list(
      apply = function(W) {
        W <- as.matrix(W)
        out <- matrix(0, p, ncol(W))
        if (!is.null(last_sol) && identical(dim(last_sol), dim(out))) {
          out[,] <- last_sol
        }

        for (j in seq_len(ncol(W))) {
          out[, j] <- .rrr_cg_solve(
            A_mul, W[, j],
            M_inv = M_inv,
            tol = cg_tol,
            maxiter = cg_maxiter,
            x0 = out[, j]
          )
        }

        last_sol <<- out
        out
      },
      strategy = "matrix_free_cg",
      uses_woodbury = FALSE
    ))
  }

  if (p <= n) {
    if (verbose) message("Using direct Cholesky solve from X without precomputing Sx upstream.")
    solve_gram <- solve_from_factor(crossprod(X) / n + rho * diag(p))
    return(list(
      apply = solve_gram,
      strategy = "direct",
      uses_woodbury = FALSE
    ))
  }

  if (verbose) message("Using Woodbury solve for ADMM update.")
  M <- diag(n) + tcrossprod(X) / (n * rho)
  R <- tryCatch(chol(M + ridge * diag(n)), error = function(e) NULL)

  solve_M <- if (!is.null(R)) {
    function(B) backsolve(R, forwardsolve(t(R), B))
  } else {
    if (verbose) message("Cholesky failed in Woodbury solve; using solve().")
    function(B) solve(M, B)
  }

  list(
    apply = function(W) {
      XW <- (X %*% W) / n
      tmp <- solve_M(XW)
      (1 / rho) * W - (1 / rho^2) * crossprod(X, tmp)
    },
    strategy = "woodbury",
    uses_woodbury = TRUE
  )
}



# Helper: ADMM-based group sparse solver
solve_rrr_admm <- function(X, tilde_Y, Sx = NULL, lambda, rho=1, niter=10, thresh, verbose = FALSE, thresh_0 = 0,
                           invSx_apply = NULL,
                           matrix_free_threshold = 4000L,
                           cg_tol = 1e-6,
                           cg_maxiter = NULL) {
  p <- ncol(X); 
  q <- ncol(tilde_Y)
  n <- nrow(X)

  if (is.null(invSx_apply)) {
    invSx_info <- make_invSx_apply(
      X, Sx, rho, verbose = verbose,
      matrix_free_threshold = matrix_free_threshold,
      cg_tol = cg_tol,
      cg_maxiter = cg_maxiter
    )
    invSx_apply <- invSx_info$apply
    if (verbose) message("ADMM linear solver strategy: ", invSx_info$strategy)
  }

  U <- Z <- matrix(0, p, q)
  
  prod_xy <- crossprod(X, tilde_Y) / n ### size is p x q, so okay since q <<< n
  for (i in seq_len(niter)) {
    U_old <- U; Z_old <- Z
    B <- invSx_apply(prod_xy + rho * (Z - U))
    Z <- B + U
    norm_col <- sqrt(rowSums(Z^2))
    shrinkage <- pmax(0, 1 - (lambda / rho) / norm_col)
    shrinkage[is.nan(shrinkage)] <- 0
    Z <- sweep(Z, 1, shrinkage, '*')
    U <- U + B - Z
    if (verbose) cat("ADMM iter", i, "Primal: ", norm(Z - B), "Dual:", norm(Z_old - Z), "\n")
    if (max(c(norm(Z - B) / sqrt(p), norm(Z_old - Z) / sqrt(p))) < thresh) break
  }
  B_opt <- B
  
  B_opt[abs(B_opt) < thresh_0] <- 0
  return(B_opt)
}


# Helper: CVXR-based group sparse solver
solve_rrr_cvxr <- function(X, tilde_Y, lambda, thresh_0=1e-6,
                           rho = 1, niter = 1e4, thresh = 1e-4,
                           verbose = FALSE,
                           matrix_free_threshold = 4000L,
                           cg_tol = 1e-6,
                           cg_maxiter = NULL) {
  if (!nzchar(system.file(package = "CVXR"))) {
    stop("Package 'CVXR' must be installed to use the CVX solver.",
         call. = FALSE)
  }

  p <- ncol(X); q <- ncol(tilde_Y)
  n <- nrow(X)

  fallback_admm <- function(reason = NULL) {
    if (verbose && !is.null(reason)) {
      message("Falling back to ADMM from CVXR: ", reason)
    }
    solve_rrr_admm(
      X, tilde_Y, Sx = NULL, lambda = lambda, rho = rho,
      niter = niter, thresh = thresh, thresh_0 = thresh_0,
      verbose = verbose,
      matrix_free_threshold = matrix_free_threshold,
      cg_tol = cg_tol,
      cg_maxiter = cg_maxiter
    )
  }

  # CVXR becomes impractical quickly because canonicalization densifies the
  # quadratic objective. For larger problems, the ADMM path is both faster and
  # dramatically more memory efficient.
  if (n * p * q > 5e6) {
    return(fallback_admm("problem size is too large for CVXR canonicalization"))
  }

  cvxr_variable <- getExportedValue("CVXR", "Variable")
  cvxr_norm <- getExportedValue("CVXR", "norm")
  cvxr_minimize <- getExportedValue("CVXR", "Minimize")
  cvxr_sum_squares <- getExportedValue("CVXR", "sum_squares")
  cvxr_problem <- getExportedValue("CVXR", "Problem")
  cvxr_psolve <- getExportedValue("CVXR", "psolve")
  cvxr_solver <- .cvxr_preferred_solver()

  B <- cvxr_variable(c(p, q))
  
  row_penalty <- Reduce(
    `+`,
    lapply(seq_len(p), function(j) cvxr_norm(B[j, , drop = FALSE], 2))
  )
  objective <- cvxr_minimize(
    1 / n * cvxr_sum_squares(tilde_Y - X %*% B) + lambda * row_penalty
  )
  B_opt <- tryCatch({
    problem <- cvxr_problem(objective)
    result <- if (is.null(cvxr_solver)) {
      cvxr_psolve(problem)
    } else {
      cvxr_psolve(problem, solver = cvxr_solver)
    }
    result$getValue(B)
  }, error = function(e) {
    msg <- conditionMessage(e)
    if (grepl("memory|vector memory|sparse->dense|cannot allocate", msg, ignore.case = TRUE)) {
      return(fallback_admm(msg))
    }
    if (.cvxr_mosek_error(e)) {
      .stop_unusable_mosek()
    }
    stop(e)
  })
  
  B_opt[abs(B_opt) < thresh_0] <- 0
  return(B_opt)
}


#' Canonical Correlation Analysis via Reduced Rank Regression (RRR)
#'
#' Estimates canonical directions using various RRR solvers and penalties.
#'
#' @param X Matrix of predictors.
#' @param Y Matrix of responses.
#' @param Sx Optional X covariance matrix.
#' @param Sy Optional Y covariance matrix.
#' @param lambda Regularization parameter.
#' @param r Rank of the solution.
#' @param mode Mode for postprocessing the RRR solution. One of `"sqrtm_norm"`
#'   (default) or `"product_norm"`. Legacy aliases `"new"` and `"old"` are
#'   also accepted. The former whitens the canonical variates to have identity
#'   covariance, while the latter does not whiten and instead returns the raw
#'   SVD factors of the RRR solution. The `"product_norm"` mode may be more
#'   interpretable in some cases but can yield canonical variates with very
#'   different scales and is not guaranteed to be numerically stable when the
#'   RRR solution is very low-rank or nearly low-rank.
#' @param highdim Boolean for high-dimensional regime.
#' @param solver Solver type: "rrr", "CVX", or "ADMM". The `"CVX"` backend
#'   requires the optional package `CVXR`.
#' @param LW_Sy Whether to use Ledoit-Wolf shrinkage for Sy.
#' @param standardize Backward-compatible preprocessing flag: TRUE = `"scale"`, FALSE = `"center"`.
#' @param preprocess Preprocessing mode. One of `"scale"` (center + scale), `"center"` (center only), or `"none"`.
#'   Logical values are still accepted for backward compatibility: `TRUE` uses `standardize`, `FALSE` skips preprocessing.
#' @param rho ADMM parameter.
#' @param niter Maximum number of iterations for ADMM.
#' @param thresh Convergence threshold.
#' @param thresh_0 For the ADMM solver: Set entries whose absolute value is below this to 0 (default 1e-6).
#' @param matrix_free_threshold For ADMM: when both `n` and `p` are at least this value, use a matrix-free conjugate-gradient solve instead of forming a dense linear system.
#' @param cg_tol Relative tolerance for the matrix-free conjugate-gradient solve used in ADMM.
#' @param cg_maxiter Maximum iterations for the matrix-free conjugate-gradient solve. Defaults to `min(p, 1000)`.
#' @param verbose Logical for verbose output.
#'
#' @return A list with elements:
#' \itemize{
#'   \item U: Canonical direction matrix for X (p x r)
#'   \item V: Canonical direction matrix for Y (q x r)
#'   \item cor: Canonical covariances
#'   \item loss: The prediction error 1/n * || XU - YV ||^2
#' }

#' @export
cca_rrr <- function(X, Y, Sx=NULL, Sy=NULL,
                    lambda = 0, 
                    r, highdim=TRUE, 
                    solver="ADMM",
                    LW_Sy = TRUE,
                    mode = "sqrtm_norm",
                    standardize = FALSE,
                    preprocess = NULL,
                    rho=1,
                    niter=1e4,
                    thresh = 1e-4, thresh_0 = 0,
                    matrix_free_threshold = 4000L,
                    cg_tol = 1e-6,
                    cg_maxiter = NULL,
                    verbose=FALSE) {
  
  preprocess_mode <- .resolve_preprocess_mode(standardize = standardize, preprocess = preprocess)
  mode <- .rrr_match_mode(mode)
  X <- .preprocess_matrix(X, preprocess_mode)
  Y <- .preprocess_matrix(Y, preprocess_mode)

  n <- nrow(X)
  p <- ncol(X)
  
  if (is.null(Sx) && !highdim) Sx <- crossprod(X) / n
  if (is.null(Sy)) {
    Sy <- crossprod(Y) / n
    if (LW_Sy) Sy <- as.matrix(corpcor::cov.shrink(Y, verbose=verbose))
  }
  
  sqrt_inv_Sy <- compute_sqrt_inv(Sy)
  tilde_Y <- Y %*% sqrt_inv_Sy
  
  if (!highdim) {
    if (verbose) print("Not using highdim")
    B_fit <- solve(Sx, crossprod(X, tilde_Y) / n)
  } else {
    if (solver == "CVX") {
      if(verbose){ print("Using CVX solver")}
      B_fit <- solve_rrr_cvxr(
        X, tilde_Y, lambda, thresh_0 = thresh_0,
        rho = rho, niter = niter, thresh = thresh, verbose = verbose,
        matrix_free_threshold = matrix_free_threshold,
        cg_tol = cg_tol,
        cg_maxiter = cg_maxiter
      )
    } else if (solver == "ADMM") {
      if(verbose){print("Using ADMM solver")}
      B_fit <- solve_rrr_admm(X, tilde_Y, Sx = Sx,
                              lambda=lambda, rho=rho, niter=niter, 
                              thresh=thresh, thresh_0=thresh_0,
                              verbose = verbose,
                              matrix_free_threshold = matrix_free_threshold,
                              cg_tol = cg_tol,
                              cg_maxiter = cg_maxiter)
    } else {
      if(verbose){print("Using gglasso solver")}
        if (!requireNamespace("rrpack", quietly = TRUE)) {
          stop("Package 'rrpack' must be installed to use the rrpack solver.",
              call. = FALSE)
        }
      fit <- rrpack::cv.srrr(tilde_Y, X, nrank = r, method = "glasso", nfold = 2,
                              modstr = list("lamA" = rep(lambda, 10), "nlam" = 10))
      B_fit <- fit$coef
    }
  }
  
  B_fit[abs(B_fit) < thresh_0] <- 0
  if(mode =="product_norm"){
    post <- .postprocess_rrr_fit_old(
    B_fit, X, Y, sqrt_inv_Sy, r,
    ridge_whiten = max(thresh_0, 1e-18),
    thresh_0 = thresh_0,
    verbose = verbose
  )

  }else{
    post <- .postprocess_rrr_fit(
    B_fit, X, Y, sqrt_inv_Sy, r,
    ridge_whiten = max(thresh_0, 1e-18),
    thresh_0 = thresh_0,
    verbose = verbose
  )

  }


  list(U = post$U, V = post$V, Lambda = post$Lambda, loss = post$loss, cor = post$cor, B_opt = B_fit)
}




#' Cross-validated Canonical Correlation Analysis via RRR
#'
#' Performs cross-validation to select optimal lambda, fits CCA_rrr.
#' Canonical Correlation Analysis via Reduced Rank Regression (RRR)
#' @param X Matrix of predictors.
#' @param Y Matrix of responses.
#' @param r Rank of the solution.
#' @param kfolds Number of folds for cross-validation.
#' @param lambdas Sequence of lambda values for cross-validation.
#' @param parallelize Logical; should cross-validation be parallelized?
#' @param mode Mode for postprocessing the RRR solution. One of `"sqrtm_norm"`
#'   (default) or `"product_norm"`. Legacy aliases `"new"` and `"old"` are
#'   also accepted. The former whitens the canonical variates to have identity
#'   covariance, while the latter does not whiten and instead returns the raw
#'   SVD factors of the RRR solution. The `"product_norm"` mode may be more
#'   interpretable in some cases but can yield canonical variates with very
#'   different scales and is not guaranteed to be numerically stable when the
#'   RRR solution is very low-rank or nearly low-rank.
#' @param solver Solver type: "rrr", "CVX", or "ADMM". The `"CVX"` backend
#'   requires the optional package `CVXR`.
#' @param LW_Sy Whether to use Ledoit-Wolf shrinkage for Sy.
#' @param standardize Backward-compatible preprocessing flag: TRUE = `"scale"`, FALSE = `"center"`.
#' @param preprocess Preprocessing mode. One of `"scale"` (center + scale), `"center"` (center only), or `"none"`.
#'   Logical values are still accepted for backward compatibility: `TRUE` uses `standardize`, `FALSE` skips preprocessing.
#' Data are preprocessed once up front, and fold fits reuse those transformed matrices without re-centering or re-scaling inside each fold.
#' @param rho ADMM parameter.
#' @param niter Maximum number of iterations for ADMM.
#' @param thresh Convergence threshold.
#' @param cv_metric Cross-validation metric. Use `"mse"` to minimize held-out prediction
#'   error or `"correlation"` to maximize held-out association between `X %*% U`
#'   and `Y %*% V`.
#' @param verbose Logical for verbose output.
#' @param thresh_0 tolerance for declaring entries non-zero
#' @param matrix_free_threshold For ADMM: when both `n` and `p` are at least this value, use a matrix-free conjugate-gradient solve instead of forming a dense linear system.
#' @param cg_tol Relative tolerance for the matrix-free conjugate-gradient solve used in ADMM.
#' @param cg_maxiter Maximum iterations for the matrix-free conjugate-gradient solve. Defaults to `min(p, 1000)`.
#' @param nb_cores Number of cores to use for parallelization. Defaults to min(kfolds, available cores minus 1).
#'
#' @return A list with elements:
#' \describe{
#'   \item{U}{Canonical direction matrix for X (p x r)}
#'   \item{V}{Canonical direction matrix for Y (q x r)}
#'   \item{lambda}{Optimal regularisation parameter lambda chosen by CV}
#'   \item{rmse}{Backward-compatible optimization objective. For `cv_metric = "mse"`
#'   this is the held-out mean squared error; for `cv_metric = "correlation"` it is
#'   `-cv_score` so that smaller still means better.}
#'   \item{cv_score}{Raw held-out cross-validation score averaged across folds.}
#'   \item{cv_metric}{The metric used to score lambdas during cross-validation.}
#'   \item{cor}{Canonical correlations at the selected lambda}
#'   \item{lambda_x}{Alias of the selected `lambda`}
#'   \item{lambda_x_se}{Foldwise standard error at the selected `lambda`}
#'   \item{lambda_y}{Placeholder for symmetry with two-penalty interfaces}
#'   \item{lambda_y_se}{Placeholder for symmetry with two-penalty interfaces}
#'   \item{resultsx}{Backward-compatible alias of `cv_summary`}
#'   \item{cv_summary}{Data frame with one row per lambda containing the mean CV score
#'   and its foldwise standard error.}
#'   \item{cv_folds}{Data frame with fold-level CV scores for each lambda.}
#'   \item{Lambda}{Canonical correlations from the final fit}
#'   \item{B}{Estimated reduced-rank coefficient matrix from the final fit}
#'   \item{fit}{Final fit at the selected lambda}
#' }
#' @importFrom foreach foreach %dopar%
#' @importFrom foreach foreach %do%
#' @export
cca_rrr_cv <- function(X, Y, 
                       r=2, 
                       lambdas=10^seq(-3, 1.5, length.out = 100),
                       kfolds=10,
                       solver="ADMM",
                       mode = "sqrtm_norm",
                       parallelize = FALSE,
                       LW_Sy = TRUE,
                       standardize = FALSE,
                       preprocess = NULL,
                       cv_metric = "mse",
                       rho=1,
                       thresh_0=0,
                       niter=1e4,
                       matrix_free_threshold = 4000L,
                       cg_tol = 1e-6,
                       cg_maxiter = NULL,
                       thresh = 1e-4, verbose=FALSE,
                       nb_cores = NULL) {
  preprocess_mode <- .resolve_preprocess_mode(standardize = standardize, preprocess = preprocess)
  mode <- .rrr_match_mode(mode)
  cv_metric <- .rrr_match_cv_metric(cv_metric)
  X <- .preprocess_matrix(X, preprocess_mode)
  Y <- .preprocess_matrix(Y, preprocess_mode)

  n <- nrow(X)
  Sx <- NULL
  Sy <- if (LW_Sy) NULL else crossprod(Y) / n
  folds <- .create_cv_folds(n, kfolds)
  safe_mean_na <- .safe_mean_na
  safe_sd_na <- .safe_sd_na
  cv_function <- function(lambda) {
    fold_values <- cca_rrr_cv_folds(
      X, Y, Sx = Sx, Sy = Sy, kfolds = kfolds,
      LW_Sy = LW_Sy,
      preprocess = "none",
      cv_metric = cv_metric,
      mode = mode,
      lambda = lambda, r = r, solver = solver,
      rho = rho, niter = niter, thresh = thresh,
      matrix_free_threshold = matrix_free_threshold,
      cg_tol = cg_tol,
      cg_maxiter = cg_maxiter,
      thresh_0 = thresh_0, folds = folds,
      return_fold_values = TRUE
    )

    fold_objective <- if (cv_metric == "mse") fold_values else -fold_values

    list(
      summary = data.frame(
        lambda = lambda,
        rmse = safe_mean_na(fold_objective),
        cv_score = safe_mean_na(fold_values),
        cv_metric = cv_metric,
        se = safe_sd_na(fold_values) / sqrt(sum(!is.na(fold_values))),
        stringsAsFactors = FALSE
      ),
      fold = data.frame(
        lambda = lambda,
        fold = seq_along(fold_values),
        rmse = fold_objective,
        cv_score = fold_values,
        cv_metric = cv_metric,
        stringsAsFactors = FALSE
      )
    )
  }

  foreach <- foreach::foreach
  `%dopar%` <- foreach::`%dopar%`
  
  if (parallelize && solver %in% c("CVX", "CVXR", "ADMM")) {

    if (!requireNamespace("doParallel", quietly = TRUE)) {
    stop("Package 'doParallel' must be installed to use the parallelization option.",
         call. = FALSE)
    }

    if (!requireNamespace("crayon", quietly = TRUE)) {
    stop("Package 'crayon' must be installed to use the parallelization option.",
         call. = FALSE)
    }

    # --- GRACEFUL PARALLEL SETUP ---
      available_cores_str <- Sys.getenv("SLURM_CPUS_PER_TASK")
      available_cores <- if (available_cores_str == "") {
        parallel::detectCores()
      } else {
        as.integer(available_cores_str)
      }
      available_cores <- max(1L, available_cores)
      if (is.null(nb_cores)) {
        nb_cores_effective <- min(kfolds, length(lambdas), available_cores - 1L)
      } else {
        nb_cores_effective <- min(as.integer(nb_cores), kfolds, length(lambdas))
      }
      nb_cores_effective <- max(1L, nb_cores_effective)
      nb_cores_effective <- .limit_parallel_cores(nb_cores_effective)
      cl <- setup_parallel_backend(nb_cores_effective)
      
      if (!is.null(cl)) {
        # If the cluster was created successfully, register it and plan to stop it
        if (verbose){
          cat(crayon::green("Parallel backend successfully registered.\n"))
        }
        initialize_parallel_workers(cl, verbose = verbose)
        doParallel::registerDoParallel(cl)
        on.exit(cleanup_parallel_backend(cl), add = TRUE)
      } else {
        # If setup_parallel_backend returned NULL, print a warning and proceed serially
        warning("All parallel setup attempts failed. Proceeding in serial mode.", immediate. = TRUE)
        parallelize <- FALSE # Ensure %dopar% runs serially
    }
   }

  if (parallelize && solver %in% c("CVX", "CVXR", "ADMM")) {
    rrr_parallel_exports <- c(
      "cca_rrr_cv_folds", "cca_rrr", "solve_rrr_admm", "solve_rrr_cvxr",
      "make_invSx_apply", ".rrr_cg_solve", ".rrr_chol_jitter", ".resolve_preprocess_mode",
      ".preprocess_matrix", ".create_cv_folds",
      ".rrr_invsqrt_psd", ".rrr_whiten_factor", ".rrr_safe_rank_svd",
      ".postprocess_rrr_fit", "compute_sqrt_inv", ".rrr_match_cv_metric",
      ".rrr_correlation_score",
      ".safe_mean_na", ".safe_sd_na"
    )
    if (solver  %in% c("CVX", "CVXR" )){
        if (!nzchar(system.file(package = "CVXR"))) {
         stop("Package 'CVXR' must be installed to use the CVXR/CVX solver.",
              call. = FALSE)
       }

      results_list <- foreach(
        lambda = lambdas,
        .packages = c("CVXR", "Matrix"),
        .export = rrr_parallel_exports
      ) %dopar% {
        cv_function(lambda)
      }

    }else{
      results_list <- foreach(
        lambda = lambdas,
        .export = rrr_parallel_exports
      ) %dopar% {
        cv_function(lambda)
      }
    }
    
  } else {
    results_list <- lapply(lambdas, cv_function)
  }
  
  cv_summary <- dplyr::bind_rows(lapply(results_list, `[[`, "summary"))
  cv_folds <- dplyr::bind_rows(lapply(results_list, `[[`, "fold"))

  if (cv_metric == "mse") {
    cv_summary <- cv_summary %>% 
      dplyr::mutate(rmse = ifelse(is.na(rmse) | rmse == 0, 1e8, rmse)) %>%
      dplyr::filter(rmse > 1e-5)
  } else {
    cv_summary <- cv_summary[is.finite(cv_summary$cv_score), , drop = FALSE]
  }
  
  resultsx <- cv_summary

  if (nrow(cv_summary) == 0L) {
    stop("No valid lambda values were found during cross-validation.", call. = FALSE)
  }

  opt_lambda <- if (cv_metric == "mse") {
    cv_summary$lambda[which.min(cv_summary$rmse)]
  } else {
    cv_summary$lambda[which.max(cv_summary$cv_score)]
  }
  opt_lambda <- ifelse(is.na(opt_lambda), 0.1, opt_lambda)

  final <- cca_rrr(X, Y, Sx=NULL, Sy=NULL, 
                   lambda=opt_lambda, r=r,
                   highdim=TRUE, solver=solver,
                   mode =mode,
                   standardize=FALSE, preprocess="none",
                   LW_Sy=LW_Sy, rho=rho, niter=niter, 
                   matrix_free_threshold = matrix_free_threshold,
                   cg_tol = cg_tol,
                   cg_maxiter = cg_maxiter,
                   thresh=thresh, thresh_0=thresh_0,
                   verbose=verbose)


  return(list(U = final$U, 
       V = final$V,
       lambda = opt_lambda,
       rmse = cv_summary$rmse,
       cv_score = cv_summary$cv_score,
       cv_metric = cv_metric,
       cor = final$cor,
       lambda_x = opt_lambda,
       lambda_x_se = cv_summary$se[match(opt_lambda, cv_summary$lambda)],
       lambda_y = NA_real_,
       lambda_y_se = NA_real_,
       resultsx = resultsx,
       cv_summary = cv_summary,
       cv_folds = cv_folds,
       Lambda = final$Lambda,
       B = final$B_opt,
       fit = final
       ))
}




cca_rrr_cv_folds <- function(X, Y, Sx, Sy, kfolds=5,
                             lambda=0.01,
                             r=2,
                             solver = "ADMM",
                             mode = "sqrtm_norm",
                             standardize = FALSE,
                             preprocess = "none",
                             cv_metric = "mse",
                             rho=1,
                             LW_Sy = TRUE,
                             niter=1e4,
                             thresh_0=0,
                             matrix_free_threshold = 4000L,
                             cg_tol = 1e-6,
                             cg_maxiter = NULL,
                             thresh = 1e-4,
                             folds = NULL,
                             return_fold_values = FALSE) {
  if (is.null(folds)) {
    folds <- .create_cv_folds(nrow(Y), kfolds)
  }

  preprocess_mode <- .resolve_preprocess_mode(standardize = FALSE, preprocess = preprocess)
  mode <- .rrr_match_mode(mode)
  cv_metric <- .rrr_match_cv_metric(cv_metric)
  X <- .preprocess_matrix(X, preprocess_mode)
  Y <- .preprocess_matrix(Y, preprocess_mode)

  rmse <- vapply(seq_along(folds), function(i) {
    n <- nrow(X)
    X_train <- X[-folds[[i]], ]; Y_train <- Y[-folds[[i]], ]
    X_val <- X[folds[[i]], ]; Y_val <- Y[folds[[i]], ]
    n_train <- n - nrow(X_val)
    
    if (is.null(Sx) == FALSE) {
      Sx_train <- (n * Sx - crossprod(X_val)) / n_train
    } else {
      Sx_train <- NULL
    }

    if (is.null(Sy) == FALSE) {
      Sy_train <- (n * Sy - crossprod(Y_val)) / n_train
    } else {
      Sy_train <- NULL
    }
    
    tryCatch({
      final <- cca_rrr(X_train, Y_train, 
                       Sx=Sx_train, Sy=Sy_train, highdim=TRUE,
                       mode =mode,
                       lambda=lambda, r=r, solver=solver,
                       LW_Sy=LW_Sy, standardize=FALSE, preprocess="none",
                       rho=rho, niter=niter, 
                       matrix_free_threshold = matrix_free_threshold,
                       cg_tol = cg_tol,
                       cg_maxiter = cg_maxiter,
                       thresh=thresh, thresh_0=thresh_0,
                       verbose=FALSE)
      if (cv_metric == "mse") {
        mean((X_val %*% final$U - Y_val %*% final$V)^2)
      } else {
        .rrr_correlation_score(X_val %*% final$U, Y_val %*% final$V)
      }
    }, error = function(e) {
      message("Error in fold ", i, ": ", conditionMessage(e))
      NA_real_
    })
  }, numeric(1))
  
  if (return_fold_values) return(rmse)
  if (all(is.na(rmse))) {
    return(if (cv_metric == "mse") 1e8 else NA_real_)
  }
  mean(rmse, na.rm = TRUE)
}

#' Function to perform regular (low dimensional) canonical correlation analysis (CCA
#' @param X Matrix of predictors (n x p)
#' @param Y Matrix of responses (n x q)
#' @param rank Number of canonical components to extract
#' @return A list with elements:
#' \describe{
#'   \item{U}{Canonical direction matrix for X (p x r)}
#'   \item{V}{Canonical direction matrix for Y (q x r)}
#'   \item{cor}{Canonical covariances}
#' }
#' @export
regular_cca <- function(X, Y, rank){
  Sigma_x = t(X)%*% X
  Sigma_y = t(Y)%*% Y
  Sigma_xy = t(X)%*% Y
  svd_x = svd(Sigma_x)
  inv_sqrt_Sigma_x = svd_x$u %*% diag(sapply(svd_x$d, function(x){ifelse(x<1e-5, 0, 1/sqrt(x))})) %*% t(svd_x$v)
  svd_y = svd(Sigma_y)
  inv_sqrt_Sigma_y = svd_y$u %*% diag(sapply(svd_y$d, function(x){ifelse(x<1e-5, 0, 1/sqrt(x))})) %*% t(svd_y$v)
  svd_for_cca = svd(inv_sqrt_Sigma_x %*% Sigma_xy %*% inv_sqrt_Sigma_y)
  xcoef = inv_sqrt_Sigma_x %*% svd_for_cca$u[, 1:rank]
  ycoef = inv_sqrt_Sigma_y %*% svd_for_cca$v[, 1:rank]
  cancor <- svd_for_cca$d
  return(list(xcoef=xcoef, ycoef=ycoef, cancor = cancor))
}

# -----------------------------
# Helper: robust chol with jitter
# -----------------------------
chol_jitter <- function(G, jitter = 1e-8, max_tries = 6) {
  p <- nrow(G)
  for (k in 0:max_tries) {
    out <- tryCatch(
      chol(G + (10^k) * jitter * diag(p)),
      error = function(e) NULL
    )
    if (!is.null(out)) return(out)
  }
  stop("chol() failed even with jitter.")
}

# -----------------------------
# Helper: inverse square root for PSD matrices (eigen fallback)
# -----------------------------
invsqrt_psd <- function(G, eps = 1e-8) {
  G <- (G + t(G)) / 2
  ed <- eigen(G, symmetric = TRUE)
  vals <- pmax(ed$values, eps)
  ed$vectors %*% diag(1 / sqrt(vals), nrow = length(vals)) %*% t(ed$vectors)
}

# -----------------------------
# Helper: whitening factor G^{-1/2}, with chol then eigen fallback
# -----------------------------
whiten_factor <- function(G, ridge = 1e-8, verbose = FALSE) {
  if (any(!is.finite(G))) {
    if (verbose) cat("\nNon-finite entries in whitening Gram matrix; sanitizing.\n")
    G[!is.finite(G)] <- 0
    G <- (G + t(G)) / 2 + ridge * diag(nrow(G))
  }
  tryCatch({
    R <- chol_jitter(G, jitter = ridge)
    backsolve(R, diag(nrow(G)))
  }, error = function(e) {
    if (verbose) cat("\nchol_jitter failed; using eigen whitening fallback.\n")
    invsqrt_psd(G, eps = ridge)
  })
}

# -----------------------------
# Helper: stable rank-k SVD with fallback
# -----------------------------
safe_rank_svd <- function(M, k, prefer_sparse = FALSE) {
  if (k < 1) return(NULL)
  nr <- nrow(M); nc <- ncol(M)
  if (nr < 1 || nc < 1) return(NULL)
  k <- min(k, nr, nc)

  out <- NULL

  if (requireNamespace("RSpectra", quietly = TRUE) && k < min(nr, nc)) {
    out <- tryCatch({
      if (prefer_sparse && requireNamespace("Matrix", quietly = TRUE)) {
        RSpectra::svds(Matrix::Matrix(M, sparse = TRUE), k)
      } else {
        RSpectra::svds(M, k)
      }
    }, error = function(e) NULL)

    if (!is.null(out)) {
      ok <- all(is.finite(out$u)) && all(is.finite(out$v)) && all(is.finite(out$d))
      if (!ok) out <- NULL
    }
  }

  if (is.null(out)) {
    out <- tryCatch(svd(as.matrix(M), nu = k, nv = k), error = function(e) NULL)
    if (is.null(out)) return(NULL)
    out$u <- out$u[, seq_len(k), drop = FALSE]
    out$v <- out$v[, seq_len(k), drop = FALSE]
    out$d <- out$d[seq_len(k)]
  }

  out
}

#' Efficient CCA for Two High-Dimensional Views
#'
#' Fits sparse canonical directions with an ADMM-based reduced-rank regression
#' formulation tailored to the setting where both views are high-dimensional.
#'
#' @param X Predictor matrix (n x p).
#' @param Y Response matrix (n x q).
#' @param lambda Regularization parameter.
#' @param groups Optional group structure for blockwise sparsity.
#' @param r Target rank.
#' @param standardize Whether to scale variables after centering.
#' @param rho ADMM penalty parameter.
#' @param B0 Optional warm start for the coefficient matrix.
#' @param eps Convergence tolerance for ADMM.
#' @param maxiter Maximum number of ADMM iterations.
#' @param verbose Whether to print diagnostics.
#' @param epsilon_sv Numerical threshold used to discard near-zero singular values.
#' @param ridge_whiten Ridge added when whitening Gram matrices.
#'
#' @return A list containing the estimated canonical directions, canonical
#'   correlations, the fitted coefficient matrix, preprocessing metadata, and
#'   convergence information.
#' @export
ecca <- function(
  X, Y,
  lambda = 0,
  groups = NULL,
  r = 2,
  standardize = FALSE,
  rho = 1,
  B0 = NULL,
  eps = 1e-4,
  maxiter = 500,
  verbose = TRUE,
  epsilon_sv = 1e-8,
  ridge_whiten = 1e-8
) {
  if (!is.matrix(X)) X <- as.matrix(X)
  if (!is.matrix(Y)) Y <- as.matrix(Y)
  if (nrow(X) != nrow(Y)) stop("X and Y must have the same number of rows.")

  n  <- nrow(X)
  p0 <- ncol(X)
  q0 <- ncol(Y)
  x_names <- colnames(X)
  y_names <- colnames(Y)

  if (verbose) cat("eps =", eps, " maxiter =", maxiter, "\n")

  # ---- preprocess: center, drop 0-var, optional scale
  x_center <- colMeans(X)
  y_center <- colMeans(Y)
  X <- sweep(X, 2, x_center, "-")
  Y <- sweep(Y, 2, y_center, "-")

  sx <- matrixStats::colSds(X)
  sy <- matrixStats::colSds(Y)

  keepX <- which(is.finite(sx) & sx > 0)
  keepY <- which(is.finite(sy) & sy > 0)

  dropX <- setdiff(seq_len(p0), keepX)
  dropY <- setdiff(seq_len(q0), keepY)

  X <- X[, keepX, drop = FALSE]
  Y <- Y[, keepY, drop = FALSE]

  x_scale <- rep(1, p0)
  y_scale <- rep(1, q0)
  if (standardize) {
    X <- sweep(X, 2, sx[keepX], "/")
    Y <- sweep(Y, 2, sy[keepY], "/")
    x_scale[keepX] <- sx[keepX]
    y_scale[keepY] <- sy[keepY]
  }

  p <- ncol(X)
  q <- ncol(Y)

  if (p < 1 || q < 1) {
    return(list(
      U = matrix(NA_real_, p0, r),
      V = matrix(NA_real_, q0, r),
      cor = rep(0, r),
      loss = Inf,
      Bhat = matrix(0, p0, q0),
      keepX = keepX, keepY = keepY, dropX = dropX, dropY = dropY,
      center = list(X = x_center, Y = y_center),
      scale  = list(X = x_scale,  Y = y_scale),
      converged = FALSE
    ))
  }

  # ---- map coord groups from ORIGINAL coords to filtered coords
  if (!is.null(groups)) {
    mapX <- integer(p0); mapX[keepX] <- seq_along(keepX)
    mapY <- integer(q0); mapY[keepY] <- seq_along(keepY)

    groups <- lapply(groups, function(idx) {
      if (is.matrix(idx) && ncol(idx) == 2) {
        ix <- mapX[idx[, 1]]
        iy <- mapY[idx[, 2]]
        ok <- (ix > 0) & (iy > 0)
        if (!any(ok)) return(matrix(integer(0), ncol = 2))
        cbind(ix[ok], iy[ok])
      } else {
        idx
      }
    })
  }

  # ---- linearize groups once + precompute sqrt(|g|)
  group_sqrt <- NULL
  if (!is.null(groups)) {
    groups_lin <- vector("list", length(groups))
    group_sqrt <- numeric(length(groups))
    for (g in seq_along(groups)) {
      idx <- groups[[g]]
      if (is.matrix(idx) && ncol(idx) == 2) {
        groups_lin[[g]] <- as.integer(idx[, 1] + (idx[, 2] - 1L) * p)
        group_sqrt[g] <- sqrt(nrow(idx))
      } else {
        groups_lin[[g]] <- idx
        group_sqrt[g] <- sqrt(length(idx))
      }
    }
    groups <- groups_lin
  }

  # ---- reduced bases via SVD
  EDx <- svd(X, nu = 0, nv = min(n, p))
  EDy <- svd(Y, nu = 0, nv = min(n, q))

  Ux_all <- EDx$v
  Uy_all <- EDy$v
  Lx_all <- (EDx$d^2) / n
  Ly_all <- (EDy$d^2) / n

  idx_x <- which(Lx_all > epsilon_sv)
  idx_y <- which(Ly_all > epsilon_sv)
  if (length(idx_x) < 1 || length(idx_y) < 1) {
    U_full <- matrix(0, p0, r)
    V_full <- matrix(0, q0, r)
    if (!is.null(x_names)) rownames(U_full) <- x_names
    if (!is.null(y_names)) rownames(V_full) <- y_names
    return(list(
      U = U_full, V = V_full, cor = rep(0, r), loss = Inf,
      Bhat = matrix(0, p0, q0),
      keepX = keepX, keepY = keepY, dropX = dropX, dropY = dropY,
      center = list(X = x_center, Y = y_center),
      scale  = list(X = x_scale,  Y = y_scale),
      converged = FALSE
    ))
  }

  Ux <- Ux_all[, idx_x, drop = FALSE]; Lx <- Lx_all[idx_x]
  Uy <- Uy_all[, idx_y, drop = FALSE]; Ly <- Ly_all[idx_y]
  UyT <- t(Uy)

  kx <- length(Lx)
  ky <- length(Ly)

  b <- outer(Lx, Ly) + rho

  # B1 = (X Ux)^T (Y Uy) / n
  XUx <- X %*% Ux
  YUy <- Y %*% Uy
  B1  <- crossprod(XUx, YUy) / n
  rm(XUx, YUy)

  # choose multiplication order once
  use_left_Low    <- (p * ky <= kx * q)
  use_left_Ztilde <- (kx * q <= p * ky)

  # ---- ADMM (low-memory)
  if (!is.null(B0)) {
    Z <- B0[keepX, keepY, drop = FALSE]
  } else {
    Z <- matrix(0, p, q)
  }

  # Ztilde = Ux^T Z Uy with order choice
  if (use_left_Ztilde) {
    Ztilde <- crossprod(Ux, Z) %*% Uy
  } else {
    Ztilde <- crossprod(Ux, Z %*% Uy)
  }
  Htilde <- matrix(0, kx, ky)

  converged <- FALSE
  tau_base <- lambda / rho   

  for (iter in 1:maxiter) {

    proj   <- Ztilde - Htilde
    Btilde <- (B1 + rho * proj) / b
    M      <- Btilde - proj

    # Low = Ux M Uy^T
    if (use_left_Low) {
      tmp <- Ux %*% M
      Low <- tmp %*% UyT
    } else {
      tmp <- M %*% UyT
      Low <- Ux %*% tmp
    }

    # Z_in = Low + Z   (no Z_old copy)
    Z_in <- Low + Z

    if (is.null(groups)) {
      Z <- soft_thresh(Z_in, tau_base)
    } else {
      Z <- Z_in
      for (g in seq_along(groups)) {
        idx <- groups[[g]]
        if (length(idx) == 0) next
        Z[idx] <- soft_thresh2(Z[idx], group_sqrt[g] * tau_base)
      }
    }

    # update Ztilde
    if (use_left_Ztilde) {
      Ztilde_new <- crossprod(Ux, Z) %*% Uy
    } else {
      Ztilde_new <- crossprod(Ux, Z %*% Uy)
    }

    Htilde <- Htilde + (Btilde - Ztilde_new)

    r_norm <- sqrt(sum((Btilde - Ztilde_new)^2))
    s_norm <- rho * sqrt(sum((Ztilde_new - Ztilde)^2))

    eps_pri  <- eps * (1 + max(sqrt(sum(Btilde^2)), sqrt(sum(Ztilde_new^2))))
    eps_dual <- eps * (1 + sqrt(sum(Htilde^2)))

    if (verbose && (iter %% 50 == 0)) {
      cat("\niter:", iter, " r:", signif(r_norm, 3), " s:", signif(s_norm, 3))
    }

    Ztilde <- Ztilde_new
    if (r_norm < eps_pri && s_norm < eps_dual) {
      converged <- TRUE
      break
    }
  }

  if (verbose) {
    if (!converged) cat("\nADMM did not converge (hit maxiter).\n")
    else cat("\nADMM converged.\n")
  }

  # ---- postprocessing (small SVD)
  r_eff <- min(r, nrow(Ztilde), ncol(Ztilde))
  if (r_eff < 1) {
    U_full <- matrix(0, p0, r)
    V_full <- matrix(0, q0, r)
    if (!is.null(x_names)) rownames(U_full) <- x_names
    if (!is.null(y_names)) rownames(V_full) <- y_names
    Bhat_full <- matrix(0, p0, q0); Bhat_full[keepX, keepY] <- Z
    return(list(
      U = U_full, V = V_full, cor = rep(0, r), loss = Inf,
      Bhat = Bhat_full,
      keepX = keepX, keepY = keepY, dropX = dropX, dropY = dropY,
      center = list(X = x_center, Y = y_center),
      scale  = list(X = x_scale,  Y = y_scale),
      converged = converged
    ))
  }

  # Csmall <- Ztilde
  # Csmall <- sweep(Csmall, 1, sqrt(Lx), "*")
  # Csmall <- sweep(Csmall, 2, sqrt(Ly), "*")

  # --- SVD on sparse Z (preserves row sparsity) ---
  SVDs <- safe_rank_svd(Z, r_eff, prefer_sparse = TRUE)
  if (is.null(SVDs)) {
    if (verbose) cat("\nSVD failed in ecca(). Returning NA directions.\n")
    U_full <- matrix(NA_real_, p0, r)
    V_full <- matrix(NA_real_, q0, r)
    if (!is.null(x_names)) rownames(U_full) <- x_names
    if (!is.null(y_names)) rownames(V_full) <- y_names
    Bhat_full <- matrix(0, p0, q0)
    Bhat_full[keepX, keepY] <- Z
    return(list(
      U = U_full,
      V = V_full,
      cor = rep(0, r),
      loss = Inf,
      Bhat = Bhat_full,
      keepX = keepX, keepY = keepY, dropX = dropX, dropY = dropY,
      center = list(X = x_center, Y = y_center),
      scale  = list(X = x_scale,  Y = y_scale),
      converged = FALSE
    ))
  }

  U0 <- SVDs$u
  V0 <- SVDs$v

  # --- whitening ---
  XU0 <- X %*% U0
  YV0 <- Y %*% V0

  GX <- crossprod(XU0) / n + ridge_whiten * diag(r_eff)
  GY <- crossprod(YV0) / n + ridge_whiten * diag(r_eff)

  WGX <- tryCatch(
    whiten_factor(GX, ridge = ridge_whiten, verbose = verbose),
    error = function(e) NULL
  )
  WGY <- tryCatch(
    whiten_factor(GY, ridge = ridge_whiten, verbose = verbose),
    error = function(e) NULL
  )
  if (is.null(WGX) || is.null(WGY)) {
    if (verbose) cat("\nWhitening failed in ecca(). Returning NA directions.\n")
    U_full <- matrix(NA_real_, p0, r)
    V_full <- matrix(NA_real_, q0, r)
    if (!is.null(x_names)) rownames(U_full) <- x_names
    if (!is.null(y_names)) rownames(V_full) <- y_names
    Bhat_full <- matrix(0, p0, q0)
    Bhat_full[keepX, keepY] <- Z
    return(list(
      U = U_full,
      V = V_full,
      cor = rep(0, r),
      loss = Inf,
      Bhat = Bhat_full,
      keepX = keepX, keepY = keepY, dropX = dropX, dropY = dropY,
      center = list(X = x_center, Y = y_center),
      scale  = list(X = x_scale,  Y = y_scale),
      converged = FALSE
    ))
  }

  U <- U0 %*% WGX
  V <- V0 %*% WGY

  XU <- X %*% U
  YV <- Y %*% V

  cor <- diag(crossprod(XU, YV) / n)

  # if (requireNamespace("RSpectra", quietly = TRUE) && r_eff < min(dim(Csmall))) {
  #   SVDs <- RSpectra::svds(Csmall, r_eff)
  # } else {
  #   SVDs <- svd(Csmall, nu = r_eff, nv = r_eff)
  # }

  # U0 <- Ux %*% SVDs$u
  # V0 <- Uy %*% SVDs$v

  # GX <- crossprod(X %*% U0) / n + ridge_whiten * diag(r_eff)
  # GY <- crossprod(Y %*% V0) / n + ridge_whiten * diag(r_eff)

  # Rx <- chol_jitter(GX, jitter = ridge_whiten)
  # Ry <- chol_jitter(GY, jitter = ridge_whiten)

  # U <- U0 %*% backsolve(Rx, diag(r_eff))
  # V <- V0 %*% backsolve(Ry, diag(r_eff))

  # XU <- X %*% U
  # YV <- Y %*% V
  # cor <- diag(crossprod(XU, YV) / n)

  neg <- cor < 0
  if (any(neg)) {
    V[, neg] <- -V[, neg, drop = FALSE]
    cor[neg] <- -cor[neg]
    YV[, neg] <- -YV[, neg, drop = FALSE]
  }

  ord <- order(cor, decreasing = TRUE)
  cor <- cor[ord]
  U <- U[, ord, drop = FALSE]
  V <- V[, ord, drop = FALSE]
  XU <- XU[, ord, drop = FALSE]
  YV <- YV[, ord, drop = FALSE]

  loss <- sum((XU - YV)^2) / n

  # ---- inflate to original coords
  U_full <- matrix(0, p0, r_eff)
  V_full <- matrix(0, q0, r_eff)
  U_full[keepX, ] <- U
  V_full[keepY, ] <- V

  if (!is.null(x_names)) rownames(U_full) <- x_names
  if (!is.null(y_names)) rownames(V_full) <- y_names
  colnames(U_full) <- paste0("comp", seq_len(r_eff))
  colnames(V_full) <- paste0("comp", seq_len(r_eff))

  Bhat_full <- matrix(0, p0, q0)
  Bhat_full[keepX, keepY] <- Z

  list(
    U = U_full,
    V = V_full,
    Ux = Ux, Uy = Uy, Lx = Lx, Ly = Ly,
    cor = cor,
    loss = loss,
    Bhat = Bhat_full,
    keepX = keepX, keepY = keepY, dropX = dropX, dropY = dropY,
    center = list(X = x_center, Y = y_center),
    scale  = list(X = x_scale,  Y = y_scale),
    converged = converged
  )
}

# -----------------------------
# ecca_across_lambdas (lambda path, CV-friendly)
# -----------------------------
ecca_across_lambdas <- function(
  X, Y, lambdas = 0, groups = NULL, r = 2,
  Sx = NULL, Sy = NULL, Sxy = NULL,                 # kept for compatibility; ignored
  standardize = TRUE,
  rho = 1, B0 = NULL, eps = 1e-4, maxiter = 500, verbose = TRUE,
  dense = TRUE, optimized = FALSE,                  # kept for compatibility; ignored
  epsilon_sv = 1e-8, ridge_whiten = 1e-8,
  warm_start = TRUE,
  lambda_order = c("decreasing", "increasing", "given"),
  log_prefix = "",
  admm_print_every = 50L,
  lambda_print = TRUE,
  collect_logs = FALSE,
  preprocess = TRUE,
  return_uv = TRUE,
  X_val = NULL, Y_val = NULL,
  scoring_method = c("mse", "trace")
){
  scoring_method <- match.arg(scoring_method)
  lambda_order <- match.arg(lambda_order)

  if (!is.matrix(X)) X <- as.matrix(X)
  if (!is.matrix(Y)) Y <- as.matrix(Y)
  if (nrow(X) != nrow(Y)) stop("X and Y must have same nrow().")
  logs <- character(0)

  lambdas <- as.numeric(lambdas)
  L <- length(lambdas)
  if (L < 1) stop("lambdas must have length >= 1")

  ord <- switch(
    lambda_order,
    decreasing = order(lambdas, decreasing = TRUE),
    increasing = order(lambdas, decreasing = FALSE),
    given      = seq_along(lambdas)
  )
  inv_ord <- order(ord)
  lambdas_run <- lambdas[ord]

  p0 <- ncol(X); q0 <- ncol(Y)
  x_names <- colnames(X); y_names <- colnames(Y)

  keepX <- seq_len(p0); keepY <- seq_len(q0)
  centerX <- rep(0, p0); centerY <- rep(0, q0)
  scaleX  <- rep(1, p0); scaleY  <- rep(1, q0)

  if (preprocess) {
    centerX <- colMeans(X); centerY <- colMeans(Y)
    X <- sweep(X, 2, centerX, "-")
    Y <- sweep(Y, 2, centerY, "-")

    sx <- matrixStats::colSds(X)
    sy <- matrixStats::colSds(Y)

    keepX <- which(is.finite(sx) & sx > 0)
    keepY <- which(is.finite(sy) & sy > 0)

    X <- X[, keepX, drop = FALSE]
    Y <- Y[, keepY, drop = FALSE]

    if (standardize) {
      X <- sweep(X, 2, sx[keepX], "/")
      Y <- sweep(Y, 2, sy[keepY], "/")
      scaleX[keepX] <- sx[keepX]
      scaleY[keepY] <- sy[keepY]
    }

    # map coord groups (original -> filtered)
    if (!is.null(groups)) {
      mapX <- integer(p0); mapX[keepX] <- seq_along(keepX)
      mapY <- integer(q0); mapY[keepY] <- seq_along(keepY)

      groups <- lapply(groups, function(idx) {
        if (is.matrix(idx) && ncol(idx) == 2) {
          ix <- mapX[idx[, 1]]
          iy <- mapY[idx[, 2]]
          ok <- (ix > 0) & (iy > 0)
          if (!any(ok)) return(matrix(integer(0), ncol = 2))
          cbind(ix[ok], iy[ok])
        } else {
          idx
        }
      })
    }

    # preprocess validation too
    if (!is.null(X_val) && !is.null(Y_val)) {
      if (!is.matrix(X_val)) X_val <- as.matrix(X_val)
      if (!is.matrix(Y_val)) Y_val <- as.matrix(Y_val)
      X_val <- sweep(X_val, 2, centerX, "-")
      Y_val <- sweep(Y_val, 2, centerY, "-")
      X_val <- X_val[, keepX, drop = FALSE]
      Y_val <- Y_val[, keepY, drop = FALSE]
      if (standardize) {
        X_val <- sweep(X_val, 2, scaleX[keepX], "/")
        Y_val <- sweep(Y_val, 2, scaleY[keepY], "/")
      }
    }
  } else {
    if (!is.null(X_val) && !is.null(Y_val)) {
      if (!is.matrix(X_val)) X_val <- as.matrix(X_val)
      if (!is.matrix(Y_val)) Y_val <- as.matrix(Y_val)
    }
  }

  n <- nrow(X)
  p <- ncol(X); q <- ncol(Y)

  # linearize groups once + precompute sqrt(|g|)
  group_sqrt <- NULL
  if (!is.null(groups)) {
    groups_lin <- vector("list", length(groups))
    group_sqrt <- numeric(length(groups))
    for (g in seq_along(groups)) {
      idx <- groups[[g]]
      if (is.matrix(idx) && ncol(idx) == 2) {
        groups_lin[[g]] <- as.integer(idx[, 1] + (idx[, 2] - 1L) * p)
        group_sqrt[g] <- sqrt(nrow(idx))
      } else {
        groups_lin[[g]] <- idx
        group_sqrt[g] <- sqrt(length(idx))
      }
    }
    groups <- groups_lin
  }

  if (p < 1 || q < 1) {
    out_scores <- rep(Inf, L)
    if (!return_uv) return(list(scores = out_scores[inv_ord], lambdas = lambdas))
    Uout <- lapply(seq_len(L), function(.) matrix(NA_real_, p0, r))
    Vout <- lapply(seq_len(L), function(.) matrix(NA_real_, q0, r))
    return(list(U = Uout, V = Vout, scores = out_scores[inv_ord], lambdas = lambdas))
  }

  # reduced bases
  EDx <- svd(X, nu = 0, nv = min(n, p))
  EDy <- svd(Y, nu = 0, nv = min(n, q))

  Ux_all <- EDx$v
  Uy_all <- EDy$v
  Lx_all <- (EDx$d^2) / n
  Ly_all <- (EDy$d^2) / n

  idx_x <- which(Lx_all > epsilon_sv)
  idx_y <- which(Ly_all > epsilon_sv)
  if (length(idx_x) < 1 || length(idx_y) < 1) {
    out_scores <- rep(Inf, L)
    if (!return_uv) return(list(scores = out_scores[inv_ord], lambdas = lambdas))
    Uout <- lapply(seq_len(L), function(.) matrix(NA_real_, p0, r))
    Vout <- lapply(seq_len(L), function(.) matrix(NA_real_, q0, r))
    return(list(U = Uout, V = Vout, scores = out_scores[inv_ord], lambdas = lambdas))
  }

  Ux <- Ux_all[, idx_x, drop = FALSE]; Lx <- Lx_all[idx_x]
  Uy <- Uy_all[, idx_y, drop = FALSE]; Ly <- Ly_all[idx_y]
  UyT <- t(Uy)

  kx <- length(Lx); ky <- length(Ly)
  b  <- outer(Lx, Ly) + rho

  # B1 once
  XUx <- X %*% Ux
  YUy <- Y %*% Uy
  B1  <- crossprod(XUx, YUy) / n
  rm(XUx, YUy)

  use_left_Low    <- (p * ky <= kx * q)
  use_left_Ztilde <- (kx * q <= p * ky)

  compute_Ztilde <- function(Zmat) {
    if (use_left_Ztilde) crossprod(Ux, Zmat) %*% Uy
    else crossprod(Ux, Zmat %*% Uy)
  }
  compute_Low <- function(M) {
    if (use_left_Low) (Ux %*% M) %*% UyT
    else Ux %*% (M %*% UyT)
  }

  out_scores <- rep(NA_real_, L)
  out_cor    <- vector("list", L)
  out_loss   <- rep(NA_real_, L)
  out_U      <- if (return_uv) vector("list", L) else NULL
  out_V      <- if (return_uv) vector("list", L) else NULL

  if (!is.null(B0)) {
    Z <- if (preprocess) B0[keepX, keepY, drop = FALSE] else B0
  } else {
    Z <- matrix(0, p, q)
  }
  admm_print_every <- as.integer(admm_print_every)
  if (!is.finite(admm_print_every) || admm_print_every < 1L) admm_print_every <- 50L

  for (t in seq_len(L)) {
    lam <- lambdas_run[t]
    t_lambda0 <- Sys.time()
    if (verbose && lambda_print) {
      cat(sprintf("\nStarting lambda %d/%d: %g", t, L, lam))
      cat(sprintf("\n%s lambda %d/%d = %g", log_prefix, t, L, lam))
      utils::flush.console()
      if (collect_logs) {
        logs <- c(logs,
                  sprintf("Starting lambda %d/%d: %g", t, L, lam),
                  sprintf("%s lambda %d/%d = %g", log_prefix, t, L, lam))
      }
    }


    if (!warm_start && t > 1) Z <- matrix(0, p, q)

    Ztilde <- compute_Ztilde(Z)
    Htilde <- matrix(0, kx, ky)

    tau_base <- lam / rho
    converged <- FALSE

    for (iter in seq_len(maxiter)) {
      proj   <- Ztilde - Htilde
      Btilde <- (B1 + rho * proj) / b
      M      <- Btilde - proj

      Low <- compute_Low(M)

      Z_in <- Low + Z

      if (is.null(groups)) {
        Z <- soft_thresh(Z_in, tau_base)
      } else {
        Z <- Z_in
        for (g in seq_along(groups)) {
          idx <- groups[[g]]
          if (length(idx) == 0) next
          Z[idx] <- soft_thresh2(Z[idx], group_sqrt[g] * tau_base)
        }
      }

      Ztilde_new <- compute_Ztilde(Z)
      Htilde <- Htilde + (Btilde - Ztilde_new)

      r_norm <- sqrt(sum((Btilde - Ztilde_new)^2))
      s_norm <- rho * sqrt(sum((Ztilde_new - Ztilde)^2))

      eps_pri  <- eps * (1 + max(sqrt(sum(Btilde^2)), sqrt(sum(Ztilde_new^2))))
      eps_dual <- eps * (1 + sqrt(sum(Htilde^2)))

      if (verbose && iter %% 100 == 0) {
        cat("\nlambda", signif(lam, 3), "iter", iter,
            "r", signif(r_norm, 3), "s", signif(s_norm, 3))
        if (collect_logs) {
          logs <- c(logs, sprintf("lambda %s iter %d r %s s %s",
                                  signif(lam, 3), iter, signif(r_norm, 3), signif(s_norm, 3)))
        }
      }
          # inside ADMM iter loop
      if (verbose && (iter %% admm_print_every == 0L)) {
        cat(sprintf("\n%s   iter %d  r=%g  s=%g", log_prefix, iter, r_norm, s_norm))
        utils::flush.console()
        if (collect_logs) {
          logs <- c(logs, sprintf("%s   iter %d  r=%g  s=%g", log_prefix, iter, r_norm, s_norm))
        }
      }

      Ztilde <- Ztilde_new
      if (r_norm < eps_pri && s_norm < eps_dual) {
        converged <- TRUE
        break
      }
    }

    # postprocessing small SVD
    Csmall <- Ztilde
    Csmall <- sweep(Csmall, 1, sqrt(Lx), "*")
    Csmall <- sweep(Csmall, 2, sqrt(Ly), "*")

    r_eff <- min(r, nrow(Csmall), ncol(Csmall))
    if (r_eff < 1) {
      out_scores[t] <- Inf
      out_loss[t]   <- Inf
      out_cor[[t]]  <- rep(0, r)
      if (return_uv) {
        out_U[[t]] <- matrix(0, p0, r)
        out_V[[t]] <- matrix(0, q0, r)
      }
      next
    }

    SVDs <- safe_rank_svd(Z, r_eff, prefer_sparse = TRUE)
    if (is.null(SVDs)) {
      out_scores[t] <- Inf
      out_loss[t]   <- Inf
      out_cor[[t]]  <- rep(0, r)
      if (return_uv) {
        out_U[[t]] <- matrix(0, p0, r)
        out_V[[t]] <- matrix(0, q0, r)
      }
      if (collect_logs) {
        logs <- c(logs, sprintf("%s SVD failed at lambda=%g; assigned Inf score", log_prefix, lam))
      }
      next
    }

    Uhat <- SVDs$u
    Vhat <- SVDs$v

    # --- whitening ---
    XU0 <- X %*% Uhat
    YV0 <- Y %*% Vhat

    GX <- crossprod(XU0) / n + ridge_whiten * diag(r_eff)
    GY <- crossprod(YV0) / n + ridge_whiten * diag(r_eff)

    WGX <- tryCatch(
      whiten_factor(GX, ridge = ridge_whiten, verbose = FALSE),
      error = function(e) NULL
    )
    WGY <- tryCatch(
      whiten_factor(GY, ridge = ridge_whiten, verbose = FALSE),
      error = function(e) NULL
    )
    if (is.null(WGX) || is.null(WGY)) {
      out_scores[t] <- Inf
      out_loss[t]   <- Inf
      out_cor[[t]]  <- rep(0, r)
      if (return_uv) {
        out_U[[t]] <- matrix(0, p0, r)
        out_V[[t]] <- matrix(0, q0, r)
      }
      if (collect_logs) {
        logs <- c(logs, sprintf("%s whitening failed at lambda=%g; assigned Inf score", log_prefix, lam))
      }
      next
    }

    Uhat <- Uhat %*% WGX
    Vhat <- Vhat %*% WGY

    XU <- X %*% Uhat
    YV <- Y %*% Vhat

    cor <- diag(crossprod(XU, YV) / n)

    neg <- cor < 0
    if (any(neg)) {
      Vhat[, neg] <- -Vhat[, neg, drop = FALSE]
      cor[neg] <- -cor[neg]
      YV[, neg] <- -YV[, neg, drop = FALSE]
    }

    ord_cor <- order(cor, decreasing = TRUE)
    cor <- cor[ord_cor]
    Uhat <- Uhat[, ord_cor, drop = FALSE]
    Vhat <- Vhat[, ord_cor, drop = FALSE]
    XU <- XU[, ord_cor, drop = FALSE]
    YV <- YV[, ord_cor, drop = FALSE]

    if (!is.null(X_val) && !is.null(Y_val)) {
      XU_val <- X_val %*% Uhat
      YV_val <- Y_val %*% Vhat
      if (scoring_method == "mse") {
        out_scores[t] <- mean((XU_val - YV_val)^2)
      } else {
        # Faster than cor(): compute only paired (diagonal) correlations.
        XUc <- scale(XU_val, center = TRUE, scale = FALSE)
        YVc <- scale(YV_val, center = TRUE, scale = FALSE)
        num <- colSums(XUc * YVc)
        den <- sqrt(colSums(XUc^2) * colSums(YVc^2))
        den[den < 1e-12 | !is.finite(den)] <- 1e-12
        corr_diag <- num / den
        out_scores[t] <- -mean(corr_diag)
      }
    } else {
      out_scores[t] <- NA_real_
    }

    out_loss[t]  <- sum((XU - YV)^2) / n
    out_cor[[t]] <- cor

    if (return_uv) {
      U_full <- matrix(0, p0, r_eff)
      V_full <- matrix(0, q0, r_eff)
      U_full[keepX, ] <- Uhat
      V_full[keepY, ] <- Vhat
      if (!is.null(x_names)) rownames(U_full) <- x_names
      if (!is.null(y_names)) rownames(V_full) <- y_names
      out_U[[t]] <- U_full
      out_V[[t]] <- V_full
    }
    if (verbose && lambda_print) {
      cat(sprintf("\nFinished lambda %d/%d in %.1fs",
                  t, L, as.numeric(difftime(Sys.time(), t_lambda0, units="secs"))))
      if (collect_logs) {
        logs <- c(logs, sprintf("Finished lambda %d/%d in %.1fs",
                                t, L, as.numeric(difftime(Sys.time(), t_lambda0, units="secs"))))
      }
    }
  }

  # restore original lambda order
  out_scores <- out_scores[inv_ord]
  out_loss   <- out_loss[inv_ord]
  out_cor    <- out_cor[inv_ord]
  if (return_uv) {
    out_U <- out_U[inv_ord]
    out_V <- out_V[inv_ord]
  }

  if (!return_uv) {
    out <- list(scores = out_scores, lambdas = lambdas)
    if (collect_logs) out$logs <- logs
    return(out)
  }

  if (length(lambdas) == 1) {
    out <- list(U = out_U[[1]], V = out_V[[1]],
                cor = out_cor[[1]], loss = out_loss[1],
                scores = out_scores, lambdas = lambdas,
                center = list(X = centerX, Y = centerY),
                scale  = list(X = scaleX,  Y = scaleY),
                keepX = keepX, keepY = keepY)
    if (collect_logs) out$logs <- logs
    return(out)
  }

  out <- list(U = out_U, V = out_V,
              cor = out_cor, loss = out_loss,
              scores = out_scores, lambdas = lambdas,
              center = list(X = centerX, Y = centerY),
              scale  = list(X = scaleX,  Y = scaleY),
              keepX = keepX, keepY = keepY,
              Ux = Ux, Uy = Uy, Lx = Lx, Ly = Ly)
  if (collect_logs) out$logs <- logs
  out
}

# -----------------------------
# ecca.eval (CV) - with seamless group mapping + faster parallel behavior
# -----------------------------
ecca.eval <- function(
  X, Y, lambdas = 0, groups = NULL, r = 2,
  standardize = TRUE,
  rho = 1, B0 = NULL, nfold = 5,
  eps = 1e-4, maxiter = 500, verbose = TRUE,
  parallel = TRUE, nb_cores = NULL, set_seed_cv = NULL,
  scoring_method = c("mse", "trace"),
  cv_use_median = FALSE,
  dense = TRUE, optimized = FALSE,
  epsilon_sv = 1e-8, ridge_whiten = 1e-8
){
  scoring_method <- match.arg(scoring_method)
  if (!is.matrix(X)) X <- as.matrix(X)
  if (!is.matrix(Y)) Y <- as.matrix(Y)
  if (nrow(X) != nrow(Y)) stop("X and Y must have same nrow().")

  n <- nrow(X)
  if (n < 2 * nfold) {
    if (verbose) cat("\nWarning: n too small for CV; using first lambda.\n")
    scores <- data.frame(lambda = lambdas, mse = NA_real_, se = NA_real_)
    return(list(scores = scores, lambda.min = lambdas[1], lambda.1se = lambdas[1]))
  }

  # global preprocessing
  p0 <- ncol(X); q0 <- ncol(Y)

  X <- scale(X, center = TRUE, scale = FALSE)
  Y <- scale(Y, center = TRUE, scale = FALSE)

  sx <- matrixStats::colSds(X)
  sy <- matrixStats::colSds(Y)
  keepX <- which(is.finite(sx) & sx > 0)
  keepY <- which(is.finite(sy) & sy > 0)

  X <- X[, keepX, drop = FALSE]
  Y <- Y[, keepY, drop = FALSE]

  if (standardize) {
    X <- sweep(X, 2, sx[keepX], "/")
    Y <- sweep(Y, 2, sy[keepY], "/")
  }

  n <- nrow(X)

  # IMPORTANT: map groups to the filtered coordinates ONCE (since preprocess=FALSE in folds)
  if (!is.null(groups)) {
    mapX <- integer(p0); mapX[keepX] <- seq_along(keepX)
    mapY <- integer(q0); mapY[keepY] <- seq_along(keepY)

    groups <- lapply(groups, function(idx) {
      if (is.matrix(idx) && ncol(idx) == 2) {
        ix <- mapX[idx[, 1]]
        iy <- mapY[idx[, 2]]
        ok <- (ix > 0) & (iy > 0)
        if (!any(ok)) return(matrix(integer(0), ncol = 2))
        cbind(ix[ok], iy[ok])
      } else {
        idx
      }
    })
  }

  if (!is.null(set_seed_cv)) set.seed(set_seed_cv)
  folds <- .create_cv_folds(n, nfold)

  # parallel setup
  if (parallel) {
    cores <- if (is.null(nb_cores)) max(1L, parallel::detectCores() - 1L) else as.integer(nb_cores)
    cores <- min(cores, length(folds))
    Sys.setenv(
      OMP_NUM_THREADS = 1,
      OPENBLAS_NUM_THREADS = 1,
      MKL_NUM_THREADS = 1,
      VECLIB_MAXIMUM_THREADS = 1
    )

    cl <- setup_parallel_backend(cores, verbose = verbose)
    if (is.null(cl)) {
      warning("All parallel setup attempts failed. Proceeding in serial mode.", immediate. = TRUE)
      parallel <- FALSE
    }
  }

  if (parallel) {
    initialize_parallel_workers(cl, verbose = verbose)
    on.exit(cleanup_parallel_backend(cl), add = TRUE)
    if (verbose) {
      cat("\nRunning CV in parallel on", cores, "cores\n")
    }
    ecca_across_lambdas_current <- ecca_across_lambdas
    safe_rank_svd_current <- safe_rank_svd
    whiten_factor_current <- whiten_factor
    chol_jitter_current <- chol_jitter
    invsqrt_psd_current <- invsqrt_psd
    soft_thresh_current <- soft_thresh
    soft_thresh2_current <- soft_thresh2
    safe_rank_svd <- safe_rank_svd_current
    whiten_factor <- whiten_factor_current
    chol_jitter <- chol_jitter_current
    invsqrt_psd <- invsqrt_psd_current
    soft_thresh <- soft_thresh_current
    soft_thresh2 <- soft_thresh2_current
    collect_logs_parallel <- isTRUE(verbose)
    ecca_parallel_fold_worker <- function(fold_id) {
      tryCatch({
        fold <- folds[[fold_id]]
        pid <- Sys.getpid()
        prefix <- sprintf("[pid=%d fold=%d]", pid, fold_id)

        tr <- setdiff(seq_len(n), fold)
        Xtr <- X[tr, , drop = FALSE]
        Ytr <- Y[tr, , drop = FALSE]
        Xva <- X[fold, , drop = FALSE]
        Yva <- Y[fold, , drop = FALSE]

        fit <- ecca_across_lambdas_current(
          Xtr, Ytr, lambdas = lambdas, groups = groups, r = r,
          standardize = FALSE, rho = rho, B0 = B0,
          eps = eps, maxiter = maxiter,
          verbose = collect_logs_parallel,
          log_prefix = prefix,
          admm_print_every = 50L,
          lambda_print = collect_logs_parallel,
          epsilon_sv = epsilon_sv, ridge_whiten = ridge_whiten,
          preprocess = FALSE,
          return_uv = FALSE,
          collect_logs = collect_logs_parallel,
          X_val = Xva, Y_val = Yva,
          scoring_method = scoring_method
        )

        list(
          fold_id = fold_id,
          scores = fit$scores,
          logs = fit$logs
        )
      }, error = function(e) e)
    }

    parallel::clusterExport(
      cl,
      varlist = c(
        "X", "Y", "folds", "n",
        "lambdas", "groups", "r",
        "rho", "B0", "eps", "maxiter",
        "epsilon_sv", "ridge_whiten", "scoring_method",
        "collect_logs_parallel",
        "ecca_across_lambdas_current",
        "safe_rank_svd",
        "whiten_factor",
        "chol_jitter",
        "invsqrt_psd",
        "soft_thresh",
        "soft_thresh2",
        "ecca_parallel_fold_worker"
      ),
      envir = environment()
    )

    cv_list <- parallel::parLapplyLB(cl, seq_along(folds), ecca_parallel_fold_worker)

    if (verbose) {
      ok_folds <- cv_list[!vapply(cv_list, inherits, logical(1), "error")]
      if (length(ok_folds) > 0) {
        for (res in ok_folds) {
          if (!is.null(res$logs) && length(res$logs) > 0) {
            cat("\n")
            cat(paste(res$logs, collapse = "\n"))
            cat("\n")
          }
        }
        utils::flush.console()
      }
    }

    is_err <- vapply(cv_list, inherits, logical(1), "error")

    if (all(is_err)) {
      msgs <- unique(vapply(cv_list, function(e) {
        msg <- tryCatch(conditionMessage(e), error = function(...) "")
        if (!nzchar(msg)) msg <- paste(class(e), collapse = "/")
        msg
      }, character(1)))
      stop("All CV folds failed. Example error(s):\n- ",
           paste(utils::head(msgs, 5), collapse = "\n- "))
    }

    if (any(is_err)) warning(sum(is_err), " folds failed.")
    cv_list <- lapply(cv_list[!is_err], function(x) x$scores)
    if (length(cv_list) == 0) stop("All CV folds failed.")
    scores.cv <- do.call(cbind, cv_list)
    n_success <- ncol(scores.cv)

    

  } else {
    cv_list <- lapply(folds, function(fold) {
      tryCatch({
        tr <- setdiff(seq_len(n), fold)
        Xtr <- X[tr, , drop = FALSE]
        Ytr <- Y[tr, , drop = FALSE]
        Xva <- X[fold, , drop = FALSE]
        Yva <- Y[fold, , drop = FALSE]

        fit <- ecca_across_lambdas(
          Xtr, Ytr, lambdas = lambdas, groups = groups, r = r,
          standardize = FALSE, rho = rho, B0 = B0,
          eps = eps, maxiter = maxiter, verbose = TRUE,
          epsilon_sv = epsilon_sv, ridge_whiten = ridge_whiten,
          preprocess = FALSE,
          return_uv = FALSE,
          X_val = Xva, Y_val = Yva,
          scoring_method = scoring_method
        )
        fit$scores
      }, error = function(e) e)
    })

    is_err <- vapply(cv_list, inherits, logical(1), "error")
    if (any(is_err)) warning(sum(is_err), " folds failed.")
    cv_list <- cv_list[!is_err]
    if (length(cv_list) == 0) stop("All CV folds failed.")
    scores.cv <- do.call(cbind, cv_list)
    n_success <- ncol(scores.cv)
  }

  if (!cv_use_median) {
    mse <- rowMeans(scores.cv, na.rm = TRUE)
  } else {
    mse <- apply(scores.cv, 1, median, na.rm = TRUE)
  }
  se <- matrixStats::rowSds(scores.cv, na.rm = TRUE) / sqrt(n_success)
  se[!is.finite(se)] <- NA_real_
  scores <- data.frame(lambda = lambdas, mse = mse, se = se)

  lambda.min <- scores$lambda[which.min(scores$mse)]
  upper <- scores$mse[which.min(scores$mse)] + scores$se[which.min(scores$mse)]
  lambda.1se <- max(scores$lambda[scores$mse <= upper])

  list(scores = scores, lambda.min = lambda.min, lambda.1se = lambda.1se)
}

#' Cross-Validated Efficient CCA
#'
#' Selects a regularization parameter for [ecca()] by cross-validation and
#' refits the final model at the selected value.
#'
#' @param X Predictor matrix (n x p).
#' @param Y Response matrix (n x q).
#' @param lambdas Candidate regularization values.
#' @param groups Optional group structure for blockwise sparsity.
#' @param r Target rank.
#' @param standardize Whether to scale variables after centering.
#' @param rho ADMM penalty parameter.
#' @param B0 Optional warm start for the coefficient matrix.
#' @param nfold Number of cross-validation folds.
#' @param select Selection rule for the final lambda. One of `"lambda.min"` or `"lambda.1se"`.
#' @param eps Convergence tolerance for the final ADMM refit.
#' @param maxiter Maximum iterations for the final ADMM refit.
#' @param verbose Whether to print diagnostics.
#' @param maxiter_cv Maximum iterations used inside the cross-validation fits.
#' @param parallel Whether to parallelize cross-validation.
#' @param nb_cores Number of worker processes to use when `parallel = TRUE`.
#' @param set_seed_cv Optional random seed for fold generation.
#' @param scoring_method Cross-validation score to optimize. One of `"mse"` or `"trace"`.
#' @param cv_use_median Whether to aggregate fold scores with the median instead of the mean.
#' @param dense Retained for backward compatibility.
#' @param optimized Retained for backward compatibility.
#' @param epsilon_sv Numerical threshold used to discard near-zero singular values.
#' @param ridge_whiten Ridge added when whitening Gram matrices.
#'
#' @return A list with the final fit, selected lambda, and cross-validation
#'   scores when more than one lambda is supplied.
#' @export
ecca.cv <- function(
  X, Y, lambdas = 0, groups = NULL, r = 2, standardize = FALSE,
  rho = 1, B0 = NULL, nfold = 5, select = "lambda.min",
  eps = 1e-3, maxiter = 1000, verbose = FALSE, maxiter_cv = 300,
  parallel = FALSE, nb_cores = NULL, set_seed_cv = NULL,
  scoring_method = c("mse", "trace"), cv_use_median = FALSE,
  dense = TRUE, optimized = FALSE,
  epsilon_sv = 1e-8, ridge_whiten = 1e-8
){
  scoring_method <- match.arg(scoring_method)

  eval <- NULL
  if (length(lambdas) > 1) {
    eval <- ecca.eval(
      X, Y, lambdas = lambdas, groups = groups, r = r,
      standardize = standardize,
      rho = rho, B0 = B0, nfold = nfold,
      eps = eps, maxiter = maxiter_cv, verbose = verbose,
      parallel = parallel, nb_cores = nb_cores, set_seed_cv = set_seed_cv,
      scoring_method = scoring_method, cv_use_median = cv_use_median,
      epsilon_sv = epsilon_sv, ridge_whiten = ridge_whiten
    )
    lambda.opt <- if (select == "lambda.1se") eval$lambda.1se else eval$lambda.min
  } else {
    lambda.opt <- as.numeric(lambdas)
  }

  if (verbose) cat("\nselected lambda:", lambda.opt, "\n")

  final_verbose <- TRUE
  cat("\nRefitting at selected lambda with ADMM logs...\n")
  utils::flush.console()

  fit <- ecca_across_lambdas(
    X, Y, lambdas = lambda.opt, groups = groups, r = r,
    standardize = standardize,
    rho = rho, B0 = B0, eps = eps, maxiter = maxiter, verbose = final_verbose,
    admm_print_every = 1L,
    lambda_print = TRUE,
    epsilon_sv = epsilon_sv, ridge_whiten = ridge_whiten,
    preprocess = TRUE,
    return_uv = TRUE
  )
  cat("\nFinal ADMM refit completed.\n")
  utils::flush.console()

  out <- list(
    U = fit$U,
    V = fit$V,
    cor = fit$cor,
    loss = fit$loss,
    lambda.opt = lambda.opt
  )
  if (!is.null(eval)) out$cv.scores <- eval$scores
  out
}

library(ccar3)


test_that("SMUT multiplication works if available", {
  skip_if_not_installed("SMUT")
  A <- matrix(rnorm(100), 10)
  B <- matrix(rnorm(100), 10)
  expect_equal(SMUT::eigenMapMatMult(A, B), A %*% B, tolerance = 1e-10)
})


test_that("ecca returns correct output structure", {
  
  
  set.seed(123)
  gen = generate_example_none_trivial_pca(n=100, p1=50, p2=50, r_pca = 3, nnzeros = 5, 
                                           theta = diag(c(.9, .9)), overlapping_amount = 1,
                                           lambda_pca = 1, r = 2)
  result <- ecca(gen$X, gen$Y, r = 2)
  
  expect_type(result, "list")
  expect_true(all(c("U", "V", "loss", "cor") %in% names(result)))
  expect_equal(dim(result$U)[2], 2)
  expect_equal(dim(result$V)[2], 2)
  expect_equal(dim(result$U)[1], dim(gen$X)[2])
  expect_equal(dim(result$V)[1], dim(gen$Y)[2])
})


test_that("ecca returns the correct answer", {
  
  ##### Generate toy example data
  set.seed(123)
  r = 2
  gen = generate_example_none_trivial_pca(n=700, p1=50, p2=50, r_pca = 3, nnzeros = 5, 
                                          theta = diag(c(.9, .9)), overlapping_amount = 1,
                                          lambda_pca = 1, r = r)
  
  
  result <- ecca(gen$X, gen$Y, r = r, lambda = 0.01)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.3)  
  expect_true(subdistance(result$V, gen$v) < 0.3)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
})





test_that("ecca.cv returns the correct answer", {
  
  ##### Generate toy example data
  set.seed(123)
  r = 2
  gen = generate_example_none_trivial_pca(n=700, p1=50, p2=50, r_pca = 3, nnzeros = 5, 
                                          theta = diag(c(.9, .9)), overlapping_amount = 1,
                                          lambda_pca = 1, r = r)
  
  
  result <- ecca.cv(gen$X, gen$Y, r = r, lambdas = c(10^(-4), 10^(-3), 0.01, 0.05, 0.075, 0.1, 0.5),, parallel=F)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.3)  
  expect_true(subdistance(result$V, gen$v) < 0.3)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
})
  


test_that("ecca.cv parallelization works", {
  skip_if_not_installed("doParallel")
  ##### Generate toy example data
  set.seed(123)
  r = 2
  gen = generate_example_none_trivial_pca(n=700, p1=50, p2=50, r_pca = 3, nnzeros = 5, 
                                          theta = diag(c(.9, .9)), overlapping_amount = 1,
                                          lambda_pca = 1, r = r)
  
  
  result <- ecca.cv(gen$X, gen$Y, r = r, lambdas = c(10^(-4), 10^(-3), 0.01, 0.05, 0.075, 0.1, 0.5), parallel=T)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.3)  
  expect_true(subdistance(result$V, gen$v) < 0.3)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
})

library(ccar3)


test_that("SMUT multiplication works if available", {
  skip_if_not_installed("SMUT")
  A <- matrix(rnorm(100), 10)
  B <- matrix(rnorm(100), 10)
  expect_equal(SMUT::eigenMapMatMult(A, B), A %*% B, tolerance = 1e-10)
})


test_that("ecca returns correct output structure", {
  
  
  set.seed(123)
  gen = generate_example_none_trivial_pca(n=100, p1=50, p2=50, r_pca = 3, nnzeros = 5, 
                                           theta = diag(c(.9, .9)), overlapping_amount = 1,
                                           lambda_pca = 1, r = 2)
  result <- ecca(gen$X, gen$Y, r = 2)
  
  expect_type(result, "list")
  expect_true(all(c("U", "V", "loss", "cor") %in% names(result)))
  expect_equal(dim(result$U)[2], 2)
  expect_equal(dim(result$V)[2], 2)
  expect_equal(dim(result$U)[1], dim(gen$X)[2])
  expect_equal(dim(result$V)[1], dim(gen$Y)[2])
})


test_that("ecca returns the correct answer", {
  
  ##### Generate toy example data
  set.seed(123)
  r = 2
  gen = generate_example_none_trivial_pca(n=700, p1=50, p2=50, r_pca = 3, nnzeros = 5, 
                                          theta = diag(c(.9, .9)), overlapping_amount = 1,
                                          lambda_pca = 1, r = r)
  
  
  result <- ecca(gen$X, gen$Y, r = r, lambda = 0.01)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.3)  
  expect_true(subdistance(result$V, gen$v) < 0.3)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
})





test_that("ecca.cv returns the correct answer", {
  
  ##### Generate toy example data
  set.seed(123)
  r = 2
  gen = generate_example_none_trivial_pca(n=700, p1=50, p2=50, r_pca = 3, nnzeros = 5, 
                                          theta = diag(c(.9, .9)), overlapping_amount = 1,
                                          lambda_pca = 1, r = r)
  
  
  result <- ecca.cv(gen$X, gen$Y, r = r, lambdas = c(10^(-4), 10^(-3), 0.01, 0.05, 0.075, 0.1, 0.5),, parallel=F)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.3)  
  expect_true(subdistance(result$V, gen$v) < 0.3)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
})
  


test_that("ecca.cv parallelization works", {
  skip_if_not_installed("doParallel")
  skip_if_cran_limits_cores()
  ##### Generate toy example data
  set.seed(123)
  r = 2
  gen = generate_example_none_trivial_pca(n=700, p1=50, p2=50, r_pca = 3, nnzeros = 5, 
                                          theta = diag(c(.9, .9)), overlapping_amount = 1,
                                          lambda_pca = 1, r = r)
  
  
  result <- ecca.cv(gen$X, gen$Y, r = r, lambdas = c(10^(-4), 10^(-3), 0.01, 0.05, 0.075, 0.1, 0.5), parallel=T)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.3)  
  expect_true(subdistance(result$V, gen$v) < 0.3)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
})

# -----------------------------
# Helper: robust chol with jitter
# -----------------------------
chol_jitter <- function(G, jitter = 1e-8, max_tries = 6) {
  p <- nrow(G)
  for (k in 0:max_tries) {
    out <- tryCatch(
      chol(G + (10^k) * jitter * diag(p)),
      error = function(e) NULL
    )
    if (!is.null(out)) return(out)
  }
  stop("chol() failed even with jitter.")
}

# -----------------------------
# Helper: inverse square root for PSD matrices (eigen fallback)
# -----------------------------
invsqrt_psd <- function(G, eps = 1e-8) {
  G <- (G + t(G)) / 2
  ed <- eigen(G, symmetric = TRUE)
  vals <- pmax(ed$values, eps)
  ed$vectors %*% diag(1 / sqrt(vals), nrow = length(vals)) %*% t(ed$vectors)
}

# -----------------------------
# Helper: whitening factor G^{-1/2}, with chol then eigen fallback
# -----------------------------
whiten_factor <- function(G, ridge = 1e-8, verbose = FALSE) {
  if (any(!is.finite(G))) {
    if (verbose) cat("\nNon-finite entries in whitening Gram matrix; sanitizing.\n")
    G[!is.finite(G)] <- 0
    G <- (G + t(G)) / 2 + ridge * diag(nrow(G))
  }
  tryCatch({
    R <- chol_jitter(G, jitter = ridge)
    backsolve(R, diag(nrow(G)))
  }, error = function(e) {
    if (verbose) cat("\nchol_jitter failed; using eigen whitening fallback.\n")
    invsqrt_psd(G, eps = ridge)
  })
}

# -----------------------------
# Helper: stable rank-k SVD with fallback
# -----------------------------
safe_rank_svd <- function(M, k, prefer_sparse = FALSE) {
  if (k < 1) return(NULL)
  nr <- nrow(M); nc <- ncol(M)
  if (nr < 1 || nc < 1) return(NULL)
  k <- min(k, nr, nc)

  out <- NULL

  if (requireNamespace("RSpectra", quietly = TRUE) && k < min(nr, nc)) {
    out <- tryCatch({
      if (prefer_sparse && requireNamespace("Matrix", quietly = TRUE)) {
        RSpectra::svds(Matrix::Matrix(M, sparse = TRUE), k)
      } else {
        RSpectra::svds(M, k)
      }
    }, error = function(e) NULL)

    if (!is.null(out)) {
      ok <- all(is.finite(out$u)) && all(is.finite(out$v)) && all(is.finite(out$d))
      if (!ok) out <- NULL
    }
  }

  if (is.null(out)) {
    out <- tryCatch(svd(as.matrix(M), nu = k, nv = k), error = function(e) NULL)
    if (is.null(out)) return(NULL)
    out$u <- out$u[, seq_len(k), drop = FALSE]
    out$v <- out$v[, seq_len(k), drop = FALSE]
    out$d <- out$d[seq_len(k)]
  }

  out
}

#' Efficient CCA for Two High-Dimensional Views
#'
#' Fits sparse canonical directions with an ADMM-based reduced-rank regression
#' formulation tailored to the setting where both views are high-dimensional.
#'
#' @param X Predictor matrix (n x p).
#' @param Y Response matrix (n x q).
#' @param lambda Regularization parameter.
#' @param groups Optional group structure for blockwise sparsity.
#' @param r Target rank.
#' @param standardize Whether to scale variables after centering.
#' @param rho ADMM penalty parameter.
#' @param B0 Optional warm start for the coefficient matrix.
#' @param eps Convergence tolerance for ADMM.
#' @param maxiter Maximum number of ADMM iterations.
#' @param verbose Whether to print diagnostics.
#' @param epsilon_sv Numerical threshold used to discard near-zero singular values.
#' @param ridge_whiten Ridge added when whitening Gram matrices.
#'
#' @return A list containing the estimated canonical directions, canonical
#'   correlations, the fitted coefficient matrix, preprocessing metadata, and
#'   convergence information.
#' @export
ecca <- function(
  X, Y,
  lambda = 0,
  groups = NULL,
  r = 2,
  standardize = FALSE,
  rho = 1,
  B0 = NULL,
  eps = 1e-4,
  maxiter = 500,
  verbose = TRUE,
  epsilon_sv = 1e-8,
  ridge_whiten = 1e-8
) {
  if (!is.matrix(X)) X <- as.matrix(X)
  if (!is.matrix(Y)) Y <- as.matrix(Y)
  if (nrow(X) != nrow(Y)) stop("X and Y must have the same number of rows.")

  n  <- nrow(X)
  p0 <- ncol(X)
  q0 <- ncol(Y)
  x_names <- colnames(X)
  y_names <- colnames(Y)

  if (verbose) cat("eps =", eps, " maxiter =", maxiter, "\n")

  # ---- preprocess: center, drop 0-var, optional scale
  x_center <- colMeans(X)
  y_center <- colMeans(Y)
  X <- sweep(X, 2, x_center, "-")
  Y <- sweep(Y, 2, y_center, "-")

  sx <- matrixStats::colSds(X)
  sy <- matrixStats::colSds(Y)

  keepX <- which(is.finite(sx) & sx > 0)
  keepY <- which(is.finite(sy) & sy > 0)

  dropX <- setdiff(seq_len(p0), keepX)
  dropY <- setdiff(seq_len(q0), keepY)

  X <- X[, keepX, drop = FALSE]
  Y <- Y[, keepY, drop = FALSE]

  x_scale <- rep(1, p0)
  y_scale <- rep(1, q0)
  if (standardize) {
    X <- sweep(X, 2, sx[keepX], "/")
    Y <- sweep(Y, 2, sy[keepY], "/")
    x_scale[keepX] <- sx[keepX]
    y_scale[keepY] <- sy[keepY]
  }

  p <- ncol(X)
  q <- ncol(Y)

  if (p < 1 || q < 1) {
    return(list(
      U = matrix(NA_real_, p0, r),
      V = matrix(NA_real_, q0, r),
      cor = rep(0, r),
      loss = Inf,
      Bhat = matrix(0, p0, q0),
      keepX = keepX, keepY = keepY, dropX = dropX, dropY = dropY,
      center = list(X = x_center, Y = y_center),
      scale  = list(X = x_scale,  Y = y_scale),
      converged = FALSE
    ))
  }

  # ---- map coord groups from ORIGINAL coords to filtered coords
  if (!is.null(groups)) {
    mapX <- integer(p0); mapX[keepX] <- seq_along(keepX)
    mapY <- integer(q0); mapY[keepY] <- seq_along(keepY)

    groups <- lapply(groups, function(idx) {
      if (is.matrix(idx) && ncol(idx) == 2) {
        ix <- mapX[idx[, 1]]
        iy <- mapY[idx[, 2]]
        ok <- (ix > 0) & (iy > 0)
        if (!any(ok)) return(matrix(integer(0), ncol = 2))
        cbind(ix[ok], iy[ok])
      } else {
        idx
      }
    })
  }

  # ---- linearize groups once + precompute sqrt(|g|)
  group_sqrt <- NULL
  if (!is.null(groups)) {
    groups_lin <- vector("list", length(groups))
    group_sqrt <- numeric(length(groups))
    for (g in seq_along(groups)) {
      idx <- groups[[g]]
      if (is.matrix(idx) && ncol(idx) == 2) {
        groups_lin[[g]] <- as.integer(idx[, 1] + (idx[, 2] - 1L) * p)
        group_sqrt[g] <- sqrt(nrow(idx))
      } else {
        groups_lin[[g]] <- idx
        group_sqrt[g] <- sqrt(length(idx))
      }
    }
    groups <- groups_lin
  }

  # ---- reduced bases via SVD
  EDx <- svd(X, nu = 0, nv = min(n, p))
  EDy <- svd(Y, nu = 0, nv = min(n, q))

  Ux_all <- EDx$v
  Uy_all <- EDy$v
  Lx_all <- (EDx$d^2) / n
  Ly_all <- (EDy$d^2) / n

  idx_x <- which(Lx_all > epsilon_sv)
  idx_y <- which(Ly_all > epsilon_sv)
  if (length(idx_x) < 1 || length(idx_y) < 1) {
    U_full <- matrix(0, p0, r)
    V_full <- matrix(0, q0, r)
    if (!is.null(x_names)) rownames(U_full) <- x_names
    if (!is.null(y_names)) rownames(V_full) <- y_names
    return(list(
      U = U_full, V = V_full, cor = rep(0, r), loss = Inf,
      Bhat = matrix(0, p0, q0),
      keepX = keepX, keepY = keepY, dropX = dropX, dropY = dropY,
      center = list(X = x_center, Y = y_center),
      scale  = list(X = x_scale,  Y = y_scale),
      converged = FALSE
    ))
  }

  Ux <- Ux_all[, idx_x, drop = FALSE]; Lx <- Lx_all[idx_x]
  Uy <- Uy_all[, idx_y, drop = FALSE]; Ly <- Ly_all[idx_y]
  UyT <- t(Uy)

  kx <- length(Lx)
  ky <- length(Ly)

  b <- outer(Lx, Ly) + rho

  # B1 = (X Ux)^T (Y Uy) / n
  XUx <- X %*% Ux
  YUy <- Y %*% Uy
  B1  <- crossprod(XUx, YUy) / n
  rm(XUx, YUy)

  # choose multiplication order once
  use_left_Low    <- (p * ky <= kx * q)
  use_left_Ztilde <- (kx * q <= p * ky)

  # ---- ADMM (low-memory)
  if (!is.null(B0)) {
    Z <- B0[keepX, keepY, drop = FALSE]
  } else {
    Z <- matrix(0, p, q)
  }

  # Ztilde = Ux^T Z Uy with order choice
  if (use_left_Ztilde) {
    Ztilde <- crossprod(Ux, Z) %*% Uy
  } else {
    Ztilde <- crossprod(Ux, Z %*% Uy)
  }
  Htilde <- matrix(0, kx, ky)

  converged <- FALSE
  tau_base <- lambda / rho   

  for (iter in 1:maxiter) {

    proj   <- Ztilde - Htilde
    Btilde <- (B1 + rho * proj) / b
    M      <- Btilde - proj

    # Low = Ux M Uy^T
    if (use_left_Low) {
      tmp <- Ux %*% M
      Low <- tmp %*% UyT
    } else {
      tmp <- M %*% UyT
      Low <- Ux %*% tmp
    }

    # Z_in = Low + Z   (no Z_old copy)
    Z_in <- Low + Z

    if (is.null(groups)) {
      Z <- soft_thresh(Z_in, tau_base)
    } else {
      Z <- Z_in
      for (g in seq_along(groups)) {
        idx <- groups[[g]]
        if (length(idx) == 0) next
        Z[idx] <- soft_thresh2(Z[idx], group_sqrt[g] * tau_base)
      }
    }

    # update Ztilde
    if (use_left_Ztilde) {
      Ztilde_new <- crossprod(Ux, Z) %*% Uy
    } else {
      Ztilde_new <- crossprod(Ux, Z %*% Uy)
    }

    Htilde <- Htilde + (Btilde - Ztilde_new)

    r_norm <- sqrt(sum((Btilde - Ztilde_new)^2))
    s_norm <- rho * sqrt(sum((Ztilde_new - Ztilde)^2))

    eps_pri  <- eps * (1 + max(sqrt(sum(Btilde^2)), sqrt(sum(Ztilde_new^2))))
    eps_dual <- eps * (1 + sqrt(sum(Htilde^2)))

    if (verbose && (iter %% 50 == 0)) {
      cat("\niter:", iter, " r:", signif(r_norm, 3), " s:", signif(s_norm, 3))
    }

    Ztilde <- Ztilde_new
    if (r_norm < eps_pri && s_norm < eps_dual) {
      converged <- TRUE
      break
    }
  }

  if (verbose) {
    if (!converged) cat("\nADMM did not converge (hit maxiter).\n")
    else cat("\nADMM converged.\n")
  }

  # ---- postprocessing (small SVD)
  r_eff <- min(r, nrow(Ztilde), ncol(Ztilde))
  if (r_eff < 1) {
    U_full <- matrix(0, p0, r)
    V_full <- matrix(0, q0, r)
    if (!is.null(x_names)) rownames(U_full) <- x_names
    if (!is.null(y_names)) rownames(V_full) <- y_names
    Bhat_full <- matrix(0, p0, q0); Bhat_full[keepX, keepY] <- Z
    return(list(
      U = U_full, V = V_full, cor = rep(0, r), loss = Inf,
      Bhat = Bhat_full,
      keepX = keepX, keepY = keepY, dropX = dropX, dropY = dropY,
      center = list(X = x_center, Y = y_center),
      scale  = list(X = x_scale,  Y = y_scale),
      converged = converged
    ))
  }

  # Csmall <- Ztilde
  # Csmall <- sweep(Csmall, 1, sqrt(Lx), "*")
  # Csmall <- sweep(Csmall, 2, sqrt(Ly), "*")

  # --- SVD on sparse Z (preserves row sparsity) ---
  SVDs <- safe_rank_svd(Z, r_eff, prefer_sparse = TRUE)
  if (is.null(SVDs)) {
    if (verbose) cat("\nSVD failed in ecca(). Returning NA directions.\n")
    U_full <- matrix(NA_real_, p0, r)
    V_full <- matrix(NA_real_, q0, r)
    if (!is.null(x_names)) rownames(U_full) <- x_names
    if (!is.null(y_names)) rownames(V_full) <- y_names
    Bhat_full <- matrix(0, p0, q0)
    Bhat_full[keepX, keepY] <- Z
    return(list(
      U = U_full,
      V = V_full,
      cor = rep(0, r),
      loss = Inf,
      Bhat = Bhat_full,
      keepX = keepX, keepY = keepY, dropX = dropX, dropY = dropY,
      center = list(X = x_center, Y = y_center),
      scale  = list(X = x_scale,  Y = y_scale),
      converged = FALSE
    ))
  }

  U0 <- SVDs$u
  V0 <- SVDs$v

  # --- whitening ---
  XU0 <- X %*% U0
  YV0 <- Y %*% V0

  GX <- crossprod(XU0) / n + ridge_whiten * diag(r_eff)
  GY <- crossprod(YV0) / n + ridge_whiten * diag(r_eff)

  WGX <- tryCatch(
    whiten_factor(GX, ridge = ridge_whiten, verbose = verbose),
    error = function(e) NULL
  )
  WGY <- tryCatch(
    whiten_factor(GY, ridge = ridge_whiten, verbose = verbose),
    error = function(e) NULL
  )
  if (is.null(WGX) || is.null(WGY)) {
    if (verbose) cat("\nWhitening failed in ecca(). Returning NA directions.\n")
    U_full <- matrix(NA_real_, p0, r)
    V_full <- matrix(NA_real_, q0, r)
    if (!is.null(x_names)) rownames(U_full) <- x_names
    if (!is.null(y_names)) rownames(V_full) <- y_names
    Bhat_full <- matrix(0, p0, q0)
    Bhat_full[keepX, keepY] <- Z
    return(list(
      U = U_full,
      V = V_full,
      cor = rep(0, r),
      loss = Inf,
      Bhat = Bhat_full,
      keepX = keepX, keepY = keepY, dropX = dropX, dropY = dropY,
      center = list(X = x_center, Y = y_center),
      scale  = list(X = x_scale,  Y = y_scale),
      converged = FALSE
    ))
  }

  U <- U0 %*% WGX
  V <- V0 %*% WGY

  XU <- X %*% U
  YV <- Y %*% V

  cor <- diag(crossprod(XU, YV) / n)

  # if (requireNamespace("RSpectra", quietly = TRUE) && r_eff < min(dim(Csmall))) {
  #   SVDs <- RSpectra::svds(Csmall, r_eff)
  # } else {
  #   SVDs <- svd(Csmall, nu = r_eff, nv = r_eff)
  # }

  # U0 <- Ux %*% SVDs$u
  # V0 <- Uy %*% SVDs$v

  # GX <- crossprod(X %*% U0) / n + ridge_whiten * diag(r_eff)
  # GY <- crossprod(Y %*% V0) / n + ridge_whiten * diag(r_eff)

  # Rx <- chol_jitter(GX, jitter = ridge_whiten)
  # Ry <- chol_jitter(GY, jitter = ridge_whiten)

  # U <- U0 %*% backsolve(Rx, diag(r_eff))
  # V <- V0 %*% backsolve(Ry, diag(r_eff))

  # XU <- X %*% U
  # YV <- Y %*% V
  # cor <- diag(crossprod(XU, YV) / n)

  neg <- cor < 0
  if (any(neg)) {
    V[, neg] <- -V[, neg, drop = FALSE]
    cor[neg] <- -cor[neg]
    YV[, neg] <- -YV[, neg, drop = FALSE]
  }

  ord <- order(cor, decreasing = TRUE)
  cor <- cor[ord]
  U <- U[, ord, drop = FALSE]
  V <- V[, ord, drop = FALSE]
  XU <- XU[, ord, drop = FALSE]
  YV <- YV[, ord, drop = FALSE]

  loss <- sum((XU - YV)^2) / n

  # ---- inflate to original coords
  U_full <- matrix(0, p0, r_eff)
  V_full <- matrix(0, q0, r_eff)
  U_full[keepX, ] <- U
  V_full[keepY, ] <- V

  if (!is.null(x_names)) rownames(U_full) <- x_names
  if (!is.null(y_names)) rownames(V_full) <- y_names
  colnames(U_full) <- paste0("comp", seq_len(r_eff))
  colnames(V_full) <- paste0("comp", seq_len(r_eff))

  Bhat_full <- matrix(0, p0, q0)
  Bhat_full[keepX, keepY] <- Z

  list(
    U = U_full,
    V = V_full,
    Ux = Ux, Uy = Uy, Lx = Lx, Ly = Ly,
    cor = cor,
    loss = loss,
    Bhat = Bhat_full,
    keepX = keepX, keepY = keepY, dropX = dropX, dropY = dropY,
    center = list(X = x_center, Y = y_center),
    scale  = list(X = x_scale,  Y = y_scale),
    converged = converged
  )
}

# -----------------------------
# ecca_across_lambdas (lambda path, CV-friendly)
# -----------------------------
ecca_across_lambdas <- function(
  X, Y, lambdas = 0, groups = NULL, r = 2,
  Sx = NULL, Sy = NULL, Sxy = NULL,                 # kept for compatibility; ignored
  standardize = TRUE,
  rho = 1, B0 = NULL, eps = 1e-4, maxiter = 500, verbose = TRUE,
  dense = TRUE, optimized = FALSE,                  # kept for compatibility; ignored
  epsilon_sv = 1e-8, ridge_whiten = 1e-8,
  warm_start = TRUE,
  lambda_order = c("decreasing", "increasing", "given"),
  log_prefix = "",
  admm_print_every = 50L,
  lambda_print = TRUE,
  collect_logs = FALSE,
  preprocess = TRUE,
  return_uv = TRUE,
  X_val = NULL, Y_val = NULL,
  scoring_method = c("mse", "trace")
){
  scoring_method <- match.arg(scoring_method)
  lambda_order <- match.arg(lambda_order)

  if (!is.matrix(X)) X <- as.matrix(X)
  if (!is.matrix(Y)) Y <- as.matrix(Y)
  if (nrow(X) != nrow(Y)) stop("X and Y must have same nrow().")
  logs <- character(0)

  lambdas <- as.numeric(lambdas)
  L <- length(lambdas)
  if (L < 1) stop("lambdas must have length >= 1")

  ord <- switch(
    lambda_order,
    decreasing = order(lambdas, decreasing = TRUE),
    increasing = order(lambdas, decreasing = FALSE),
    given      = seq_along(lambdas)
  )
  inv_ord <- order(ord)
  lambdas_run <- lambdas[ord]

  p0 <- ncol(X); q0 <- ncol(Y)
  x_names <- colnames(X); y_names <- colnames(Y)

  keepX <- seq_len(p0); keepY <- seq_len(q0)
  centerX <- rep(0, p0); centerY <- rep(0, q0)
  scaleX  <- rep(1, p0); scaleY  <- rep(1, q0)

  if (preprocess) {
    centerX <- colMeans(X); centerY <- colMeans(Y)
    X <- sweep(X, 2, centerX, "-")
    Y <- sweep(Y, 2, centerY, "-")

    sx <- matrixStats::colSds(X)
    sy <- matrixStats::colSds(Y)

    keepX <- which(is.finite(sx) & sx > 0)
    keepY <- which(is.finite(sy) & sy > 0)

    X <- X[, keepX, drop = FALSE]
    Y <- Y[, keepY, drop = FALSE]

    if (standardize) {
      X <- sweep(X, 2, sx[keepX], "/")
      Y <- sweep(Y, 2, sy[keepY], "/")
      scaleX[keepX] <- sx[keepX]
      scaleY[keepY] <- sy[keepY]
    }

    # map coord groups (original -> filtered)
    if (!is.null(groups)) {
      mapX <- integer(p0); mapX[keepX] <- seq_along(keepX)
      mapY <- integer(q0); mapY[keepY] <- seq_along(keepY)

      groups <- lapply(groups, function(idx) {
        if (is.matrix(idx) && ncol(idx) == 2) {
          ix <- mapX[idx[, 1]]
          iy <- mapY[idx[, 2]]
          ok <- (ix > 0) & (iy > 0)
          if (!any(ok)) return(matrix(integer(0), ncol = 2))
          cbind(ix[ok], iy[ok])
        } else {
          idx
        }
      })
    }

    # preprocess validation too
    if (!is.null(X_val) && !is.null(Y_val)) {
      if (!is.matrix(X_val)) X_val <- as.matrix(X_val)
      if (!is.matrix(Y_val)) Y_val <- as.matrix(Y_val)
      X_val <- sweep(X_val, 2, centerX, "-")
      Y_val <- sweep(Y_val, 2, centerY, "-")
      X_val <- X_val[, keepX, drop = FALSE]
      Y_val <- Y_val[, keepY, drop = FALSE]
      if (standardize) {
        X_val <- sweep(X_val, 2, scaleX[keepX], "/")
        Y_val <- sweep(Y_val, 2, scaleY[keepY], "/")
      }
    }
  } else {
    if (!is.null(X_val) && !is.null(Y_val)) {
      if (!is.matrix(X_val)) X_val <- as.matrix(X_val)
      if (!is.matrix(Y_val)) Y_val <- as.matrix(Y_val)
    }
  }

  n <- nrow(X)
  p <- ncol(X); q <- ncol(Y)

  # linearize groups once + precompute sqrt(|g|)
  group_sqrt <- NULL
  if (!is.null(groups)) {
    groups_lin <- vector("list", length(groups))
    group_sqrt <- numeric(length(groups))
    for (g in seq_along(groups)) {
      idx <- groups[[g]]
      if (is.matrix(idx) && ncol(idx) == 2) {
        groups_lin[[g]] <- as.integer(idx[, 1] + (idx[, 2] - 1L) * p)
        group_sqrt[g] <- sqrt(nrow(idx))
      } else {
        groups_lin[[g]] <- idx
        group_sqrt[g] <- sqrt(length(idx))
      }
    }
    groups <- groups_lin
  }

  if (p < 1 || q < 1) {
    out_scores <- rep(Inf, L)
    if (!return_uv) return(list(scores = out_scores[inv_ord], lambdas = lambdas))
    Uout <- lapply(seq_len(L), function(.) matrix(NA_real_, p0, r))
    Vout <- lapply(seq_len(L), function(.) matrix(NA_real_, q0, r))
    return(list(U = Uout, V = Vout, scores = out_scores[inv_ord], lambdas = lambdas))
  }

  # reduced bases
  EDx <- svd(X, nu = 0, nv = min(n, p))
  EDy <- svd(Y, nu = 0, nv = min(n, q))

  Ux_all <- EDx$v
  Uy_all <- EDy$v
  Lx_all <- (EDx$d^2) / n
  Ly_all <- (EDy$d^2) / n

  idx_x <- which(Lx_all > epsilon_sv)
  idx_y <- which(Ly_all > epsilon_sv)
  if (length(idx_x) < 1 || length(idx_y) < 1) {
    out_scores <- rep(Inf, L)
    if (!return_uv) return(list(scores = out_scores[inv_ord], lambdas = lambdas))
    Uout <- lapply(seq_len(L), function(.) matrix(NA_real_, p0, r))
    Vout <- lapply(seq_len(L), function(.) matrix(NA_real_, q0, r))
    return(list(U = Uout, V = Vout, scores = out_scores[inv_ord], lambdas = lambdas))
  }

  Ux <- Ux_all[, idx_x, drop = FALSE]; Lx <- Lx_all[idx_x]
  Uy <- Uy_all[, idx_y, drop = FALSE]; Ly <- Ly_all[idx_y]
  UyT <- t(Uy)

  kx <- length(Lx); ky <- length(Ly)
  b  <- outer(Lx, Ly) + rho

  # B1 once
  XUx <- X %*% Ux
  YUy <- Y %*% Uy
  B1  <- crossprod(XUx, YUy) / n
  rm(XUx, YUy)

  use_left_Low    <- (p * ky <= kx * q)
  use_left_Ztilde <- (kx * q <= p * ky)

  compute_Ztilde <- function(Zmat) {
    if (use_left_Ztilde) crossprod(Ux, Zmat) %*% Uy
    else crossprod(Ux, Zmat %*% Uy)
  }
  compute_Low <- function(M) {
    if (use_left_Low) (Ux %*% M) %*% UyT
    else Ux %*% (M %*% UyT)
  }

  out_scores <- rep(NA_real_, L)
  out_cor    <- vector("list", L)
  out_loss   <- rep(NA_real_, L)
  out_U      <- if (return_uv) vector("list", L) else NULL
  out_V      <- if (return_uv) vector("list", L) else NULL

  if (!is.null(B0)) {
    Z <- if (preprocess) B0[keepX, keepY, drop = FALSE] else B0
  } else {
    Z <- matrix(0, p, q)
  }
  admm_print_every <- as.integer(admm_print_every)
  if (!is.finite(admm_print_every) || admm_print_every < 1L) admm_print_every <- 50L

  for (t in seq_len(L)) {
    lam <- lambdas_run[t]
    t_lambda0 <- Sys.time()
    if (verbose && lambda_print) {
      cat(sprintf("\nStarting lambda %d/%d: %g", t, L, lam))
      cat(sprintf("\n%s lambda %d/%d = %g", log_prefix, t, L, lam))
      utils::flush.console()
      if (collect_logs) {
        logs <- c(logs,
                  sprintf("Starting lambda %d/%d: %g", t, L, lam),
                  sprintf("%s lambda %d/%d = %g", log_prefix, t, L, lam))
      }
    }


    if (!warm_start && t > 1) Z <- matrix(0, p, q)

    Ztilde <- compute_Ztilde(Z)
    Htilde <- matrix(0, kx, ky)

    tau_base <- lam / rho
    converged <- FALSE

    for (iter in seq_len(maxiter)) {
      proj   <- Ztilde - Htilde
      Btilde <- (B1 + rho * proj) / b
      M      <- Btilde - proj

      Low <- compute_Low(M)

      Z_in <- Low + Z

      if (is.null(groups)) {
        Z <- soft_thresh(Z_in, tau_base)
      } else {
        Z <- Z_in
        for (g in seq_along(groups)) {
          idx <- groups[[g]]
          if (length(idx) == 0) next
          Z[idx] <- soft_thresh2(Z[idx], group_sqrt[g] * tau_base)
        }
      }

      Ztilde_new <- compute_Ztilde(Z)
      Htilde <- Htilde + (Btilde - Ztilde_new)

      r_norm <- sqrt(sum((Btilde - Ztilde_new)^2))
      s_norm <- rho * sqrt(sum((Ztilde_new - Ztilde)^2))

      eps_pri  <- eps * (1 + max(sqrt(sum(Btilde^2)), sqrt(sum(Ztilde_new^2))))
      eps_dual <- eps * (1 + sqrt(sum(Htilde^2)))

      if (verbose && iter %% 100 == 0) {
        cat("\nlambda", signif(lam, 3), "iter", iter,
            "r", signif(r_norm, 3), "s", signif(s_norm, 3))
        if (collect_logs) {
          logs <- c(logs, sprintf("lambda %s iter %d r %s s %s",
                                  signif(lam, 3), iter, signif(r_norm, 3), signif(s_norm, 3)))
        }
      }
          # inside ADMM iter loop
      if (verbose && (iter %% admm_print_every == 0L)) {
        cat(sprintf("\n%s   iter %d  r=%g  s=%g", log_prefix, iter, r_norm, s_norm))
        utils::flush.console()
        if (collect_logs) {
          logs <- c(logs, sprintf("%s   iter %d  r=%g  s=%g", log_prefix, iter, r_norm, s_norm))
        }
      }

      Ztilde <- Ztilde_new
      if (r_norm < eps_pri && s_norm < eps_dual) {
        converged <- TRUE
        break
      }
    }

    # postprocessing small SVD
    Csmall <- Ztilde
    Csmall <- sweep(Csmall, 1, sqrt(Lx), "*")
    Csmall <- sweep(Csmall, 2, sqrt(Ly), "*")

    r_eff <- min(r, nrow(Csmall), ncol(Csmall))
    if (r_eff < 1) {
      out_scores[t] <- Inf
      out_loss[t]   <- Inf
      out_cor[[t]]  <- rep(0, r)
      if (return_uv) {
        out_U[[t]] <- matrix(0, p0, r)
        out_V[[t]] <- matrix(0, q0, r)
      }
      next
    }

    SVDs <- safe_rank_svd(Z, r_eff, prefer_sparse = TRUE)
    if (is.null(SVDs)) {
      out_scores[t] <- Inf
      out_loss[t]   <- Inf
      out_cor[[t]]  <- rep(0, r)
      if (return_uv) {
        out_U[[t]] <- matrix(0, p0, r)
        out_V[[t]] <- matrix(0, q0, r)
      }
      if (collect_logs) {
        logs <- c(logs, sprintf("%s SVD failed at lambda=%g; assigned Inf score", log_prefix, lam))
      }
      next
    }

    Uhat <- SVDs$u
    Vhat <- SVDs$v

    # --- whitening ---
    XU0 <- X %*% Uhat
    YV0 <- Y %*% Vhat

    GX <- crossprod(XU0) / n + ridge_whiten * diag(r_eff)
    GY <- crossprod(YV0) / n + ridge_whiten * diag(r_eff)

    WGX <- tryCatch(
      whiten_factor(GX, ridge = ridge_whiten, verbose = FALSE),
      error = function(e) NULL
    )
    WGY <- tryCatch(
      whiten_factor(GY, ridge = ridge_whiten, verbose = FALSE),
      error = function(e) NULL
    )
    if (is.null(WGX) || is.null(WGY)) {
      out_scores[t] <- Inf
      out_loss[t]   <- Inf
      out_cor[[t]]  <- rep(0, r)
      if (return_uv) {
        out_U[[t]] <- matrix(0, p0, r)
        out_V[[t]] <- matrix(0, q0, r)
      }
      if (collect_logs) {
        logs <- c(logs, sprintf("%s whitening failed at lambda=%g; assigned Inf score", log_prefix, lam))
      }
      next
    }

    Uhat <- Uhat %*% WGX
    Vhat <- Vhat %*% WGY

    XU <- X %*% Uhat
    YV <- Y %*% Vhat

    cor <- diag(crossprod(XU, YV) / n)

    neg <- cor < 0
    if (any(neg)) {
      Vhat[, neg] <- -Vhat[, neg, drop = FALSE]
      cor[neg] <- -cor[neg]
      YV[, neg] <- -YV[, neg, drop = FALSE]
    }

    ord_cor <- order(cor, decreasing = TRUE)
    cor <- cor[ord_cor]
    Uhat <- Uhat[, ord_cor, drop = FALSE]
    Vhat <- Vhat[, ord_cor, drop = FALSE]
    XU <- XU[, ord_cor, drop = FALSE]
    YV <- YV[, ord_cor, drop = FALSE]

    if (!is.null(X_val) && !is.null(Y_val)) {
      XU_val <- X_val %*% Uhat
      YV_val <- Y_val %*% Vhat
      if (scoring_method == "mse") {
        out_scores[t] <- mean((XU_val - YV_val)^2)
      } else {
        # Faster than cor(): compute only paired (diagonal) correlations.
        XUc <- scale(XU_val, center = TRUE, scale = FALSE)
        YVc <- scale(YV_val, center = TRUE, scale = FALSE)
        num <- colSums(XUc * YVc)
        den <- sqrt(colSums(XUc^2) * colSums(YVc^2))
        den[den < 1e-12 | !is.finite(den)] <- 1e-12
        corr_diag <- num / den
        out_scores[t] <- -mean(corr_diag)
      }
    } else {
      out_scores[t] <- NA_real_
    }

    out_loss[t]  <- sum((XU - YV)^2) / n
    out_cor[[t]] <- cor

    if (return_uv) {
      U_full <- matrix(0, p0, r_eff)
      V_full <- matrix(0, q0, r_eff)
      U_full[keepX, ] <- Uhat
      V_full[keepY, ] <- Vhat
      if (!is.null(x_names)) rownames(U_full) <- x_names
      if (!is.null(y_names)) rownames(V_full) <- y_names
      out_U[[t]] <- U_full
      out_V[[t]] <- V_full
    }
    if (verbose && lambda_print) {
      cat(sprintf("\nFinished lambda %d/%d in %.1fs",
                  t, L, as.numeric(difftime(Sys.time(), t_lambda0, units="secs"))))
      if (collect_logs) {
        logs <- c(logs, sprintf("Finished lambda %d/%d in %.1fs",
                                t, L, as.numeric(difftime(Sys.time(), t_lambda0, units="secs"))))
      }
    }
  }

  # restore original lambda order
  out_scores <- out_scores[inv_ord]
  out_loss   <- out_loss[inv_ord]
  out_cor    <- out_cor[inv_ord]
  if (return_uv) {
    out_U <- out_U[inv_ord]
    out_V <- out_V[inv_ord]
  }

  if (!return_uv) {
    out <- list(scores = out_scores, lambdas = lambdas)
    if (collect_logs) out$logs <- logs
    return(out)
  }

  if (length(lambdas) == 1) {
    out <- list(U = out_U[[1]], V = out_V[[1]],
                cor = out_cor[[1]], loss = out_loss[1],
                scores = out_scores, lambdas = lambdas,
                center = list(X = centerX, Y = centerY),
                scale  = list(X = scaleX,  Y = scaleY),
                keepX = keepX, keepY = keepY)
    if (collect_logs) out$logs <- logs
    return(out)
  }

  out <- list(U = out_U, V = out_V,
              cor = out_cor, loss = out_loss,
              scores = out_scores, lambdas = lambdas,
              center = list(X = centerX, Y = centerY),
              scale  = list(X = scaleX,  Y = scaleY),
              keepX = keepX, keepY = keepY,
              Ux = Ux, Uy = Uy, Lx = Lx, Ly = Ly)
  if (collect_logs) out$logs <- logs
  out
}

# -----------------------------
# ecca.eval (CV) - with seamless group mapping + faster parallel behavior
# -----------------------------
ecca.eval <- function(
  X, Y, lambdas = 0, groups = NULL, r = 2,
  standardize = TRUE,
  rho = 1, B0 = NULL, nfold = 5,
  eps = 1e-4, maxiter = 500, verbose = TRUE,
  parallel = TRUE, nb_cores = NULL, set_seed_cv = NULL,
  scoring_method = c("mse", "trace"),
  cv_use_median = FALSE,
  dense = TRUE, optimized = FALSE,
  epsilon_sv = 1e-8, ridge_whiten = 1e-8
){
  scoring_method <- match.arg(scoring_method)
  if (!is.matrix(X)) X <- as.matrix(X)
  if (!is.matrix(Y)) Y <- as.matrix(Y)
  if (nrow(X) != nrow(Y)) stop("X and Y must have same nrow().")

  n <- nrow(X)
  if (n < 2 * nfold) {
    if (verbose) cat("\nWarning: n too small for CV; using first lambda.\n")
    scores <- data.frame(lambda = lambdas, mse = NA_real_, se = NA_real_)
    return(list(scores = scores, lambda.min = lambdas[1], lambda.1se = lambdas[1]))
  }

  # global preprocessing
  p0 <- ncol(X); q0 <- ncol(Y)

  X <- scale(X, center = TRUE, scale = FALSE)
  Y <- scale(Y, center = TRUE, scale = FALSE)

  sx <- matrixStats::colSds(X)
  sy <- matrixStats::colSds(Y)
  keepX <- which(is.finite(sx) & sx > 0)
  keepY <- which(is.finite(sy) & sy > 0)

  X <- X[, keepX, drop = FALSE]
  Y <- Y[, keepY, drop = FALSE]

  if (standardize) {
    X <- sweep(X, 2, sx[keepX], "/")
    Y <- sweep(Y, 2, sy[keepY], "/")
  }

  n <- nrow(X)

  # IMPORTANT: map groups to the filtered coordinates ONCE (since preprocess=FALSE in folds)
  if (!is.null(groups)) {
    mapX <- integer(p0); mapX[keepX] <- seq_along(keepX)
    mapY <- integer(q0); mapY[keepY] <- seq_along(keepY)

    groups <- lapply(groups, function(idx) {
      if (is.matrix(idx) && ncol(idx) == 2) {
        ix <- mapX[idx[, 1]]
        iy <- mapY[idx[, 2]]
        ok <- (ix > 0) & (iy > 0)
        if (!any(ok)) return(matrix(integer(0), ncol = 2))
        cbind(ix[ok], iy[ok])
      } else {
        idx
      }
    })
  }

  if (!is.null(set_seed_cv)) set.seed(set_seed_cv)
  folds <- .create_cv_folds(n, nfold)

  # parallel setup
  if (parallel) {
    cores <- if (is.null(nb_cores)) max(1L, parallel::detectCores() - 1L) else as.integer(nb_cores)
    cores <- min(cores, length(folds))
    cores <- .limit_parallel_cores(cores)
    Sys.setenv(
      OMP_NUM_THREADS = 1,
      OPENBLAS_NUM_THREADS = 1,
      MKL_NUM_THREADS = 1,
      VECLIB_MAXIMUM_THREADS = 1
    )

    cl <- setup_parallel_backend(cores, verbose = verbose)
    if (is.null(cl)) {
      warning("All parallel setup attempts failed. Proceeding in serial mode.", immediate. = TRUE)
      parallel <- FALSE
    }
  }

  if (parallel) {
    initialize_parallel_workers(cl, verbose = verbose)
    on.exit(cleanup_parallel_backend(cl), add = TRUE)
    if (verbose) {
      cat("\nRunning CV in parallel on", cores, "cores\n")
    }
    ecca_across_lambdas_current <- ecca_across_lambdas
    safe_rank_svd_current <- safe_rank_svd
    whiten_factor_current <- whiten_factor
    chol_jitter_current <- chol_jitter
    invsqrt_psd_current <- invsqrt_psd
    soft_thresh_current <- soft_thresh
    soft_thresh2_current <- soft_thresh2
    safe_rank_svd <- safe_rank_svd_current
    whiten_factor <- whiten_factor_current
    chol_jitter <- chol_jitter_current
    invsqrt_psd <- invsqrt_psd_current
    soft_thresh <- soft_thresh_current
    soft_thresh2 <- soft_thresh2_current
    collect_logs_parallel <- isTRUE(verbose)
    ecca_parallel_fold_worker <- function(fold_id) {
      tryCatch({
        fold <- folds[[fold_id]]
        pid <- Sys.getpid()
        prefix <- sprintf("[pid=%d fold=%d]", pid, fold_id)

        tr <- setdiff(seq_len(n), fold)
        Xtr <- X[tr, , drop = FALSE]
        Ytr <- Y[tr, , drop = FALSE]
        Xva <- X[fold, , drop = FALSE]
        Yva <- Y[fold, , drop = FALSE]

        fit <- ecca_across_lambdas_current(
          Xtr, Ytr, lambdas = lambdas, groups = groups, r = r,
          standardize = FALSE, rho = rho, B0 = B0,
          eps = eps, maxiter = maxiter,
          verbose = collect_logs_parallel,
          log_prefix = prefix,
          admm_print_every = 50L,
          lambda_print = collect_logs_parallel,
          epsilon_sv = epsilon_sv, ridge_whiten = ridge_whiten,
          preprocess = FALSE,
          return_uv = FALSE,
          collect_logs = collect_logs_parallel,
          X_val = Xva, Y_val = Yva,
          scoring_method = scoring_method
        )

        list(
          fold_id = fold_id,
          scores = fit$scores,
          logs = fit$logs
        )
      }, error = function(e) e)
    }

    parallel::clusterExport(
      cl,
      varlist = c(
        "X", "Y", "folds", "n",
        "lambdas", "groups", "r",
        "rho", "B0", "eps", "maxiter",
        "epsilon_sv", "ridge_whiten", "scoring_method",
        "collect_logs_parallel",
        "ecca_across_lambdas_current",
        "safe_rank_svd",
        "whiten_factor",
        "chol_jitter",
        "invsqrt_psd",
        "soft_thresh",
        "soft_thresh2",
        "ecca_parallel_fold_worker"
      ),
      envir = environment()
    )

    cv_list <- parallel::parLapplyLB(cl, seq_along(folds), ecca_parallel_fold_worker)

    if (verbose) {
      ok_folds <- cv_list[!vapply(cv_list, inherits, logical(1), "error")]
      if (length(ok_folds) > 0) {
        for (res in ok_folds) {
          if (!is.null(res$logs) && length(res$logs) > 0) {
            cat("\n")
            cat(paste(res$logs, collapse = "\n"))
            cat("\n")
          }
        }
        utils::flush.console()
      }
    }

    is_err <- vapply(cv_list, inherits, logical(1), "error")

    if (all(is_err)) {
      msgs <- unique(vapply(cv_list, function(e) {
        msg <- tryCatch(conditionMessage(e), error = function(...) "")
        if (!nzchar(msg)) msg <- paste(class(e), collapse = "/")
        msg
      }, character(1)))
      stop("All CV folds failed. Example error(s):\n- ",
           paste(utils::head(msgs, 5), collapse = "\n- "))
    }

    if (any(is_err)) warning(sum(is_err), " folds failed.")
    cv_list <- lapply(cv_list[!is_err], function(x) x$scores)
    if (length(cv_list) == 0) stop("All CV folds failed.")
    scores.cv <- do.call(cbind, cv_list)
    n_success <- ncol(scores.cv)

    

  } else {
    cv_list <- lapply(folds, function(fold) {
      tryCatch({
        tr <- setdiff(seq_len(n), fold)
        Xtr <- X[tr, , drop = FALSE]
        Ytr <- Y[tr, , drop = FALSE]
        Xva <- X[fold, , drop = FALSE]
        Yva <- Y[fold, , drop = FALSE]

        fit <- ecca_across_lambdas(
          Xtr, Ytr, lambdas = lambdas, groups = groups, r = r,
          standardize = FALSE, rho = rho, B0 = B0,
          eps = eps, maxiter = maxiter, verbose = TRUE,
          epsilon_sv = epsilon_sv, ridge_whiten = ridge_whiten,
          preprocess = FALSE,
          return_uv = FALSE,
          X_val = Xva, Y_val = Yva,
          scoring_method = scoring_method
        )
        fit$scores
      }, error = function(e) e)
    })

    is_err <- vapply(cv_list, inherits, logical(1), "error")
    if (any(is_err)) warning(sum(is_err), " folds failed.")
    cv_list <- cv_list[!is_err]
    if (length(cv_list) == 0) stop("All CV folds failed.")
    scores.cv <- do.call(cbind, cv_list)
    n_success <- ncol(scores.cv)
  }

  if (!cv_use_median) {
    mse <- rowMeans(scores.cv, na.rm = TRUE)
  } else {
    mse <- apply(scores.cv, 1, median, na.rm = TRUE)
  }
  se <- matrixStats::rowSds(scores.cv, na.rm = TRUE) / sqrt(n_success)
  se[!is.finite(se)] <- NA_real_
  scores <- data.frame(lambda = lambdas, mse = mse, se = se)

  lambda.min <- scores$lambda[which.min(scores$mse)]
  upper <- scores$mse[which.min(scores$mse)] + scores$se[which.min(scores$mse)]
  lambda.1se <- max(scores$lambda[scores$mse <= upper])

  list(scores = scores, lambda.min = lambda.min, lambda.1se = lambda.1se)
}

#' Cross-Validated Efficient CCA
#'
#' Selects a regularization parameter for [ecca()] by cross-validation and
#' refits the final model at the selected value.
#'
#' @param X Predictor matrix (n x p).
#' @param Y Response matrix (n x q).
#' @param lambdas Candidate regularization values.
#' @param groups Optional group structure for blockwise sparsity.
#' @param r Target rank.
#' @param standardize Whether to scale variables after centering.
#' @param rho ADMM penalty parameter.
#' @param B0 Optional warm start for the coefficient matrix.
#' @param nfold Number of cross-validation folds.
#' @param select Selection rule for the final lambda. One of `"lambda.min"` or `"lambda.1se"`.
#' @param eps Convergence tolerance for the final ADMM refit.
#' @param maxiter Maximum iterations for the final ADMM refit.
#' @param verbose Whether to print diagnostics.
#' @param maxiter_cv Maximum iterations used inside the cross-validation fits.
#' @param parallel Whether to parallelize cross-validation.
#' @param nb_cores Number of worker processes to use when `parallel = TRUE`.
#' @param set_seed_cv Optional random seed for fold generation.
#' @param scoring_method Cross-validation score to optimize. One of `"mse"` or `"trace"`.
#' @param cv_use_median Whether to aggregate fold scores with the median instead of the mean.
#' @param dense Retained for backward compatibility.
#' @param optimized Retained for backward compatibility.
#' @param epsilon_sv Numerical threshold used to discard near-zero singular values.
#' @param ridge_whiten Ridge added when whitening Gram matrices.
#'
#' @return A list with the final fit, selected lambda, and cross-validation
#'   scores when more than one lambda is supplied.
#' @export
ecca.cv <- function(
  X, Y, lambdas = 0, groups = NULL, r = 2, standardize = FALSE,
  rho = 1, B0 = NULL, nfold = 5, select = "lambda.min",
  eps = 1e-3, maxiter = 1000, verbose = FALSE, maxiter_cv = 300,
  parallel = FALSE, nb_cores = NULL, set_seed_cv = NULL,
  scoring_method = c("mse", "trace"), cv_use_median = FALSE,
  dense = TRUE, optimized = FALSE,
  epsilon_sv = 1e-8, ridge_whiten = 1e-8
){
  scoring_method <- match.arg(scoring_method)

  eval <- NULL
  if (length(lambdas) > 1) {
    eval <- ecca.eval(
      X, Y, lambdas = lambdas, groups = groups, r = r,
      standardize = standardize,
      rho = rho, B0 = B0, nfold = nfold,
      eps = eps, maxiter = maxiter_cv, verbose = verbose,
      parallel = parallel, nb_cores = nb_cores, set_seed_cv = set_seed_cv,
      scoring_method = scoring_method, cv_use_median = cv_use_median,
      epsilon_sv = epsilon_sv, ridge_whiten = ridge_whiten
    )
    lambda.opt <- if (select == "lambda.1se") eval$lambda.1se else eval$lambda.min
  } else {
    lambda.opt <- as.numeric(lambdas)
  }

  if (verbose) cat("\nselected lambda:", lambda.opt, "\n")

  final_verbose <- TRUE
  cat("\nRefitting at selected lambda with ADMM logs...\n")
  utils::flush.console()

  fit <- ecca_across_lambdas(
    X, Y, lambdas = lambda.opt, groups = groups, r = r,
    standardize = standardize,
    rho = rho, B0 = B0, eps = eps, maxiter = maxiter, verbose = final_verbose,
    admm_print_every = 1L,
    lambda_print = TRUE,
    epsilon_sv = epsilon_sv, ridge_whiten = ridge_whiten,
    preprocess = TRUE,
    return_uv = TRUE
  )
  cat("\nFinal ADMM refit completed.\n")
  utils::flush.console()

  out <- list(
    U = fit$U,
    V = fit$V,
    cor = fit$cor,
    loss = fit$loss,
    lambda.opt = lambda.opt
  )
  if (!is.null(eval)) out$cv.scores <- eval$scores
  out
}

.resolve_preprocess_mode <- function(standardize = NULL, preprocess = NULL) {
  if (is.logical(preprocess) && length(preprocess) == 1L) {
    if (isTRUE(preprocess)) {
      preprocess <- NULL
    } else {
      return("none")
    }
  }

  if (!is.null(preprocess)) {
    return(match.arg(preprocess, c("scale", "center", "none")))
  }

  if (is.character(standardize)) {
    return(match.arg(standardize, c("scale", "center", "none")))
  }

  if (is.logical(standardize) && length(standardize) == 1L) {
    return(if (isTRUE(standardize)) "scale" else "center")
  }

  if (is.numeric(standardize) && length(standardize) == 1L && is.finite(standardize)) {
    return(if (isTRUE(as.logical(standardize))) "scale" else "center")
  }

  if (is.null(standardize)) {
    return("center")
  }

  stop("`standardize` must be logical or one of: 'scale', 'center', 'none'.", call. = FALSE)
}

.preprocess_matrix <- function(M, mode) {
  M <- as.matrix(M)
  if (mode == "none") {
    return(M)
  }

  M <- scale(M, center = TRUE, scale = identical(mode, "scale"))
  M <- as.matrix(M)
  M[!is.finite(M)] <- 0
  M
}

.as_double_matrix <- function(M) {
  M <- as.matrix(M)
  storage.mode(M) <- "double"
  M
}

.graph_null_basis <- function(Gamma, tol = 1e-10) {
  p <- ncol(Gamma)
  if (p < 1) {
    return(matrix(0, 0, 0))
  }

  parent <- seq_len(p)
  rank <- integer(p)

  find_root <- function(x) {
    root <- x
    while (parent[root] != root) {
      root <- parent[root]
    }
    while (parent[x] != x) {
      next_x <- parent[x]
      parent[x] <<- root
      x <- next_x
    }
    root
  }

  union_nodes <- function(a, b) {
    ra <- find_root(a)
    rb <- find_root(b)
    if (ra == rb) {
      return(invisible(NULL))
    }

    if (rank[ra] < rank[rb]) {
      parent[ra] <<- rb
    } else if (rank[ra] > rank[rb]) {
      parent[rb] <<- ra
    } else {
      parent[rb] <<- ra
      rank[ra] <<- rank[ra] + 1L
    }

    invisible(NULL)
  }

  nz_by_row <- if (inherits(Gamma, "sparseMatrix")) {
    gamma_summary <- Matrix::summary(Gamma)
    gamma_summary <- gamma_summary[abs(gamma_summary$x) > tol, c("i", "j"), drop = FALSE]
    split(gamma_summary$j, gamma_summary$i)
  } else {
    nz_idx <- which(abs(Gamma) > tol, arr.ind = TRUE)
    split(nz_idx[, "col"], nz_idx[, "row"])
  }

  for (cols in nz_by_row) {
    cols <- unique(as.integer(cols))
    if (length(cols) < 2L) {
      next
    }
    anchor <- cols[1]
    for (col_idx in cols[-1]) {
      union_nodes(anchor, col_idx)
    }
  }

  roots <- vapply(seq_len(p), find_root, integer(1))
  components <- split(seq_len(p), roots)
  basis <- matrix(0, p, length(components))

  for (k in seq_along(components)) {
    idx <- components[[k]]
    basis[idx, k] <- 1 / sqrt(length(idx))
  }

  basis
}

.make_graph_invSx_apply <- function(A, rho, ridge = 1e-8, verbose = FALSE) {
  A <- .as_double_matrix(A)
  n <- nrow(A)
  p <- ncol(A)

  if (p <= n) {
    gram <- crossprod(A) + rho * diag(p)
    R <- tryCatch(chol(gram + ridge * diag(p)), error = function(e) NULL)

    solve_gram <- if (!is.null(R)) {
      function(B) backsolve(R, forwardsolve(t(R), B))
    } else {
      if (verbose) message("Cholesky failed in graph ADMM solve; using solve().")
      function(B) solve(gram, B)
    }

    return(list(
      apply = function(rhs) solve_gram(.as_double_matrix(rhs)),
      uses_woodbury = FALSE
    ))
  }

  K <- diag(n) + tcrossprod(A) / rho
  R <- tryCatch(chol(K + ridge * diag(n)), error = function(e) NULL)

  solve_K <- if (!is.null(R)) {
    function(B) backsolve(R, forwardsolve(t(R), B))
  } else {
    if (verbose) message("Cholesky failed in graph Woodbury solve; using solve().")
    function(B) solve(K, B)
  }

  list(
    apply = function(rhs) {
      rhs <- .as_double_matrix(rhs)
      middle <- solve_K(A %*% rhs)
      rhs / rho - crossprod(A, middle) / (rho^2)
    },
    uses_woodbury = TRUE
  )
}

.prepare_graph_cache <- function(Gamma, gamma_ridge, Gamma_dagger = NULL) {
  if (is.null(Gamma_dagger)) {
    Gamma <- if (inherits(Gamma, "dgCMatrix")) {
      Gamma
    } else {
      methods::as(Matrix::Matrix(Gamma, sparse = TRUE), "dgCMatrix")
    }

    L_eps <- Matrix::crossprod(Gamma) + gamma_ridge * Matrix::Diagonal(n = ncol(Gamma))
    L_fac <- Matrix::Cholesky(L_eps, LDL = FALSE)

    solve_L <- function(rhs) {
      as.matrix(Matrix::solve(L_fac, .as_double_matrix(rhs), system = "A"))
    }

    return(list(
      mode = "regularized",
      Gamma = Gamma,
      gamma_ridge = gamma_ridge,
      solve_L = solve_L,
      null_basis = .graph_null_basis(Gamma)
    ))
  }

  list(
    mode = "explicit",
    Gamma = Gamma,
    Gamma_dagger = Gamma_dagger,
    gamma_ridge = gamma_ridge
  )
}

.transform_graph_design <- function(graph_cache, X = NULL, solved_tX = NULL) {
  if (graph_cache$mode == "regularized") {
    if (is.null(solved_tX)) {
      if (is.null(X)) stop("Either `X` or `solved_tX` must be provided.", call. = FALSE)
      solved_tX <- graph_cache$solve_L(t(X))
    }

    return(list(
      XGamma_dagger = as.matrix(Matrix::t(graph_cache$Gamma %*% solved_tX)),
      XPi = if (is.null(X)) NULL else .as_double_matrix(X %*% graph_cache$null_basis),
      solved_tX = solved_tX
    ))
  }

  if (is.null(X)) {
    stop("`X` must be provided when using an explicit Gamma_dagger.", call. = FALSE)
  }

  XGamma_dagger <- as.matrix(X %*% graph_cache$Gamma_dagger)
  list(
    XGamma_dagger = XGamma_dagger,
    XPi = as.matrix(X - XGamma_dagger %*% graph_cache$Gamma)
  )
}

.reconstruct_graph_coefficients <- function(graph_cache, B, Projection) {
  if (graph_cache$mode == "regularized") {
    return(list(
      gamma_dagger_B = graph_cache$solve_L(Matrix::crossprod(graph_cache$Gamma, .as_double_matrix(B))),
      Pi_projection = .as_double_matrix(graph_cache$null_basis %*% Projection)
    ))
  }

  list(
    gamma_dagger_B = graph_cache$Gamma_dagger %*% B,
    Pi_projection = Projection - as.matrix(graph_cache$Gamma_dagger %*% (graph_cache$Gamma %*% Projection))
  )
}

.project_graph_rrr_coefficients <- function(Bhat, X, r, Sx = NULL) {
  r_eff <- min(r, nrow(Bhat), ncol(Bhat))
  if (r_eff < 1 || r_eff >= ncol(Bhat)) {
    return(Bhat)
  }

  if (is.null(Sx)) {
    sx_svd <- svd(X / sqrt(nrow(X)), nu = 0, nv = min(nrow(X), ncol(X)))
    sqrt_Sx_B <- sx_svd$v %*% (sx_svd$d * crossprod(sx_svd$v, Bhat))
  } else {
    svd_Sx <- svd(as.matrix(Sx), nu = 0, nv = ncol(Sx))
    sqrt_Sx_B <- svd_Sx$v %*% (sqrt(pmax(svd_Sx$d, 0)) * crossprod(svd_Sx$v, Bhat))
  }

  weighted_svd <- .rrr_safe_rank_svd(sqrt_Sx_B, r_eff, prefer_sparse = FALSE)
  if (is.null(weighted_svd)) {
    return(Bhat)
  }

  Vr <- weighted_svd$v[, seq_len(r_eff), drop = FALSE]
  Bhat %*% Vr %*% t(Vr)
}

.fit_graph_rrr_preprocessed <- function(X, Y,
                                        Sx = NULL, Sy = NULL,
                                        lambda = 0, r,
                                        LW_Sy = TRUE, rho = 10,
                                        niter = 1e4, thresh = 1e-4,
                                        thresh_0 = 1e-6,
                                        verbose = FALSE,
                                        graph_cache,
                                        graph_design = NULL) {
  n <- nrow(X)
  p <- ncol(X)
  q <- ncol(Y)

  if (q > n) {
    stop("The X/Y swapping heuristic (for q > n) is not compatible with the graph constraint Gamma. Swap X and Y in your arguments.")
  }

  if (n < min(p, q)) {
    warning("Both X and Y are high dimensional; method may be unstable.")
  }

  fro_norm <- function(M) sqrt(sum(M^2))

  if (is.null(Sy)) {
    Sy <- crossprod(Y) / n
    if (LW_Sy) Sy <- as.matrix(corpcor::cov.shrink(Y))
  }

  svd_Sy <- svd(as.matrix(Sy))
  sqrt_inv_Sy <- svd_Sy$u %*% diag(ifelse(svd_Sy$d > 1e-4, 1 / sqrt(svd_Sy$d), 0)) %*% t(svd_Sy$v)
  tilde_Y <- Y %*% sqrt_inv_Sy

  if (is.null(graph_design)) {
    graph_design <- .transform_graph_design(graph_cache, X = X)
  }

  XGamma_dagger <- graph_design$XGamma_dagger
  XPi <- if (is.null(graph_design$XPi) && graph_cache$mode == "regularized") {
    .as_double_matrix(X %*% graph_cache$null_basis)
  } else {
    graph_design$XPi
  }

  Projection <- tryCatch(
    qr.solve(XPi, tilde_Y),
    error = function(e) {
      XtX <- .as_double_matrix(crossprod(XPi) + graph_cache$gamma_ridge * diag(ncol(XPi)))
      rhs_proj <- .as_double_matrix(crossprod(XPi, tilde_Y))
      solve(XtX, rhs_proj)
    }
  )
  new_Ytilde <- tilde_Y - XPi %*% Projection

  if (verbose) {
    cat("Norm of Projection:", fro_norm(Projection), "\n")
  }

  new_p <- ncol(XGamma_dagger)
  prod_xy <- crossprod(XGamma_dagger, new_Ytilde) / n
  invSx_info <- .make_graph_invSx_apply(XGamma_dagger / sqrt(n), rho, verbose = verbose)
  solve_invSx <- invSx_info$apply

  U_dual <- matrix(0, new_p, q)
  Z <- matrix(0, new_p, q)
  B <- matrix(0, new_p, q)

  for (i in seq_len(niter)) {
    Z_old <- Z

    B <- solve_invSx(prod_xy + rho * (Z - U_dual))
    Z <- B + U_dual

    norms <- sqrt(rowSums(Z^2))
    shrink_factors <- pmax(0, 1 - lambda / (rho * norms))
    shrink_factors[is.nan(shrink_factors)] <- 0
    Z <- Z * shrink_factors

    U_dual <- U_dual + B - Z

    primal <- fro_norm(Z - B)
    dual <- fro_norm(Z_old - Z)
    if (verbose) {
      cat("ADMM iter:", i, "Primal:", primal, "Dual:", dual, "\n")
    }
    if (max(primal / sqrt(new_p), dual / sqrt(new_p)) < thresh) break
  }

  reconstruction <- .reconstruct_graph_coefficients(graph_cache, B, Projection)
  gamma_dagger_B <- reconstruction$gamma_dagger_B
  Pi_projection <- reconstruction$Pi_projection

  B_opt <- gamma_dagger_B + Pi_projection
  B_opt[abs(B_opt) < thresh_0] <- 0

  if (verbose) {
    cat("Norm of Pi Projection:", fro_norm(Pi_projection), "\n")
    cat("Norm of Gamma_dagger %*% B:", fro_norm(gamma_dagger_B), "\n")
    cat("Norm of XPi:", fro_norm(XPi), "\n")
    cat("rank of XPi:", qr(XPi)$rank, "\n")
  }

  B_post <- .project_graph_rrr_coefficients(B_opt, X, r, Sx = Sx)
  post <- .postprocess_rrr_fit(
    B_post, X, Y, sqrt_inv_Sy, r,
    ridge_whiten = max(thresh_0, 1e-8),
    thresh_0 = thresh_0,
    verbose = verbose
  )
  regression_loss <- mean((X %*% B_post - tilde_Y)^2)

  list(
    U = post$U,
    V = post$V,
    loss = post$loss,
    regression_loss = regression_loss,
    cor = post$cor,
    Lambda = post$Lambda,
    B_opt = B_post,
    sqrt_inv_Sy = sqrt_inv_Sy
  )
}


cca_graph_rrr_cv_folds <- function(X, Y, Gamma,
                                Sx = NULL, Sy = NULL, kfolds = 5,
                                lambda = 0.01, r = 2, 
                                standardize = FALSE,
                                preprocess = NULL,
                                LW_Sy = TRUE, rho = 10,
                                niter = 1e4, thresh = 1e-4,
                                thresh_0 = 1e-6,
                                Gamma_dagger = NULL,
                                graph_cache = NULL,
                                solved_tX = NULL,
                                folds = NULL,
                                return_fold_values = FALSE) {
  preprocess_mode <- .resolve_preprocess_mode(standardize = standardize, preprocess = preprocess)
  X <- .preprocess_matrix(X, preprocess_mode)
  Y <- .preprocess_matrix(Y, preprocess_mode)
  if (preprocess_mode != "none") {
    solved_tX <- NULL
  }
  if (is.null(folds)) {
    folds <- .create_cv_folds(nrow(Y), kfolds)
  }

  rmse <- vapply(seq_along(folds), function(i) {
    n_full <- nrow(X)
    X_train <- X[-folds[[i]], ]
    Y_train <- Y[-folds[[i]], ]
    X_val <- X[folds[[i]], ]
    Y_val <- Y[folds[[i]], ]

   # We don't even need to create X_train and Y_train explicitly for the covariance calculation
    n_train <- n_full - nrow(X_val)

    # Reuse downdated empirical covariances only when shrinkage is disabled.
    # For LW_Sy = TRUE, each fold recomputes the shrinkage estimate from Y_train.
    Sx_train <- if (is.null(Sx)) NULL else (n_full * Sx - crossprod(X_val)) / n_train
    Sy_train <- if (LW_Sy || is.null(Sy)) NULL else (n_full * Sy - crossprod(Y_val)) / n_train


    
    tryCatch({
      if (!is.null(graph_cache)) {
        graph_design <- if (!is.null(solved_tX) && graph_cache$mode == "regularized") {
          .transform_graph_design(
            graph_cache,
            X = X_train,
            solved_tX = solved_tX[, -folds[[i]], drop = FALSE]
          )
        } else {
          NULL
        }

        fit <- .fit_graph_rrr_preprocessed(
          X_train, Y_train,
          Sx = Sx_train, Sy = Sy_train,
          lambda = lambda, r = r,
          LW_Sy = LW_Sy, rho = rho,
          niter = niter, thresh = thresh,
          thresh_0 = thresh_0,
          verbose = FALSE,
          graph_cache = graph_cache,
          graph_design = graph_design
        )
      } else {
        fit <- cca_graph_rrr(X_train, Y_train, Gamma,
                             Sx = Sx_train, Sy = Sy_train,
                             lambda = lambda, r = r,
                             standardize = FALSE, preprocess = "none",
                             LW_Sy = LW_Sy, rho = rho, niter = niter, thresh = thresh,
                             thresh_0 = thresh_0,
                             Gamma_dagger = Gamma_dagger)
      }

      # Score the same regression target optimized during training:
      # X B approximates Y whitened by the training-fold Sy estimate.
      tilde_Y_val <- Y_val %*% fit$sqrt_inv_Sy
      mean((X_val %*% fit$B_opt - tilde_Y_val)^2)
    }, error = function(e) {
      message("Error in fold ", i, ": ", conditionMessage(e))
      NA_real_
    })
  }, numeric(1))

  if (return_fold_values) {
    return(rmse)
  }

  if (all(is.na(rmse))) return(1e8)
  mean(rmse, na.rm = TRUE)
}

#' Graph-regularized Reduced-Rank Regression for Canonical Correlation Analysis with cross validation
#'
#' Solves a sparse canonical correlation problem using a graph-constrained reduced-rank regression formulation.
#' The problem is solved via an ADMM approach.
#'
#' @param X Matrix of predictors (n x p)
#' @param Y Matrix of responses (n x q)
#' @param Gamma Graph constraint matrix (g x p)
#' @param kfolds Number of folds for cross-validation
#' @param parallelize Whether to parallelize cross-validation
#' @param lambdas Grid of regularization parameters to test for sparsity
#' @param r Target rank
#' @param standardize Backward-compatible preprocessing flag: TRUE = `"scale"`, FALSE = `"center"`.
#' @param preprocess Preprocessing mode. One of `"scale"` (center + scale), `"center"` (center only), or `"none"`.
#' @param LW_Sy Whether to apply Ledoit-Wolf shrinkage to Sy
#' @param rho ADMM penalty parameter
#' @param niter Maximum number of ADMM iterations
#' @param thresh Convergence threshold for ADMM
#' @param thresh_0 Threshold for small values in the coefficient matrix (default 1e-6)
#' @param verbose Whether to print diagnostic output
#' @param Gamma_dagger Optional pseudoinverse of Gamma (computed if NULL)
#' @param nb_cores Number of cores to use for parallelization. Defaults to min(kfolds, available cores minus 1).
#'
#' @return A list with elements:
#' \describe{
#'   \item{U}{Canonical direction matrix for X (p x r)}
#'   \item{V}{Canonical direction matrix for Y (q x r)}
#'   \item{lambda}{Optimal regularisation parameter lambda chosen by CV}
#'   \item{rmse}{Mean squared error of prediction (as computed in the CV)}
#'   \item{cor}{Canonical covariances}
#'   \item{lambda_x}{Alias of the selected `lambda`}
#'   \item{lambda_x_se}{Foldwise standard error at the selected `lambda`}
#'   \item{lambda_y}{Placeholder for symmetry with two-penalty interfaces}
#'   \item{lambda_y_se}{Placeholder for symmetry with two-penalty interfaces}
#'   \item{resultsx}{Backward-compatible alias of `cv_summary`}
#'   \item{cv_summary}{Data frame with one row per lambda containing mean RMSE and its foldwise standard error}
#'   \item{cv_folds}{Data frame with fold-level RMSE values for each lambda}
#'   \item{Lambda}{Canonical correlations from the final fit}
#'   \item{B}{Estimated reduced-rank coefficient matrix from the final fit}
#'   \item{fit}{Final fit at the selected lambda}
#' }
#' @importFrom foreach foreach %dopar%

#' @export
cca_graph_rrr_cv <- function(X, Y, Gamma,
                             r = 2, 
                             lambdas = 10^seq(-3, 1.5, length.out = 10),
                             kfolds = 5, parallelize = FALSE,
                             standardize = TRUE, LW_Sy = TRUE,
                             preprocess = NULL,
                             rho = 10, niter = 1e4, thresh = 1e-4,
                             thresh_0 = 1e-6, verbose = FALSE,
                             Gamma_dagger = NULL, nb_cores= NULL) {
  preprocess_mode <- .resolve_preprocess_mode(standardize = standardize, preprocess = preprocess)

  if (nrow(X) < min(ncol(X), ncol(Y))) {
    warning("Both X and Y are high dimensional; method may fail.")
  }

  # Apply preprocessing once before CV. Each fold then reuses these transformed matrices.
  X <- .preprocess_matrix(X, preprocess_mode)
  Y <- .preprocess_matrix(Y, preprocess_mode)

  n <- nrow(X)
  Sx <- NULL
  Sy <- if (LW_Sy) NULL else crossprod(Y) / n
  graph_cache <- .prepare_graph_cache(Gamma, max(thresh_0, 1e-6), Gamma_dagger = Gamma_dagger)
  solved_tX <- if (graph_cache$mode == "regularized") graph_cache$solve_L(t(X)) else NULL
  full_graph_design <- if (!is.null(solved_tX)) .transform_graph_design(graph_cache, X = X, solved_tX = solved_tX) else NULL

  folds <- .create_cv_folds(nrow(Y), kfolds)
  .rrr_chol_jitter <- .rrr_chol_jitter
  .rrr_invsqrt_psd <- .rrr_invsqrt_psd
  .rrr_whiten_factor <- .rrr_whiten_factor
  .rrr_safe_rank_svd <- .rrr_safe_rank_svd
  .postprocess_rrr_fit <- .postprocess_rrr_fit
  .project_graph_rrr_coefficients <- .project_graph_rrr_coefficients
  .resolve_preprocess_mode <- .resolve_preprocess_mode
  .preprocess_matrix <- .preprocess_matrix
  .create_cv_folds <- .create_cv_folds
  .make_graph_invSx_apply <- .make_graph_invSx_apply
  .prepare_graph_cache <- .prepare_graph_cache
  .transform_graph_design <- .transform_graph_design
  .reconstruct_graph_coefficients <- .reconstruct_graph_coefficients
  .fit_graph_rrr_preprocessed <- .fit_graph_rrr_preprocessed
  cca_graph_rrr <- cca_graph_rrr

  environment(.postprocess_rrr_fit) <- environment()
  environment(.project_graph_rrr_coefficients) <- environment()
  environment(.fit_graph_rrr_preprocessed) <- environment()
  environment(cca_graph_rrr) <- environment()

  graph_rrr_cv_folds <- cca_graph_rrr_cv_folds
  environment(graph_rrr_cv_folds) <- environment()
  safe_mean_na <- .safe_mean_na
  safe_sd_na <- .safe_sd_na

  run_cv <- function(lambda) {
    fold_values <- graph_rrr_cv_folds(
      X, Y, Gamma,
      Sx = Sx, Sy = Sy,
      kfolds = kfolds,
      lambda = lambda, r = r,
      preprocess = "none",
      LW_Sy = LW_Sy, rho = rho,
      niter = niter, thresh = thresh,
      thresh_0 = thresh_0,
      Gamma_dagger = Gamma_dagger,
      graph_cache = graph_cache,
      solved_tX = solved_tX,
      folds = folds,
      return_fold_values = TRUE
    )

    list(
      summary = data.frame(
        lambda = lambda,
        rmse = safe_mean_na(fold_values),
        se = safe_sd_na(fold_values) / sqrt(sum(!is.na(fold_values))),
        stringsAsFactors = FALSE
      ),
      fold = data.frame(
        lambda = lambda,
        fold = seq_along(fold_values),
        rmse = fold_values,
        stringsAsFactors = FALSE
      )
    )
  }

    if (parallelize) {
      if (!requireNamespace("doParallel", quietly = TRUE)) {
      stop("Package 'doParallel' must be installed to use the parallelization option.",
          call. = FALSE)
      }

      if (!requireNamespace("crayon", quietly = TRUE)) {
      stop("Package 'crayon' must be installed to use the parallelization option.",
          call. = FALSE)
      }
    # --- GRACEFUL PARALLEL SETUP ---
      available_cores_str <- Sys.getenv("SLURM_CPUS_PER_TASK")
      available_cores <- if (available_cores_str == "") {
        parallel::detectCores()
      } else {
        as.integer(available_cores_str)
      }
      available_cores <- max(1L, available_cores)
      if (is.null(nb_cores)) {
        nb_cores_effective <- min(kfolds, length(lambdas), available_cores - 1L)
      } else {
        nb_cores_effective <- min(as.integer(nb_cores), kfolds, length(lambdas))
      }
      nb_cores_effective <- max(1L, nb_cores_effective)
      cl <- setup_parallel_backend(nb_cores_effective)
      
      if (!is.null(cl)) {
        # If the cluster was created successfully, register it and plan to stop it
        if (verbose)  { cat(crayon::green("Parallel backend successfully registered.\n")) }

        initialize_parallel_workers(cl, verbose = verbose)
        doParallel::registerDoParallel(cl)
        on.exit(cleanup_parallel_backend(cl), add = TRUE)
      } else {
        # If setup_parallel_backend returned NULL, print a warning and proceed serially
        warning("All parallel setup attempts failed. Proceeding in serial mode.", immediate. = TRUE)
        parallelize <- FALSE # Ensure %dopar% runs serially
    }
   }

  graph_rrr_parallel_exports <- c(
    ".rrr_chol_jitter", ".rrr_invsqrt_psd",
    ".rrr_whiten_factor", ".rrr_safe_rank_svd",
    ".postprocess_rrr_fit", ".project_graph_rrr_coefficients",
    ".resolve_preprocess_mode", ".preprocess_matrix", ".create_cv_folds",
    ".make_graph_invSx_apply", ".prepare_graph_cache",
    ".transform_graph_design", ".reconstruct_graph_coefficients",
    ".fit_graph_rrr_preprocessed"
  )

   if (parallelize) {
    results_list <- foreach(
      lambda = lambdas,
      .packages = "Matrix",
      .export = graph_rrr_parallel_exports
    ) %dopar% run_cv(lambda)
  } else {
    results_list <- lapply(lambdas, run_cv)
  }

  cv_summary <- dplyr::bind_rows(lapply(results_list, `[[`, "summary"))
  cv_folds <- dplyr::bind_rows(lapply(results_list, `[[`, "fold"))

  cv_summary$rmse[is.na(cv_summary$rmse) | cv_summary$rmse == 0] <- 1e8
  cv_summary <- cv_summary %>% dplyr::filter(rmse > 1e-5)

  opt_lambda <- cv_summary$lambda[which.min(cv_summary$rmse)]
  opt_lambda <- ifelse(is.na(opt_lambda), 0.1, opt_lambda)

  final <- .fit_graph_rrr_preprocessed(
    X, Y,
    Sx = Sx, Sy = Sy,
    lambda = opt_lambda, r = r,
    LW_Sy = LW_Sy, rho = rho, niter = niter,
    thresh = thresh, thresh_0 = thresh_0,
    verbose = verbose,
    graph_cache = graph_cache,
    graph_design = full_graph_design
  )

  list(
    U = final$U,
    V = final$V,
    lambda = opt_lambda,
    rmse = cv_summary$rmse,
    cor = final$cor,
    lambda_x = opt_lambda,
    lambda_x_se = cv_summary$se[match(opt_lambda, cv_summary$lambda)],
    lambda_y = NA_real_,
    lambda_y_se = NA_real_,
    resultsx = cv_summary,
    cv_summary = cv_summary,
    cv_folds = cv_folds,
    Lambda = final$Lambda,
    B = final$B_opt,
    fit = final
  )
}



#' Graph-regularized Reduced-Rank Regression for Canonical Correlation Analysis
#'
#' Solves a sparse canonical correlation problem using a graph-constrained reduced-rank regression formulation.
#' The problem is solved via an ADMM approach.
#'
#' @param X Matrix of predictors (n x p)
#' @param Y Matrix of responses (n x q)
#' @param Gamma Graph constraint matrix (g x p)
#' @param Sx Optional covariance matrix for X. Kept for backward compatibility; the graph fit now postprocesses directly from `X` and does not need to form `Sx`.
#' @param Sy Optional covariance matrix for Y. If NULL, computed similarly; optionally shrunk via Ledoit-Wolf
#' @param Sxy Optional cross-covariance matrix (not currently used)
#' @param lambda Regularization parameter for sparsity
#' @param r Target rank
#' @param standardize Backward-compatible preprocessing flag: TRUE = `"scale"`, FALSE = `"center"`.
#' @param preprocess Preprocessing mode. One of `"scale"` (center + scale), `"center"` (center only), or `"none"`.
#' @param LW_Sy Whether to apply Ledoit-Wolf shrinkage to Sy
#' @param rho ADMM penalty parameter
#' @param niter Maximum number of ADMM iterations
#' @param thresh Convergence threshold for ADMM
#' @param verbose Whether to print diagnostic output
#' @param thresh_0 Threshold for small values in the coefficient matrix (default 1e-6)
#' @param Gamma_dagger Optional pseudoinverse of Gamma (computed if NULL)
#'
#' @return A list with elements:
#' \describe{
#'   \item{U}{Canonical direction matrix for X (p x r)}
#'   \item{V}{Canonical direction matrix for Y (q x r)}
#'   \item{cor}{Canonical covariances}
#'   \item{loss}{The prediction error 1/n * \| XU - YV\|^2}
#'   \item{Lambda}{Canonical correlations}
#'   \item{B_opt}{Estimated reduced-rank coefficient matrix}
#' }
#' @export
cca_graph_rrr <- function(X, Y, Gamma,
                          Sx = NULL, Sy = NULL, Sxy = NULL,
                          lambda = 0, r,
                          standardize = FALSE,
                          preprocess = NULL,
                          LW_Sy = TRUE, rho = 10,
                          niter = 1e4, thresh = 1e-4,
                          thresh_0 = 1e-6,
                          verbose = FALSE, Gamma_dagger = NULL) {
  preprocess_mode <- .resolve_preprocess_mode(standardize = standardize, preprocess = preprocess)
  X <- .preprocess_matrix(X, preprocess_mode)
  Y <- .preprocess_matrix(Y, preprocess_mode)
  graph_cache <- .prepare_graph_cache(Gamma, max(thresh_0, 1e-6), Gamma_dagger = Gamma_dagger)
  .fit_graph_rrr_preprocessed(
    X, Y,
    Sx = Sx, Sy = Sy,
    lambda = lambda, r = r,
    LW_Sy = LW_Sy, rho = rho,
    niter = niter, thresh = thresh,
    thresh_0 = thresh_0,
    verbose = verbose,
    graph_cache = graph_cache
  )
}

library(ccar3)

expected_cv_names <- c(
  "U", "V", "lambda", "rmse", "cor",
  "lambda_x", "lambda_x_se", "lambda_y", "lambda_y_se",
  "resultsx", "cv_summary", "cv_folds", "Lambda", "B", "fit"
)

test_that("cca_graph_rrr_cv returns correct output structure", {
  set.seed(123)
  skip_if_not_installed("igraph")

  r = 2
  gen = generate_example_graph(n=500, p1=10, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  Gamma = gen$Gamma
  
  result <- cca_graph_rrr_cv(X, Y, Gamma=Gamma, r = r, kfolds=3, parallelize = FALSE)
  
  expect_type(result, "list")
  expect_identical(names(result), expected_cv_names)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], dim(X)[2])
  expect_equal(dim(result$V)[1], dim(Y)[2])
})



test_that("cca_graph_rrr_cv can run in parallel", {
  skip_if_not_installed("igraph")

  r = 2
  gen = generate_example_graph(n=500, p1=10, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  Gamma = gen$Gamma
  result <- cca_graph_rrr_cv(X, Y, Gamma=Gamma, r = 2, kfolds=3,  parallelize = TRUE)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.7)  
  expect_true(subdistance(result$V, gen$v) < 0.7)
})


test_that("cca_graph_rrr_cv returns the correct answer", {
  skip_if_not_installed("igraph")

  
  ##### Generate toy example data
  set.seed(123)
  r = 2
  gen = generate_example_graph(n=500, p1=10, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  Gamma = gen$Gamma
  
  result <- cca_graph_rrr_cv(X, Y, Gamma=Gamma, r = r, kfolds=3, parallelize =TRUE)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.5)  
  expect_true(subdistance(result$V, gen$v) < 0.5)
})

test_that("cca_graph_rrr_cv accepts preprocessing mode and sparse Gamma", {
  skip_if_not_installed("igraph")
  skip_if_not_installed("Matrix")

  set.seed(42)
  r <- 2
  gen <- generate_example_graph(
    n = 250, p1 = 10, p2 = 5,
    r_pca = 3,
    nnzeros = 5,
    theta = diag(c(0.9, 0.8)),
    lambda_pca = 1,
    r = r, overlapping_amount = 1,
    normalize_diagonal = TRUE,
    n_new = 1000
  )

  Gamma_sparse <- Matrix::Matrix(gen$Gamma, sparse = TRUE)
  result <- cca_graph_rrr_cv(
    gen$X, gen$Y, Gamma = Gamma_sparse, r = r,
    kfolds = 3, parallelize = FALSE, preprocess = "center"
  )

  expect_type(result, "list")
  expect_identical(names(result), expected_cv_names)
})

test_that("cca_graph_rrr_cv_folds honors preprocessing modes on raw inputs", {
  set.seed(88)
  X <- matrix(rnorm(60 * 8), 60, 8)
  Y <- matrix(rnorm(60 * 4), 60, 4)
  Gamma <- diag(ncol(X) - 1)
  Gamma <- cbind(Gamma, 0) - cbind(0, diag(ncol(X) - 1))
  folds <- ccar3:::.create_cv_folds(nrow(X), 3)

  fold_values <- ccar3:::cca_graph_rrr_cv_folds(
    X, Y, Gamma = Gamma,
    Sx = NULL, Sy = NULL,
    kfolds = 3, lambda = 0.01, r = 2,
    preprocess = "center",
    LW_Sy = FALSE,
    niter = 200,
    folds = folds,
    return_fold_values = TRUE
  )

  expect_length(fold_values, length(folds))
  expect_false(all(is.na(fold_values)))
})

library(ccar3)

expected_cv_names <- c(
  "U", "V", "lambda", "rmse", "cor",
  "lambda_x", "lambda_x_se", "lambda_y", "lambda_y_se",
  "resultsx", "cv_summary", "cv_folds", "Lambda", "B", "fit"
)

test_that("cca_graph_rrr_cv returns correct output structure", {
  set.seed(123)
  skip_if_not_installed("igraph")

  r = 2
  gen = generate_example_graph(n=500, p1=10, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  Gamma = gen$Gamma
  
  result <- cca_graph_rrr_cv(X, Y, Gamma=Gamma, r = r, kfolds=3, parallelize = FALSE)
  
  expect_type(result, "list")
  expect_identical(names(result), expected_cv_names)
  expect_equal(dim(result$U)[2], r)
  expect_equal(dim(result$V)[2], r)
  expect_equal(dim(result$U)[1], dim(X)[2])
  expect_equal(dim(result$V)[1], dim(Y)[2])
})



test_that("cca_graph_rrr_cv can run in parallel", {
  skip_if_not_installed("igraph")
  skip_if_cran_limits_cores()

  r = 2
  gen = generate_example_graph(n=500, p1=10, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  Gamma = gen$Gamma
  result <- cca_graph_rrr_cv(X, Y, Gamma=Gamma, r = 2, kfolds=3,  parallelize = TRUE)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.7)  
  expect_true(subdistance(result$V, gen$v) < 0.7)
})


test_that("cca_graph_rrr_cv returns the correct answer", {
  skip_if_not_installed("igraph")
  skip_if_cran_limits_cores()

  
  ##### Generate toy example data
  set.seed(123)
  r = 2
  gen = generate_example_graph(n=500, p1=10, p2=5, 
                               r_pca = 3,
                               nnzeros = 5,
                               theta = diag(c(0.9,  0.8)),
                               lambda_pca = 1,
                               r = r, overlapping_amount = 1,
                               normalize_diagonal = TRUE,
                               n_new = 5000)
  
  X = gen$X
  Y = gen$Y
  Gamma = gen$Gamma
  
  result <- cca_graph_rrr_cv(X, Y, Gamma=Gamma, r = r, kfolds=3, parallelize =TRUE)
  
  expect_type(result, "list")
  expect_true(subdistance(result$U, gen$u) < 0.5)  
  expect_true(subdistance(result$V, gen$v) < 0.5)
})

test_that("cca_graph_rrr_cv accepts preprocessing mode and sparse Gamma", {
  skip_if_not_installed("igraph")
  skip_if_not_installed("Matrix")

  set.seed(42)
  r <- 2
  gen <- generate_example_graph(
    n = 250, p1 = 10, p2 = 5,
    r_pca = 3,
    nnzeros = 5,
    theta = diag(c(0.9, 0.8)),
    lambda_pca = 1,
    r = r, overlapping_amount = 1,
    normalize_diagonal = TRUE,
    n_new = 1000
  )

  Gamma_sparse <- Matrix::Matrix(gen$Gamma, sparse = TRUE)
  result <- cca_graph_rrr_cv(
    gen$X, gen$Y, Gamma = Gamma_sparse, r = r,
    kfolds = 3, parallelize = FALSE, preprocess = "center"
  )

  expect_type(result, "list")
  expect_identical(names(result), expected_cv_names)
})

test_that("cca_graph_rrr_cv_folds honors preprocessing modes on raw inputs", {
  set.seed(88)
  X <- matrix(rnorm(60 * 8), 60, 8)
  Y <- matrix(rnorm(60 * 4), 60, 4)
  Gamma <- diag(ncol(X) - 1)
  Gamma <- cbind(Gamma, 0) - cbind(0, diag(ncol(X) - 1))
  folds <- ccar3:::.create_cv_folds(nrow(X), 3)

  fold_values <- ccar3:::cca_graph_rrr_cv_folds(
    X, Y, Gamma = Gamma,
    Sx = NULL, Sy = NULL,
    kfolds = 3, lambda = 0.01, r = 2,
    preprocess = "center",
    LW_Sy = FALSE,
    niter = 200,
    folds = folds,
    return_fold_values = TRUE
  )

  expect_length(fold_values, length(folds))
  expect_false(all(is.na(fold_values)))
})

.resolve_preprocess_mode <- function(standardize = NULL, preprocess = NULL) {
  if (is.logical(preprocess) && length(preprocess) == 1L) {
    if (isTRUE(preprocess)) {
      preprocess <- NULL
    } else {
      return("none")
    }
  }

  if (!is.null(preprocess)) {
    return(match.arg(preprocess, c("scale", "center", "none")))
  }

  if (is.character(standardize)) {
    return(match.arg(standardize, c("scale", "center", "none")))
  }

  if (is.logical(standardize) && length(standardize) == 1L) {
    return(if (isTRUE(standardize)) "scale" else "center")
  }

  if (is.numeric(standardize) && length(standardize) == 1L && is.finite(standardize)) {
    return(if (isTRUE(as.logical(standardize))) "scale" else "center")
  }

  if (is.null(standardize)) {
    return("center")
  }

  stop("`standardize` must be logical or one of: 'scale', 'center', 'none'.", call. = FALSE)
}

.preprocess_matrix <- function(M, mode) {
  M <- as.matrix(M)
  if (mode == "none") {
    return(M)
  }

  M <- scale(M, center = TRUE, scale = identical(mode, "scale"))
  M <- as.matrix(M)
  M[!is.finite(M)] <- 0
  M
}

.as_double_matrix <- function(M) {
  M <- as.matrix(M)
  storage.mode(M) <- "double"
  M
}

.graph_null_basis <- function(Gamma, tol = 1e-10) {
  p <- ncol(Gamma)
  if (p < 1) {
    return(matrix(0, 0, 0))
  }

  parent <- seq_len(p)
  rank <- integer(p)

  find_root <- function(x) {
    root <- x
    while (parent[root] != root) {
      root <- parent[root]
    }
    while (parent[x] != x) {
      next_x <- parent[x]
      parent[x] <<- root
      x <- next_x
    }
    root
  }

  union_nodes <- function(a, b) {
    ra <- find_root(a)
    rb <- find_root(b)
    if (ra == rb) {
      return(invisible(NULL))
    }

    if (rank[ra] < rank[rb]) {
      parent[ra] <<- rb
    } else if (rank[ra] > rank[rb]) {
      parent[rb] <<- ra
    } else {
      parent[rb] <<- ra
      rank[ra] <<- rank[ra] + 1L
    }

    invisible(NULL)
  }

  nz_by_row <- if (inherits(Gamma, "sparseMatrix")) {
    gamma_summary <- Matrix::summary(Gamma)
    gamma_summary <- gamma_summary[abs(gamma_summary$x) > tol, c("i", "j"), drop = FALSE]
    split(gamma_summary$j, gamma_summary$i)
  } else {
    nz_idx <- which(abs(Gamma) > tol, arr.ind = TRUE)
    split(nz_idx[, "col"], nz_idx[, "row"])
  }

  for (cols in nz_by_row) {
    cols <- unique(as.integer(cols))
    if (length(cols) < 2L) {
      next
    }
    anchor <- cols[1]
    for (col_idx in cols[-1]) {
      union_nodes(anchor, col_idx)
    }
  }

  roots <- vapply(seq_len(p), find_root, integer(1))
  components <- split(seq_len(p), roots)
  basis <- matrix(0, p, length(components))

  for (k in seq_along(components)) {
    idx <- components[[k]]
    basis[idx, k] <- 1 / sqrt(length(idx))
  }

  basis
}

.make_graph_invSx_apply <- function(A, rho, ridge = 1e-8, verbose = FALSE) {
  A <- .as_double_matrix(A)
  n <- nrow(A)
  p <- ncol(A)

  if (p <= n) {
    gram <- crossprod(A) + rho * diag(p)
    R <- tryCatch(chol(gram + ridge * diag(p)), error = function(e) NULL)

    solve_gram <- if (!is.null(R)) {
      function(B) backsolve(R, forwardsolve(t(R), B))
    } else {
      if (verbose) message("Cholesky failed in graph ADMM solve; using solve().")
      function(B) solve(gram, B)
    }

    return(list(
      apply = function(rhs) solve_gram(.as_double_matrix(rhs)),
      uses_woodbury = FALSE
    ))
  }

  K <- diag(n) + tcrossprod(A) / rho
  R <- tryCatch(chol(K + ridge * diag(n)), error = function(e) NULL)

  solve_K <- if (!is.null(R)) {
    function(B) backsolve(R, forwardsolve(t(R), B))
  } else {
    if (verbose) message("Cholesky failed in graph Woodbury solve; using solve().")
    function(B) solve(K, B)
  }

  list(
    apply = function(rhs) {
      rhs <- .as_double_matrix(rhs)
      middle <- solve_K(A %*% rhs)
      rhs / rho - crossprod(A, middle) / (rho^2)
    },
    uses_woodbury = TRUE
  )
}

.prepare_graph_cache <- function(Gamma, gamma_ridge, Gamma_dagger = NULL) {
  if (is.null(Gamma_dagger)) {
    Gamma <- if (inherits(Gamma, "dgCMatrix")) {
      Gamma
    } else {
      methods::as(Matrix::Matrix(Gamma, sparse = TRUE), "dgCMatrix")
    }

    L_eps <- Matrix::crossprod(Gamma) + gamma_ridge * Matrix::Diagonal(n = ncol(Gamma))
    L_fac <- Matrix::Cholesky(L_eps, LDL = FALSE)

    solve_L <- function(rhs) {
      as.matrix(Matrix::solve(L_fac, .as_double_matrix(rhs), system = "A"))
    }

    return(list(
      mode = "regularized",
      Gamma = Gamma,
      gamma_ridge = gamma_ridge,
      solve_L = solve_L,
      null_basis = .graph_null_basis(Gamma)
    ))
  }

  list(
    mode = "explicit",
    Gamma = Gamma,
    Gamma_dagger = Gamma_dagger,
    gamma_ridge = gamma_ridge
  )
}

.transform_graph_design <- function(graph_cache, X = NULL, solved_tX = NULL) {
  if (graph_cache$mode == "regularized") {
    if (is.null(solved_tX)) {
      if (is.null(X)) stop("Either `X` or `solved_tX` must be provided.", call. = FALSE)
      solved_tX <- graph_cache$solve_L(t(X))
    }

    return(list(
      XGamma_dagger = as.matrix(Matrix::t(graph_cache$Gamma %*% solved_tX)),
      XPi = if (is.null(X)) NULL else .as_double_matrix(X %*% graph_cache$null_basis),
      solved_tX = solved_tX
    ))
  }

  if (is.null(X)) {
    stop("`X` must be provided when using an explicit Gamma_dagger.", call. = FALSE)
  }

  XGamma_dagger <- as.matrix(X %*% graph_cache$Gamma_dagger)
  list(
    XGamma_dagger = XGamma_dagger,
    XPi = as.matrix(X - XGamma_dagger %*% graph_cache$Gamma)
  )
}

.reconstruct_graph_coefficients <- function(graph_cache, B, Projection) {
  if (graph_cache$mode == "regularized") {
    return(list(
      gamma_dagger_B = graph_cache$solve_L(Matrix::crossprod(graph_cache$Gamma, .as_double_matrix(B))),
      Pi_projection = .as_double_matrix(graph_cache$null_basis %*% Projection)
    ))
  }

  list(
    gamma_dagger_B = graph_cache$Gamma_dagger %*% B,
    Pi_projection = Projection - as.matrix(graph_cache$Gamma_dagger %*% (graph_cache$Gamma %*% Projection))
  )
}

.project_graph_rrr_coefficients <- function(Bhat, X, r, Sx = NULL) {
  r_eff <- min(r, nrow(Bhat), ncol(Bhat))
  if (r_eff < 1 || r_eff >= ncol(Bhat)) {
    return(Bhat)
  }

  if (is.null(Sx)) {
    sx_svd <- svd(X / sqrt(nrow(X)), nu = 0, nv = min(nrow(X), ncol(X)))
    sqrt_Sx_B <- sx_svd$v %*% (sx_svd$d * crossprod(sx_svd$v, Bhat))
  } else {
    svd_Sx <- svd(as.matrix(Sx), nu = 0, nv = ncol(Sx))
    sqrt_Sx_B <- svd_Sx$v %*% (sqrt(pmax(svd_Sx$d, 0)) * crossprod(svd_Sx$v, Bhat))
  }

  weighted_svd <- .rrr_safe_rank_svd(sqrt_Sx_B, r_eff, prefer_sparse = FALSE)
  if (is.null(weighted_svd)) {
    return(Bhat)
  }

  Vr <- weighted_svd$v[, seq_len(r_eff), drop = FALSE]
  Bhat %*% Vr %*% t(Vr)
}

.fit_graph_rrr_preprocessed <- function(X, Y,
                                        Sx = NULL, Sy = NULL,
                                        lambda = 0, r,
                                        LW_Sy = TRUE, rho = 10,
                                        niter = 1e4, thresh = 1e-4,
                                        thresh_0 = 1e-6,
                                        verbose = FALSE,
                                        graph_cache,
                                        graph_design = NULL) {
  n <- nrow(X)
  p <- ncol(X)
  q <- ncol(Y)

  if (q > n) {
    stop("The X/Y swapping heuristic (for q > n) is not compatible with the graph constraint Gamma. Swap X and Y in your arguments.")
  }

  if (n < min(p, q)) {
    warning("Both X and Y are high dimensional; method may be unstable.")
  }

  fro_norm <- function(M) sqrt(sum(M^2))

  if (is.null(Sy)) {
    Sy <- crossprod(Y) / n
    if (LW_Sy) Sy <- as.matrix(corpcor::cov.shrink(Y))
  }

  svd_Sy <- svd(as.matrix(Sy))
  sqrt_inv_Sy <- svd_Sy$u %*% diag(ifelse(svd_Sy$d > 1e-4, 1 / sqrt(svd_Sy$d), 0)) %*% t(svd_Sy$v)
  tilde_Y <- Y %*% sqrt_inv_Sy

  if (is.null(graph_design)) {
    graph_design <- .transform_graph_design(graph_cache, X = X)
  }

  XGamma_dagger <- graph_design$XGamma_dagger
  XPi <- if (is.null(graph_design$XPi) && graph_cache$mode == "regularized") {
    .as_double_matrix(X %*% graph_cache$null_basis)
  } else {
    graph_design$XPi
  }

  Projection <- tryCatch(
    qr.solve(XPi, tilde_Y),
    error = function(e) {
      XtX <- .as_double_matrix(crossprod(XPi) + graph_cache$gamma_ridge * diag(ncol(XPi)))
      rhs_proj <- .as_double_matrix(crossprod(XPi, tilde_Y))
      solve(XtX, rhs_proj)
    }
  )
  new_Ytilde <- tilde_Y - XPi %*% Projection

  if (verbose) {
    cat("Norm of Projection:", fro_norm(Projection), "\n")
  }

  new_p <- ncol(XGamma_dagger)
  prod_xy <- crossprod(XGamma_dagger, new_Ytilde) / n
  invSx_info <- .make_graph_invSx_apply(XGamma_dagger / sqrt(n), rho, verbose = verbose)
  solve_invSx <- invSx_info$apply

  U_dual <- matrix(0, new_p, q)
  Z <- matrix(0, new_p, q)
  B <- matrix(0, new_p, q)

  for (i in seq_len(niter)) {
    Z_old <- Z

    B <- solve_invSx(prod_xy + rho * (Z - U_dual))
    Z <- B + U_dual

    norms <- sqrt(rowSums(Z^2))
    shrink_factors <- pmax(0, 1 - lambda / (rho * norms))
    shrink_factors[is.nan(shrink_factors)] <- 0
    Z <- Z * shrink_factors

    U_dual <- U_dual + B - Z

    primal <- fro_norm(Z - B)
    dual <- fro_norm(Z_old - Z)
    if (verbose) {
      cat("ADMM iter:", i, "Primal:", primal, "Dual:", dual, "\n")
    }
    if (max(primal / sqrt(new_p), dual / sqrt(new_p)) < thresh) break
  }

  reconstruction <- .reconstruct_graph_coefficients(graph_cache, B, Projection)
  gamma_dagger_B <- reconstruction$gamma_dagger_B
  Pi_projection <- reconstruction$Pi_projection

  B_opt <- gamma_dagger_B + Pi_projection
  B_opt[abs(B_opt) < thresh_0] <- 0

  if (verbose) {
    cat("Norm of Pi Projection:", fro_norm(Pi_projection), "\n")
    cat("Norm of Gamma_dagger %*% B:", fro_norm(gamma_dagger_B), "\n")
    cat("Norm of XPi:", fro_norm(XPi), "\n")
    cat("rank of XPi:", qr(XPi)$rank, "\n")
  }

  B_post <- .project_graph_rrr_coefficients(B_opt, X, r, Sx = Sx)
  post <- .postprocess_rrr_fit(
    B_post, X, Y, sqrt_inv_Sy, r,
    ridge_whiten = max(thresh_0, 1e-8),
    thresh_0 = thresh_0,
    verbose = verbose
  )
  regression_loss <- mean((X %*% B_post - tilde_Y)^2)

  list(
    U = post$U,
    V = post$V,
    loss = post$loss,
    regression_loss = regression_loss,
    cor = post$cor,
    Lambda = post$Lambda,
    B_opt = B_post,
    sqrt_inv_Sy = sqrt_inv_Sy
  )
}


cca_graph_rrr_cv_folds <- function(X, Y, Gamma,
                                Sx = NULL, Sy = NULL, kfolds = 5,
                                lambda = 0.01, r = 2, 
                                standardize = FALSE,
                                preprocess = NULL,
                                LW_Sy = TRUE, rho = 10,
                                niter = 1e4, thresh = 1e-4,
                                thresh_0 = 1e-6,
                                Gamma_dagger = NULL,
                                graph_cache = NULL,
                                solved_tX = NULL,
                                folds = NULL,
                                return_fold_values = FALSE) {
  preprocess_mode <- .resolve_preprocess_mode(standardize = standardize, preprocess = preprocess)
  X <- .preprocess_matrix(X, preprocess_mode)
  Y <- .preprocess_matrix(Y, preprocess_mode)
  if (preprocess_mode != "none") {
    solved_tX <- NULL
  }
  if (is.null(folds)) {
    folds <- .create_cv_folds(nrow(Y), kfolds)
  }

  rmse <- vapply(seq_along(folds), function(i) {
    n_full <- nrow(X)
    X_train <- X[-folds[[i]], ]
    Y_train <- Y[-folds[[i]], ]
    X_val <- X[folds[[i]], ]
    Y_val <- Y[folds[[i]], ]

   # We don't even need to create X_train and Y_train explicitly for the covariance calculation
    n_train <- n_full - nrow(X_val)

    # Reuse downdated empirical covariances only when shrinkage is disabled.
    # For LW_Sy = TRUE, each fold recomputes the shrinkage estimate from Y_train.
    Sx_train <- if (is.null(Sx)) NULL else (n_full * Sx - crossprod(X_val)) / n_train
    Sy_train <- if (LW_Sy || is.null(Sy)) NULL else (n_full * Sy - crossprod(Y_val)) / n_train


    
    tryCatch({
      if (!is.null(graph_cache)) {
        graph_design <- if (!is.null(solved_tX) && graph_cache$mode == "regularized") {
          .transform_graph_design(
            graph_cache,
            X = X_train,
            solved_tX = solved_tX[, -folds[[i]], drop = FALSE]
          )
        } else {
          NULL
        }

        fit <- .fit_graph_rrr_preprocessed(
          X_train, Y_train,
          Sx = Sx_train, Sy = Sy_train,
          lambda = lambda, r = r,
          LW_Sy = LW_Sy, rho = rho,
          niter = niter, thresh = thresh,
          thresh_0 = thresh_0,
          verbose = FALSE,
          graph_cache = graph_cache,
          graph_design = graph_design
        )
      } else {
        fit <- cca_graph_rrr(X_train, Y_train, Gamma,
                             Sx = Sx_train, Sy = Sy_train,
                             lambda = lambda, r = r,
                             standardize = FALSE, preprocess = "none",
                             LW_Sy = LW_Sy, rho = rho, niter = niter, thresh = thresh,
                             thresh_0 = thresh_0,
                             Gamma_dagger = Gamma_dagger)
      }

      # Score the same regression target optimized during training:
      # X B approximates Y whitened by the training-fold Sy estimate.
      tilde_Y_val <- Y_val %*% fit$sqrt_inv_Sy
      mean((X_val %*% fit$B_opt - tilde_Y_val)^2)
    }, error = function(e) {
      message("Error in fold ", i, ": ", conditionMessage(e))
      NA_real_
    })
  }, numeric(1))

  if (return_fold_values) {
    return(rmse)
  }

  if (all(is.na(rmse))) return(1e8)
  mean(rmse, na.rm = TRUE)
}

#' Graph-regularized Reduced-Rank Regression for Canonical Correlation Analysis with cross validation
#'
#' Solves a sparse canonical correlation problem using a graph-constrained reduced-rank regression formulation.
#' The problem is solved via an ADMM approach.
#'
#' @param X Matrix of predictors (n x p)
#' @param Y Matrix of responses (n x q)
#' @param Gamma Graph constraint matrix (g x p)
#' @param kfolds Number of folds for cross-validation
#' @param parallelize Whether to parallelize cross-validation
#' @param lambdas Grid of regularization parameters to test for sparsity
#' @param r Target rank
#' @param standardize Backward-compatible preprocessing flag: TRUE = `"scale"`, FALSE = `"center"`.
#' @param preprocess Preprocessing mode. One of `"scale"` (center + scale), `"center"` (center only), or `"none"`.
#' @param LW_Sy Whether to apply Ledoit-Wolf shrinkage to Sy
#' @param rho ADMM penalty parameter
#' @param niter Maximum number of ADMM iterations
#' @param thresh Convergence threshold for ADMM
#' @param thresh_0 Threshold for small values in the coefficient matrix (default 1e-6)
#' @param verbose Whether to print diagnostic output
#' @param Gamma_dagger Optional pseudoinverse of Gamma (computed if NULL)
#' @param nb_cores Number of cores to use for parallelization. Defaults to min(kfolds, available cores minus 1).
#'
#' @return A list with elements:
#' \describe{
#'   \item{U}{Canonical direction matrix for X (p x r)}
#'   \item{V}{Canonical direction matrix for Y (q x r)}
#'   \item{lambda}{Optimal regularisation parameter lambda chosen by CV}
#'   \item{rmse}{Mean squared error of prediction (as computed in the CV)}
#'   \item{cor}{Canonical covariances}
#'   \item{lambda_x}{Alias of the selected `lambda`}
#'   \item{lambda_x_se}{Foldwise standard error at the selected `lambda`}
#'   \item{lambda_y}{Placeholder for symmetry with two-penalty interfaces}
#'   \item{lambda_y_se}{Placeholder for symmetry with two-penalty interfaces}
#'   \item{resultsx}{Backward-compatible alias of `cv_summary`}
#'   \item{cv_summary}{Data frame with one row per lambda containing mean RMSE and its foldwise standard error}
#'   \item{cv_folds}{Data frame with fold-level RMSE values for each lambda}
#'   \item{Lambda}{Canonical correlations from the final fit}
#'   \item{B}{Estimated reduced-rank coefficient matrix from the final fit}
#'   \item{fit}{Final fit at the selected lambda}
#' }
#' @importFrom foreach foreach %dopar%

#' @export
cca_graph_rrr_cv <- function(X, Y, Gamma,
                             r = 2, 
                             lambdas = 10^seq(-3, 1.5, length.out = 10),
                             kfolds = 5, parallelize = FALSE,
                             standardize = TRUE, LW_Sy = TRUE,
                             preprocess = NULL,
                             rho = 10, niter = 1e4, thresh = 1e-4,
                             thresh_0 = 1e-6, verbose = FALSE,
                             Gamma_dagger = NULL, nb_cores= NULL) {
  preprocess_mode <- .resolve_preprocess_mode(standardize = standardize, preprocess = preprocess)

  if (nrow(X) < min(ncol(X), ncol(Y))) {
    warning("Both X and Y are high dimensional; method may fail.")
  }

  # Apply preprocessing once before CV. Each fold then reuses these transformed matrices.
  X <- .preprocess_matrix(X, preprocess_mode)
  Y <- .preprocess_matrix(Y, preprocess_mode)

  n <- nrow(X)
  Sx <- NULL
  Sy <- if (LW_Sy) NULL else crossprod(Y) / n
  graph_cache <- .prepare_graph_cache(Gamma, max(thresh_0, 1e-6), Gamma_dagger = Gamma_dagger)
  solved_tX <- if (graph_cache$mode == "regularized") graph_cache$solve_L(t(X)) else NULL
  full_graph_design <- if (!is.null(solved_tX)) .transform_graph_design(graph_cache, X = X, solved_tX = solved_tX) else NULL

  folds <- .create_cv_folds(nrow(Y), kfolds)
  .rrr_chol_jitter <- .rrr_chol_jitter
  .rrr_invsqrt_psd <- .rrr_invsqrt_psd
  .rrr_whiten_factor <- .rrr_whiten_factor
  .rrr_safe_rank_svd <- .rrr_safe_rank_svd
  .postprocess_rrr_fit <- .postprocess_rrr_fit
  .project_graph_rrr_coefficients <- .project_graph_rrr_coefficients
  .resolve_preprocess_mode <- .resolve_preprocess_mode
  .preprocess_matrix <- .preprocess_matrix
  .create_cv_folds <- .create_cv_folds
  .make_graph_invSx_apply <- .make_graph_invSx_apply
  .prepare_graph_cache <- .prepare_graph_cache
  .transform_graph_design <- .transform_graph_design
  .reconstruct_graph_coefficients <- .reconstruct_graph_coefficients
  .fit_graph_rrr_preprocessed <- .fit_graph_rrr_preprocessed
  cca_graph_rrr <- cca_graph_rrr

  environment(.postprocess_rrr_fit) <- environment()
  environment(.project_graph_rrr_coefficients) <- environment()
  environment(.fit_graph_rrr_preprocessed) <- environment()
  environment(cca_graph_rrr) <- environment()

  graph_rrr_cv_folds <- cca_graph_rrr_cv_folds
  environment(graph_rrr_cv_folds) <- environment()
  safe_mean_na <- .safe_mean_na
  safe_sd_na <- .safe_sd_na

  run_cv <- function(lambda) {
    fold_values <- graph_rrr_cv_folds(
      X, Y, Gamma,
      Sx = Sx, Sy = Sy,
      kfolds = kfolds,
      lambda = lambda, r = r,
      preprocess = "none",
      LW_Sy = LW_Sy, rho = rho,
      niter = niter, thresh = thresh,
      thresh_0 = thresh_0,
      Gamma_dagger = Gamma_dagger,
      graph_cache = graph_cache,
      solved_tX = solved_tX,
      folds = folds,
      return_fold_values = TRUE
    )

    list(
      summary = data.frame(
        lambda = lambda,
        rmse = safe_mean_na(fold_values),
        se = safe_sd_na(fold_values) / sqrt(sum(!is.na(fold_values))),
        stringsAsFactors = FALSE
      ),
      fold = data.frame(
        lambda = lambda,
        fold = seq_along(fold_values),
        rmse = fold_values,
        stringsAsFactors = FALSE
      )
    )
  }

    if (parallelize) {
      if (!requireNamespace("doParallel", quietly = TRUE)) {
      stop("Package 'doParallel' must be installed to use the parallelization option.",
          call. = FALSE)
      }

      if (!requireNamespace("crayon", quietly = TRUE)) {
      stop("Package 'crayon' must be installed to use the parallelization option.",
          call. = FALSE)
      }
    # --- GRACEFUL PARALLEL SETUP ---
      available_cores_str <- Sys.getenv("SLURM_CPUS_PER_TASK")
      available_cores <- if (available_cores_str == "") {
        parallel::detectCores()
      } else {
        as.integer(available_cores_str)
      }
      available_cores <- max(1L, available_cores)
      if (is.null(nb_cores)) {
        nb_cores_effective <- min(kfolds, length(lambdas), available_cores - 1L)
      } else {
        nb_cores_effective <- min(as.integer(nb_cores), kfolds, length(lambdas))
      }
      nb_cores_effective <- max(1L, nb_cores_effective)
      nb_cores_effective <- .limit_parallel_cores(nb_cores_effective)
      cl <- setup_parallel_backend(nb_cores_effective)
      
      if (!is.null(cl)) {
        # If the cluster was created successfully, register it and plan to stop it
        if (verbose)  { cat(crayon::green("Parallel backend successfully registered.\n")) }

        initialize_parallel_workers(cl, verbose = verbose)
        doParallel::registerDoParallel(cl)
        on.exit(cleanup_parallel_backend(cl), add = TRUE)
      } else {
        # If setup_parallel_backend returned NULL, print a warning and proceed serially
        warning("All parallel setup attempts failed. Proceeding in serial mode.", immediate. = TRUE)
        parallelize <- FALSE # Ensure %dopar% runs serially
    }
   }

  graph_rrr_parallel_exports <- c(
    ".rrr_chol_jitter", ".rrr_invsqrt_psd",
    ".rrr_whiten_factor", ".rrr_safe_rank_svd",
    ".postprocess_rrr_fit", ".project_graph_rrr_coefficients",
    ".resolve_preprocess_mode", ".preprocess_matrix", ".create_cv_folds",
    ".make_graph_invSx_apply", ".prepare_graph_cache",
    ".transform_graph_design", ".reconstruct_graph_coefficients",
    ".fit_graph_rrr_preprocessed"
  )

   if (parallelize) {
    results_list <- foreach(
      lambda = lambdas,
      .packages = "Matrix",
      .export = graph_rrr_parallel_exports
    ) %dopar% run_cv(lambda)
  } else {
    results_list <- lapply(lambdas, run_cv)
  }

  cv_summary <- dplyr::bind_rows(lapply(results_list, `[[`, "summary"))
  cv_folds <- dplyr::bind_rows(lapply(results_list, `[[`, "fold"))

  cv_summary$rmse[is.na(cv_summary$rmse) | cv_summary$rmse == 0] <- 1e8
  cv_summary <- cv_summary %>% dplyr::filter(rmse > 1e-5)

  opt_lambda <- cv_summary$lambda[which.min(cv_summary$rmse)]
  opt_lambda <- ifelse(is.na(opt_lambda), 0.1, opt_lambda)

  final <- .fit_graph_rrr_preprocessed(
    X, Y,
    Sx = Sx, Sy = Sy,
    lambda = opt_lambda, r = r,
    LW_Sy = LW_Sy, rho = rho, niter = niter,
    thresh = thresh, thresh_0 = thresh_0,
    verbose = verbose,
    graph_cache = graph_cache,
    graph_design = full_graph_design
  )

  list(
    U = final$U,
    V = final$V,
    lambda = opt_lambda,
    rmse = cv_summary$rmse,
    cor = final$cor,
    lambda_x = opt_lambda,
    lambda_x_se = cv_summary$se[match(opt_lambda, cv_summary$lambda)],
    lambda_y = NA_real_,
    lambda_y_se = NA_real_,
    resultsx = cv_summary,
    cv_summary = cv_summary,
    cv_folds = cv_folds,
    Lambda = final$Lambda,
    B = final$B_opt,
    fit = final
  )
}



#' Graph-regularized Reduced-Rank Regression for Canonical Correlation Analysis
#'
#' Solves a sparse canonical correlation problem using a graph-constrained reduced-rank regression formulation.
#' The problem is solved via an ADMM approach.
#'
#' @param X Matrix of predictors (n x p)
#' @param Y Matrix of responses (n x q)
#' @param Gamma Graph constraint matrix (g x p)
#' @param Sx Optional covariance matrix for X. Kept for backward compatibility; the graph fit now postprocesses directly from `X` and does not need to form `Sx`.
#' @param Sy Optional covariance matrix for Y. If NULL, computed similarly; optionally shrunk via Ledoit-Wolf
#' @param Sxy Optional cross-covariance matrix (not currently used)
#' @param lambda Regularization parameter for sparsity
#' @param r Target rank
#' @param standardize Backward-compatible preprocessing flag: TRUE = `"scale"`, FALSE = `"center"`.
#' @param preprocess Preprocessing mode. One of `"scale"` (center + scale), `"center"` (center only), or `"none"`.
#' @param LW_Sy Whether to apply Ledoit-Wolf shrinkage to Sy
#' @param rho ADMM penalty parameter
#' @param niter Maximum number of ADMM iterations
#' @param thresh Convergence threshold for ADMM
#' @param verbose Whether to print diagnostic output
#' @param thresh_0 Threshold for small values in the coefficient matrix (default 1e-6)
#' @param Gamma_dagger Optional pseudoinverse of Gamma (computed if NULL)
#'
#' @return A list with elements:
#' \describe{
#'   \item{U}{Canonical direction matrix for X (p x r)}
#'   \item{V}{Canonical direction matrix for Y (q x r)}
#'   \item{cor}{Canonical covariances}
#'   \item{loss}{The prediction error 1/n * \| XU - YV\|^2}
#'   \item{Lambda}{Canonical correlations}
#'   \item{B_opt}{Estimated reduced-rank coefficient matrix}
#' }
#' @export
cca_graph_rrr <- function(X, Y, Gamma,
                          Sx = NULL, Sy = NULL, Sxy = NULL,
                          lambda = 0, r,
                          standardize = FALSE,
                          preprocess = NULL,
                          LW_Sy = TRUE, rho = 10,
                          niter = 1e4, thresh = 1e-4,
                          thresh_0 = 1e-6,
                          verbose = FALSE, Gamma_dagger = NULL) {
  preprocess_mode <- .resolve_preprocess_mode(standardize = standardize, preprocess = preprocess)
  X <- .preprocess_matrix(X, preprocess_mode)
  Y <- .preprocess_matrix(Y, preprocess_mode)
  graph_cache <- .prepare_graph_cache(Gamma, max(thresh_0, 1e-6), Gamma_dagger = Gamma_dagger)
  .fit_graph_rrr_preprocessed(
    X, Y,
    Sx = Sx, Sy = Sy,
    lambda = lambda, r = r,
    LW_Sy = LW_Sy, rho = rho,
    niter = niter, thresh = thresh,
    thresh_0 = thresh_0,
    verbose = verbose,
    graph_cache = graph_cache
  )
}

#' Return the edge incidence matrix of an igraph graph
#'
#'
#' @param g igraph graph object.
#' @param weight edge weights.
#' @return Edge incidence matrix of the graph g, with +weight for the source node and -weight for the target node.
#' @export
get_edge_incidence <- function(g, weight = 1){
  n_nodes = igraph::vcount(g)
  d_max = max(igraph::degree(g))
  #d_max = 1
  edges = data.frame(igraph::as_edgelist(g)) %>%
    dplyr::arrange(X1, X2)
  Gamma = matrix(0, nrow(edges), n_nodes)
  
  # Make beta_v into a matrix
  names_st = unique(c(edges$X1, edges$X2))
  for (e in 1:nrow(edges)){
    ind1 = which( edges$X1[e] == names_st)
    ind2 = which( edges$X2[e] == names_st)
    Gamma[e, ind1] = weight
    Gamma[e, ind2] = - weight
  }
  return(Gamma)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alt_regular_CCA.R
\name{regular_cca}
\alias{regular_cca}
\title{Function to perform regular (low dimensional) canonical correlation analysis (CCA}
\usage{
regular_cca(X, Y, rank)
}
\arguments{
\item{X}{Matrix of predictors (n x p)}

\item{Y}{Matrix of responses (n x q)}

\item{rank}{Number of canonical components to extract}
}
\value{
A list with elements:
\describe{
\item{U}{Canonical direction matrix for X (p x r)}
\item{V}{Canonical direction matrix for Y (q x r)}
\item{cor}{Canonical covariances}
}
}
\description{
Function to perform regular (low dimensional) canonical correlation analysis (CCA
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph_reduced_rank_regression.R
\name{cca_graph_rrr}
\alias{cca_graph_rrr}
\title{Graph-regularized Reduced-Rank Regression for Canonical Correlation Analysis}
\usage{
cca_graph_rrr(
  X,
  Y,
  Gamma,
  Sx = NULL,
  Sy = NULL,
  Sxy = NULL,
  lambda = 0,
  r,
  standardize = FALSE,
  preprocess = NULL,
  LW_Sy = TRUE,
  rho = 10,
  niter = 10000,
  thresh = 1e-04,
  thresh_0 = 1e-06,
  verbose = FALSE,
  Gamma_dagger = NULL
)
}
\arguments{
\item{X}{Matrix of predictors (n x p)}

\item{Y}{Matrix of responses (n x q)}

\item{Gamma}{Graph constraint matrix (g x p)}

\item{Sx}{Optional covariance matrix for X. Kept for backward compatibility; the graph fit now postprocesses directly from \code{X} and does not need to form \code{Sx}.}

\item{Sy}{Optional covariance matrix for Y. If NULL, computed similarly; optionally shrunk via Ledoit-Wolf}

\item{Sxy}{Optional cross-covariance matrix (not currently used)}

\item{lambda}{Regularization parameter for sparsity}

\item{r}{Target rank}

\item{standardize}{Backward-compatible preprocessing flag: TRUE = \code{"scale"}, FALSE = \code{"center"}.}

\item{preprocess}{Preprocessing mode. One of \code{"scale"} (center + scale), \code{"center"} (center only), or \code{"none"}.}

\item{LW_Sy}{Whether to apply Ledoit-Wolf shrinkage to Sy}

\item{rho}{ADMM penalty parameter}

\item{niter}{Maximum number of ADMM iterations}

\item{thresh}{Convergence threshold for ADMM}

\item{thresh_0}{Threshold for small values in the coefficient matrix (default 1e-6)}

\item{verbose}{Whether to print diagnostic output}

\item{Gamma_dagger}{Optional pseudoinverse of Gamma (computed if NULL)}
}
\value{
A list with elements:
\describe{
\item{U}{Canonical direction matrix for X (p x r)}
\item{V}{Canonical direction matrix for Y (q x r)}
\item{cor}{Canonical covariances}
\item{loss}{The prediction error 1/n * \| XU - YV\|^2}
\item{Lambda}{Canonical correlations}
\item{B_opt}{Estimated reduced-rank coefficient matrix}
}
}
\description{
Solves a sparse canonical correlation problem using a graph-constrained reduced-rank regression formulation.
The problem is solved via an ADMM approach.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/benchmarks.r
\name{sparse_CCA_benchmarks}
\alias{sparse_CCA_benchmarks}
\title{Additional Benchmarks for Sparse CCA Methods}
\usage{
sparse_CCA_benchmarks(
  X_train,
  Y_train,
  S = NULL,
  rank = 2,
  kfolds = 5,
  method.type = "FIT_SAR_CV",
  lambdax = 10^seq(from = -3, to = 2, length = 10),
  lambday = c(0, 1e-07, 1e-06, 1e-05),
  standardize = TRUE
)
}
\arguments{
\item{X_train}{Matrix of predictors (n x p)}

\item{Y_train}{Matrix of responses (n x q)}

\item{S}{Optional covariance matrix (default is NULL, which computes it from X_train and Y_train)}

\item{rank}{Target rank for the CCA (default is 2)}

\item{kfolds}{Number of cross-validation folds (default is 5)}

\item{method.type}{Type of method to use for Sparse CCA (default is "FIT_SAR_CV"). Choices include "FIT_SAR_BIC", "FIT_SAR_CV", "Witten_Perm", "Witten.CV",  and "SCCA_Parkhomenko".}

\item{lambdax}{Vector of sparsity parameters for X (default is a sequence from 0 to 1 with step 0.1)}

\item{lambday}{Vector of sparsity parameters for Y (default is a sequence from 0 to 1 with step 0.1)}

\item{standardize}{Standardize (center and scale) the data matrices X and Y (default is TRUE) before analysis}
}
\value{
A matrix (p+q)x r containing the canonical directions for X and Y.
}
\description{
Additional Benchmarks for Sparse CCA Methods
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alt_Parkhomenko.R
\name{SCCA_Parkhomenko}
\alias{SCCA_Parkhomenko}
\title{Function to perform Sparse CCA based on Waaijenborg et al. (2008)
REFERENCE Parkhomenko et al. (2009), "Sparse Canonical Correlation Anlaysis with Application to Genomic Data Integration" in  Statistical Applications in Genetics and Molecular Biology, Volume 8, Issue 1, Article 1}
\usage{
SCCA_Parkhomenko(
  x.data,
  y.data,
  n.cv = 5,
  lambda.v.seq = seq(0, 0.2, by = 0.02),
  lambda.u.seq = seq(0, 0.2, by = 0.02),
  Krank = 1,
  standardize = TRUE
)
}
\arguments{
\item{x.data}{Matrix of predictors (n x p)}

\item{y.data}{Matrix of responses (n x q)}

\item{n.cv}{Number of cross-validation folds (default is 5)}

\item{lambda.v.seq}{Vector of sparsity parameters for Y (default is a sequence from 0 to 1 with step 0.1)}

\item{lambda.u.seq}{Vector of sparsity parameters for X (default is a sequence from 0 to 1 with step 0.1)}

\item{Krank}{Number of canonical components to extract}

\item{standardize}{Standardize (center and scale) the data matrices X and Y (default is TRUE) before analysis}
}
\value{
A list with elements:
\describe{
\item{U}{Canonical direction matrix for X (p x r)}
\item{V}{Canonical direction matrix for Y (q x r)}
\item{cor}{Canonical correlations}
}
}
\description{
Function to perform Sparse CCA based on Waaijenborg et al. (2008)
REFERENCE Parkhomenko et al. (2009), "Sparse Canonical Correlation Anlaysis with Application to Genomic Data Integration" in  Statistical Applications in Genetics and Molecular Biology, Volume 8, Issue 1, Article 1
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reduced_rank_regression.R
\name{cca_rrr_cv}
\alias{cca_rrr_cv}
\title{Cross-validated Canonical Correlation Analysis via RRR}
\usage{
cca_rrr_cv(
  X,
  Y,
  r = 2,
  lambdas = 10^seq(-3, 1.5, length.out = 100),
  kfolds = 10,
  solver = "ADMM",
  mode = "sqrtm_norm",
  parallelize = FALSE,
  LW_Sy = TRUE,
  standardize = FALSE,
  preprocess = NULL,
  cv_metric = "mse",
  rho = 1,
  thresh_0 = 0,
  niter = 10000,
  matrix_free_threshold = 4000L,
  cg_tol = 1e-06,
  cg_maxiter = NULL,
  thresh = 1e-04,
  verbose = FALSE,
  nb_cores = NULL
)
}
\arguments{
\item{X}{Matrix of predictors.}

\item{Y}{Matrix of responses.}

\item{r}{Rank of the solution.}

\item{lambdas}{Sequence of lambda values for cross-validation.}

\item{kfolds}{Number of folds for cross-validation.}

\item{solver}{Solver type: "rrr", "CVX", or "ADMM". The \code{"CVX"} backend
requires the optional package \code{CVXR}.}

\item{mode}{Mode for postprocessing the RRR solution. One of \code{"sqrtm_norm"}
(default) or \code{"product_norm"}. Legacy aliases \code{"new"} and \code{"old"} are
also accepted. The former whitens the canonical variates to have identity
covariance, while the latter does not whiten and instead returns the raw
SVD factors of the RRR solution. The \code{"product_norm"} mode may be more
interpretable in some cases but can yield canonical variates with very
different scales and is not guaranteed to be numerically stable when the
RRR solution is very low-rank or nearly low-rank.}

\item{parallelize}{Logical; should cross-validation be parallelized?}

\item{LW_Sy}{Whether to use Ledoit-Wolf shrinkage for Sy.}

\item{standardize}{Backward-compatible preprocessing flag: TRUE = \code{"scale"}, FALSE = \code{"center"}.}

\item{preprocess}{Preprocessing mode. One of \code{"scale"} (center + scale), \code{"center"} (center only), or \code{"none"}.
Logical values are still accepted for backward compatibility: \code{TRUE} uses \code{standardize}, \code{FALSE} skips preprocessing.
Data are preprocessed once up front, and fold fits reuse those transformed matrices without re-centering or re-scaling inside each fold.}

\item{cv_metric}{Cross-validation metric. Use \code{"mse"} to minimize held-out prediction
error or \code{"correlation"} to maximize held-out association between \code{X \%*\% U}
and \code{Y \%*\% V}.}

\item{rho}{ADMM parameter.}

\item{thresh_0}{tolerance for declaring entries non-zero}

\item{niter}{Maximum number of iterations for ADMM.}

\item{matrix_free_threshold}{For ADMM: when both \code{n} and \code{p} are at least this value, use a matrix-free conjugate-gradient solve instead of forming a dense linear system.}

\item{cg_tol}{Relative tolerance for the matrix-free conjugate-gradient solve used in ADMM.}

\item{cg_maxiter}{Maximum iterations for the matrix-free conjugate-gradient solve. Defaults to \code{min(p, 1000)}.}

\item{thresh}{Convergence threshold.}

\item{verbose}{Logical for verbose output.}

\item{nb_cores}{Number of cores to use for parallelization. Defaults to min(kfolds, available cores minus 1).}
}
\value{
A list with elements:
\describe{
\item{U}{Canonical direction matrix for X (p x r)}
\item{V}{Canonical direction matrix for Y (q x r)}
\item{lambda}{Optimal regularisation parameter lambda chosen by CV}
\item{rmse}{Backward-compatible optimization objective. For \code{cv_metric = "mse"}
this is the held-out mean squared error; for \code{cv_metric = "correlation"} it is
\code{-cv_score} so that smaller still means better.}
\item{cv_score}{Raw held-out cross-validation score averaged across folds.}
\item{cv_metric}{The metric used to score lambdas during cross-validation.}
\item{cor}{Canonical correlations at the selected lambda}
\item{lambda_x}{Alias of the selected \code{lambda}}
\item{lambda_x_se}{Foldwise standard error at the selected \code{lambda}}
\item{lambda_y}{Placeholder for symmetry with two-penalty interfaces}
\item{lambda_y_se}{Placeholder for symmetry with two-penalty interfaces}
\item{resultsx}{Backward-compatible alias of \code{cv_summary}}
\item{cv_summary}{Data frame with one row per lambda containing the mean CV score
and its foldwise standard error.}
\item{cv_folds}{Data frame with fold-level CV scores for each lambda.}
\item{Lambda}{Canonical correlations from the final fit}
\item{B}{Estimated reduced-rank coefficient matrix from the final fit}
\item{fit}{Final fit at the selected lambda}
}
}
\description{
Performs cross-validation to select optimal lambda, fits CCA_rrr.
Canonical Correlation Analysis via Reduced Rank Regression (RRR)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.R
\docType{package}
\name{ccar3-package}
\alias{ccar3}
\alias{ccar3-package}
\title{ccar3: Canonical Correlation Analysis via Reduced Rank Regression}
\description{
Canonical correlation analysis (CCA) via reduced-rank regression with support for regularization and cross-validation. Several methods for estimating CCA in high-dimensional settings are implemented. The first set of methods, cca_rrr() (and variants: cca_group_rrr() and cca_graph_rrr()), assumes that one dataset is high-dimensional and the other is low-dimensional, while the second, ecca() (for Efficient CCA) assumes that both datasets are high-dimensional. For both methods, standard l1 regularization as well as group-lasso regularization are available. cca_graph_rrr further supports total variation regularization when there is a known graph structure among the variables of the high-dimensional dataset. In this case, the loadings of the canonical directions of the high-dimensional dataset are assumed to be smooth on the graph. For more details see Donnat and Tuzhilina (2024) \doi{10.48550/arXiv.2405.19539} and Wu, Tuzhilina and Donnat (2025) \doi{10.48550/arXiv.2507.11160}.
}
\author{
\strong{Maintainer}: Claire Donnat \email{cdonnat@uchicago.edu} (\href{https://orcid.org/0000-0001-7079-8060}{ORCID})

Authors:
\itemize{
  \item Elena Tuzhilina \email{elena.tuzhilina@utoronto.ca} (\href{https://orcid.org/0000-0002-1898-6010}{ORCID})
  \item Zixuan Wu \email{zixuanwu@uchicago.edu} (\href{https://orcid.org/0009-0006-4745-0000}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{subdistance}
\alias{subdistance}
\title{Subdistance between subspaces}
\usage{
subdistance(A, B)
}
\arguments{
\item{A}{A matrix whose columns span a subspace.}

\item{B}{A matrix whose columns span a subspace.}
}
\value{
subdistance between the two subspaces spanned by the matrices A and B, defined as min(O orthogonal) ||AO-B||_F
}
\description{
Calculate subdistance between subspace spanned by the columns of a and the subspace spanned by the columns of b
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{TPR}
\alias{TPR}
\title{True Positive Rate (TPR)}
\usage{
TPR(A, B, tol = 1e-04)
}
\arguments{
\item{A}{A matrix (the estimator).}

\item{B}{A matrix (assumed to be the ground truth).}

\item{tol}{tolerance for declaring the entries non zero.}
}
\value{
True Positive Rate (nb of values that are non zero in both A and B / (nb of values that are non zero in A))
}
\description{
This is a function that compares the structure of two matrices A and B.  It outputs the number of entries that A and B have in common that are different from zero. A and B need to have the same number of rows and columns
}
\examples{
A <- matrix(c(1, 0, 0, 1, 1, 0), nrow = 2)
B <- matrix(c(1, 0, 1, 1, 0, 0), nrow = 2)
TPR(A, B)  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alt_Witten_CrossValidation.R
\name{Witten.CV}
\alias{Witten.CV}
\title{Sparse CCA by Witten and Tibshirani (2009)}
\usage{
Witten.CV(
  X,
  Y,
  n.cv = 5,
  rank,
  lambdax = matrix(seq(from = 0, to = 1, by = 0.1), nrow = 1),
  lambday = matrix(seq(from = 0, to = 1, by = 0.1), nrow = 1),
  standardize = TRUE
)
}
\arguments{
\item{X}{Matrix of predictors (n x p)}

\item{Y}{Matrix of responses (n x q)}

\item{n.cv}{Number of cross-validation folds (default is 5)}

\item{rank}{Number of canonical components to extract}

\item{lambdax}{Vector of sparsity parameters for X (default is a sequence from 0 to 1 with step 0.1)}

\item{lambday}{Vector of sparsity parameters for Y (default is a sequence from 0 to 1 with step 0.1)}

\item{standardize}{Standardize (center and scale) the data matrices X and Y (default is TRUE) before analysis}
}
\value{
the appropriate levels of regularisation
}
\description{
Sparse CCA by Witten and Tibshirani (2009)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph_reduced_rank_regression.R
\name{cca_graph_rrr_cv}
\alias{cca_graph_rrr_cv}
\title{Graph-regularized Reduced-Rank Regression for Canonical Correlation Analysis with cross validation}
\usage{
cca_graph_rrr_cv(
  X,
  Y,
  Gamma,
  r = 2,
  lambdas = 10^seq(-3, 1.5, length.out = 10),
  kfolds = 5,
  parallelize = FALSE,
  standardize = TRUE,
  LW_Sy = TRUE,
  preprocess = NULL,
  rho = 10,
  niter = 10000,
  thresh = 1e-04,
  thresh_0 = 1e-06,
  verbose = FALSE,
  Gamma_dagger = NULL,
  nb_cores = NULL
)
}
\arguments{
\item{X}{Matrix of predictors (n x p)}

\item{Y}{Matrix of responses (n x q)}

\item{Gamma}{Graph constraint matrix (g x p)}

\item{r}{Target rank}

\item{lambdas}{Grid of regularization parameters to test for sparsity}

\item{kfolds}{Number of folds for cross-validation}

\item{parallelize}{Whether to parallelize cross-validation}

\item{standardize}{Backward-compatible preprocessing flag: TRUE = \code{"scale"}, FALSE = \code{"center"}.}

\item{LW_Sy}{Whether to apply Ledoit-Wolf shrinkage to Sy}

\item{preprocess}{Preprocessing mode. One of \code{"scale"} (center + scale), \code{"center"} (center only), or \code{"none"}.}

\item{rho}{ADMM penalty parameter}

\item{niter}{Maximum number of ADMM iterations}

\item{thresh}{Convergence threshold for ADMM}

\item{thresh_0}{Threshold for small values in the coefficient matrix (default 1e-6)}

\item{verbose}{Whether to print diagnostic output}

\item{Gamma_dagger}{Optional pseudoinverse of Gamma (computed if NULL)}

\item{nb_cores}{Number of cores to use for parallelization. Defaults to min(kfolds, available cores minus 1).}
}
\value{
A list with elements:
\describe{
\item{U}{Canonical direction matrix for X (p x r)}
\item{V}{Canonical direction matrix for Y (q x r)}
\item{lambda}{Optimal regularisation parameter lambda chosen by CV}
\item{rmse}{Mean squared error of prediction (as computed in the CV)}
\item{cor}{Canonical covariances}
\item{lambda_x}{Alias of the selected \code{lambda}}
\item{lambda_x_se}{Foldwise standard error at the selected \code{lambda}}
\item{lambda_y}{Placeholder for symmetry with two-penalty interfaces}
\item{lambda_y_se}{Placeholder for symmetry with two-penalty interfaces}
\item{resultsx}{Backward-compatible alias of \code{cv_summary}}
\item{cv_summary}{Data frame with one row per lambda containing mean RMSE and its foldwise standard error}
\item{cv_folds}{Data frame with fold-level RMSE values for each lambda}
\item{Lambda}{Canonical correlations from the final fit}
\item{B}{Estimated reduced-rank coefficient matrix from the final fit}
\item{fit}{Final fit at the selected lambda}
}
}
\description{
Solves a sparse canonical correlation problem using a graph-constrained reduced-rank regression formulation.
The problem is solved via an ADMM approach.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ecca.r
\name{ecca}
\alias{ecca}
\title{Efficient CCA for Two High-Dimensional Views}
\usage{
ecca(
  X,
  Y,
  lambda = 0,
  groups = NULL,
  r = 2,
  standardize = FALSE,
  rho = 1,
  B0 = NULL,
  eps = 1e-04,
  maxiter = 500,
  verbose = TRUE,
  epsilon_sv = 1e-08,
  ridge_whiten = 1e-08
)
}
\arguments{
\item{X}{Predictor matrix (n x p).}

\item{Y}{Response matrix (n x q).}

\item{lambda}{Regularization parameter.}

\item{groups}{Optional group structure for blockwise sparsity.}

\item{r}{Target rank.}

\item{standardize}{Whether to scale variables after centering.}

\item{rho}{ADMM penalty parameter.}

\item{B0}{Optional warm start for the coefficient matrix.}

\item{eps}{Convergence tolerance for ADMM.}

\item{maxiter}{Maximum number of ADMM iterations.}

\item{verbose}{Whether to print diagnostics.}

\item{epsilon_sv}{Numerical threshold used to discard near-zero singular values.}

\item{ridge_whiten}{Ridge added when whitening Gram matrices.}
}
\value{
A list containing the estimated canonical directions, canonical
correlations, the fitted coefficient matrix, preprocessing metadata, and
convergence information.
}
\description{
Fits sparse canonical directions with an ADMM-based reduced-rank regression
formulation tailored to the setting where both views are high-dimensional.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alt_SAR.R
\name{SparseCCA}
\alias{SparseCCA}
\title{Function to perform Sparse CCA based on Wilms and Croux (2018)
REFERENCE Wilms, I., & Croux, C. (2018). Sparse canonical correlation analysis using alternating regressions. Journal of Computational and Graphical Statistics, 27(1), 1-10.}
\usage{
SparseCCA(
  X,
  Y,
  lambdaAseq = seq(from = 1, to = 0.01, by = -0.01),
  lambdaBseq = seq(from = 1, to = 0.01, by = -0.01),
  rank,
  selection.criterion = 1,
  n.cv = 5,
  A.initial = NULL,
  B.initial = NULL,
  max.iter = 20,
  conv = 10^-2,
  standardize = TRUE
)
}
\arguments{
\item{X}{Matrix of predictors (n x p)}

\item{Y}{Matrix of responses (n x q)}

\item{lambdaAseq}{Vector of sparsity parameters for X (default is a sequence from 0 to 1 with step 0.1)}

\item{lambdaBseq}{Vector of sparsity parameters for Y (default is a sequence from 0 to 1 with step 0.1)}

\item{rank}{Number of canonical components to extract}

\item{selection.criterion}{Criterion for selecting the optimal tuning parameter (1 for minimizing difference between test and training sample correlation, 2 for maximizing test sample correlation)}

\item{n.cv}{Number of cross-validation folds (default is 5)}

\item{A.initial}{Initial value for the canonical vector A (default is NULL, which uses a canonical ridge solution)}

\item{B.initial}{Initial value for the canonical vector B (default is NULL, which uses a canonical ridge solution)}

\item{max.iter}{Maximum number of iterations for convergence (default is 20)}

\item{conv}{Convergence threshold (default is 1e-2)}

\item{standardize}{Standardize (center and scale) the data matrices X and Y (default is TRUE) before analysis}
}
\value{
A list with elements:
\describe{
\item{U}{Canonical direction matrix for X (p x r)}
\item{V}{Canonical direction matrix for Y (q x r)}
\item{loss}{Mean squared error of prediction}
\item{cor}{Canonical covariances}
}
}
\description{
Function to perform Sparse CCA based on Wilms and Croux (2018)
REFERENCE Wilms, I., & Croux, C. (2018). Sparse canonical correlation analysis using alternating regressions. Journal of Computational and Graphical Statistics, 27(1), 1-10.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{principal_angles}
\alias{principal_angles}
\title{Metrics for subspaces}
\usage{
principal_angles(a, b)
}
\arguments{
\item{a}{A matrix whose columns span a subspace.}

\item{b}{A matrix whose columns span a subspace.}
}
\value{
a vector of principal angles (in radians)
}
\description{
Calculate principal angles between subspace spanned by the columns of a and the subspace spanned by the columns of b
}
\examples{
a <- matrix(rnorm(9), 3, 3)
b <- matrix(rnorm(9), 3, 3)
principal_angles(a, b)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph_utils.R
\name{get_edge_incidence}
\alias{get_edge_incidence}
\title{Return the edge incidence matrix of an igraph graph}
\usage{
get_edge_incidence(g, weight = 1)
}
\arguments{
\item{g}{igraph graph object.}

\item{weight}{edge weights.}
}
\value{
Edge incidence matrix of the graph g, with +weight for the source node and -weight for the target node.
}
\description{
Return the edge incidence matrix of an igraph graph
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{FPR}
\alias{FPR}
\title{False Positive Rate (TPR)}
\usage{
FPR(A, B, tol = 1e-04)
}
\arguments{
\item{A}{A matrix.}

\item{B}{A matrix (assumed to be the ground truth).}

\item{tol}{tolerance for declaring the entries non zero.}
}
\value{
False Positive Rate (nb of values that are non zero in A and  zero in B / (nb of values that are non zero in A))
}
\description{
This is a function that compares the structure of two matrices A and B. It outputs the number of entries where A  is not zero but Bis. A and B need to have the same number of rows and columns
}
\examples{
A <- matrix(c(1, 0, 0, 1, 1, 0), nrow = 2)
B <- matrix(c(1, 0, 1, 1, 0, 0), nrow = 2)
FPR(A, B)  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_reduced_rank_regression.R
\name{cca_group_rrr}
\alias{cca_group_rrr}
\title{Group-Sparse Canonical Correlation via Reduced-Rank Regression}
\usage{
cca_group_rrr(
  X,
  Y,
  groups,
  Sx = NULL,
  Sy = NULL,
  Sxy = NULL,
  lambda = 0,
  r,
  standardize = FALSE,
  preprocess = NULL,
  LW_Sy = TRUE,
  solver = "ADMM",
  rho = 1,
  niter = 10000,
  thresh = 1e-04,
  thresh_0 = 1e-06,
  matrix_free_threshold = 4000L,
  cg_tol = 1e-06,
  cg_maxiter = NULL,
  verbose = FALSE
)
}
\arguments{
\item{X}{Predictor matrix (n x p)}

\item{Y}{Response matrix (n x q)}

\item{groups}{List of index vectors defining groups of predictors}

\item{Sx}{Optional covariance matrix for X; if NULL computed internally}

\item{Sy}{Optional covariance matrix for Y; if NULL computed internally}

\item{Sxy}{Optional cross covariance matrix for X and Y; if NULL computed internally}

\item{lambda}{Regularization parameter}

\item{r}{Target rank}

\item{standardize}{Backward-compatible preprocessing flag: TRUE = \code{"scale"}, FALSE = \code{"center"}.}

\item{preprocess}{Preprocessing mode. One of \code{"scale"} (center + scale), \code{"center"} (center only), or \code{"none"}.}

\item{LW_Sy}{Whether to apply Ledoit-Wolf shrinkage to Sy (default TRUE)}

\item{solver}{Either "ADMM" or "CVXR". The \code{"CVXR"} backend requires the
optional package \code{CVXR}.}

\item{rho}{ADMM parameter}

\item{niter}{Maximum number of ADMM iterations}

\item{thresh}{Convergence threshold for ADMM}

\item{thresh_0}{tolerance for declaring entries non-zero}

\item{matrix_free_threshold}{For ADMM: when both \code{n} and \code{p} are at least this value, use a matrix-free conjugate-gradient solve instead of forming a dense linear system.}

\item{cg_tol}{Relative tolerance for the matrix-free conjugate-gradient solve used in ADMM.}

\item{cg_maxiter}{Maximum iterations for the matrix-free conjugate-gradient solve. Defaults to \code{min(p, 1000)}.}

\item{verbose}{Print diagnostics}
}
\value{
A list with elements:
\describe{
\item{U}{Canonical direction matrix for X (p x r)}
\item{V}{Canonical direction matrix for Y (q x r)}
\item{cor}{Canonical covariances}
\item{loss}{The prediction error 1/n * \| XU - YV\|^2}
\item{Lambda}{Canonical correlations}
\item{B_opt}{Estimated reduced-rank coefficient matrix}
}
}
\description{
Performs group-sparse reduced-rank regression for CCA using either ADMM or CVXR solvers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{TNR}
\alias{TNR}
\title{True Negative Rate (TNR)}
\usage{
TNR(A, B, tol = 1e-04)
}
\arguments{
\item{A}{A matrix.}

\item{B}{A matrix (assumed to be the ground truth)..}

\item{tol}{tolerance for declaring the entries non zero.}
}
\value{
True Negative Rate (nb of values that are zero in A and  zero in B / (nb of values that are  zero in A))
}
\description{
This is a function that compares the structure of two matrices A and B. It outputs the number of entries where A and B are both 0. A and B need to have the same number of rows and columns
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reduced_rank_regression.R
\name{cca_rrr}
\alias{cca_rrr}
\title{Canonical Correlation Analysis via Reduced Rank Regression (RRR)}
\usage{
cca_rrr(
  X,
  Y,
  Sx = NULL,
  Sy = NULL,
  lambda = 0,
  r,
  highdim = TRUE,
  solver = "ADMM",
  LW_Sy = TRUE,
  mode = "sqrtm_norm",
  standardize = FALSE,
  preprocess = NULL,
  rho = 1,
  niter = 10000,
  thresh = 1e-04,
  thresh_0 = 0,
  matrix_free_threshold = 4000L,
  cg_tol = 1e-06,
  cg_maxiter = NULL,
  verbose = FALSE
)
}
\arguments{
\item{X}{Matrix of predictors.}

\item{Y}{Matrix of responses.}

\item{Sx}{Optional X covariance matrix.}

\item{Sy}{Optional Y covariance matrix.}

\item{lambda}{Regularization parameter.}

\item{r}{Rank of the solution.}

\item{highdim}{Boolean for high-dimensional regime.}

\item{solver}{Solver type: "rrr", "CVX", or "ADMM". The \code{"CVX"} backend
requires the optional package \code{CVXR}.}

\item{LW_Sy}{Whether to use Ledoit-Wolf shrinkage for Sy.}

\item{mode}{Mode for postprocessing the RRR solution. One of \code{"sqrtm_norm"}
(default) or \code{"product_norm"}. Legacy aliases \code{"new"} and \code{"old"} are
also accepted. The former whitens the canonical variates to have identity
covariance, while the latter does not whiten and instead returns the raw
SVD factors of the RRR solution. The \code{"product_norm"} mode may be more
interpretable in some cases but can yield canonical variates with very
different scales and is not guaranteed to be numerically stable when the
RRR solution is very low-rank or nearly low-rank.}

\item{standardize}{Backward-compatible preprocessing flag: TRUE = \code{"scale"}, FALSE = \code{"center"}.}

\item{preprocess}{Preprocessing mode. One of \code{"scale"} (center + scale), \code{"center"} (center only), or \code{"none"}.
Logical values are still accepted for backward compatibility: \code{TRUE} uses \code{standardize}, \code{FALSE} skips preprocessing.}

\item{rho}{ADMM parameter.}

\item{niter}{Maximum number of iterations for ADMM.}

\item{thresh}{Convergence threshold.}

\item{thresh_0}{For the ADMM solver: Set entries whose absolute value is below this to 0 (default 1e-6).}

\item{matrix_free_threshold}{For ADMM: when both \code{n} and \code{p} are at least this value, use a matrix-free conjugate-gradient solve instead of forming a dense linear system.}

\item{cg_tol}{Relative tolerance for the matrix-free conjugate-gradient solve used in ADMM.}

\item{cg_maxiter}{Maximum iterations for the matrix-free conjugate-gradient solve. Defaults to \code{min(p, 1000)}.}

\item{verbose}{Logical for verbose output.}
}
\value{
A list with elements:
\itemize{
\item U: Canonical direction matrix for X (p x r)
\item V: Canonical direction matrix for Y (q x r)
\item cor: Canonical covariances
\item loss: The prediction error 1/n * || XU - YV ||^2
}
}
\description{
Estimates canonical directions using various RRR solvers and penalties.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ecca.r
\name{ecca.cv}
\alias{ecca.cv}
\title{Cross-Validated Efficient CCA}
\usage{
ecca.cv(
  X,
  Y,
  lambdas = 0,
  groups = NULL,
  r = 2,
  standardize = FALSE,
  rho = 1,
  B0 = NULL,
  nfold = 5,
  select = "lambda.min",
  eps = 0.001,
  maxiter = 1000,
  verbose = FALSE,
  maxiter_cv = 300,
  parallel = FALSE,
  nb_cores = NULL,
  set_seed_cv = NULL,
  scoring_method = c("mse", "trace"),
  cv_use_median = FALSE,
  dense = TRUE,
  optimized = FALSE,
  epsilon_sv = 1e-08,
  ridge_whiten = 1e-08
)
}
\arguments{
\item{X}{Predictor matrix (n x p).}

\item{Y}{Response matrix (n x q).}

\item{lambdas}{Candidate regularization values.}

\item{groups}{Optional group structure for blockwise sparsity.}

\item{r}{Target rank.}

\item{standardize}{Whether to scale variables after centering.}

\item{rho}{ADMM penalty parameter.}

\item{B0}{Optional warm start for the coefficient matrix.}

\item{nfold}{Number of cross-validation folds.}

\item{select}{Selection rule for the final lambda. One of \code{"lambda.min"} or \code{"lambda.1se"}.}

\item{eps}{Convergence tolerance for the final ADMM refit.}

\item{maxiter}{Maximum iterations for the final ADMM refit.}

\item{verbose}{Whether to print diagnostics.}

\item{maxiter_cv}{Maximum iterations used inside the cross-validation fits.}

\item{parallel}{Whether to parallelize cross-validation.}

\item{nb_cores}{Number of worker processes to use when \code{parallel = TRUE}.}

\item{set_seed_cv}{Optional random seed for fold generation.}

\item{scoring_method}{Cross-validation score to optimize. One of \code{"mse"} or \code{"trace"}.}

\item{cv_use_median}{Whether to aggregate fold scores with the median instead of the mean.}

\item{dense}{Retained for backward compatibility.}

\item{optimized}{Retained for backward compatibility.}

\item{epsilon_sv}{Numerical threshold used to discard near-zero singular values.}

\item{ridge_whiten}{Ridge added when whitening Gram matrices.}
}
\value{
A list with the final fit, selected lambda, and cross-validation
scores when more than one lambda is supplied.
}
\description{
Selects a regularization parameter for \code{\link[=ecca]{ecca()}} by cross-validation and
refits the final model at the selected value.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{sinTheta}
\alias{sinTheta}
\title{SinTheta distance between subspaces}
\usage{
sinTheta(U, V)
}
\arguments{
\item{U}{A matrix whose columns span a subspace.}

\item{V}{A matrix whose columns span a subspace.}
}
\value{
sinTheta distance between the two subspaces spanned by the matrices A and B, defined as ||UU^T - VV^T||_F / sqrt(2)
}
\description{
Calculate the distance spanned by the columns of A and the subspace spanned by the columns of B, defined as ||UU^T - VV^T||_F / sqrt(2)
}
